package items;

import java.util.logging.Logger;

import l2r.extensions.scripts.ScriptFile;
import l2r.gameserver.ThreadPoolManager;
import l2r.gameserver.cache.Msg;
import l2r.gameserver.geodata.network.geoclient.GeoClient;
import l2r.gameserver.handler.IItemHandler;
import l2r.gameserver.handler.ItemHandler;
import l2r.gameserver.model.L2PetData;
import l2r.gameserver.model.L2Playable;
import l2r.gameserver.model.L2Player;
import l2r.gameserver.model.L2Skill;
import l2r.gameserver.model.L2World;
import l2r.gameserver.model.instances.L2ItemInstance;
import l2r.gameserver.model.instances.L2PetBabyInstance;
import l2r.gameserver.model.instances.L2PetInstance;
import l2r.gameserver.serverpackets.MagicSkillLaunched;
import l2r.gameserver.serverpackets.MagicSkillUse;
import l2r.gameserver.serverpackets.SystemMessage;
import l2r.gameserver.tables.NpcTable;
import l2r.gameserver.tables.PetDataTable;
import l2r.gameserver.tables.SkillTable;
import l2r.gameserver.templates.L2NpcTemplate;

public class PetSummon implements IItemHandler, ScriptFile
{
	protected static Logger _log = Logger.getLogger(PetSummon.class.getName());

	// all the items ids that this handler knowns
	private static final short[] _itemIds = PetDataTable._itemControlIds;
	private static final int _skillId = 2046;

	static final SystemMessage YOU_MAY_NOT_USE_MULTIPLE_PETS_OR_SERVITORS_AT_THE_SAME_TIME = new SystemMessage(SystemMessage.YOU_MAY_NOT_USE_MULTIPLE_PETS_OR_SERVITORS_AT_THE_SAME_TIME);
	static final SystemMessage A_STRIDER_CANNOT_BE_RIDDEN_WHILE_IN_BATTLE = new SystemMessage(SystemMessage.A_STRIDER_CANNOT_BE_RIDDEN_WHILE_IN_BATTLE);
	static final SystemMessage SUMMON_A_PET = new SystemMessage(SystemMessage.SUMMON_A_PET);

	public void useItem(L2Playable playable, L2ItemInstance item)
	{
		if(playable == null || !playable.isPlayer)
			return;
		L2Player player = (L2Player) playable;

		// pets resummon fast fix
		if(System.currentTimeMillis() <= player.lastDiceThrown)
			return;
		player.lastDiceThrown = System.currentTimeMillis() + 4000L;

		if(player.isTransactionInProgress())
			return;

		if(player.isSitting())
		{
			player.sendPacket(Msg.YOU_CANNOT_MOVE_WHILE_SITTING);
			return;
		}

		if(player.isInOlympiadMode())
		{
			player.sendPacket(Msg.THIS_ITEM_IS_NOT_AVAILABLE_FOR_THE_OLYMPIAD_EVENT);
			return;
		}

		if(player.isCastingNow() || player.isActionsDisabled())
			return;

		L2Skill skill = SkillTable.getInstance().getInfo(_skillId, 1);
		if(!skill.checkCondition(player, player, false, true, true))
			return;

		if(player.getPet() != null)
		{
			player.sendPacket(YOU_MAY_NOT_USE_MULTIPLE_PETS_OR_SERVITORS_AT_THE_SAME_TIME);
			return;
		}

		if(player.isMounted() || player.isInBoat())
			return;

		if(player.isCursedWeaponEquipped() || player.isInCombat())
		{
			// You can't mount while weilding a cursed weapon
			player.sendPacket(A_STRIDER_CANNOT_BE_RIDDEN_WHILE_IN_BATTLE);
			return;
		}

		int npcId = PetDataTable.getSummonId(item);
		if(npcId == 0)
			return;

		L2NpcTemplate petTemplate = NpcTable.getTemplate(npcId);

		L2PetInstance newpet;
		if(L2PetData.isBaby(npcId))
			newpet = L2PetBabyInstance.spawnPet(petTemplate, player, item);
		else
			newpet = L2PetInstance.spawnPet(petTemplate, player, item);

		if(newpet == null)
			return;

		newpet.setTitle(player.getName());

		if(!newpet.isRespawned())
			try
			{
				newpet.setCurrentHp(newpet.getMaxHp());
				newpet.setCurrentMp(newpet.getMaxMp());
				newpet.setExp(newpet.getExpForThisLevel());
				newpet.setCurrentFed(newpet.getMaxFed());
			}
			catch(NullPointerException e)
			{
				_log.warning("PetSummon: failed set stats for summon " + npcId + ".");
				return;
			}

		newpet.setRunning();

		if(!newpet.isRespawned())
			newpet.store();

		player.setPet(newpet);

		player.broadcastPacket(new MagicSkillUse(player, player, _skillId, 1, 1000, 600000));
		player.sendPacket(SUMMON_A_PET);
		L2World.addObject(newpet);

		newpet.spawnMe(GeoClient.getInstance().findPointToStay(player.getX(), player.getY(), player.getZ(), 100, 150));
		newpet.broadcastPetInfo();
		newpet.setShowSpawnAnimation(false);
		newpet.startFeed(false);

		if(newpet instanceof L2PetBabyInstance)
			((L2PetBabyInstance) newpet).startHealTask();

		// continue execution in 1 seconds
		ThreadPoolManager.getInstance().scheduleGeneral(new SummonFinalizer(player, newpet), 900);
	}

	static class SummonFinalizer implements Runnable
	{
		private L2Player _activeChar;
		private L2PetInstance _newpet;

		SummonFinalizer(L2Player activeChar, L2PetInstance newpet)
		{
			_activeChar = activeChar;
			_newpet = newpet;
		}

		public void run()
		{
			try
			{
				_activeChar.sendPacket(new MagicSkillLaunched(_activeChar.getObjectId(), 2046, 1, _newpet));
				_newpet.setFollowStatus(true);
			}
			catch(Throwable e)
			{
				_log.severe(e.toString());
			}
		}
	}

	public final short[] getItemIds()
	{
		return _itemIds;
	}

	public void onLoad()
	{
		ItemHandler.getInstance().registerItemHandler(this);
	}

	public void onReload()
	{}

	public void onShutdown()
	{}
}