package l2r.gameserver;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Iterator;
import java.util.logging.Logger;

import l2r.Config;
import l2r.extensions.listeners.PropertyCollection;
import l2r.extensions.listeners.engine.DefaultListenerEngine;
import l2r.extensions.listeners.engine.ListenerEngine;
import l2r.gameserver.cache.Msg;
import l2r.gameserver.instancemanager.DayNightSpawnManager;
import l2r.gameserver.model.L2Character;
import l2r.gameserver.model.L2Player;
import l2r.gameserver.model.L2World;
import l2r.gameserver.model.L2WorldRegion;
import l2r.gameserver.model.entity.SevenSigns;
import l2r.gameserver.serverpackets.L2GameServerPacket;

@SuppressWarnings( { "nls", "unqualified-field-access", "boxing" })
public class GameTimeController
{
	private static final Logger _log = Logger.getLogger(GameTimeController.class.getName());

	public static final int TICKS_PER_SECOND = 10;
	public static final int MILLIS_IN_TICK = 1000 / TICKS_PER_SECOND;

	private static GameTimeController _instance = new GameTimeController();

	private static int _gameTicks;
	private static long _gameStartTime;
	private static boolean _isNight;

	private TimerThread _timer;

	public long _startMoveTime;

	private final ListenerEngine<GameTimeController> listenerEngine = new DefaultListenerEngine<GameTimeController>(this);

	/**
	 * one ingame day is 240 real minutes
	 */
	public static GameTimeController getInstance()
	{
		return _instance;
	}

	private GameTimeController()
	{
		_gameStartTime = System.currentTimeMillis() - 3600000; // offset so that the server starts a day begin
		_gameTicks = 3600000 / MILLIS_IN_TICK; // offset so that the server starts a day begin

		_timer = new TimerThread();
		_timer.start();

		ThreadPoolManager.getInstance().scheduleGeneralAtFixedRate(new TimerWatcher(), 0, 1000);
		ThreadPoolManager.getInstance().scheduleGeneralAtFixedRate(new BroadcastSunState(), 120000, 600000);
	}

	public boolean isNowNight()
	{
		return _isNight;
	}

	public int getGameTime()
	{
		return _gameTicks / (TICKS_PER_SECOND * 10);
	}

	public static int getGameTicks()
	{
		return _gameTicks;
	}

	private void moveObjects()
	{
		Iterator<L2WorldRegion> it = L2World.getActiveRegions().iterator();
		while(it.hasNext())
		{
			L2WorldRegion region = it.next();
			if(region != null && region.isActive() && region.getObjectsSize() > 0)
				for(L2Character cha : region.getCharactersList(-1))
					if(cha != null)
					{
						cha.updateTerritories();
						if(!cha.isMoving || !cha.hasAI() || cha.getAI().isGlobalAI())
							continue;
						cha.updatePosition(_gameTicks);
					}
		}

		for(L2Character cha : L2World.getCustomMoveObjects())
		{
			if(cha == null || !cha.isMoving || !cha.hasAI() || !cha.getAI().isGlobalAI())
				continue;
			cha.updatePosition(_gameTicks);
		}
	}

	public void stopTimer()
	{
		_timer._stop = true;
	}

	class TimerThread extends Thread
	{
		protected boolean _stop;
		protected Exception _error;

		public TimerThread()
		{
			super("TimerThread");
			setDaemon(true);
			setPriority(MAX_PRIORITY);
			_stop = false;
			_error = null;
		}

		@Override
		public void run()
		{
			try
			{
				while(!_stop)
				{
					long currentTime = System.currentTimeMillis();
					_startMoveTime = currentTime;
					_gameTicks = (int) ((currentTime - _gameStartTime) / MILLIS_IN_TICK);

					moveObjects();

					currentTime = System.currentTimeMillis();

					// Пересчитываем _gameTicks еще раз, для большей точности, т.к. moveObjects() мог занять много времени
					_gameTicks = (int) ((currentTime - _gameStartTime) / MILLIS_IN_TICK);

					// Интервалы между вызовами минимум Config.MOVE_DELAY мсек
					long sleepTime = Config.MOVE_DELAY - (currentTime - _startMoveTime);
					if(sleepTime > 0)
						sleep(sleepTime);
				}
			}
			catch(Exception e)
			{
				_error = e;
			}

			_log.info("TimerThread was canceled");
		}
	}

	class TimerWatcher implements Runnable
	{
		public void run()
		{
			if(!_timer.isAlive())
			{
				String time = (new SimpleDateFormat("HH:mm:ss")).format(new Date());
				_log.warning(time + " TimerThread stop with following error. restart it.");
				if(_timer._error != null)
					_timer._error.printStackTrace();

				_timer = new TimerThread();
				_timer.start();
			}
		}
	}

	public class BroadcastSunState implements Runnable
	{
		public void run()
		{
			int h = getGameTime() / 60 % 24; // Time in hour
			boolean tempIsNight = h < Config.DAY_STATUS_SUN_RISE_AT || h >= Config.DAY_STATUS_SUN_SET_AT;

			if(tempIsNight != isNowNight()) // If diff day/night state
				_isNight = tempIsNight; // Set current day/night varible to value of temp varible
			else
				return; // Do nothing if same state

			DayNightSpawnManager.getInstance().notifyChangeMode();

			getListenerEngine().firePropertyChanged(PropertyCollection.GameTimeControllerDayNightChange, getInstance(), !_isNight, _isNight);

			for(L2Player player : L2World.getAllPlayers())
				player.checkDayNightMessages();

			if(!Config.DAY_STATUS_FORCE_CLIENT_UPDATE) // Do not force client to update their day/night status
				return;

			if(SevenSigns.getInstance().isSealValidationPeriod()) // Do nothing if in Seal validation period
				return;

			// Force client day/night cycle to sync everyone's day/night state
			L2GameServerPacket packet;
			if(isNowNight()) // If is now night
			{
				packet = Msg.SunSet; // Set client day/night state to night
				if(Config.DEBUG)
					_log.fine("SunSet");
			}
			else
			{
				packet = Msg.SunRise; // Set client day/night state to day
				if(Config.DEBUG)
					_log.fine("SunRise");
			}

			for(L2Player player : L2World.getAllPlayers())
				player.sendPacket(packet);
		}
	}

	public ListenerEngine<GameTimeController> getListenerEngine()
	{
		return listenerEngine;
	}
}