package l2r.gameserver;

import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Logger;

import l2r.Config;
import l2r.extensions.network.ReceivablePacket;
import l2r.gameserver.network.L2GameClient;

/**
 * <p>This class is made to handle all the ThreadPools used in l2r.</p>
 * <p>Scheduled Tasks can either be sent to a {@link #_generalScheduledThreadPool "general"} or {@link #_effectsScheduledThreadPool "effects"} {@link ScheduledThreadPoolExecutor ScheduledThreadPool}:
 * The "effects" one is used for every effects (skills, hp/mp regen ...) while the "general" one is used for
 * everything else that needs to be scheduled.<br>
 * There also is an {@link #_aiScheduledThreadPool "ai"} {@link ScheduledThreadPoolExecutor ScheduledThreadPool} used for AI Tasks.</p>
 * <p>Tasks can be sent to {@link ScheduledThreadPoolExecutor ScheduledThreadPool} either with:
 * <ul>
 * <li>{@link #scheduleEffect(Runnable,long)} : for effects Tasks that needs to be executed only once.</li>
 * <li>{@link #scheduleGeneral(Runnable,long)} : for scheduled Tasks that needs to be executed once.</li>
 * <li>{@link #scheduleAi(Runnable,long)} : for AI Tasks that needs to be executed once</li>
 * </ul>
 * or
 * <ul>
 * <li>{@link #scheduleEffectAtFixedRate(Runnable,long,long)(Runnable, long)} : for effects Tasks that needs to be executed periodicaly.</li>
 * <li>{@link #scheduleGeneralAtFixedRate(Runnable,long,long)(Runnable, long)} : for scheduled Tasks that needs to be executed periodicaly.</li>
 * <li>{@link #scheduleAiAtFixedRate(Runnable,long,long)(Runnable, long)} : for AI Tasks that needs to be executed periodicaly</li>
 * </ul></p>
 *
 * <p>For all Tasks that should be executed with no delay asynchronously in a ThreadPool there also are usual {@link ThreadPoolExecutor ThreadPools}
 * that can grow/shrink according to their load.:
 * <ul>
 * <li>{@link #_generalPacketsThreadPool GeneralPackets} where most packets handler are executed.</li>
 * <li>{@link #_urgentPacketsThreadPool UrgentPackets} where all the urgent packets are executed.</li>
 * <li>There will be an AI ThreadPool where AI events should be executed</li>
 * <li>A general ThreadPool where everything else that needs to run asynchronously with no delay should be executed ({@link l2r.gameserver.model.actor.knownlist KnownList} updates, SQL updates/inserts...)?</li>
 * </ul>
 * </p>
 * @author -Wooden-
 *
 */
@SuppressWarnings({"unchecked"})
public class ThreadPoolManager
{
	private static final Logger _log = Logger.getLogger(ThreadPoolManager.class.getName());

	private static ThreadPoolManager _instance;

	private ScheduledThreadPoolExecutor _effectsScheduledThreadPool;
	private ScheduledThreadPoolExecutor _generalScheduledThreadPool;

	private ThreadPoolExecutor _generalPacketsThreadPool;
	private ThreadPoolExecutor _ioPacketsThreadPool;

	private ThreadPoolExecutor _LsGsExecutor;

	// temp
	private ScheduledThreadPoolExecutor _aiScheduledThreadPool;

	private boolean _shutdown;

	public static ThreadPoolManager getInstance()
	{
		if(_instance == null)
			_instance = new ThreadPoolManager();
		return _instance;
	}

	private ThreadPoolManager()
	{
		_effectsScheduledThreadPool = new ScheduledThreadPoolExecutor(Config.THREAD_P_EFFECTS, new PriorityThreadFactory("EffectsSTPool", Thread.MIN_PRIORITY));
		_generalScheduledThreadPool = new ScheduledThreadPoolExecutor(Config.THREAD_P_GENERAL, new PriorityThreadFactory("GerenalSTPool", Thread.NORM_PRIORITY));
		_aiScheduledThreadPool = new ScheduledThreadPoolExecutor(Config.AI_MAX_THREAD, new PriorityThreadFactory("AISTPool", Thread.NORM_PRIORITY));

		_ioPacketsThreadPool = new ThreadPoolExecutor(Config.URGENT_PACKET_THREAD_CORE_SIZE, Integer.MAX_VALUE, 5L, TimeUnit.SECONDS, new LinkedBlockingQueue<Runnable>(), new PriorityThreadFactory("High Packet Pool", Thread.NORM_PRIORITY + 3));
		_generalPacketsThreadPool = new ThreadPoolExecutor(Config.GENERAL_PACKET_THREAD_CORE_SIZE, Config.GENERAL_PACKET_THREAD_CORE_SIZE + 2, 15L, TimeUnit.SECONDS, new LinkedBlockingQueue<Runnable>(), new PriorityThreadFactory("Normal Packet Pool", Thread.NORM_PRIORITY + 1));

		_LsGsExecutor = new ThreadPoolExecutor(1, 6, 5L, TimeUnit.MILLISECONDS, new LinkedBlockingQueue<Runnable>(), new PriorityThreadFactory("LS/GS Communications", Thread.MAX_PRIORITY - 2));
	}

	public <T extends Runnable> ScheduledFuture<T> scheduleEffect(T r, long delay)
	{
		try
		{
			if(delay < 0)
				delay = 0;
			return (ScheduledFuture<T>) _effectsScheduledThreadPool.schedule(r, delay, TimeUnit.MILLISECONDS);
		}
		catch(RejectedExecutionException e)
		{
			if(!isShutdown())
			{
				_log.warning("ThreadPoolManager[99]: Failed executing new task!");
				Thread.dumpStack();
			}
			return null; /* shutdown, ignore */
		}
	}

	public <T extends Runnable> ScheduledFuture<T> scheduleEffectAtFixedRate(T r, long initial, long delay)
	{
		try
		{
			if(delay < 0)
				delay = 0;
			if(initial < 0)
				initial = 0;
			return (ScheduledFuture<T>) _effectsScheduledThreadPool.scheduleAtFixedRate(r, initial, delay, TimeUnit.MILLISECONDS);
		}
		catch(RejectedExecutionException e)
		{
			if(!isShutdown())
			{
				_log.warning("ThreadPoolManager[120]: Failed executing new task!");
				Thread.dumpStack();
			}
			return null; /* shutdown, ignore */
		}
	}

	public <T extends Runnable> ScheduledFuture<T> scheduleGeneral(T r, long delay)
	{
		try
		{
			if(delay < 0)
				delay = 0;
			return (ScheduledFuture<T>) _generalScheduledThreadPool.schedule(r, delay, TimeUnit.MILLISECONDS);
		}
		catch(RejectedExecutionException e)
		{
			if(!isShutdown())
			{
				_log.warning("ThreadPoolManager[139]: Failed executing new task!");
				Thread.dumpStack();
			}
			return null; /* shutdown, ignore */
		}
	}

	public <T extends Runnable> ScheduledFuture<T> scheduleGeneralAtFixedRate(T r, long initial, long delay)
	{
		try
		{
			if(delay <= 0)
				delay = 1;
			return (ScheduledFuture<T>) _generalScheduledThreadPool.scheduleAtFixedRate(r, initial, delay, TimeUnit.MILLISECONDS);
		}
		catch(RejectedExecutionException e)
		{
			if(!isShutdown())
			{
				_log.warning("ThreadPoolManager[158]: Failed executing new task!");
				Thread.dumpStack();
			}
			return null; /* shutdown, ignore */
		}
	}

	public <T extends Runnable> ScheduledFuture<T> scheduleAi(T r, long delay, boolean isPlayer)
	{
		try
		{
			if(delay < 0)
				delay = 0;
			if(isPlayer)
				return (ScheduledFuture<T>) _aiScheduledThreadPool.schedule(r, delay, TimeUnit.MILLISECONDS);
			return (ScheduledFuture<T>) _aiScheduledThreadPool.schedule(r, delay, TimeUnit.MILLISECONDS);
		}
		catch(RejectedExecutionException e)
		{
			if(!isShutdown())
			{
				_log.warning("ThreadPoolManager[177]: Failed executing new task!");
				Thread.dumpStack();
			}
			return null; /* shutdown, ignore */
		}
	}

	public <T extends Runnable> ScheduledFuture<T> scheduleAiAtFixedRate(T r, long initial, long delay)
	{
		try
		{
			if(delay < 0)
				delay = 0;
			if(initial < 0)
				initial = 0;
			return (ScheduledFuture<T>) _aiScheduledThreadPool.scheduleAtFixedRate(r, initial, delay, TimeUnit.MILLISECONDS);
		}
		catch(RejectedExecutionException e)
		{
			if(!isShutdown())
			{
				_log.warning("ThreadPoolManager[198]: Failed executing new task!");
				Thread.dumpStack();
			}
			return null; /* shutdown, ignore */
		}
	}

	public void executePacket(ReceivablePacket<L2GameClient> pkt)
	{
		_generalPacketsThreadPool.execute(pkt);
	}

	public void executeIOPacket(ReceivablePacket<L2GameClient> pkt)
	{
		_ioPacketsThreadPool.execute(pkt);
	}

	public void executeLSGSPacket(Runnable r)
	{
		_LsGsExecutor.execute(r);
	}

	public String[] getStats()
	{
		return new String[] {
				"Scheduled Thread Pools:",
				" + Effects:",
				" |- ActiveThreads:   " + _effectsScheduledThreadPool.getActiveCount(),
				" |- getCorePoolSize: " + _effectsScheduledThreadPool.getCorePoolSize(),
				" |- PoolSize:        " + _effectsScheduledThreadPool.getPoolSize(),
				" |- MaximumPoolSize: " + _effectsScheduledThreadPool.getMaximumPoolSize(),
				" |- CompletedTasks:  " + _effectsScheduledThreadPool.getCompletedTaskCount(),
				" |- ScheduledTasks:  " + (_effectsScheduledThreadPool.getTaskCount() - _effectsScheduledThreadPool.getCompletedTaskCount()),
				" |- Shutdown:  " + _effectsScheduledThreadPool.isShutdown(),
				" | -------",
				" + General:",
				" |- ActiveThreads:   " + _generalScheduledThreadPool.getActiveCount(),
				" |- getCorePoolSize: " + _generalScheduledThreadPool.getCorePoolSize(),
				" |- PoolSize:        " + _generalScheduledThreadPool.getPoolSize(),
				" |- MaximumPoolSize: " + _generalScheduledThreadPool.getMaximumPoolSize(),
				" |- CompletedTasks:  " + _generalScheduledThreadPool.getCompletedTaskCount(),
				" |- ScheduledTasks:  " + (_generalScheduledThreadPool.getTaskCount() - _generalScheduledThreadPool.getCompletedTaskCount()),
				" |- Shutdown:  " + _generalScheduledThreadPool.isShutdown(),
				" | -------",
				" + AI:",
				" |- ActiveThreads:   " + _aiScheduledThreadPool.getActiveCount(),
				" |- getCorePoolSize: " + _aiScheduledThreadPool.getCorePoolSize(),
				" |- PoolSize:        " + _aiScheduledThreadPool.getPoolSize(),
				" |- MaximumPoolSize: " + _aiScheduledThreadPool.getMaximumPoolSize(),
				" |- CompletedTasks:  " + _aiScheduledThreadPool.getCompletedTaskCount(),
				" |- ScheduledTasks:  " + (_aiScheduledThreadPool.getTaskCount() - _aiScheduledThreadPool.getCompletedTaskCount()),
				" |- Shutdown:  " + _aiScheduledThreadPool.isShutdown(),
				" | -------",
				"Thread Pools:",
				" + Packets:",
				" |- ActiveThreads:   " + _generalPacketsThreadPool.getActiveCount(),
				" |- getCorePoolSize: " + _generalPacketsThreadPool.getCorePoolSize(),
				" |- MaximumPoolSize: " + _generalPacketsThreadPool.getMaximumPoolSize(),
				" |- LargestPoolSize: " + _generalPacketsThreadPool.getLargestPoolSize(),
				" |- PoolSize:        " + _generalPacketsThreadPool.getPoolSize(),
				" |- CompletedTasks:  " + _generalPacketsThreadPool.getCompletedTaskCount(),
				" |- QueuedTasks:     " + _generalPacketsThreadPool.getQueue().size(),
				" |- Shutdown:  " + _generalPacketsThreadPool.isShutdown(),
				" | -------",
				" + Urgent Packets:",
				" |- ActiveThreads:   " + _ioPacketsThreadPool.getActiveCount(),
				" |- getCorePoolSize: " + _ioPacketsThreadPool.getCorePoolSize(),
				" |- MaximumPoolSize: " + _ioPacketsThreadPool.getMaximumPoolSize(),
				" |- LargestPoolSize: " + _ioPacketsThreadPool.getLargestPoolSize(),
				" |- PoolSize:        " + _ioPacketsThreadPool.getPoolSize(),
				" |- CompletedTasks:  " + _ioPacketsThreadPool.getCompletedTaskCount(),
				" |- QueuedTasks:     " + _ioPacketsThreadPool.getQueue().size(),
				" |- Shutdown:  " + _ioPacketsThreadPool.isShutdown(),
				" | -------",
				" + LS/GS Packets:",
				" |- ActiveThreads:   " + _LsGsExecutor.getActiveCount(),
				" |- getCorePoolSize: " + _LsGsExecutor.getCorePoolSize(),
				" |- MaximumPoolSize: " + _LsGsExecutor.getMaximumPoolSize(),
				" |- LargestPoolSize: " + _LsGsExecutor.getLargestPoolSize(),
				" |- PoolSize:        " + _LsGsExecutor.getPoolSize(),
				" |- CompletedTasks:  " + _LsGsExecutor.getCompletedTaskCount(),
				" |- QueuedTasks:     " + _LsGsExecutor.getQueue().size(),
				" |- Shutdown:  " + _LsGsExecutor.isShutdown(),
				" | -------", };
	}

	public static class PriorityThreadFactory implements ThreadFactory
	{
		private int _prio;
		private String _name;
		private AtomicInteger _threadNumber = new AtomicInteger(1);
		private ThreadGroup _group;

		public PriorityThreadFactory(String name, int prio)
		{
			_prio = prio;
			_name = name;
			_group = new ThreadGroup(_name);
		}

		@Override
		public Thread newThread(Runnable r)
		{
			Thread t = new Thread(_group, r);
			t.setName(_name + "-" + _threadNumber.getAndIncrement());
			t.setPriority(_prio);
			return t;
		}

		public ThreadGroup getGroup()
		{
			return _group;
		}
	}

	/**
	 *
	 */
	public void shutdown()
	{
		_shutdown = true;
		try
		{
			_effectsScheduledThreadPool.awaitTermination(1, TimeUnit.SECONDS);
			_generalScheduledThreadPool.awaitTermination(1, TimeUnit.SECONDS);
			_generalPacketsThreadPool.awaitTermination(1, TimeUnit.SECONDS);
			_ioPacketsThreadPool.awaitTermination(1, TimeUnit.SECONDS);
			_aiScheduledThreadPool.awaitTermination(1, TimeUnit.SECONDS);
			_LsGsExecutor.awaitTermination(1, TimeUnit.SECONDS);
			_effectsScheduledThreadPool.shutdown();
			_generalScheduledThreadPool.shutdown();
			_generalPacketsThreadPool.shutdown();
			_ioPacketsThreadPool.shutdown();
			_aiScheduledThreadPool.shutdown();
			_LsGsExecutor.shutdown();
			System.out.println("All ThreadPools are now stoped.");

		}
		catch(InterruptedException e)
		{
			e.printStackTrace();
		}
	}

	/**
	 * Tries to remove from the work queue all {@link java.util.concurrent.Future}
	 * tasks that have been cancelled. This method can be useful as a
	 * storage reclamation operation, that has no other impact on
	 * functionality. Cancelled tasks are never executed, but may
	 * accumulate in work queues until worker threads can actively
	 * remove them. Invoking this method instead tries to remove them now.
	 * However, this method may fail to remove tasks in
	 * the presence of interference by other threads.
	 */
	public void purge()
	{
		_effectsScheduledThreadPool.purge();
		_generalScheduledThreadPool.purge();
		_aiScheduledThreadPool.purge();
		_ioPacketsThreadPool.purge();
		_generalPacketsThreadPool.purge();
		_LsGsExecutor.purge();
	}

	public String getPacketStats()
	{
		return getThreadPoolStats(_generalPacketsThreadPool, "general packets");
	}

	public String getGeneralPoolStats()
	{
		return getThreadPoolStats(_generalScheduledThreadPool, "general");
	}

	public String getAIPoolStats()
	{
		return getThreadPoolStats(_aiScheduledThreadPool, "ai");
	}

	private String getThreadPoolStats(ThreadPoolExecutor pool, String poolname)
	{
		ThreadFactory tf = pool.getThreadFactory();
		if(!(tf instanceof PriorityThreadFactory))
			return "This should not be seen, pool " + poolname;

		StringBuilder res = new StringBuilder();
		PriorityThreadFactory ptf = (PriorityThreadFactory) tf;
		int count = ptf.getGroup().activeCount();
		Thread[] threads = new Thread[count + 2];
		ptf.getGroup().enumerate(threads);

		res.append("\nThread Pool: ").append(poolname);
		res.append("\nTasks in the queue: ").append(pool.getQueue().size());
		res.append("\nThreads stack trace:");
		res.append("\nThere should be ").append(count).append(" threads\n");

		for(Thread t : threads)
		{
			if(t == null)
				continue;
			res.append("\n").append(t.getName());
			for(StackTraceElement ste : t.getStackTrace())
				res.append("\n").append(ste.toString());
		}

		return res.toString();
	}

	public boolean isShutdown()
	{
		return _shutdown;
	}
}
