package l2r.gameserver.ai;

import static l2r.gameserver.ai.CtrlIntention.AI_INTENTION_ACTIVE;
import static l2r.gameserver.ai.CtrlIntention.AI_INTENTION_ATTACK;
import static l2r.gameserver.ai.CtrlIntention.AI_INTENTION_IDLE;

import java.util.ArrayList;
import java.util.Vector;
import java.util.concurrent.Future;
import java.util.concurrent.ScheduledFuture;
import java.util.logging.Logger;

import l2r.Config;
import l2r.gameserver.GameTimeController;
import l2r.gameserver.ThreadPoolManager;
import l2r.gameserver.geodata.network.geoclient.GeoClient;
import l2r.gameserver.model.*;
import l2r.gameserver.model.L2Character.HateInfo;
import l2r.gameserver.model.L2Skill.SkillType;
import l2r.gameserver.model.instances.*;
import l2r.gameserver.model.quest.QuestState;
import l2r.gameserver.serverpackets.MagicSkillUse;
import l2r.gameserver.tables.TerritoryTable;
import l2r.gameserver.templates.StatsSet;
import l2r.util.Location;
import l2r.util.Rnd;

public class DefaultAI extends L2CharacterAI implements Runnable
{
	protected StatsSet this_params = null;

	public static enum TaskType
	{
		MOVE,
		ATTACK,
		CAST,
		BUFF
	}

	public static class Task
	{
		public TaskType type;
		public L2Skill skill;
		public L2Character target;
		public Location loc;
	}

	public class Teleport implements Runnable
	{
		Location _destination;

		public Teleport(Location destination)
		{
			_destination = destination;
		}

		public void run()
		{
			_thisActor.teleToLocation(_destination);
		}
	}

	public class RunningTask implements Runnable
	{
		public void run()
		{
			_thisActor.setRunning();
			_runningTask = null;
		}
	}

	protected static Logger _log = Logger.getLogger(DefaultAI.class.getName());

	protected int MAX_ATTACK_TIMEOUT = 40000 / GameTimeController.MILLIS_IN_TICK; // int millis->ticks
	protected int TELEPORT_TIMEOUT = 10000 / GameTimeController.MILLIS_IN_TICK; // int millis->ticks
	protected int MAX_PATHFIND_FAILS = 5;
	protected int AI_TASK_DELAY = 1000;
	protected int MAX_PURSUE_RANGE = Config.MAX_PURSUE_RANGE;

	/** Current L2NpcInstance */
	protected L2NpcInstance _thisActor;

	/** The L2NpcInstance AI task executed every 1s (call onEvtThink method)*/
	protected Future _aiTask;

	protected ScheduledFuture _runningTask;

	/** The flag used to indicate that a thinking action is in progress */
	private boolean _thinking = false;

	/** The L2NpcInstance aggro counter */
	protected int _globalAggro;

	protected long _randomAnimationEnd;
	protected boolean _isMobile = true;
	protected int _pathfind_fails;

	/** Список заданий */
	protected Vector<Task> _task_list = new Vector<Task>();

	/** Показывает, есть ли задания */
	protected boolean _def_think = false;

	public L2Skill[] _heal;
	public L2Skill[] _buff;
	public L2Skill[] _teleport;
	public L2Skill[] _mute;
	public L2Skill[] _paralyze;
	public L2Skill[] _root;
	public L2Skill[] _sleep;
	public L2Skill[] _stun;
	public L2Skill[] _cancel;
	public L2Skill[] _debuff;
	public L2Skill[] _dot;
	public L2Skill[] _poison;
	public L2Skill[] _bleed;
	public L2Skill[] _mdot;
	public L2Skill[] _manadam;
	public L2Skill[] _mdam;
	public L2Skill[] _pdam;
	public L2Skill[] _drain;
	public L2Skill[] _drainsoul;
	public L2Skill[] _healpercent;
	public L2Skill[] _hot;

	public L2Skill[] _dot_skills;
	public L2Skill[] _buff_skills;
	public L2Skill[] _dam_skills;
	public L2Skill[] _debuff_skills;

	/**
	 * Сложение массивов скиллов
	 */
	public static L2Skill[] msum(L2Skill[] array1, L2Skill[] array2)
	{
		L2Skill[] result = new L2Skill[array1.length + array2.length];

		System.arraycopy(array1, 0, result, 0, array1.length);

		System.arraycopy(array2, 0, result, array1.length, array2.length);

		return result;
	}

	public DefaultAI(L2Character actor)
	{
		super(actor);

		_globalAggro = -10; // 10 seconds timeout of ATTACK after respawn
		_thisActor = (L2NpcInstance) _actor;
		_thisActor.setAttackTimeout(Integer.MAX_VALUE);

		_heal = _thisActor.getTemplate().getSkillsByType(SkillType.HEAL);
		_buff = _thisActor.getTemplate().getSkillsByType(SkillType.BUFF);
		_teleport = _thisActor.getTemplate().getSkillsByType(SkillType.TELEPORT_NPC);
		_mute = _thisActor.getTemplate().getSkillsByType(SkillType.MUTE);
		_paralyze = _thisActor.getTemplate().getSkillsByType(SkillType.PARALYZE);
		_root = _thisActor.getTemplate().getSkillsByType(SkillType.ROOT);
		_sleep = _thisActor.getTemplate().getSkillsByType(SkillType.SLEEP);
		_stun = _thisActor.getTemplate().getSkillsByType(SkillType.STUN);
		_cancel = _thisActor.getTemplate().getSkillsByType(SkillType.CANCEL);
		_debuff = _thisActor.getTemplate().getSkillsByType(SkillType.DEBUFF);
		_dot = _thisActor.getTemplate().getSkillsByType(SkillType.DOT);
		_poison = _thisActor.getTemplate().getSkillsByType(SkillType.POISON);
		_bleed = _thisActor.getTemplate().getSkillsByType(SkillType.BLEED);
		_mdot = _thisActor.getTemplate().getSkillsByType(SkillType.MDOT);
		_manadam = _thisActor.getTemplate().getSkillsByType(SkillType.MANADAM);
		_mdam = _thisActor.getTemplate().getSkillsByType(SkillType.MDAM);
		_pdam = _thisActor.getTemplate().getSkillsByType(SkillType.PDAM);
		_drain = _thisActor.getTemplate().getSkillsByType(SkillType.DRAIN);
		_drainsoul = _thisActor.getTemplate().getSkillsByType(SkillType.DRAIN_SOUL);
		_healpercent = _thisActor.getTemplate().getSkillsByType(SkillType.HEAL_PERCENT);
		_hot = _thisActor.getTemplate().getSkillsByType(SkillType.HOT);

		_dot_skills = msum(_dot, _poison);
		_dot_skills = msum(_dot_skills, _bleed);
		_dot_skills = msum(_dot_skills, _mdot);

		_buff_skills = msum(_heal, _buff);
		_buff_skills = msum(_buff_skills, _healpercent);

		_dam_skills = msum(_manadam, _mdam);
		_dam_skills = msum(_dam_skills, _pdam);
		_dam_skills = msum(_dam_skills, _drain);
		_dam_skills = msum(_dam_skills, _drainsoul);

		_debuff_skills = msum(_teleport, _mute);
		_debuff_skills = msum(_debuff_skills, _paralyze);
		_debuff_skills = msum(_debuff_skills, _root);
		_debuff_skills = msum(_debuff_skills, _sleep);
		_debuff_skills = msum(_debuff_skills, _cancel);
		_debuff_skills = msum(_debuff_skills, _debuff);
	}

	public void run()
	{
		onEvtThink();
	}

	@Override
	public void startAITask()
	{
		if(_aiTask == null)
		{
			_aiTask = ThreadPoolManager.getInstance().scheduleAiAtFixedRate(this, Rnd.get(AI_TASK_DELAY), AI_TASK_DELAY);
			setIntention(CtrlIntention.AI_INTENTION_ACTIVE);
		}
	}

	@Override
	public void stopAITask()
	{
		if(_aiTask != null)
		{
			setIntention(CtrlIntention.AI_INTENTION_IDLE);
			_aiTask.cancel(false);
			_aiTask = null;
		}
	}

	/**
	 * Определяет, может ли этот тип АИ видеть персонажей в режиме Silent Move.
	 * @param target L2Playable цель
	 * @return true если цель не видна в режиме Silent Move
	 */
	protected boolean isSilentMoveNotVisible(L2Playable target)
	{
		return target.isSilentMoving();
	}

	@Override
	public void checkAggression(L2Character target)
	{
		if(_intention != AI_INTENTION_ACTIVE || _globalAggro < 0)
			return;
		if(_thisActor instanceof L2FestivalMonsterInstance && !target.getOwner().isFestivalParticipant())
			return;
		if(target.getHateList().get(_thisActor) == null && (!_thisActor.isAggressive() || !_thisActor.isInRange(target, _thisActor.getAggroRange())))
			return;
		if(Math.abs(target.getZ() - _actor.getZ()) > 400)
			return;
		if(target instanceof L2Playable && isSilentMoveNotVisible((L2Playable) target))
			return;
		if(_thisActor.getFactionId().equalsIgnoreCase("varka_silenos_clan") && target.getOwner().getVarka() > 0)
			return;
		if(_thisActor.getFactionId().equalsIgnoreCase("ketra_orc_clan") && target.getOwner().getKetra() > 0)
			return;
		if(!GeoClient.getInstance().canSeeTarget(_actor, target))
			return;
		if(target.isInZonePeace())
			return;
		if(target.isPlayer)
			if(((L2Player) target).getPlayerAccess().IsGM && ((L2Player) target).isInvisible())
				return;
		if(!_thisActor.canAttackCharacter(target))
			return;
		if(_actor.isRaid || _actor.isMinion && ((L2MinionInstance) _actor).getLeader().isRaid)
			for(L2Effect e : target.getAllEffects())
			{
				L2Skill skill = e.getSkill();
				if(skill.getId() == 4215 || skill.getId() == 4515)
					return;
			}

		if((target.isSummon || target.isPet) && target.getOwner() != null)
			target.getOwner().addDamageHate(_thisActor, 0, 1);

		target.addDamageHate(_thisActor, 0, 2);

		startRunningTask(2000);
		setIntention(CtrlIntention.AI_INTENTION_ATTACK, target);
	}

	protected boolean randomAnimation()
	{
		if(!_actor.isMoving)
			// Анимации мобов
			if(_actor.hasRandomAnimation() && Rnd.chance(Config.RND_ANIMATION_RATE))
			{
				_randomAnimationEnd = System.currentTimeMillis();
				_thisActor.onRandomAnimation();
				return true;
			}
		return false;
	}

	protected boolean randomWalk()
	{
		if(!_actor.isMoving)
			if(maybeMoveToHome())
				return true;
		return false;
	}

	protected boolean thinkActive()
	{
		if(_thisActor.isDead())
			return true;

		// Update every 1s the _globalAggro counter to come close to 0
		if(_globalAggro < 0)
			_globalAggro++;
		else if(_globalAggro > 0)
			_globalAggro--;

		if(_randomAnimationEnd > System.currentTimeMillis())
			return true;

		if(_def_think)
		{
			doTask();
			return true;
		}

		// If this is a festival monster or chest, then it remains in the same location
		if(_thisActor instanceof L2FestivalMonsterInstance || _thisActor instanceof L2ChestInstance || _thisActor instanceof L2TamedBeastInstance)
			return false;

		if(_thisActor instanceof L2MinionInstance)
		{
			L2MonsterInstance leader = ((L2MinionInstance) _thisActor).getLeader();
			if(leader == null)
				return false;
			double distance = _thisActor.getDistance(leader.getX(), leader.getY());
			Location pos = leader.getMinionPosition();
			if(distance > 1000)
				_thisActor.teleToLocation(pos);
			else if(distance > 200)
				_thisActor.moveToLocation(pos, 0, true);
			return false;
		}

		if(randomAnimation())
			return true;

		if(randomWalk())
			return true;

		return false;
	}

	@Override
	protected void onIntentionActive()
	{
		_thisActor.setAttackTimeout(Integer.MAX_VALUE);
		super.onIntentionActive();
	}

	protected boolean checkTarget(L2Character target, int range)
	{
		if(target == null || target == _thisActor || target.isAlikeDead() || !_thisActor.isInRange(target, range))
			return true;

		if(_actor.isConfused() || target instanceof L2DecoyInstance)
			return false;
		if(target.getHateList().get(_thisActor) == null || target.getHateList().get(_thisActor).hate == 0)
			return true;

		if(_actor.isRaid || _actor.isMinion && ((L2MinionInstance) _actor).getLeader().isRaid)
			for(L2Effect e : target.getAllEffects())
			{
				L2Skill skill = e.getSkill();
				if(skill.getId() == 4215 || skill.getId() == 4515)
					return true;
			}
		if(_thisActor.getAttackTimeout() < GameTimeController.getGameTicks())
		{
			if(_thisActor.isRunning() && _thisActor.getAggroListSize() == 1)
			{
				_thisActor.setWalking();
				_thisActor.setAttackTimeout(MAX_ATTACK_TIMEOUT / 4 + GameTimeController.getGameTicks());
				return false;
			}
			target.removeFromHatelist(_thisActor);
			return true;
		}

		return false;
	}

	protected void thinkAttack()
	{
		if(!_actor.HaveIndividualReturnTask)//But not uses for Antharas
			if(!_actor.isInRange(_actor.getLoc(), MAX_PURSUE_RANGE))
			{
				clearTasks();
				teleportHome();
				return;
			}
        else if((getAttackTarget() == null || getAttackTarget().isDead()))
			{
				setIntention(AI_INTENTION_ACTIVE);
				onEvtThink();
			}
        else if(_actor.getDistance(getAttackTarget()) > Config.MAX_PURSUE_RANGE)
				{
					setIntention(AI_INTENTION_ACTIVE);
					onEvtThink();
					return;
				}

		if(doTask() && !_actor.isAttackingNow() && !_actor.isCastingNow())
			createNewTask();
	}

	@Override
	protected void onEvtReadyToAct()
	{
		onEvtThink();
	}

	@Override
	protected void onEvtArrivedTarget()
	{
		onEvtThink();
	}

	protected boolean tryMoveToTarget(L2Character target)
	{
		if(!GeoClient.getInstance().canSeeTarget(_actor, target))
		{
			if(!_thisActor.followToCharacter(target, 0, true))
				_pathfind_fails++;
		}
		else if(!_thisActor.followToCharacter(target, _thisActor.getPhysicalAttackRange(), true))
			_pathfind_fails++;

		if(_pathfind_fails >= MAX_PATHFIND_FAILS && GameTimeController.getGameTicks() - (_thisActor.getAttackTimeout() - MAX_ATTACK_TIMEOUT) < TELEPORT_TIMEOUT && _thisActor.isInRange(target, 2000))
		{
			_pathfind_fails = 0;
			HateInfo hate = target.getHateList().get(_thisActor);
			if(hate == null || hate.damage < 100 && hate.hate < 100)
				return false;
			_thisActor.broadcastPacketToOthers(new MagicSkillUse(_thisActor, _thisActor, 2036, 1, 500, 600000));
			ThreadPoolManager.getInstance().scheduleAi(new Teleport(GeoClient.getInstance().moveCheckForAI(target, _thisActor)), 500, false);
			return false;
		}

		return true;
	}
	protected boolean doTask()
	{
		if(_task_list.size() == 0 || _task_list.get(0) == null)
			clearTasks();

		if(!_def_think)
			return true;

		L2Character _temp_attack_target = null;

		Task currentTask = _task_list.get(0);

		switch(currentTask.type)
		{
			// Задание "прибежать в заданные координаты"
			case MOVE:
			{
				if(_actor.isMoving)
					return false;

				if(!_thisActor.isInRange(currentTask.loc, 50))
				{
					if(_actor.isMovementDisabled())
					{
						_def_think = false;
						return true;
					}

					if(!_thisActor.moveToLocation(currentTask.loc, 0, true))
					{
						_thisActor.broadcastPacketToOthers(new MagicSkillUse(_thisActor, _thisActor, 2036, 1, 500, 600000));
						ThreadPoolManager.getInstance().scheduleAi(new Teleport(currentTask.loc), 500, false);
						return false;
					}
				}
				else
				{
					// Следующее задание
					_task_list.remove(currentTask);
					// Если заданий больше нет - определить новое
					if(_task_list.size() == 0)
					{
						_def_think = false;
						return true;
					}
				}
			}
				break;
			// Задание "добежать - ударить"
			case ATTACK:
			{
				_temp_attack_target = currentTask.target;
				if(checkTarget(_temp_attack_target, 2000))
					return true;
				setAttackTarget(_temp_attack_target);
				_attack_target = _temp_attack_target;

				if(_actor.isMoving)
				{
					// Периодически прекращать преследовать и заного оценивать обстановку
					if(Rnd.chance(10))
					{
						_def_think = false;
						return true;
					}
					return false;
				}

				// Преследовать до тех пор пока не приблизиться на расстояние атаки, потом атаковать
				if(_thisActor.getRealDistance(_temp_attack_target) > _thisActor.getPhysicalAttackRange() + 40 || !GeoClient.getInstance().canSeeTarget(_actor, _temp_attack_target))
				{
					if(_thisActor.isMovementDisabled())
					{
						_def_think = false;
						return true;
					}

					if(!_isMobile)
					{
						_def_think = false;
						return true;
					}

					if(!tryMoveToTarget(_temp_attack_target))
						return false;
				}
				else
				{
					_pathfind_fails = 0;

					// Обновление таймаута
					_thisActor.setAttackTimeout(MAX_ATTACK_TIMEOUT + GameTimeController.getGameTicks());

					clientStopMoving();

					// Атаковать
					_thisActor.doAttack(_temp_attack_target);

					// Следующее задание
					_task_list.remove(currentTask);
					// Если заданий больше нет - определить новое
					if(_task_list.size() == 0)
					{
						_def_think = false;
						return true;
					}
				}
			}
				break;
			// Задание "добежать - атаковать скиллом"
			case CAST:
			{
				_temp_attack_target = currentTask.target;

				if(currentTask.skill.getTargetType() == L2Skill.SkillTargetType.TARGET_AURA)
				{
					_pathfind_fails = 0;

					// Обновление таймаута
					_thisActor.setAttackTimeout(MAX_ATTACK_TIMEOUT + GameTimeController.getGameTicks());

					clientStopMoving();

					// Использовать скилл на цель
					_thisActor.doCast(currentTask.skill, _temp_attack_target, !(_temp_attack_target instanceof L2Playable));

					// Следующее задание
					_task_list.remove(currentTask);
					// Если заданий больше нет - определить новое
					if(_task_list.size() == 0)
					{
						_def_think = false;
						return true;
					}

					break;
				}

				if(checkTarget(_temp_attack_target, 3000))
					return true;
				setAttackTarget(_temp_attack_target);

				if(_actor.isMoving)
				{
					// Периодически прекращать преследовать и заного оценивать обстановку
					if(Rnd.chance(10))
					{
						_def_think = false;
						return true;
					}
					return false;
				}

				// Преследовать до тех пор пока не приблизиться на расстояние атаки, потом атаковать
				if(_thisActor.getRealDistance(_temp_attack_target) > currentTask.skill.getCastRange() + 40 || !GeoClient.getInstance().canSeeTarget(_actor, _temp_attack_target))
				{
					if(_thisActor.isMovementDisabled())
					{
						_def_think = false;
						return true;
					}

					if(!_isMobile)
					{
						_def_think = false;
						return true;
					}

					if(!tryMoveToTarget(_temp_attack_target))
						return false;
				}
				else
				{
					_pathfind_fails = 0;

					// Обновление таймаута
					_thisActor.setAttackTimeout(MAX_ATTACK_TIMEOUT + GameTimeController.getGameTicks());

					clientStopMoving();

					// Использовать скилл на цель
					_thisActor.doCast(currentTask.skill, _temp_attack_target, !(_temp_attack_target instanceof L2Playable));

					// Следующее задание
					_task_list.remove(currentTask);
					// Если заданий больше нет - определить новое
					if(_task_list.size() == 0)
					{
						_def_think = false;
						return true;
					}
				}
			}
				break;
			// Задание "добежать - применить скилл"
			case BUFF:
			{
				_temp_attack_target = currentTask.target;

				if(currentTask.skill.getTargetType() == L2Skill.SkillTargetType.TARGET_AURA)
				{
					_pathfind_fails = 0;

					clientStopMoving();

					// Использовать скилл на цель
					_thisActor.doCast(currentTask.skill, _temp_attack_target, !(_temp_attack_target instanceof L2Playable));

					// Следующее задание
					_task_list.remove(currentTask);
					// Если заданий больше нет - определить новое
					if(_task_list.size() == 0)
					{
						_def_think = false;
						return true;
					}

					break;
				}

				if(_temp_attack_target == null || _temp_attack_target.isAlikeDead() || !_thisActor.isInRange(_temp_attack_target, 2000))
					return true;

				if(_actor.isMoving)
					return false;

				// Преследовать до тех пор пока не приблизиться на расстояние атаки, потом атаковать
				if(_thisActor.getRealDistance(_temp_attack_target) > currentTask.skill.getCastRange() + 40 || !GeoClient.getInstance().canSeeTarget(_actor, _temp_attack_target))
				{
					if(_thisActor.isMovementDisabled())
					{
						_def_think = false;
						return true;
					}

					if(!_isMobile)
					{
						_def_think = false;
						return true;
					}

					if(!tryMoveToTarget(_temp_attack_target))
						return false;
				}
				else
				{
					_pathfind_fails = 0;

					clientStopMoving();

					// Использовать скилл на цель
					_thisActor.doCast(currentTask.skill, _temp_attack_target, !(_temp_attack_target instanceof L2Playable));

					// Следующее задание
					_task_list.remove(currentTask);
					// Если заданий больше нет - определить новое
					if(_task_list.size() == 0)
					{
						_def_think = false;
						return true;
					}
				}
			}
				break;
		}

		return false;
	}

	protected boolean createNewTask()
	{
		return false;
	}

	@Override
	protected void onIntentionAttack(L2Character target)
	{
		// Удаляем все задания
		clearTasks();

		_attack_target = target;
		changeIntention(AI_INTENTION_ATTACK, target, null);
		clientStopMoving();
		onEvtThink();
	}

	@Override
	protected void onEvtThink()
	{
		if(_thinking || _thisActor.isActionsDisabled() || _thisActor.isAfraid() || _thisActor.isDead())
			return;

		if(_randomAnimationEnd > System.currentTimeMillis())
			return;

		_thinking = true;
		try
		{
			if(_intention == AI_INTENTION_ACTIVE || _intention == AI_INTENTION_IDLE)
				thinkActive();
			else if(_intention == AI_INTENTION_ATTACK)
				thinkAttack();
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		finally
		{
			_thinking = false;
		}
	}

	@Override
	protected void onEvtDead()
	{
		stopAITask();

		_thisActor.clearAggroList();

		// 10 seconds timeout of ATTACK after respawn
		_globalAggro = -10;

		_thisActor.setAttackTimeout(Integer.MAX_VALUE);

		// Удаляем все задания
		clearTasks();

		super.onEvtDead();
	}

	@Override
	protected void onEvtAttacked(L2Character attacker, int damage)
	{
		if(attacker == null)
			return;

		if(!_thisActor.canAttackCharacter(attacker))
			return;
		L2Player _player = attacker.getPlayer();
		if(_player != null)
		{
			ArrayList<QuestState> quests = _player.getQuestsForAttacks(_thisActor);
			if(quests != null)
				for(QuestState qs : quests)
					qs.getQuest().notifyAttack(_thisActor, qs);

			// 1 хейт добавляется хозяину суммона, чтобы после смерти суммона моб накинулся на хозяина.
			if(damage > 0 && attacker.getOwner() != null && (attacker.isSummon || attacker.isPet))
				attacker.getOwner().addDamageHate(_thisActor, 0, 1);
		}

		_thisActor.callFriends(attacker);
		_thisActor.setAttackTimeout(MAX_ATTACK_TIMEOUT + GameTimeController.getGameTicks());
		_globalAggro = 0;

		attacker.addDamage(_thisActor, damage);

		if(!_actor.isRunning())
			startRunningTask(1000);

		if(_intention != CtrlIntention.AI_INTENTION_ATTACK)
			setIntention(CtrlIntention.AI_INTENTION_ATTACK, attacker);
	}

	@Override
	protected void onEvtAggression(L2Character attacker, int aggro)
	{
		if(attacker == null || attacker.getPlayer() == null)
			return;

		L2Player _player = attacker.getPlayer();
		if(!_thisActor.canAttackCharacter(attacker))
			return;

		_thisActor.setAttackTimeout(MAX_ATTACK_TIMEOUT + GameTimeController.getGameTicks());
		_globalAggro = 0;

		// 1 хейт добавляется хозяину суммона, чтобы после смерти суммона моб накинулся на хозяина.
		if(_player != null && aggro > 0 && attacker.getOwner() != null && (attacker.isSummon || attacker.isPet))
			attacker.getOwner().addDamageHate(_thisActor, 0, 1);

		attacker.addDamageHate(_thisActor, 0, aggro);

		if(!_actor.isRunning())
			startRunningTask(1000);

		if(_intention != CtrlIntention.AI_INTENTION_ATTACK)
		{
			// Показываем анимацию зарядки шотов, если есть таковые.
			switch(_thisActor.getTemplate().shots)
			{
				case SOUL:
					_thisActor.unChargeShots(false);
					break;
				case SPIRIT:
				case BSPIRIT:
					_thisActor.unChargeShots(true);
					break;
				case SOUL_SPIRIT:
				case SOUL_BSPIRIT:
					_thisActor.unChargeShots(false);
					_thisActor.unChargeShots(true);
					break;
			}

			setIntention(CtrlIntention.AI_INTENTION_ATTACK, attacker);
		}
	}

	protected boolean maybeMoveToHome()
	{
		boolean randomWalk = _actor.hasRandomWalk();
		if(_thisActor.getSpawnedLoc() == null)
			return false;

		// Random walk or not?
		if(randomWalk && (!Config.RND_WALK || !Rnd.chance(Config.RND_WALK_RATE)))
			return false;

		if(!randomWalk && _thisActor.isInRangeZ(_thisActor.getSpawnedLoc(), Config.MAX_DRIFT_RANGE))
			return false;

		int x = _thisActor.getSpawnedLoc().getX() + Rnd.get(-Config.MAX_DRIFT_RANGE, Config.MAX_DRIFT_RANGE);
		int y = _thisActor.getSpawnedLoc().getY() + Rnd.get(-Config.MAX_DRIFT_RANGE, Config.MAX_DRIFT_RANGE);
		int z = GeoClient.getInstance().getHeight(x, y, _thisActor.getSpawnedLoc().getZ());

		if(_thisActor.getSpawnedLoc().getZ() - z > 200)
			return false;

		L2Spawn spawn = _thisActor.getSpawn();
		if(spawn != null && spawn.getLocation() != 0 && !TerritoryTable.getInstance().getLocation(spawn.getLocation()).isInside(x, y))
			return false;

		_thisActor.setWalking();
		_thisActor.moveToLocation(x, y, z, 0, true);

		return true;
	}

	public void returnHome()
	{
		_thisActor.clearAggroList();
		setIntention(AI_INTENTION_ACTIVE);

		// Удаляем все задания
		clearTasks();

		// Прибежать в заданную точку и переключиться в состояние AI_INTENTION_ACTIVE
		Task task = new Task();
		task.type = TaskType.MOVE;
		task.loc = _thisActor.getSpawnedLoc();
		_task_list.add(task);
		_def_think = true;
	}

	@Override
	public void teleportHome()
	{
		if(_thisActor.getSpawn() != null && _thisActor.getSpawnedLoc() != null)
		{
			_thisActor.broadcastPacketToOthers(new MagicSkillUse(_thisActor, _thisActor, 2036, 1, 500, 0));
			_thisActor.teleToLocation(_thisActor.getSpawnedLoc().getX(), _thisActor.getSpawnedLoc().getY(), GeoClient.getInstance().getHeight(_thisActor.getSpawnedLoc()));
			setIntention(AI_INTENTION_ACTIVE);
		}
	}

	@Override
	public boolean isActive()
	{
		return _aiTask != null;
	}

	public void clearTasks()
	{
		_def_think = false;
		_task_list.clear();
	}

	/**
	 * Шедюлит мобу переход в режим бега через определенный интервал времени
	 * @param interval
	 */
	public void startRunningTask(int interval)
	{
		if(_runningTask == null && !_thisActor.isRunning())
			_runningTask = ThreadPoolManager.getInstance().scheduleAi(new RunningTask(), interval, false);
	}

	@Override
	public boolean isGlobalAggro()
	{
		return _globalAggro >= 0;
	}

	@Override
	public void setGlobalAggro(int value)
	{
		_globalAggro = value;
	}

	private StatsSet getAIParams()
	{
		if(this_params != null)
			return this_params;
		if(_thisActor.getTemplate().getAIParams() != null)
			return _thisActor.getTemplate().getAIParams();
		this_params = new StatsSet();
		return this_params;
	}

	private StatsSet getThisAIParams()
	{
		if(this_params == null)
			this_params = _thisActor.getTemplate().getAIParams() == null ? new StatsSet() : _thisActor.getTemplate().getAIParams().clone();
		return this_params;
	}

	public boolean getBool(String name)
	{
		return getAIParams().getBool(name);
	}

	public boolean getBool(String name, boolean _defult)
	{
		return getAIParams().getBool(name, _defult);
	}

	public int getInt(String name)
	{
		return getAIParams().getInteger(name);
	}

	public int getInt(String name, int _defult)
	{
		return getAIParams().getInteger(name, _defult);
	}

	public long getLong(String name)
	{
		return getAIParams().getLong(name);
	}

	public long getLong(String name, long _defult)
	{
		return getAIParams().getLong(name, _defult);
	}

	public float getFloat(String name)
	{
		return getAIParams().getFloat(name);
	}

	public float getFloat(String name, float _defult)
	{
		return getAIParams().getFloat(name, _defult);
	}

	public String getString(String name)
	{
		return getAIParams().getString(name);
	}

	public String getString(String name, String _defult)
	{
		return getAIParams().getString(name, _defult);
	}

	public void set(String name, boolean value)
	{
		getThisAIParams().set(name, value);
	}

	public void set(String name, int value)
	{
		getThisAIParams().set(name, value);
	}

	public void set(String name, long value)
	{
		getThisAIParams().set(name, value);
	}

	public void set(String name, double value)
	{
		getThisAIParams().set(name, value);
	}

	public void set(String name, String value)
	{
		getThisAIParams().set(name, value);
	}
}