package l2r.gameserver.geodata.network.geoclient;

import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.Vector;
import java.util.concurrent.ScheduledFuture;
import java.util.logging.Logger;

import l2r.gameserver.ThreadPoolManager;
import l2r.gameserver.geodata.GeoEngine;
import l2r.gameserver.geodata.network.GeoConfig;
import l2r.gameserver.geodata.network.geoserver.GeoServerInterface;
import l2r.gameserver.model.L2Object;
import l2r.gameserver.model.L2Territory;
import l2r.gameserver.model.instances.L2DoorInstance;
import l2r.gameserver.tables.DoorTable;
import l2r.util.Location;
import l2r.util.Rnd;

/**
 * Менеджер управлющий работой геодаты.<BR>
 *
 * @Author: Death
 * @Date: 23/11/2007
 * @Time: 12:13:11
 */
@SuppressWarnings("serial")
public class GeoClient extends UnicastRemoteObject implements GeoClientInterface
{
	public static final Logger log = Logger.getLogger(GeoClient.class.getName());

	public static final int MODE_NO_GEO = 0;
	public static final int MODE_LOCAL_GEO = 1;
	public static final int MODE_RMI_GEO = 2;

	private static GeoClient instance;

	// Не гоняем лишние даные по сети
	private transient GeoEngine geoEngine;
	private transient GeoServerInterface geoServer;
	private transient ScheduledFuture reconnectTask;

	public static GeoClient getInstance()
	{
		if(instance == null)
			try
			{
				GeoConfig.load();
				new GeoClient();
			}
			catch(Exception e)
			{
				e.printStackTrace();
				System.out.println("Can't init geoclient.");
				System.exit(0);
			}
		return instance;
	}

	public GeoClient() throws RemoteException
	{
		instance = this;

		switch(GeoConfig.GEODATA_MODE)
		{
			case MODE_NO_GEO:
				initFake();
				break;
			case MODE_LOCAL_GEO:
				initLocal();
				break;
			case MODE_RMI_GEO:
				initRemote();
				break;
		}
	}

	/**
	 * Фэйковый режим геодаты.<BR>
	 * Геодата не грузится.
	 */
	public void initFake()
	{
		log.info("GeoData: Disabled");
	}

	/**
	 * Логальный режим геодаты.<BR>
	 * Геодата загружается в гейм сервер.
	 */
	public void initLocal()
	{
		log.info("GeoData: Running at combomode");
		GeoEngine.loadGeo();
		geoEngine = new GeoEngine();

		geoEngine.initPathFindBuffers(GeoConfig.PATH_FIND_BUFFERS);

		for(L2DoorInstance door : DoorTable.getInstance().getDoors())
			if(!door.isOpen())
			{
				closeDoor(door.pos);
				door.geoOpen = false;
			}
	}

	/**
	 * Удаленный режим геодаты средствами RMI.<BR>
	 * Геодата крутится на отдельном сервере.<BR><BR>
	 * <b><font color="red">УДАЛЕННЫЙ СЕРВЕР ДОЛЖЕН БЫТЬ ЗАПУЩЕН РАНЬШЕ ГЕЙМА</color></b>
	 */
	public void initRemote()
	{
		try
		{
			Registry r = LocateRegistry.getRegistry(GeoConfig.CLIENT_TARGET_HOST, GeoConfig.PORT);
			geoServer = (GeoServerInterface) r.lookup("geoServer");
			geoServer.addClient(this);
			GeoConfig.GEODATA_MODE = MODE_RMI_GEO;
		}
		catch(Exception e)
		{
			e.printStackTrace();
			log.severe("GeoData: Connection Failed.");
			onConnectionLost();
		}
		log.info("GeoData: Using remote communication");

		for(L2DoorInstance door : DoorTable.getInstance().getDoors())
			if(!door.isOpen())
			{
				closeDoor(door.pos);
				door.geoOpen = false;
			}
	}

	/**
	 * Даный метод вызывается при потере связи гейм сервера с RMI сервером геодаты.
	 */
	public void onConnectionLost()
	{
		if(reconnectTask != null)
			return;

		log.info("GeoData: Connection with server lost, switching to no geodata mode.");

		GeoConfig.GEODATA_MODE = MODE_NO_GEO;
		geoServer = null;

		log.info("GeoData: Starting reconnect task");

		reconnectTask = ThreadPoolManager.getInstance().scheduleGeneral(new Runnable(){
			public void run()
			{
				reconnectTask = null;
				log.info("GeoData: Reconecting to GeoServer...");
				initRemote();
			}
		}, 5000);
	}

	/**
	 * Просто пустой метод, вызывается удаленном сервером как проверка коннекта.
	 *
	 * @throws RemoteException
	 */
	public void testConnection()
	{}

	//////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////

	/**
	 * Далее следуют методы которые используются для эмуляции геодаты при ее отсутсвии,
	 * а также для работы с удаленным сервером геодаты.
	 */

	/**
	 * Даный метод производит поиск пути.
	 *
	 * @param x   коррдината игрока
	 * @param y   коррдината игрока
	 * @param z   коррдината игрока
	 * @param pos координаты цели
	 *
	 * @return обнаруженый путь, либо пустой Vector
	 */
	public Vector<Location> checkMovement(int x, int y, int z, Location pos, L2Object obj)
	{
		switch(GeoConfig.GEODATA_MODE)
		{
			case MODE_LOCAL_GEO:
				return geoEngine.getGeoMove().findPath(x, y, z, pos, obj.isPlayer || obj.isPet || obj.isSummon, false);
			case MODE_RMI_GEO:
				Vector<Location> v = null;
				try
				{
					v = geoServer.checkMovement(this, x, y, z, pos, obj.isPlayer || obj.isPet || obj.isSummon);
				}
				catch(RemoteException e)
				{
					e.printStackTrace();
					onConnectionLost();
				}
				return v != null ? v : new Vector<Location>();
			case MODE_NO_GEO:
			default:
				return new Vector<Location>();
		}
	}

	/**
	 * Проверка на видимость между двумя L2Object.<BR>
	 * Если нет связи с геосервером, то цель видна.<BR>
	 *
	 * @param x  исходная координата
	 * @param y  исходная координата
	 * @param z  исходная координата
	 * @param tx координата цели
	 * @param ty координата цели
	 * @param tz координата цели
	 *
	 * @return видна ли цель
	 */
	public boolean canSeeCoord(int x, int y, int z, int tx, int ty, int tz)
	{
		switch(GeoConfig.GEODATA_MODE)
		{
			case MODE_LOCAL_GEO:
				return geoEngine.canSeeCoord(x, y, z, tx, ty, tz);
			case MODE_RMI_GEO:
				try
				{
					return geoServer.canSeeCoord(this, x, y, z, tx, ty, tz);
				}
				catch(RemoteException e)
				{
					e.printStackTrace();
					onConnectionLost();
					return true;
				}
			case MODE_NO_GEO:
			default:
				return true;
		}
	}

	public boolean canSeeCoord(L2Object actor, int tx, int ty, int tz)
	{
		switch(GeoConfig.GEODATA_MODE)
		{
			case MODE_LOCAL_GEO:
				return geoEngine.canSeeCoord(actor, tx, ty, tz);
			case MODE_RMI_GEO:
				try
				{
					return geoServer.canSeeCoord(this, actor.getX(), actor.getY(), actor.getZ(), tx, ty, tz);
				}
				catch(RemoteException e)
				{
					e.printStackTrace();
					onConnectionLost();
					return true;
				}
			case MODE_NO_GEO:
			default:
				return true;
		}
	}

	/**
	 * Проверка на видимость между двумя L2Object.<BR>
	 * Если один из обьектов null, то цель не видна.<BR>
	 * Если нет связи с геосервером, то цель видна.<BR>
	 *
	 * @param actor  цель от имени которой проверяем
	 * @param target цель до которой проверяем
	 *
	 * @return видима ли цель
	 */

	public boolean canSeeTarget(L2Object actor, L2Object target)
	{
		switch(GeoConfig.GEODATA_MODE)
		{
			case MODE_LOCAL_GEO:
				return geoEngine.canSeeTarget(actor, target);
			case MODE_RMI_GEO:
				try
				{
					return  geoServer.canSeeCoord(this, actor.getX(), actor.getY(), actor.getZ(), target.getX(), target.getY(), target.getZ());
				}
				catch(RemoteException e)
				{
					e.printStackTrace();
					onConnectionLost();
					return true;
				}
			case MODE_NO_GEO:
			default:
				return true;
		}
	}

	boolean canSeeCoord(GeoClientInterface client, L2Object actor, int x, int y, int z)
	{
		switch(GeoConfig.GEODATA_MODE)
		{
			case MODE_LOCAL_GEO:
				return geoEngine.canSeeCoord(actor, x, y, z);
			case MODE_RMI_GEO:
				try
				{
					return geoServer.canSeeCoord(this, actor, x, y, z);
				}
				catch(RemoteException e)
				{
					e.printStackTrace();
					onConnectionLost();
					return true;
				}
			case MODE_NO_GEO:
			default:
				return true;
		}
	}

	/**
	 * Возвращает z координату по геодате.<BR>
	 * При отсутсвующей геодате возвращается передаваемая z.
	 *
	 * @param x координата
	 * @param y координата
	 * @param z координата
	 *
	 * @return Z по геодате, или передаваемую z при ее отсутсвии.
	 */
	public int getHeight(int x, int y, int z)
	{
		switch(GeoConfig.GEODATA_MODE)
		{
			case MODE_LOCAL_GEO:
				return geoEngine.getHeight(x, y, z);
			case MODE_RMI_GEO:
				try
				{
					return geoServer.getHeight(this, x, y, z);
				}
				catch(RemoteException e)
				{
					e.printStackTrace();
					onConnectionLost();
					return z;
				}
			case MODE_NO_GEO:
			default:
				return z;
		}
	}

	public int getHeight(Location loc)
	{
		return getHeight(loc.getX(), loc.getY(), loc.getZ());
	}

	public Location moveCheck(int x, int y, int z, int tx, int ty, int tz)
	{
		switch(GeoConfig.GEODATA_MODE)
		{
			case MODE_LOCAL_GEO:
				return geoEngine.moveCheck(x, y, z, tx, ty);
			case MODE_RMI_GEO:
				try
				{
					return geoServer.moveCheck(this, x, y, z, tx, ty);
				}
				catch(RemoteException e)
				{
					e.printStackTrace();
					onConnectionLost();
					return new Location(tx, ty, tz);
				}
			case MODE_NO_GEO:
			default:
				return new Location(tx, ty, tz);
		}
	}

	public Location moveInWaterCheck(int x, int y, int z, int tx, int ty, int tz)
	{
		switch(GeoConfig.GEODATA_MODE)
		{
			case MODE_LOCAL_GEO:
				return geoEngine.moveInWaterCheck(x, y, z, tx, ty, tz);
			case MODE_RMI_GEO:
				try
				{
					return geoServer.moveInWaterCheck(this, x, y, z, tx, ty, tz);
				}
				catch(RemoteException e)
				{
					e.printStackTrace();
					onConnectionLost();
					return new Location(tx, ty, tz);
				}
			case MODE_NO_GEO:
			default:
				return new Location(tx, ty, tz);
		}
	}

	/**
	 * Проверяет возможность движения с одних координат в другие.
	 *
	 * @param x  исходная координата
	 * @param y  исходная координата
	 * @param z  исходная координата
	 * @param tx координата цели
	 * @param ty координата цели
	 *
	 * @return можно ли двигатся к координатам
	 */
	public boolean canMoveToCoord(int x, int y, int z, int tx, int ty, int tz)
	{
		switch(GeoConfig.GEODATA_MODE)
		{
			case MODE_LOCAL_GEO:
				return geoEngine.canMoveToCoord(x, y, z, tx, ty, tz);
			case MODE_RMI_GEO:
				try
				{
					return geoServer.canMoveToCoord(this, x, y, z, tx, ty, tz);
				}
				catch(RemoteException e)
				{
					e.printStackTrace();
					onConnectionLost();
					return true;
				}
			case MODE_NO_GEO:
			default:
				return true;
		}
	}

	public Location moveCheckForAI(L2Object cha, L2Object target)
	{
		return moveCheckForAI(new Location(cha.getX(), cha.getY(), cha.getZ()), new Location(target.getX(), target.getY(), target.getZ()));
	}

	public Location moveCheckForAI(Location loc1, Location loc2)
	{
		switch(GeoConfig.GEODATA_MODE)
		{
			case MODE_LOCAL_GEO:
				return geoEngine.moveCheckForAI(loc1, loc2);
			case MODE_RMI_GEO:
				try
				{
					return geoServer.moveCheckForAI(this, loc1, loc2);
				}
				catch(RemoteException e)
				{
					e.printStackTrace();
					onConnectionLost();
					return loc2;
				}
			case MODE_NO_GEO:
			default:
				return loc2;
		}
	}

	/**
	 * @param x координата
	 * @param y координата
	 * @param z координата
	 *
	 * @return Geo NSWE(0-15)
	 */
	public short getNSWE(int x, int y, int z)
	{
		switch(GeoConfig.GEODATA_MODE)
		{
			case MODE_LOCAL_GEO:
				return geoEngine.getNSWE(x, y, z);
			case MODE_RMI_GEO:
				try
				{
					return geoServer.getNSWE(this, x, y, z);
				}
				catch(RemoteException e)
				{
					e.printStackTrace();
					onConnectionLost();
					return 15;
				}
			case MODE_NO_GEO:
			default:
				return 15;
		}
	}

	public boolean canMoveToCoordWithCollision(int x, int y, int z, int tx, int ty, int tz)
	{
		switch(GeoConfig.GEODATA_MODE)
		{
			case MODE_LOCAL_GEO:
				return geoEngine.canMoveWithCollision(x, y, z, tx, ty, tz);// canMoveToCoordWithCollision(x, y, z, tx, ty, tz);
			case MODE_RMI_GEO:
				try
				{
					return geoServer.canMoveToCoordWithCollision(this, x, y, z, tx, ty, tz);
				}
				catch(RemoteException e)
				{
					e.printStackTrace();
					onConnectionLost();
					return true;
				}
			case MODE_NO_GEO:
			default:
				return true;
		}
	}

	/**
	 * @param x coord
	 * @param y coord
	 *
	 * @return Geo Block Type
	 */
	public short getType(int x, int y)
	{
		switch(GeoConfig.GEODATA_MODE)
		{
			case MODE_LOCAL_GEO:
				return geoEngine.getType(x, y);
			case MODE_RMI_GEO:
				try
				{
					return geoServer.getType(this, x, y);
				}
				catch(RemoteException e)
				{
					e.printStackTrace();
					onConnectionLost();
					return 0;
				}
			case MODE_NO_GEO:
			default:
				return 0;
		}
	}

	/**
	 * Пробует открыть дверь в геодате
	 * @param pos информация о двери
	 */
	public void openDoor(L2Territory pos)
	{
		if(!GeoConfig.ALLOW_DOORS)
			return;

		switch(GeoConfig.GEODATA_MODE)
		{
			case MODE_LOCAL_GEO:
				geoEngine.openDoor(pos);
				break;
			case MODE_RMI_GEO:
				try
				{
					geoServer.openDoor(this, pos);
				}
				catch(RemoteException e)
				{
					e.printStackTrace();
					onConnectionLost();
				}
				break;
		}
	}

	/**
	 * Пробует закрыть дверь в геодате
	 * @param pos информация о двери
	 */
	public void closeDoor(L2Territory pos)
	{
		if(!GeoConfig.ALLOW_DOORS)
			return;

		switch(GeoConfig.GEODATA_MODE)
		{
			case MODE_LOCAL_GEO:
				geoEngine.closeDoor(pos);
				break;
			case MODE_RMI_GEO:
				try
				{
					geoServer.closeDoor(this, pos);
				}
				catch(RemoteException e)
				{
					e.printStackTrace();
					onConnectionLost();
				}
				break;
		}
	}

	public Location findPointToStay(int x, int y, int z, int j, int k)
	{
		Location pos = new Location(x, y, z);
		Location BasePos = new Location(x, y, z);
		boolean found = false;
		for(int i = 0; i < 100; i++)
		{
			pos = Rnd.coordsRandomize(x, y, z, 0, j, k);
			if(canMoveToCoord(x, y, z, pos.getX(), pos.getY(), pos.getZ()) && canMoveToCoord(pos.getX(), pos.getY(), pos.getZ(), x, y, z))
			{
				found = true;
				break;
			}
		}
		return found ? pos : BasePos;
	}
}