package l2r.gameserver.geodata.network.geoserver;

import java.io.File;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.Vector;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Logger;

import l2r.gameserver.geodata.GeoEngine;
import l2r.gameserver.geodata.RemoteGeoEngine;
import l2r.gameserver.geodata.network.GeoConfig;
import l2r.gameserver.geodata.network.SocketFactory;
import l2r.gameserver.geodata.network.geoclient.GeoClientInterface;
import l2r.gameserver.model.L2Territory;
import l2r.gameserver.model.L2Object;
import l2r.util.Location;

/**
 * @Author: Death
 * @Date: 23/11/2007
 * @Time: 10:57:02
 */
@SuppressWarnings("serial")
public class GeoServer extends UnicastRemoteObject implements GeoServerInterface
{
	private static final Logger log = Logger.getLogger(GeoServer.class.getName());
	private static final ConcurrentHashMap<GeoClientInterface, RemoteGeoEngine> clients = new ConcurrentHashMap<GeoClientInterface, RemoteGeoEngine>();

	public GeoServer() throws RemoteException
	{
		// Запуск трида для очистки оборвавшихся коннектов -> освобождения выделеной памяти под сервер.
		new Thread(new Runnable(){
			public void run()
			{
				while(true)
				{
					try
					{
						Thread.sleep(1000);
					}
					catch(InterruptedException e)
					{
						log.severe("GeoServer: Cleaning Thread interrupted.");
					}

					for(GeoClientInterface client : clients.keySet())
						try
						{
							// Просто вызов метода для проверки коннекта
							client.testConnection();
						}
						catch(RemoteException e)
						{
							clients.remove(client);
							log.info("GeoServer: Client disconnected.");
						}
				}
			}
		}).start();
	}

	/**
	 * Получения геоаццесора для даного удаленного клиента геодаты.<BR>
	 * При клиенте = null будет брошен RemoteException, что в свою очередь заставит клиент запустить переподключение к серверу.
	 *
	 * @param client клиент для получения геодаты
	 *
	 * @return RemoteGeoEngine актуальный для даного клиента
	 *
	 * @throws RemoteException в случае отсутвия геоаццесора, инициализирует переподключение клиента.
	 */
	private RemoteGeoEngine getGeoEngine(GeoClientInterface client) throws RemoteException
	{
		RemoteGeoEngine geo = clients.get(client);
		if(geo != null)
			return geo;
		throw new RemoteException("Null geoEngine for current geoClient!");
	}

	public void addClient(GeoClientInterface client)
	{
		clients.put(client, new RemoteGeoEngine());
	}

	public short getType(GeoClientInterface client, int x, int y) throws RemoteException
	{
		return getGeoEngine(client).getType(x, y);
	}

	public int getHeight(GeoClientInterface client, int x, int y, int z) throws RemoteException
	{
		return getGeoEngine(client).getHeight(x, y, z);
	}

	public boolean canSeeCoord(GeoClientInterface client, int x, int y, int z, int tx, int ty, int tz) throws RemoteException
	{
		return getGeoEngine(client).canSeeCoord(x, y, z, tx, ty, tz);
	}

	public boolean canSeeCoord(GeoClientInterface client, L2Object actor, int tx, int ty, int tz) throws RemoteException
	{
		return getGeoEngine(client).canSeeCoord(actor, tx, ty, tz);
	}

	public boolean canSeeTarget(GeoClientInterface client, L2Object actor, L2Object target) throws RemoteException
	{
		return getGeoEngine(client).canSeeTarget(actor, target);
	}

	public boolean canMoveToCoord(GeoClientInterface client, int x, int y, int z, int tx, int ty, int tz) throws RemoteException
	{
		return getGeoEngine(client).canMoveToCoord(x, y, z, tx, ty, tz);
	}

	public boolean canMoveToCoordWithCollision(GeoClientInterface client, int x, int y, int z, int tx, int ty, int tz) throws RemoteException
	{
		return getGeoEngine(client).canMoveWithCollision(x, y, z, tx, ty, tz);// canMoveToCoordWithCollision(x, y, z, tx, ty, tz);
	}

	public short getNSWE(GeoClientInterface client, int x, int y, int z) throws RemoteException
	{
		return getGeoEngine(client).getNSWE(x, y, z);
	}

	public Location moveCheck(GeoClientInterface client, int x, int y, int z, int tx, int ty) throws RemoteException
	{
		return getGeoEngine(client).moveCheck(x, y, z, tx, ty);
	}

	@Override
	public Location moveInWaterCheck(GeoClientInterface client, int x, int y, int z, int tx, int ty, int tz) throws RemoteException
	{
		return getGeoEngine(client).moveInWaterCheck(x, y, z, tz, tx, ty);
	}

	public Location moveCheckForAI(GeoClientInterface client, Location loc1, Location loc2) throws RemoteException
	{
		return getGeoEngine(client).moveCheckForAI(loc1, loc2);
	}

	public Vector<Location> checkMovement(GeoClientInterface client, int x, int y, int z, Location target, boolean playable) throws RemoteException
	{
		return getGeoEngine(client).getGeoMove().findPath(x, y, z, target, playable, false);// checkMovement(x, y, z, target);
	}

	public void closeDoor(GeoClientInterface client, L2Territory pos) throws RemoteException
	{
		getGeoEngine(client).closeDoor(pos);
	}

	public void openDoor(GeoClientInterface client, L2Territory pos) throws RemoteException
	{
		getGeoEngine(client).openDoor(pos);
	}

	public static void main(String[] args)
	{
		log.info("L2Final GeoServer: Starting up...");

		// Возможность установить политику безопасности на геосервер.
		// http://java.sun.com/javase/technologies/security/
		// http://java.sun.com/j2se/1.4.2/docs/guide/security/PolicyFiles.html
		File f = new File("geoserver.policy");
		if(f.exists() && f.isFile())
		{
			System.setProperty("java.security.policy", "geoserver.policy");
			System.setSecurityManager(new SecurityManager());
			log.info("L2Final GeoServer: geoserver.policy initialized.");
		}
		else
			log.info("L2Final GeoServer: geoserver.policy missing. SecurityManager setup ignored.");

		try
		{
			GeoConfig.load();
		}
		catch(Exception e)
		{
			e.printStackTrace();
			log.severe("Can't init geodata config.");
			System.exit(0);
		}

		try
		{
			GeoEngine.loadGeo();
			Registry r = LocateRegistry.createRegistry(GeoConfig.PORT, new SocketFactory(), new SocketFactory());
			GeoServer server = new GeoServer();
			r.bind("geoServer", server);
		}
		catch(Exception e)
		{
			e.printStackTrace();
			System.exit(0);
		}

		log.info("L2Final GeoServer: Running at " + GeoConfig.SERVER_BIND_HOST + ":" + GeoConfig.PORT + "");
	}
}
