package l2r.gameserver.instancemanager;

import java.awt.Point;
import java.awt.Polygon;
import java.awt.Shape;
import java.sql.ResultSet;
import java.util.logging.Logger;

import javolution.util.FastMap;
import l2r.Config;
import l2r.database.DatabaseUtils;
import l2r.database.FiltredPreparedStatement;
import l2r.database.L2DatabaseFactory;
import l2r.database.ThreadConnection;
import l2r.gameserver.model.L2Player;
import l2r.gameserver.model.entity.DimensionalRift;
import l2r.gameserver.model.instances.L2ItemInstance;
import l2r.gameserver.model.instances.L2NpcInstance;
import l2r.gameserver.serverpackets.NpcHtmlMessage;
import l2r.gameserver.serverpackets.PartyMemberPosition;
import l2r.gameserver.serverpackets.TeleportToLocation;
import l2r.util.Location;
import l2r.util.Rnd;

public class DimensionalRiftManager
{
	private static Logger _log = Logger.getLogger(DimensionalRiftManager.class.getName());
	private static DimensionalRiftManager _instance;
	private FastMap<Byte, FastMap<Byte, DimensionalRiftRoom>> _rooms = new FastMap<Byte, FastMap<Byte, DimensionalRiftRoom>>();
	private final short DIMENSIONAL_FRAGMENT_ITEM_ID = 7079;

	public static DimensionalRiftManager getInstance()
	{
		if(_instance == null)
			_instance = new DimensionalRiftManager();

		return _instance;
	}

	public DimensionalRiftManager()
	{
		load();
	}

	public DimensionalRiftRoom getRoom(byte _type, byte room)
	{
		return _rooms.get(_type).get(room);
	}

	private void load()
	{
		ThreadConnection con = null;
		FiltredPreparedStatement s = null;
		ResultSet rs = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			s = con.prepareStatement("SELECT * FROM dimensional_rift");
			rs = s.executeQuery();

			while(rs.next())
			{
				// 0 waiting room, 1 recruit, 2 soldier, 3 officer, 4 captain , 5 commander, 6 hero
				byte type = rs.getByte("type");
				byte room_id = rs.getByte("room_id");

				//coords related
				int xMin = rs.getInt("xMin");
				int xMax = rs.getInt("xMax");
				int yMin = rs.getInt("yMin");
				int yMax = rs.getInt("yMax");
				int z1 = rs.getInt("zMin");
				int z2 = rs.getInt("zMax");
				int xT = rs.getInt("xT");
				int yT = rs.getInt("yT");
				int zT = rs.getInt("zT");
				boolean isBossRoom = rs.getByte("boss") > 0;

				if(!_rooms.containsKey(type))
					_rooms.put(type, new FastMap<Byte, DimensionalRiftRoom>());

				_rooms.get(type).put(room_id, new DimensionalRiftRoom(xMin, xMax, yMin, yMax, z1, z2, xT, yT, zT, isBossRoom));
			}

		}
		catch(Exception e)
		{
			_log.warning("Can't load Dimension Rift zones.\n" + e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCSR(con, s, rs);
		}

		int typeSize = _rooms.keySet().size();
		int roomSize = 0;

		for(Byte b : _rooms.keySet())
			roomSize += _rooms.get(b).keySet().size();

		_log.info("DimensionalRiftManager: Loaded " + typeSize + " room types with " + roomSize + " rooms.");
	}

	public void reload()
	{
		for(Byte b : _rooms.keySet())
			_rooms.get(b).clear();

		_rooms.clear();
		load();
	}

	public boolean checkIfInRiftZone(Location loc, boolean ignorePeaceZone)
	{
		if(ignorePeaceZone)
			return _rooms.get((byte) 0).get((byte) 1).checkIfInZone(loc);
		return _rooms.get((byte) 0).get((byte) 1).checkIfInZone(loc) && !_rooms.get((byte) 0).get((byte) 0).checkIfInZone(loc);
	}

	public boolean checkIfInPeaceZone(Location loc)
	{
		return _rooms.get((byte) 0).get((byte) 0).checkIfInZone(loc);
	}

	public void teleportToWaitingRoom(L2Player player)
	{
		teleToLocation(player, getRoom((byte) 0, (byte) 0).getTeleportCoords());
	}

	public void start(L2Player player, byte type, L2NpcInstance npc)
	{
		if(!player.isInParty())
		{
			showHtmlFile(player, "data/html/rift/NoParty.htm", npc);
			return;
		}

		if(player.getParty().getPartyLeaderOID() != player.getObjectId())
		{
			showHtmlFile(player, "data/html/rift/NotPartyLeader.htm", npc);
			return;
		}

		if(player.getParty().isInDimensionalRift())
		{
			showHtmlFile(player, "data/html/rift/Cheater.htm", npc);

			if(!player.isGM())
				_log.warning("Player " + player.getName() + "(" + player.getObjectId() + ") was cheating in dimension rift area!");

			return;
		}

		if(player.getParty().getMemberCount() < Config.RIFT_MIN_PARTY_SIZE)
		{
			showHtmlFile(player, "data/html/rift/SmallParty.htm", npc);
			return;
		}

		for(L2Player p : player.getParty().getPartyMembers())
			if(!checkIfInPeaceZone(p.getLoc()))
			{
				showHtmlFile(player, "data/html/rift/NotInWaitingRoom.htm", npc);
				return;
			}

		L2ItemInstance i;
		for(L2Player p : player.getParty().getPartyMembers())
		{
			i = p.getInventory().getItemByItemId(DIMENSIONAL_FRAGMENT_ITEM_ID);
			if(i == null || i.getIntegerLimitedCount() < getNeededItems(type))
			{
				showHtmlFile(player, "data/html/rift/NoFragments.htm", npc);
				return;
			}
		}

		for(L2Player p : player.getParty().getPartyMembers())
		{
			i = p.getInventory().getItemByItemId(DIMENSIONAL_FRAGMENT_ITEM_ID);
			p.getInventory().destroyItem(i.getObjectId(), getNeededItems(type), true);
		}

		new DimensionalRift(player.getParty(), type, (byte) Rnd.get(1, 9));
	}

	public void killRift(DimensionalRift d)
	{
		d.getTimerTask().cancel();
		d.setTimerTask(null);

		d.getTimer().cancel();
		d.setTimer(null);
	}

	public class DimensionalRiftRoom
	{
		private final int _zMin;
		private final int _zMax;
		private final Location _teleportCoords;
		private final Shape _s;
		private final boolean _isBossRoom;

		public DimensionalRiftRoom(int xMin, int xMax, int yMin, int yMax, int zMin, int zMax, int xT, int yT, int zT, boolean isBossRoom)
		{
			_zMin = zMin;
			_zMax = zMax;
			_teleportCoords = new Location(xT, yT, zT);
			_isBossRoom = isBossRoom;
			_s = new Polygon(new int[] { xMin, xMax, xMax, xMin }, new int[] { yMin, yMin, yMax, yMax }, 4);
		}

		public Location getTeleportCoords()
		{
			return _teleportCoords;
		}

		public boolean checkIfInZone(Location loc)
		{
			return _s.contains(loc.getX(), loc.getY()) && loc.getZ() >= _zMin && loc.getZ() <= _zMax;
		}

		public boolean isBossRoom()
		{
			return _isBossRoom;
		}
	}

	private int getNeededItems(byte type)
	{
		switch(type)
		{
			case 1:
				return Config.RIFT_ENTER_COST_RECRUIT;
			case 2:
				return Config.RIFT_ENTER_COST_SOLDIER;
			case 3:
				return Config.RIFT_ENTER_COST_OFFICER;
			case 4:
				return Config.RIFT_ENTER_COST_CAPTAIN;
			case 5:
				return Config.RIFT_ENTER_COST_COMMANDER;
			case 6:
				return Config.RIFT_ENTER_COST_HERO;
			default:
				return 999999;
		}
	}

	public void showHtmlFile(L2Player player, String file, L2NpcInstance npc)
	{
		NpcHtmlMessage html = new NpcHtmlMessage(npc.getObjectId());
		html.setFile(file);
		html.replace("%t_name%", npc.getName());
		player.sendPacket(html);
	}

	public void teleToLocation(L2Player player, Location loc)
	{
		teleToLocation(player, loc.getX(), loc.getY(), loc.getZ());
	}

	public void teleToLocation(L2Player player, int x, int y, int z)
	{
		player.setTarget(null);
		player.stopMove();

		if(player.isInBoat())
		{
			player.setInBoat(false);
			player.setBoat(null);
			player.setBoatId(0);
		}

		player.decayMe();

		player.setXYZInvisible(x, y, z);
		player.setIsTeleporting(true);

		// Нужно при телепорте с более высокой точки на более низкую, иначе наносится вред от "падения"
		player.setLastClientPosition(null);
		player.setLastServerPosition(null);

		player.sendPacket(new TeleportToLocation(player, x, y, z));
		if(player.getParty() != null)
		{
			Point position = player.getLastPartyPositionSent();
			if(player.getDistance(position.x, position.y) < 500)
			{
				player.getParty().broadcastToPartyMembers(player, new PartyMemberPosition(player));
				position.move(player.getX(), player.getY());
			}
		}
	}
}