package l2r.gameserver.instancemanager;

import java.util.HashMap;
import java.util.logging.Logger;

import javolution.util.FastList;
import l2r.gameserver.model.L2Object;
import l2r.gameserver.model.L2Zone;
import l2r.gameserver.model.L2Zone.ZoneType;
import l2r.gameserver.model.entity.Fortress;

public class FortressManager
{
	protected static Logger _log = Logger.getLogger(FortressManager.class.getName());

	private static FortressManager _instance;
	private static HashMap<Integer, Fortress> _fortresses;

	public static FortressManager getInstance()
	{
		if(_instance == null)
		{
			_log.info("Initializing FortressManager");
			_instance = new FortressManager();
			_instance.load();
		}
		return _instance;
	}

	private void load()
	{
		FastList<L2Zone> zones = ZoneManager.getInstance().getZoneByType(ZoneType.Fortress);
		if(zones.size() == 0)
			_log.info("Not found zones for Fortresses!!!");
		else
			for(L2Zone zone : zones)
				getFortresses().put(zone.getIndex(), new Fortress(zone.getIndex()));
		_log.info("Loaded: " + getFortresses().size() + " fortresses.");
	}

	/**
	 * Возвращает крепость, соответствующую индексу.
	 */
	public Fortress getFortressByIndex(int index)
	{
		for(Fortress fortress : getFortresses().values())
			if(fortress != null && fortress.getId() == index)
				return fortress;
		return null;
	}

	/**
	 * Находит крепость по имени. Если такой крепости нет - возвращает null.
	 */
	public Fortress getFortressByName(String name)
	{
		int index = getFortressIndexByName(name);
		if(index > 0)
			return getFortresses().get(index);
		return null;
	}

	/**
	 * Если координаты принадлежат зоне какой-либо крепости, возвращает эту крепость.
	 * Иначе возвращает null.
	 */
	public Fortress getFortressByObject(L2Object activeObject)
	{
		return getFortressByCoord(activeObject.getX(), activeObject.getY());
	}

	/**
	 * Если обьект находится в зоне какой-либо крепости, возвращает эту крепость.
	 * Иначе возвращает null.
	 */
	public Fortress getFortressByCoord(int x, int y)
	{
		int index = getFortressIndexByCoord(x, y);
		if(index > 0)
			return getFortresses().get(index);
		return null;
	}

	/**
	 * Если обьект находится в зоне какой-либо крепости, возвращает эту крепость.
	 * Иначе возвращает -1.
	 */
	public int getFortressIndex(L2Object activeObject)
	{
		return getFortressIndexByCoord(activeObject.getX(), activeObject.getY());
	}

	/**
	 * Если координаты принадлежат зоне какой-либо крепости, возвращает индекс этой крепости.
	 * Иначе возвращает -1.
	 */
	public int getFortressIndexByCoord(int x, int y)
	{
/*
		Fortress fortress;
		for(int i = 1; i <= getFortresses().size(); i++)
		{
			fortress = getFortresses().get(i);
			if(fortress != null && fortress.checkIfInZone(x, y))
				return i;
		}
		return -1;
*/
		for(Fortress fortress : getFortresses().values())
		{
			if(fortress != null && fortress.checkIfInZone(x, y))
				return fortress.getId();
		}
		return -1;
	}

	/**
	 * Находит крепость по имени, без учета регистра.
	 * Если не найден - возвращает -1.
	 */
	public int getFortressIndexByName(String name)
	{
/*
		Fortress fortress;
		for(int i = 1; i <= getFortresses().size(); i++)
		{
			fortress = getFortresses().get(i);
			if(fortress != null && fortress.getName().equalsIgnoreCase(name.trim()))
				return i;
		}
		return -1;
*/
		for(Fortress fortress : getFortresses().values())
		{
			if(fortress != null && fortress.getName().equalsIgnoreCase(name.trim()))
				return fortress.getId();
		}
		return -1;
	}

	/**
	 * Возвращает список, содержащий все крепости.
	 */
	public static HashMap<Integer, Fortress> getFortresses()
	{
		if(_fortresses == null)
			_fortresses = new HashMap<Integer, Fortress>();
		return _fortresses;
	}

	public int findNearestFortressIndex(int x, int y, int offset)
	{
		int index = getFortressIndexByCoord(x, y);
		if(index < 0)
		{
			double closestDistance = offset;
			double distance;
			for(Fortress fortress : getFortresses().values())
			{
				if(fortress == null)
					continue;
				distance = fortress.getZone().findDistanceToZone(x, y, 0, false);
				if(closestDistance > distance)
				{
					closestDistance = distance;
					index = fortress.getId();
				}
			}
		}
		return index;
	}
}