package l2r.gameserver.instancemanager;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.logging.Logger;

import javolution.util.FastList;
import javolution.util.FastMap;
import l2r.Config;
import l2r.gameserver.model.L2Character;
import l2r.gameserver.model.L2Object;
import l2r.gameserver.model.L2Player;
import l2r.gameserver.model.entity.Fortress;
import l2r.gameserver.model.entity.siege.Siege;
import l2r.gameserver.model.entity.siege.SiegeSpawn;
import l2r.gameserver.model.entity.siege.fortress.CombatFlag;
import l2r.gameserver.model.entity.siege.fortress.FortressSiege;
import l2r.gameserver.model.instances.L2ItemInstance;
import l2r.gameserver.model.instances.L2StaticObjectInstance;
import l2r.gameserver.serverpackets.SystemMessage;
import l2r.gameserver.tables.StaticObjectsTable;

public class FortressSiegeManager extends SiegeManager
{
	protected static Logger _log = Logger.getLogger(FortressSiegeManager.class.getName());

	private static FortressSiegeManager _instance;

	public static FortressSiegeManager getInstance()
	{
		if(_instance == null)
			_instance = new FortressSiegeManager();
			
		return _instance;
	}

	public FortressSiegeManager()
	{
		load();
	}
	
	private static FastMap<Integer, FastList<SiegeSpawn>> _commanderSpawnList;
	private static FastMap<Integer, FastList<CombatFlag>> _flagList;
	private static FastMap<Integer, FastList<SiegeSpawn>> _flagPoleSpawnList;
	private static FastMap<Integer, FastList<Integer>> _guardDoorList;
	private static FastMap<Integer, FastList<Integer>> _commandCenterDoorList;

	public static void load()
	{
		try
		{
			InputStream is = new FileInputStream(new File(Config.SIEGE_FORTRESS_CONFIGURATION_FILE));
			Properties siegeSettings = new Properties();
			siegeSettings.load(is);
			is.close();

			// Siege setting
			int defenderRespawnDelay = Integer.decode(siegeSettings.getProperty("DefenderRespawn", "20000"));
			int siegeClanMinLevel = Integer.decode(siegeSettings.getProperty("SiegeClanMinLevel", "4"));
			int siegeLength = Integer.decode(siegeSettings.getProperty("SiegeLength", "60"));

			// Siege spawns settings
			_flagList = new FastMap<Integer, FastList<CombatFlag>>();
			_commanderSpawnList = new FastMap<Integer, FastList<SiegeSpawn>>();
			_flagPoleSpawnList = new FastMap<Integer, FastList<SiegeSpawn>>();
			_guardDoorList = new FastMap<Integer, FastList<Integer>>();
			_commandCenterDoorList = new FastMap<Integer, FastList<Integer>>();

			for(Fortress fortress : FortressManager.getFortresses().values())
			{
				FastList<SiegeSpawn> _commanderSpawns = new FastList<SiegeSpawn>();
				FastList<CombatFlag> _flagSpawns = new FastList<CombatFlag>();
				FastList<SiegeSpawn> _flagPoleSpawns = new FastList<SiegeSpawn>();
				FastList<Integer> _guardDoors = new FastList<Integer>();
				FastList<Integer> _commandCenterDoors = new FastList<Integer>();

				for(int i = 1; i < 0xFF; i++)
				{
					String guardDoor = siegeSettings.getProperty("N" + fortress.getId() + "GuardDoor" + i, "");

					if(guardDoor.length() == 0)
						break;

					_guardDoors.add(Integer.parseInt(guardDoor));
				}

				for(int i = 1; i < 0xFF; i++)
				{
					String commandCenter = siegeSettings.getProperty("N" + fortress.getId() + "CommandCenterDoor" + i, "");

					if(commandCenter.length() == 0)
						break;

					_commandCenterDoors.add(Integer.parseInt(commandCenter));
				}

				for(int i = 1; i < 0xFF; i++)
				{
					String _spawnParams = siegeSettings.getProperty("N" + fortress.getId() + "Commander" + i, "");

					if(_spawnParams.length() == 0)
						break;

					StringTokenizer st = new StringTokenizer(_spawnParams.trim(), ",");

					try
					{
						int x = Integer.parseInt(st.nextToken());
						int y = Integer.parseInt(st.nextToken());
						int z = Integer.parseInt(st.nextToken());
						int heading = Integer.parseInt(st.nextToken());
						int npc_id = Integer.parseInt(st.nextToken());

						_commanderSpawns.add(new SiegeSpawn(fortress.getId(), x, y, z, heading, npc_id));
					}
					catch(Exception e)
					{
						_log.warning("Error while loading commander(s) for " + fortress.getName());
					}
				}

				for(int i = 1; i < 0xFF; i++)
				{
					String _spawnParams = siegeSettings.getProperty("N" + fortress.getId() + "Flag" + i, "");

					if(_spawnParams.length() == 0)
						break;

					StringTokenizer st = new StringTokenizer(_spawnParams.trim(), ",");

					try
					{
						int x = Integer.parseInt(st.nextToken());
						int y = Integer.parseInt(st.nextToken());
						int z = Integer.parseInt(st.nextToken());
						int flag_id = Integer.parseInt(st.nextToken());

						_flagSpawns.add(new CombatFlag(fortress.getId(), x, y, z, 0, flag_id));
					}
					catch(Exception e)
					{
						_log.warning("Error while loading control flag(s) for " + fortress.getName());
					}
				}

				for(int i = 1; i < 0xFF; i++)
				{
					String _spawnParams = siegeSettings.getProperty("N" + fortress.getId() + "FlagPole" + i, "");

					if(_spawnParams.length() == 0)
						break;

					StringTokenizer st = new StringTokenizer(_spawnParams.trim(), ",");

					try
					{
						int x = Integer.parseInt(st.nextToken());
						int y = Integer.parseInt(st.nextToken());
						int z = Integer.parseInt(st.nextToken());
						int heading = Integer.parseInt(st.nextToken());
						int npc_id = Integer.parseInt(st.nextToken());

						_flagPoleSpawns.add(new SiegeSpawn(fortress.getId(), x, y, z, heading, npc_id));
					}
					catch(Exception e)
					{
						_log.warning("Error while loading FlagPole(s) for " + fortress.getName());
					}
				}

				_commanderSpawnList.put(fortress.getId(), _commanderSpawns);
				_flagList.put(fortress.getId(), _flagSpawns);
				_flagPoleSpawnList.put(fortress.getId(), _flagPoleSpawns);
				_guardDoorList.put(fortress.getId(), _guardDoors);
				_commandCenterDoorList.put(fortress.getId(), _commandCenterDoors);

				if(_commanderSpawns.size() == 0)
					_log.warning("Not found commanders for " + fortress.getName());
				if(_flagSpawns.size() == 0)
					_log.warning("Not found control flags for " + fortress.getName());
				if(_flagPoleSpawns.size() == 0)
					_log.warning("Not found flagpole for " + fortress.getName());
				if(_guardDoors.size() == 0)
					_log.warning("Not found guard doors for " + fortress.getName());
				if(_commandCenterDoors.size() == 0)
					_log.warning("Not found command center doors for " + fortress.getName());

				fortress.getSiege().setDefenderRespawnDelay(defenderRespawnDelay);
				fortress.getSiege().setSiegeClanMinLevel(siegeClanMinLevel);
				fortress.getSiege().setSiegeLength(siegeLength);

				spawnFlagPoles(fortress);
				fortress.getSiege().startAutoTask();
				fortress.getSiege().getZone().setActive(false);
			}
		}
		catch(Exception e)
		{
			System.err.println("Error while loading siege data.");
			e.printStackTrace();
		}
	}

	public final boolean checkIfOkToSummon(L2Character activeChar, boolean isCheckOnly)
	{
		if (!(activeChar instanceof L2Player))
			return false;

		SystemMessage sm = new SystemMessage(SystemMessage.S1);
		L2Player player = (L2Player) activeChar;
		Fortress fort = FortressManager.getInstance().getFortressByObject(player);

		if (fort == null || fort.getId() <= 0)
			sm.addString("You must be on fort ground to summon this");
		else if (!fort.getSiege().getIsInProgress())
			sm.addString("You can only summon this during a siege.");
		else if (player.getClanId() != 0 && fort.getSiege().getAttackerClan(player.getClanId()) == null)
			sm.addString("You can only summon this as a registered attacker.");
		else
			return true;

		if (!isCheckOnly)
			player.sendPacket(sm);
		return false;
	}
	
	public static FastList<SiegeSpawn> getCommanderSpawnList(int siegeUnitId)
	{
		return _commanderSpawnList.get(siegeUnitId);
	}

	public static FastList<CombatFlag> getFlagsList(int siegeUnitId)
	{
		return _flagList.get(siegeUnitId);
	}

	public static FastList<Integer> getGuardDoors(int siegeUnitId)
	{
		return _guardDoorList.get(siegeUnitId);
	}

	public static FastList<Integer> getCommandCenterDoors(int siegeUnitId)
	{
		return _commandCenterDoorList.get(siegeUnitId);
	}
	
	public boolean isCombat(int itemId)
	{
		return (itemId == 9819);
	}
	
	public void activateCombatFlag(L2Player player, L2ItemInstance item)
	{
//		if(!checkIfCanPickup(player))лишнее - ниже проверка
//			return;
		try
		{
			Fortress fort = FortressManager.getInstance().getFortressByObject(player);
			FastList<CombatFlag> fcf = _flagList.get(fort.getId());
			for(CombatFlag cf : fcf)
				if(cf.itemInstance == item)
				{
					cf.checkPlayer(player, item);
					break;
				}
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
	}

	public boolean checkIfCanPickup(L2Player player)
	{
		Siege siege = getSiege(player);

		if(siege == null)
		{
			player.sendMessage("You can only pickup Combat Flag during a siege.");
			return false;
		}

		else if(siege.getAttackerClan(player.getClan()) == null)
		{
			player.sendMessage("You must be an attacker to pickup Combat Flag");
			return false;
		}

		// if the player is mounted, attempt to unmount first. Only allow picking up the comabt flag if unmounting is successful.
		if(player.isMounted())
		{
			player.sendMessage("You may not pick up this item while riding in this territory");
			return false;
		}

		// Cannot own 2 combat flag
		if(player.isCombatFlagEquipped())
		{
			player.sendMessage("You already have the combat flag");
			return false;
		}

		if(!player.isClanLeader())
		{
			player.sendMessage("Only Clan Leader Can pick Up Flag");
			return false;
		}

		return true;
	}

	public void dropCombatFlag(L2Player player)
	{
		Fortress fort = FortressManager.getInstance().getFortressByObject(player);
		FastList<CombatFlag> fcf = _flagList.get(fort.getId());

		for(CombatFlag cf : fcf)
			if(cf.playerId == player.getObjectId())
			{
				cf.dropIt();
				cf.spawnMe();
			}
	}

	public static FortressSiege getSiege(L2Object activeObject)
	{
		return getSiege(activeObject.getX(), activeObject.getY());
	}

	public static FortressSiege getSiege(int x, int y)
	{
		for(Fortress fortress : FortressManager.getInstance().getFortresses().values())
			if(fortress.getSiege().checkIfInZone(x, y))
				return fortress.getSiege();
		return null;
	}

	public static void spawnFlagPoles(Fortress fortress)
	{
		for(SiegeSpawn _sp : _flagPoleSpawnList.get(fortress.getId()))
		{
			String line = "FlagPole;" + _sp.getNpcId() + ";" + _sp.getLoc().getX() + ";" + _sp.getLoc().getY() + ";" + _sp.getLoc().getZ() + ";3;none;0;0";
			L2StaticObjectInstance flagPole = StaticObjectsTable.parse(line);
			fortress.getSiege().addFlagPole(flagPole);
		}
	}
}