package l2r.gameserver.instancemanager;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.concurrent.ScheduledFuture;
import java.util.logging.Logger;

import javolution.util.FastList;
import javolution.util.FastMap;
import l2r.Config;
import l2r.database.DatabaseUtils;
import l2r.database.FiltredPreparedStatement;
import l2r.database.L2DatabaseFactory;
import l2r.database.ThreadConnection;
import l2r.database.mysql;
import l2r.gameserver.ThreadPoolManager;
import l2r.gameserver.model.L2Spawn;
import l2r.gameserver.model.instances.L2RaidBossInstance;
import l2r.gameserver.tables.GmListTable;
import l2r.gameserver.tables.NpcTable;
import l2r.gameserver.tables.SpawnTable;
import l2r.gameserver.templates.L2NpcTemplate;
import l2r.gameserver.templates.StatsSet;
import l2r.util.Util;

@SuppressWarnings( { "nls", "unqualified-field-access", "boxing" })
public class RaidBossSpawnManager
{

	private static Logger _log = Logger.getLogger(RaidBossSpawnManager.class.getName());

	private static RaidBossSpawnManager _instance;

	protected static Map<Integer, L2RaidBossInstance> _bosses;
	protected static Map<Integer, L2Spawn> _spawntable;
	protected static Map<Integer, StatsSet> _storedInfo;
	protected static Map<Integer, ScheduledFuture> _schedules;

	protected static Map<Integer, FastMap<Integer, Integer>> _points;

	public static enum StatusEnum
	{
		ALIVE,
		DEAD,
		UNDEFINED
	}

	private RaidBossSpawnManager()
	{
		_instance = this;
		if(!Config.DONTLOADSPAWN)
			reloadBosses();
	}

	public void reloadBosses()
	{
		fillSpawnTable();
		fillPointsTable();

		calculateRanking();
	}

	public static RaidBossSpawnManager getInstance()
	{
		if(_instance == null)
			new RaidBossSpawnManager();
		return _instance;
	}

	private void fillPointsTable()
	{
		_points = new FastMap<Integer, FastMap<Integer, Integer>>();

		FastList<Integer> _owners = new FastList<Integer>();

		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		ResultSet rset = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();

			//read raidboss points
			statement = con.prepareStatement("SELECT DISTINCT owner_id FROM `raidboss_points`");
			rset = statement.executeQuery();
			while(rset.next())
				_owners.add(rset.getInt("owner_id"));
			DatabaseUtils.closeDatabaseSR(statement, rset);

			statement = con.prepareStatement("SELECT * FROM `raidboss_points` WHERE `owner_id`=?");
			for(FastList.Node<Integer> n = _owners.head(), end = _owners.tail(); (n = n.getNext()) != end;)
			{
				int ownerId = n.getValue();
				FastMap<Integer, Integer> tmpScore = new FastMap<Integer, Integer>();
				statement.setInt(1, ownerId);
				rset = statement.executeQuery();
				while(rset.next())
					if(rset.getInt("boss_id") != -1 || rset.getInt("boss_id") != 0)
						tmpScore.put(rset.getInt("boss_id"), rset.getInt("points"));

				DatabaseUtils.closeResultSet(rset);
				_points.put(ownerId, tmpScore);
			}
		}
		catch(SQLException e)
		{
			_log.warning("RaidBossSpawnManager: Couldnt load raidboss points");
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		finally
		{
			DatabaseUtils.closeDatabaseCSR(con, statement, rset);
		}
	}

	private void fillSpawnTable()
	{
		_bosses = new FastMap<Integer, L2RaidBossInstance>();
		_schedules = new FastMap<Integer, ScheduledFuture>();
		_storedInfo = new FastMap<Integer, StatsSet>();
		_spawntable = new FastMap<Integer, L2Spawn>();

		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		ResultSet rset = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();

			//read raidboss status
			statement = con.prepareStatement("SELECT * FROM `raidboss_status`");
			rset = statement.executeQuery();
			while(rset.next())
			{
				StatsSet info = new StatsSet();
				info.set("current_hp", rset.getDouble("current_hp"));
				info.set("current_mp", rset.getDouble("current_mp"));

				_storedInfo.put(rset.getInt("id"), info);
			}
			DatabaseUtils.closeDatabaseSR(statement, rset);
			//read raidboss spawns
			mysql.set("DELETE FROM `kill_status` WHERE `respawn_time` < " + System.currentTimeMillis() / 1000);
			statement = con.prepareStatement("SELECT count, npc_templateid, s.locx, s.locy, s.locz, heading, respawn_delay, s.loc_id, ks.respawn_time, periodOfDay FROM `spawnlist` s LEFT JOIN `kill_status` AS ks ON(npc_templateid=ks.id and s.locx=ks.locx and s.locy=ks.locy and s.locz=ks.locz and s.loc_id=ks.loc_id) WHERE npc_templateid IN (SELECT id FROM `npc` WHERE type='L2RaidBoss' OR type='L2Boss') ORDER by npc_templateid");
			rset = statement.executeQuery();

			L2Spawn spawnDat;
			L2NpcTemplate template1;

			int npcId;

			while(rset.next())
			{
				npcId = rset.getInt("npc_templateid");
				template1 = getValidTemplate(npcId);
				if(template1 != null)
				{
					spawnDat = new L2Spawn(template1);
					spawnDat.setAmount(rset.getInt("count"));
					spawnDat.setLocx(rset.getInt("locx"));
					spawnDat.setLocy(rset.getInt("locy"));
					spawnDat.setLocz(rset.getInt("locz"));
					spawnDat.setHeading(rset.getInt("heading"));
					spawnDat.setRespawnDelay(rset.getInt("respawn_delay"));
					if(rset.getString("respawn_time") == null) 
						spawnDat.setRespawnTime(0);
					else
						spawnDat.setRespawnTime(rset.getInt("respawn_time"));
					spawnDat.setRespawnDelay(rset.getInt("respawn_delay"));
					spawnDat.setLocation(rset.getInt("loc_id"));

					addNewSpawn(spawnDat, false);
				}
				else
					_log.warning("RaidBossSpawnManager: Could not load raidboss #" + npcId + " from DB");
			}
			_log.info("RaidBossSpawnManager: Loaded " + _storedInfo.size() + " Statuses");
			_log.info("RaidBossSpawnManager: Loaded " + _bosses.size() + " Instances");
			_log.info("RaidBossSpawnManager: Scheduled " + _schedules.size() + " Instances");
		}
		catch(SQLException e)
		{
			_log.warning("RaidBossSpawnManager: Couldnt load raidboss spawnlist");
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		finally
		{
			DatabaseUtils.closeDatabaseCSR(con, statement, rset);
		}
	}

	public void calculateRanking()
	{
		FastMap<Integer, Integer> tmpRanking = new FastMap<Integer, Integer>();
		FastMap<Integer, FastMap<Integer, Integer>> tmpPoints = new FastMap<Integer, FastMap<Integer, Integer>>();

		for(int ownerId : _points.keySet())
		{
			FastMap<Integer, Integer> tmpPoint = new FastMap<Integer, Integer>();
			tmpPoint = _points.get(ownerId);
			int totalPoints = 0;

			for(int bossId : tmpPoint.keySet())
				if(bossId != -1 && bossId != 0)
					totalPoints += tmpPoint.get(bossId);

			// no need to store players w/o points
			if(totalPoints != 0)
			{
				tmpPoint.remove(0);
				tmpPoint.put(0, totalPoints);
				tmpPoints.put(ownerId, tmpPoint);

				tmpRanking.put(ownerId, totalPoints);
			}
		}

		List<Map.Entry<Integer, Integer>> list = new Vector<Map.Entry<Integer, Integer>>(tmpRanking.entrySet());

		// descending
		Collections.sort(list, new Comparator<Map.Entry<Integer, Integer>>(){
			public int compare(Map.Entry<Integer, Integer> entry, Map.Entry<Integer, Integer> entry1)
			{
				return entry.getValue().equals(entry1.getValue()) ? 0 : entry.getValue() < entry1.getValue() ? 1 : -1;
			}
		});

		int ranking = 1;
		for(Map.Entry<Integer, Integer> entry : list)
		{
			FastMap<Integer, Integer> tmpPoint = new FastMap<Integer, Integer>();
			tmpPoint = tmpPoints.get(entry.getKey());

			tmpPoint.remove(-1);
			tmpPoint.put(-1, ranking);

			tmpPoints.remove(entry.getKey());
			tmpPoints.put(entry.getKey(), tmpPoint);

			ranking++;
		}

		_points.clear();
		_points = tmpPoints;
	}

	public void addPoints(int ownerId, int bossId, int points)
	{
		FastMap<Integer, Integer> tmpPoint = new FastMap<Integer, Integer>();
		tmpPoint = _points.get(ownerId);
		_points.remove(ownerId);

		if(tmpPoint == null || tmpPoint.isEmpty())
		{
			tmpPoint = new FastMap<Integer, Integer>();
			tmpPoint.put(bossId, points);
		}
		else
		{
			int currentPoins = tmpPoint.containsKey(bossId) ? tmpPoint.get(bossId).intValue() : 0;

			tmpPoint.remove(bossId);
			tmpPoint.put(bossId, currentPoins == 0 ? points : currentPoins + points);
		}
		_points.put(ownerId, tmpPoint);
	}

	public void updateStatus(L2RaidBossInstance raidboss, boolean isBossDead)
	{
		if(_storedInfo.containsKey(raidboss.getNpcId()))
			_storedInfo.remove(raidboss.getNpcId());

		if(isBossDead)
		{
			raidboss.setRaidStatus(StatusEnum.DEAD);
			long respawn_delay = raidboss.getSpawn().getRespawnDelay() * 1000L;
			if(respawn_delay > 0)
			{
				raidboss.getSpawn().updateInDb();
				raidboss.getSpawn().setRespawnTime((int) (System.currentTimeMillis() + respawn_delay) / 1000);
				raidboss.getSpawn().updateInDb();

				ScheduledFuture futureSpawn = ThreadPoolManager.getInstance().scheduleGeneral(new SpawnSchedule(raidboss.getNpcId()), respawn_delay);
				_schedules.put(raidboss.getNpcId(), futureSpawn);

				_log.info("RaidBossSpawnManager: Scheduled " + raidboss.getName() + " for respawn in " + Util.formatTime(respawn_delay / 1000));
			}
		}
		else
		{
			raidboss.setRaidStatus(StatusEnum.ALIVE);

			if(raidboss.getCurrentHp() == raidboss.getMaxHp() && raidboss.getCurrentMp() == raidboss.getMaxMp())
				return;

			StatsSet info = new StatsSet();
			info.set("current_hp", raidboss.getCurrentHp());
			info.set("current_mp", raidboss.getCurrentMp());

			_storedInfo.put(raidboss.getNpcId(), info);
		}
	}

	private class SpawnSchedule implements Runnable
	{
		private int bossId;

		public SpawnSchedule(int npcId)
		{
			bossId = npcId;
		}

		public void run()
		{
			L2RaidBossInstance raidboss = null;

			if(bossId == 25328)
				raidboss = DayNightSpawnManager.getInstance().handleNightBoss(_spawntable.get(bossId));
			else
				raidboss = (L2RaidBossInstance) _spawntable.get(bossId).doSpawn(true);

			if(raidboss != null)
			{
				raidboss.setRaidStatus(StatusEnum.ALIVE);
				GmListTable.broadcastMessageToGMs("Spawning RaidBoss " + raidboss.getName());

				_bosses.put(bossId, raidboss);
			}
			_schedules.remove(bossId);
		}
	}

	public void addNewSpawn(L2Spawn spawnDat, boolean storeInDb)
	{
		if(spawnDat == null)
			return;

		int bossId = spawnDat.getNpcId();
		if(_spawntable.containsKey(bossId))
			return;

		SpawnTable.getInstance().addNewSpawn(spawnDat, storeInDb, null);

		if(System.currentTimeMillis() > spawnDat.getRespawnTime() * 1000L)
		{
			L2RaidBossInstance raidboss = null;

			if(bossId == 25328)
				raidboss = DayNightSpawnManager.getInstance().handleNightBoss(spawnDat);
			else
				raidboss = (L2RaidBossInstance) spawnDat.doSpawn(true);

			if(raidboss != null)
			{
				if(_storedInfo.containsKey(bossId))
				{
					StatsSet info = _storedInfo.get(bossId);
					if(info.getDouble("current_hp") == raidboss.getMaxHp() && info.getDouble("current_mp") == raidboss.getMaxMp())
						_storedInfo.remove(bossId);
					else
					{
						raidboss.setCurrentHp(info.getDouble("current_hp"));
						raidboss.setCurrentMp(info.getDouble("current_mp"));
					}
				}
				raidboss.setRaidStatus(StatusEnum.ALIVE);

				_bosses.put(bossId, raidboss);
			}
		}
		else
		{
			long spawnTime = spawnDat.getRespawnTime() * 1000L - System.currentTimeMillis();
			ScheduledFuture futureSpawn = ThreadPoolManager.getInstance().scheduleGeneral(new SpawnSchedule(bossId), spawnTime);
			_schedules.put(bossId, futureSpawn);
		}

		_spawntable.put(bossId, spawnDat);
	}

	public void deleteSpawn(L2Spawn spawnDat, boolean updateDb)
	{
		if(spawnDat == null)
			return;
		if(!_spawntable.containsKey(spawnDat.getNpcId()))
			return;

		int bossId = spawnDat.getNpcId();

		SpawnTable.getInstance().deleteSpawn(spawnDat, updateDb);
		_spawntable.remove(bossId);

		if(_bosses.containsKey(bossId))
			_bosses.remove(bossId);

		if(_schedules.containsKey(bossId))
		{
			ScheduledFuture f = _schedules.get(bossId);
			f.cancel(true);
			_schedules.remove(bossId);
		}

		if(_storedInfo.containsKey(bossId))
			_storedInfo.remove(bossId);

		if(updateDb)
		{
			ThreadConnection con = null;
			FiltredPreparedStatement statement = null;
			try
			{
				con = L2DatabaseFactory.getInstance().getConnection();
				statement = con.prepareStatement("DELETE FROM raidboss_status WHERE id=?");
				statement.setInt(1, bossId);
				statement.execute();
			}
			catch(Exception e)
			{
				// problem with deleting spawn
				_log.warning("RaidBossSpawnManager: Could not remove raidboss #" + bossId + " from DB: " + e);
			}
			finally
			{
				DatabaseUtils.closeDatabaseCS(con, statement);
			}
		}
	}

	private void emptyStatusTable()
	{
		if(!mysql.set("DELETE FROM `raidboss_status`"))
			_log.warning("RaidBossSpawnManager: Couldnt empty raidboss_status table");
	}

	private void emptyScoresTable()
	{
		if(!mysql.set("DELETE FROM `raidboss_points`"))
			_log.warning("RaidBossSpawnManager: Couldnt empty raidboss_points table");
	}

	private void updateStatusDb()
	{
		emptyStatusTable();

		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;

		for(Integer bossId : _bosses.keySet())
			try
			{
				con = L2DatabaseFactory.getInstance().getConnection();

				L2RaidBossInstance raidboss = _bosses.get(bossId);

				if(raidboss == null)
					continue;

				if(raidboss.getRaidStatus().equals(StatusEnum.ALIVE))
					updateStatus(raidboss, false);

				StatsSet info = _storedInfo.get(bossId);

				if(info == null)
					continue;

				statement = con.prepareStatement("INSERT INTO `raidboss_status` (id, current_hp, current_mp) VALUES (?,?,?)");
				statement.setInt(1, bossId);
				statement.setDouble(2, info.getDouble("current_hp"));
				statement.setDouble(3, info.getDouble("current_mp"));
				statement.execute();

				_log.info("RaidBossSpawnManager: Saved status for raidboss " + raidboss.getName());
			}
			catch(SQLException e)
			{
				_log.warning("RaidBossSpawnManager: Couldnt update raidboss_status table");
			}
			finally
			{
				DatabaseUtils.closeDatabaseCS(con, statement);
			}
	}

	private void updatePointsDb()
	{
		emptyScoresTable();

		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;

		for(int ownerId : _points.keySet())
			try
			{
				con = L2DatabaseFactory.getInstance().getConnection();

				FastMap<Integer, Integer> tmpPoint = _points.get(ownerId);
				if(tmpPoint == null || tmpPoint.isEmpty())
					continue;

				for(int bossId : tmpPoint.keySet())
				{
					if(bossId == -1 || bossId == 0)
						continue;

					int points = tmpPoint.get(bossId);
					if(points == 0)
						continue;

					statement = con.prepareStatement("INSERT INTO `raidboss_points` (owner_id, boss_id, points) VALUES (?,?,?)");
					statement.setInt(1, ownerId);
					statement.setInt(2, bossId);
					statement.setInt(3, points);
					statement.execute();
				}
			}
			catch(SQLException e)
			{
				_log.warning("RaidBossSpawnManager: Couldnt update raidboss_points table");
			}
			finally
			{
				DatabaseUtils.closeDatabaseCS(con, statement);
			}
	}

	public String[] getAllRaidBossStatus()
	{
		String[] msg = new String[_bosses == null ? 0 : _bosses.size()];

		if(_bosses == null)
		{
			msg[0] = "None";
			return msg;
		}

		int index = 0;

		for(Integer i : _bosses.keySet())
		{
			L2RaidBossInstance raidboss = _bosses.get(i);

			msg[index] = raidboss.getName() + ": " + raidboss.getRaidStatus().name();
			index++;
		}

		return msg;
	}

	public String getRaidBossStatus(int bossId)
	{
		String msg = "RaidBoss Status....\n";

		if(_bosses == null)
		{
			msg += "None";
			return msg;
		}

		if(_bosses.containsKey(bossId))
		{
			L2RaidBossInstance raidboss = _bosses.get(bossId);

			msg += raidboss.getName() + ": " + raidboss.getRaidStatus().name();
		}

		return msg;
	}

	public StatusEnum getRaidBossStatusId(int bossId)
	{
		if(_bosses.containsKey(bossId))
			return _bosses.get(bossId).getRaidStatus();
		else if(_schedules.containsKey(bossId))
			return StatusEnum.DEAD;
		else
			return StatusEnum.UNDEFINED;
	}

	public L2NpcTemplate getValidTemplate(int bossId)
	{
		L2NpcTemplate template = NpcTable.getTemplate(bossId);
		if(template == null)
			return null;
		if(!template.type.equalsIgnoreCase("L2RaidBoss") && !template.type.equalsIgnoreCase("L2Boss"))
			return null;
		return template;
	}

	public void notifySpawnNightBoss(L2RaidBossInstance raidboss)
	{
		raidboss.setRaidStatus(StatusEnum.ALIVE);
		_bosses.put(raidboss.getNpcId(), raidboss);
		GmListTable.broadcastMessageToGMs("Spawning night RaidBoss " + raidboss.getName());
	}

	public boolean isDefined(int bossId)
	{
		return _spawntable.containsKey(bossId);
	}

	public Map<Integer, L2RaidBossInstance> getBosses()
	{
		return _bosses;
	}

	public L2RaidBossInstance getBoss(int bossId)
	{
		return _bosses.get(bossId);
	}

	public Map<Integer, L2Spawn> getSpawnTable()
	{
		return _spawntable;
	}

	public Map<Integer, FastMap<Integer, Integer>> getPoints()
	{
		return _points;
	}

	public FastMap<Integer, Integer> getPointsByOwnerId(int ownerId)
	{
		return _points.get(ownerId);
	}

	/**
	 * Saves all raidboss status and then clears all info from memory,
	 * including all schedules.
	 */
	public void cleanUp()
	{
		updateStatusDb();
		updatePointsDb();

		_bosses.clear();
		_storedInfo.clear();
		_spawntable.clear();
		_points.clear();

		if(_schedules != null)
			for(Integer bossId : _schedules.keySet())
			{
				ScheduledFuture f = _schedules.get(bossId);
				f.cancel(true);
			}
		_schedules.clear();

		_log.fine("RaidBossSpawnManager: All raidboss info saved!");
	}

	public boolean isScheduled(int bossId)
	{
		return _schedules.containsKey(bossId);
	}
}