package l2r.gameserver.model;

import static l2r.gameserver.ai.CtrlIntention.AI_INTENTION_ACTIVE;

import java.awt.Point;
import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Vector;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.CopyOnWriteArraySet;
import java.util.concurrent.Future;
import java.util.logging.Level;
import java.util.logging.Logger;

import javolution.util.FastList;
import javolution.util.FastMap;
import javolution.util.FastTable;
import l2r.Config;
import l2r.extensions.listeners.MethodCollection;
import l2r.extensions.listeners.PropertyCollection;
import l2r.extensions.multilang.CustomMessage;
import l2r.gameserver.GameTimeController;
import l2r.gameserver.ThreadPoolManager;
import l2r.gameserver.ai.CtrlEvent;
import l2r.gameserver.ai.CtrlIntention;
import l2r.gameserver.ai.L2CharacterAI;
import l2r.gameserver.ai.L2PlayableAI.nextAction;
import l2r.gameserver.cache.Msg;
import l2r.gameserver.geodata.network.geoclient.GeoClient;
import l2r.gameserver.handler.IScriptHandler;
import l2r.gameserver.handler.ScriptHandler;
import l2r.gameserver.instancemanager.DimensionalRiftManager;
import l2r.gameserver.model.L2Effect.EffectType;
import l2r.gameserver.model.L2Skill.SkillTargetType;
import l2r.gameserver.model.L2Skill.SkillType;
import l2r.gameserver.model.L2Zone.ZoneType;
import l2r.gameserver.model.entity.Duel;
import l2r.gameserver.model.instances.L2ItemInstance;
import l2r.gameserver.model.instances.L2MinionInstance;
import l2r.gameserver.model.instances.L2NpcInstance;
import l2r.gameserver.model.instances.L2PetInstance;
import l2r.gameserver.model.instances.L2SummonInstance;
import l2r.gameserver.model.quest.QuestState;
import l2r.gameserver.serverpackets.Attack;
import l2r.gameserver.serverpackets.AutoAttackStart;
import l2r.gameserver.serverpackets.AutoAttackStop;
import l2r.gameserver.serverpackets.ChangeMoveType;
import l2r.gameserver.serverpackets.ChangeWaitType;
import l2r.gameserver.serverpackets.CharMoveToLocation;
import l2r.gameserver.serverpackets.FlyToLocation;
import l2r.gameserver.serverpackets.L2GameServerPacket;
import l2r.gameserver.serverpackets.MagicSkillCanceled;
import l2r.gameserver.serverpackets.MagicSkillLaunched;
import l2r.gameserver.serverpackets.MagicSkillUse;
import l2r.gameserver.serverpackets.MoveToLocationInVehicle;
import l2r.gameserver.serverpackets.MyTargetSelected;
import l2r.gameserver.serverpackets.PartyMemberPosition;
import l2r.gameserver.serverpackets.Revive;
import l2r.gameserver.serverpackets.SetupGauge;
import l2r.gameserver.serverpackets.ShortBuffStatusUpdate;
import l2r.gameserver.serverpackets.StatusUpdate;
import l2r.gameserver.serverpackets.StopMove;
import l2r.gameserver.serverpackets.SystemMessage;
import l2r.gameserver.serverpackets.TargetUnselected;
import l2r.gameserver.serverpackets.TeleportToLocation;
import l2r.gameserver.serverpackets.ValidateLocation;
import l2r.gameserver.serverpackets.FlyToLocation.FlyType;
import l2r.gameserver.skills.Calculator;
import l2r.gameserver.skills.Env;
import l2r.gameserver.skills.Formulas;
import l2r.gameserver.skills.Stats;
import l2r.gameserver.skills.effects.ResistType;
import l2r.gameserver.skills.funcs.Func;
import l2r.gameserver.templates.L2CharTemplate;
import l2r.gameserver.templates.L2NpcTemplate;
import l2r.gameserver.templates.L2Weapon;
import l2r.util.Location;
import l2r.util.Rnd;
import l2r.util.Util;

/**
 * Mother class of all character objects of the world (PC, NPC...)<BR>
 * <BR>
 * L2Character :<BR>
 * <BR>
 * <li>L2CastleGuardInstance</li>
 * <li>L2DoorInstance</li>
 * <li>L2NpcInstance</li>
 * <li>L2PlayableInstance </li>
 * <BR>
 * <BR>
 *
 * <B><U> Concept of L2CharTemplate</U> :</B><BR>
 * <BR>
 * Each L2Character owns generic and static properties (ex : all Keltir have the
 * same number of HP...). All of those properties are stored in a different
 * template for each type of L2Character. Each template is loaded once in the
 * server cache memory (reduce memory use). When a new instance of L2Character
 * is spawned, server just create a link between the instance and the template.
 * This link is stored in <B>_template</B><BR>
 */
public abstract class L2Character extends L2Object
{
    public L2Character getFollowTarget()
	{
		if(followTarget == null)
			return null;

		L2Character target = followTarget.get();
		if(target == null)
			followTarget = null;
		return target;
	}

	public void setFollowTarget(L2Character target)
	{
		followTarget = target == null ? null : new WeakReference<L2Character>(target);
	}

    public Future<?> _moveTask;
    public void stopMove(boolean validate)
	{
		if(isMoving)
		{
			synchronized (_targetRecorder)
			{
				isMoving = false;
				if(_moveTask != null)
				{
					_moveTask.cancel(false);
					_moveTask = null;
				}
				_targetRecorder.clear();
			}

			if(!isInBoat())
				isMoving = false;
			setXYZ(getX(), getY(), GeoClient.getInstance().getHeight(getX(), getY(), getZ()));
			if(!isInBoat())
				broadcastPacket(new StopMove(this));
			if(!isInBoat())
				broadcastPacket(new ValidateLocation(this));
			else
			{
				if(isPlayer)
				{
					L2Player player = this.getPlayer();
					sendPacket(player.getBoat().getCheckPacket(player));
				}
			}
		}

		if(isFollow)
		{
			isFollow = false;
		}
	}
    class AltMagicUseTask implements Runnable
	{
		public final L2Skill _skill;
		public L2Character _target;

		public AltMagicUseTask(L2Character target, L2Skill skill)
		{
			_skill = skill;
			_target = target;
		}

		public void run()
		{
			altOnMagicUseTimer(_target, _skill);
		}
	}

	public class CancelAttackStance implements Runnable
	{
		public void run()
		{
			broadcastPacket(new AutoAttackStop(getObjectId()));
			_stanceTask.cancel(false);
			_stanceTask = null;
		}
	}

	/** Task lauching the function enableSkill() */
	private class EnableSkill implements Runnable
	{
		int _skillId;

		public EnableSkill(int skillId)
		{
			_skillId = skillId;
		}

		public void run()
		{
			try
			{
				enableSkill(_skillId);
			}
			catch(Throwable e)
			{
				_log.log(Level.SEVERE, "", e);
			}
		}
	}

	private class HitTask implements Runnable
	{
		boolean _crit;
		int _damage;
		L2Character _hitTarget;
		boolean _miss;
		boolean _shld;
		boolean _soulshot;
		boolean _unchargeSS;

		public HitTask(L2Character target, int damage, boolean crit, boolean miss, boolean soulshot, boolean shld, boolean unchargeSS)
		{
			_hitTarget = target;
			_damage = damage;
			_crit = crit;
			_shld = shld;
			_miss = miss;
			_soulshot = soulshot;
			_unchargeSS = unchargeSS;
		}

		public void run()
		{
			try
			{
				onHitTimer(_hitTarget, _damage, _crit, _miss, _soulshot, _shld, _unchargeSS);
			}
			catch(Throwable e)
			{
				_log.log(Level.SEVERE, "", e);
			}
		}
	}

	/** Task lauching the function onMagicUseTimer() */
	class MagicUseTask implements Runnable
	{
		public final boolean _forceUse;

		public MagicUseTask(L2Character target, L2Skill skill, boolean forceUse)
		{
			_castingSkill = skill;
			setCastingTarget(target);
			_forceUse = forceUse;
		}

		public void run()
		{
			onMagicUseTimer(getCastingTarget(), _castingSkill, _forceUse);
		}
	}

	class MagicLaunchedTask implements Runnable
	{
		public final boolean _forceUse;
		public final L2Character _this;

		public MagicLaunchedTask(L2Character caster, boolean forceUse)
		{
			_this = caster;
			_forceUse = forceUse;
		}

		public void run()
		{
			FastList<L2Character> targets = _castingSkill.getTargets(_this, getCastingTarget(), _forceUse);
			broadcastPacket(new MagicSkillLaunched(_objectId, _castingSkill.getDisplayId(), _castingSkill.getLevel(), targets));
		}
	}

	class FlyToLocationTask implements Runnable
	{
		L2Object _target;
		L2Character _actor;
		L2Skill _skill;

		public FlyToLocationTask(L2Character actor, L2Object target, L2Skill skill)
		{
			_actor = actor;
			_target = target;
			_skill = skill;
		}

		public void run()
		{
			broadcastPacket(new FlyToLocation(_actor, _target.getLoc(), _skill.getFlyType()));
		}
	}

	public class IntentionTask implements Runnable
	{
		private final CtrlIntention _intention;
		private final Object _arg0, _arg1;

		public IntentionTask(CtrlIntention evt, Object arg0, Object arg1)
		{
			_intention = evt;
			_arg0 = arg0;
			_arg1 = arg1;
		}

		public void run()
		{
			try
			{
				getAI().setIntention(_intention, _arg0, _arg1);
			}
			catch(Throwable t)
			{
				_log.warning("Exception " + t);
				t.printStackTrace();
			}
		}
	}

	/** Task of AI notification */
	public class NotifyAITask implements Runnable
	{
		private final CtrlEvent _evt;

		public NotifyAITask(CtrlEvent evt)
		{
			_evt = evt;
		}

		public void run()
		{
			try
			{
				getAI().notifyEvent(_evt, null, null);
			}
			catch(Throwable t)
			{
				_log.warning("Exception " + t);
				t.printStackTrace();
			}
		}
	}

	/** Task of HP/MP/CP regeneration */
	class RegenTask implements Runnable
	{
		public void run()
		{
			try
			{
				if(isPlayer && !(((L2Player) L2Character.this).isConnected() || ((L2Player) L2Character.this).isInOfflineMode()))
				{
					stopHpMpRegeneration();
					return;
				}

				if(isDead() || _currentHp >= getMaxHp() && _currentMp >= getMaxMp() && _currentCp >= getMaxCp())
				{
					stopHpMpRegeneration();
					return;
				}

				double addHp = 0;
				double addMp = 0;

				// Caculate the HP regen rate for this L2Character
				if(_currentHp < getMaxHp())
					addHp += Formulas.calcHpRegen(L2Character.this);

				// Caculate the HP regen rate for this L2Character
				if(_currentMp < getMaxMp())
					addMp += Formulas.calcMpRegen(L2Character.this);

				// Added regen bonus when character is sitting
				if(isPlayer && Config.REGEN_SIT_WAIT)
				{
					L2Player pl = (L2Player) L2Character.this;
					if(pl.isSitting())
					{
						pl.updateWaitSitTime();
						if(pl.getWaitSitTime() > 5)
						{
							addHp += pl.getWaitSitTime();
							addMp += pl.getWaitSitTime();
						}
					}
				}
				else if(isRaid)
				{
					addHp *= Config.RATE_RAID_REGEN;
					addMp *= Config.RATE_RAID_REGEN;
				}

				// Modify the current HP and MP of the L2Character and broadcast
				// Server->Client packet StatusUpdate
				setCurrentHpMp(_currentHp + addHp, _currentMp + addMp);

				if(_currentCp < getMaxCp())
					// Caculate the CP regen rate for this L2Character
					// Modify the current CP of the L2Character and broadcast
					// Server->Client packet StatusUpdate
					setCurrentCp(getCurrentCp() + Formulas.calcCpRegen(L2Character.this));
			}
			catch(Throwable e)
			{
				_log.log(Level.SEVERE, "", e);
			}
		}
	}

	protected static final Logger _log = Logger.getLogger(L2Character.class.getName());

	public static final int ABNORMAL_EFFECT_BLEEDING = 0x00000001;
	public static final int ABNORMAL_EFFECT_POISON = 0x00000002;
	public static final int ABNORMAL_EFFECT_UNKNOWN_3 = 0x00000004; // блид
	public static final int ABNORMAL_EFFECT_UNKNOWN_4 = 0x00000008; // блид

	public static final int ABNORMAL_EFFECT_AFFRAID = 0x00000010;
	public static final int ABNORMAL_EFFECT_CONFUSED = 0x00000020;
	public static final int ABNORMAL_EFFECT_STUN = 0x00000040;
	public static final int ABNORMAL_EFFECT_SLEEP = 0x00000080;

	public static final int ABNORMAL_EFFECT_MUTED = 0x00000100;
	public static final int ABNORMAL_EFFECT_ROOT = 0x00000200;
	public static final int ABNORMAL_EFFECT_HOLD_1 = 0x00000400;
	public static final int ABNORMAL_EFFECT_HOLD_2 = 0x00000800;

	public static final int ABNORMAL_EFFECT_UNKNOWN_13 = 0x00001000; // пусто
	public static final int ABNORMAL_EFFECT_BIG_HEAD = 0x00002000;
	public static final int ABNORMAL_EFFECT_FLAME = 0x00004000;
	public static final int ABNORMAL_EFFECT_UNKNOWN_16 = 0x00008000; // пусто

	public static final int ABNORMAL_EFFECT_GROW = 0x00010000; // эффект роста
	public static final int ABNORMAL_EFFECT_FLOATING_ROOT = 0x00020000; // висит в воздухе, на спине
	public static final int ABNORMAL_EFFECT_DANCE_STUNNED = 0x00040000; // танцует со звездочками над головой
	public static final int ABNORMAL_EFFECT_FIREROOT_STUN = 0x00080000; // красная аура со звездочками над головой

	public static final int ABNORMAL_EFFECT_SILENT_MOVE = 0x00100000;
	public static final int ABNORMAL_EFFECT_IMPRISIONING_1 = 0x00200000; // синяя аура на уровне пояса
	public static final int ABNORMAL_EFFECT_IMPRISIONING_2 = 0x00400000; // синяя аура на уровне пояса
	public static final int ABNORMAL_EFFECT_MAGIC_CIRCLE = 0x00800000; // большой синий круг вокруг чара

	public static final int ABNORMAL_EFFECT_TEST_1_1 = 0x01000000; // небольшая ледяная аура
	public static final int ABNORMAL_EFFECT_TEST_1_2 = 0x02000000; // землетрясение
	public static final int ABNORMAL_EFFECT_TEST_1_4 = 0x04000000; // пусто
	public static final int ABNORMAL_EFFECT_TEST_1_8 = 0x08000000; // пусто

	public static final int ABNORMAL_EFFECT_TEST_2_1 = 0x10000000; // пусто
	public static final int ABNORMAL_EFFECT_TEST_2_2 = 0x20000000; // пусто
	public static final int ABNORMAL_EFFECT_TEST_2_4 = 0x40000000; // пусто
	public static final int ABNORMAL_EFFECT_TEST_2_8 = 0x80000000; // пусто

	private int _customEffect;

	public static final double HEADINGS_IN_PI = 10430.378350470452724949566316381;
	public static final int INTERACTION_DISTANCE = 150;

	/**
	 * List of all QuestState instance that needs to be notified of this
	 * character's death
	 */
	private ArrayList<QuestState> _NotifyQuestOfDeathList;

	/**
	 * Array containing all clients that need to be notified about hp/mp updates
	 * of the L2Character
	 */
	private CopyOnWriteArraySet<L2Character> _statusListener;

	@SuppressWarnings("unchecked")
	public Future _skillTask;
	@SuppressWarnings("unchecked")
	public Future _skillLaunchedTask;
	@SuppressWarnings("unchecked")
	public Future _altSkillTask;
	@SuppressWarnings("unchecked")
	public Future _regTask;
	@SuppressWarnings("unchecked")
	public Future _stanceTask;

	private long _stanceInited;

	private int _flagsRegenActive;
	private double _lastHpUpdate = -99999999;

	private static final int HP_REGEN_FLAG = 1;
	private static final int MP_REGEN_FLAG = 2;
	private static final int CP_REGEN_FLAG = 4;

	protected double _currentCp = 1;
	protected double _currentHp = 1;
	protected double _currentMp = 1;

	protected boolean _isAttackAborted;
	protected long _attackEndTime;
	protected long _attackReuseEndTime;

	/** HashMap(Integer, L2Skill) containing all skills of the L2Character */
	protected HashMap<Integer, L2Skill> _skills = new HashMap<Integer, L2Skill>();
	protected HashMap<Integer, L2Skill> _skillsOnAttack;
	protected HashMap<Integer, L2Skill> _skillsOnMagicAttack;
	protected HashMap<Integer, L2Skill> _skillsOnUnderAttack;

	public L2Skill _castingSkill;
	private WeakReference<L2Character> castingTarget;

	public int _castEndTime;
	public int _castInterruptTime;
	public int _altCastEndTime;
	public int _altCastInterruptTime;

	protected long _animationEndTime;

	/** Table containing all skillId that are disabled */
	protected List<Integer> _disabledSkills;

	protected ForceBuff _forceBuff;

	/**
	 * HashMap containing all active skills effects in progress of a
	 * L2Character. The Short key of this HashMap is the L2Skill Identifier that
	 * has created the effect.
	 */
	private ConcurrentLinkedQueue<L2Effect> _effects;

	private FastList<Stats> _blockedStats;

	/** Map 32 bits (0x00000000) containing all abnormal effect in progress */
	private int _abnormalEffects;

	private boolean _flying;
	private boolean _riding;

	private boolean _fakeDeath;
	private boolean _fishing;

	protected boolean _isInvul;

	protected boolean _isPendingRevive;

	protected boolean _isTeleporting;

	protected boolean _overloaded;

	protected boolean _killedAlready;

	private long _dropDisabled;

	private byte _isBlessedByNoblesse; // Восстанавливает все бафы после смерти
	private byte _isSalvation; // Восстанавливает все бафы после смерти и полностью CP, MP, HP
	private byte _buffImmunity; // Иммунитет к бафам/дебафам
	private final HashMap<Integer, Byte> _skillMastery = new HashMap<Integer, Byte>();

	private boolean _afraid;
	private boolean _meditated;
	private boolean _muted;
	private boolean _pmuted;
	private boolean _paralyzed;
	private boolean _rooted;
	private boolean _sleeping;
	private boolean _stunned;
	private boolean _imobilised;
	private boolean _confused;
	private boolean _blocked;
	private boolean _bigHead;

	private boolean _running;

	public boolean isMoving;
	public boolean isFollow;

	protected boolean _pathfinding;
	protected boolean _forestalling;

	protected WeakReference<L2Character> followTarget;

	protected Vector<Location> _targetRecorder = new Vector<Location>();

	protected int _xDestination;
	protected int _yDestination;
	protected int _zDestination;

	protected int _xFrom;
	protected int _yFrom;
	protected int _zFrom;

	protected int _xTo;
	protected int _yTo;
	protected int _zTo;

	protected int _offset;

	protected float _xSpeedTicks;
	protected float _ySpeedTicks;

	protected int _followTimestamp;
	protected int _startMoveTick;
	protected double _previousSpeed = -1;

	private int _heading;

	private WeakReference<L2Object> target;

	/** Table of Calculators containing all used calculator */
	private Calculator[] _calculators;

	/**
	 * The link on the L2CharTemplate object containing generic and static
	 * properties of this L2Character type (ex : Max HP, Speed...)
	 */
	protected L2CharTemplate _template;
	protected L2CharTemplate _baseTemplate;
	protected L2CharacterAI _ai;

	protected String _name;
	protected String _title;

	/**
	 * Constructor of L2Character.<BR>
	 * <BR>
	 *
	 * <B><U> Concept</U> :</B><BR>
	 * <BR>
	 * Each L2Character owns generic and static properties (ex : all Keltir have
	 * the same number of HP...). All of those properties are stored in a
	 * different template for each type of L2Character. Each template is loaded
	 * once in the server cache memory (reduce memory use). When a new instance
	 * of L2Character is spawned, server just create a link between the instance
	 * and the template This link is stored in <B>_template</B><BR>
	 * <BR>
	 *
	 * <B><U> Actions</U> :</B><BR>
	 * <BR>
	 * <li>Set the _template of the L2Character </li>
	 * <li>Set _overloaded to false (the charcater can take more items)</li>
	 * <BR>
	 * <BR>
	 *
	 * <li>If L2Character is a L2NpcInstance, copy skills from template to
	 * object</li>
	 * <li>If L2Character is a L2NpcInstance, link _calculators to
	 * NPC_STD_CALCULATOR</li>
	 * <BR>
	 * <BR>
	 *
	 * <li>If L2Character is NOT a L2NpcInstance, create an empty _skills slot</li>
	 * <li>If L2Character is a L2Player or L2Summon, copy basic Calculator set
	 * to object</li>
	 * <BR>
	 * <BR>
	 *
	 * @param objectId
	 *            Identifier of the object to initialized
	 * @param template
	 *            The L2CharTemplate to apply to the object
	 */
	public L2Character(int objectId, L2CharTemplate template)
	{
		super(objectId);

		// Set its template to the new L2Character
		_template = template;
		_baseTemplate = template;

		_calculators = new Calculator[Stats.NUM_STATS];

		if(template != null && (this instanceof L2NpcInstance || this instanceof L2Summon))
			if(((L2NpcTemplate) template).getSkills().size() > 0)
				for(L2Skill skill : ((L2NpcTemplate) template).getSkills().values())
					addSkill(skill);

		Formulas.addFuncsToNewCharacter(this);
	}

	/**
	 * Abort the attack of the L2Character and send Server->Client ActionFailed
	 * packet.<BR>
	 * <BR>
	 */
	public final void abortAttack()
	{
		if(isAttackingNow())
		{
			_attackEndTime = 0;
			_isAttackAborted = true;
			sendActionFailed();
		}
	}

	/**
	 * Abort the cast of the L2Character and send Server->Client
	 * MagicSkillCanceld/ActionFailed packet.<BR>
	 * <BR>
	 */
	public final void abortCast()
	{
		if(isCastingNow())
		{
			_castEndTime = 0;
			_castInterruptTime = 0;
			if(_castingSkill != null)
				_skillMastery.remove(_castingSkill.getId());
			_castingSkill = null;
			if(_skillTask != null)
				_skillTask.cancel(true); // cancels the skill hit scheduled task
			if(_skillLaunchedTask != null)
				_skillLaunchedTask.cancel(true); // cancels the skill hit scheduled task

			if(_forceBuff != null)
				_forceBuff.delete();

			broadcastPacket(new MagicSkillCanceled(_objectId)); // broadcast packet to stop animations client-side
			sendActionFailed(); // send an "action failed" packet to the caster
			getAI().setIntention(AI_INTENTION_ACTIVE, null, null);
		}
	}

	public void absorbAndReflect(L2Character target, int damage)
	{
		if(target.getEffectByType(EffectType.PETRIFICATION) == null)
		{
			double reflect = target.calcStat(Stats.REFLECT_DAMAGE_PERCENT, 0, null, null);
			if(reflect > 0 && target.getCurrentHp() > damage)
			{
				double rdmg = damage * reflect / 100.;
				if(this instanceof L2Playable && !target.isNpc)
					reduceCurrentHp(rdmg, this, true, true, false);
				else
					reduceCurrentHp(rdmg, target, true, true, false);
				if(target.isPlayer)
				{
					SystemMessage smsg = new SystemMessage(SystemMessage.S1_HAS_GIVEN_S2_DAMAGE_OF_S3).addString(target.getName());
					if(isNpc)
						smsg.addNpcName(getNpcId());
					else
						smsg.addString(getName());
					target.sendPacket(smsg.addNumber((int) rdmg));
				}
			}
		}

		damage = (int) (damage - target.getCurrentCp());

		if(damage <= 0)
			return;

		double absorb = calcStat(Stats.ABSORB_DAMAGE_PERCENT, 0, target, null);

		if(absorb > 0)
			setCurrentHp(_currentHp + damage * absorb / 100);

		absorb = calcStat(Stats.ABSORB_DAMAGEMP_PERCENT, 0, target, null);

		if(absorb > 0)
			setCurrentMp(_currentMp + damage * absorb / 100);
	}

	public void addBlockStats(FastList<Stats> stats)
	{
		if(_blockedStats == null)
			_blockedStats = new FastList<Stats>(stats);
		else
			_blockedStats.addAll(stats);
	}

	/**
	 * Add QuestState instance that is to be notified of character's death.<BR>
	 * <BR>
	 *
	 * @param qs
	 *            The QuestState that subscribe to this event
	 *
	 */
	public void addNotifyQuestOfDeath(QuestState qs)
	{
		if(qs == null || _NotifyQuestOfDeathList != null && _NotifyQuestOfDeathList.contains(qs))
			return;
		if(_NotifyQuestOfDeathList == null)
			_NotifyQuestOfDeathList = new ArrayList<QuestState>();
		_NotifyQuestOfDeathList.add(qs);
	}

	/**
	 * Add a skill to the L2Character _skills and its Func objects to the
	 * calculator set of the L2Character.<BR>
	 * <BR>
	 *
	 * <B><U> Concept</U> :</B><BR>
	 * <BR>
	 * All skills own by a L2Character are identified in <B>_skills</B><BR>
	 * <BR>
	 *
	 * <B><U> Actions</U> :</B><BR>
	 * <BR>
	 * <li>Replace oldSkill by newSkill or Add the newSkill </li>
	 * <li>If an old skill has been replaced, remove all its Func objects of
	 * L2Character calculator set</li>
	 * <li>Add Func objects of newSkill to the calculator set of the
	 * L2Character </li>
	 * <BR>
	 * <BR>
	 *
	 * <B><U> Overriden in </U> :</B><BR>
	 * <BR>
	 * <li> L2Player : Save update in the character_skills table of the database</li>
	 * <BR>
	 * <BR>
	 *
	 * @param newSkill
	 *            The L2Skill to add to the L2Character
	 *
	 * @return The L2Skill replaced or null if just added a new L2Skill
	 *
	 */
	public L2Skill addSkill(L2Skill newSkill)
	{
		if(newSkill == null)
			return null;

		if(_skills == null)
			_skills = new HashMap<Integer, L2Skill>();

		L2Skill oldSkill = _skills.get(newSkill.getId());

		if(oldSkill != null && oldSkill.getLevel() == newSkill.getLevel())
			return null;

		// Replace oldSkill by newSkill or Add the newSkill
		_skills.put(newSkill.getId(), newSkill);

		if(newSkill.isOnAttack() || newSkill.isOnCrit())
		{
			if(_skillsOnAttack == null)
				_skillsOnAttack = new HashMap<Integer, L2Skill>();
			_skillsOnAttack.put(newSkill.getId(), newSkill);
		}

		if(newSkill.isOnMagicAttack())
		{
			if(_skillsOnMagicAttack == null)
				_skillsOnMagicAttack = new HashMap<Integer, L2Skill>();
			_skillsOnMagicAttack.put(newSkill.getId(), newSkill);
		}

		if(newSkill.isOnUnderAttack())
		{
			if(_skillsOnUnderAttack == null)
				_skillsOnUnderAttack = new HashMap<Integer, L2Skill>();
			_skillsOnUnderAttack.put(newSkill.getId(), newSkill);
		}

		// If an old skill has been replaced, remove all its Func objects
		if(oldSkill != null)
		{
			removeStatsOwner(oldSkill);
			// Завершаем все эффекты, принадлежащие старому скиллу
			FastList<L2Effect> effects = getEffectsBySkill(oldSkill);
			if(effects != null)
				for(L2Effect effect : effects)
					effect.exit();
		}

		// Add Func objects of newSkill to the calculator set of the L2Character
		addStatFuncs(newSkill.getStatFuncs(this));

		return oldSkill;
	}

	/**
	 * Add a Func to the Calculator set of the L2Character.<BR>
	 * <BR>
	 *
	 * <B><U> Concept</U> :</B><BR>
	 * <BR>
	 * A L2Character owns a table of Calculators called <B>_calculators</B>.
	 * Each Calculator (a calculator per state) own a table of Func object. A
	 * Func object is a mathematic function that permit to calculate the
	 * modifier of a state (ex : REGENERATE_HP_RATE...). To reduce cache memory
	 * use, L2NpcInstances who don't have skills share the same Calculator set
	 * called <B>NPC_STD_CALCULATOR</B>.<BR>
	 * <BR>
	 *
	 * That's why, if a L2NpcInstance is under a skill/spell effect that modify
	 * one of its state, a copy of the NPC_STD_CALCULATOR must be create in its
	 * _calculators before addind new Func object.<BR>
	 * <BR>
	 *
	 * <B><U> Actions</U> :</B><BR>
	 * <BR>
	 * <li>If _calculators is linked to NPC_STD_CALCULATOR, create a copy of
	 * NPC_STD_CALCULATOR in _calculators</li>
	 * <li>Add the Func object to _calculators</li>
	 * <BR>
	 * <BR>
	 *
	 * @param f
	 *            The Func object to add to the Calculator corresponding to the
	 *            state affected
	 */
	public final synchronized void addStatFunc(Func f)
	{
		if(f == null)
			return;

		// Select the Calculator of the affected state in the Calculator set
		int stat = f._stat.ordinal();

		if(_calculators[stat] == null)
			_calculators[stat] = new Calculator();

		_calculators[stat].addFunc(f);
	}

	/**
	 * Add a list of Funcs to the Calculator set of the L2Character.<BR>
	 * <BR>
	 *
	 * <B><U> Concept</U> :</B><BR>
	 * <BR>
	 * A L2Character owns a table of Calculators called <B>_calculators</B>.
	 * Each Calculator (a calculator per state) own a table of Func object. A
	 * Func object is a mathematic function that permit to calculate the
	 * modifier of a state (ex : REGENERATE_HP_RATE...). <BR>
	 * <BR>
	 *
	 * <FONT COLOR=#FF0000><B> <U>Caution</U> : This method is ONLY for
	 * L2Player</B></FONT><BR>
	 * <BR>
	 *
	 * <B><U> Example of use </U> :</B><BR>
	 * <BR>
	 * <li> Equip an item from inventory</li>
	 * <li> Learn a new passive skill</li>
	 * <li> Use an active skill</li>
	 * <BR>
	 * <BR>
	 *
	 * @param funcs
	 *            The list of Func objects to add to the Calculator
	 *            corresponding to the state affected
	 */
	public final synchronized void addStatFuncs(Func[] funcs)
	{
		for(Func f : funcs)
			addStatFunc(f);
	}

	/**
	 * Add the object to the list of L2Character that must be informed of HP/MP
	 * updates of this L2Character.<BR>
	 * <BR>
	 *
	 * <B><U> Concept</U> :</B><BR>
	 * <BR>
	 * Each L2Character owns a list called <B>_statusListener</B> that contains
	 * all L2Player to inform of HP/MP updates. Players who must be informed are
	 * players that target this L2Character. When a RegenTask is in progress
	 * sever just need to go through this list to send Server->Client packet
	 * StatusUpdate.<BR>
	 * <BR>
	 *
	 * <B><U> Example of use </U> :</B><BR>
	 * <BR>
	 * <li> Target a PC or NPC</li>
	 * <BR>
	 * <BR>
	 *
	 * @param object
	 *            L2Character to add to the listener
	 */
	public void addStatusListener(L2Character object)
	{
		if(object == this)
			return;
		CopyOnWriteArraySet<L2Character> listeners;
		synchronized (this)
		{
			if(_statusListener == null)
				_statusListener = new CopyOnWriteArraySet<L2Character>();

			listeners = _statusListener;
		}
		listeners.add(object);
	}

	public void altOnMagicUseTimer(L2Character aimingTarget, L2Skill skill)
	{
		_altCastEndTime = 0;
		_altCastInterruptTime = 0;
		_altSkillTask = null;
		if(isAlikeDead())
			return;
		int magicId = skill.getDisplayId();
		int level = Math.max(1, getSkillDisplayLevel(skill.getId()));
		FastList<L2Character> targets = skill.getTargets(this, aimingTarget, true);
		broadcastPacket(new MagicSkillLaunched(_objectId, magicId, level, targets));
		double mpConsume2 = skill.getMpConsume2();
		if(mpConsume2 > 0)
		{
			if(_currentMp < mpConsume2)
			{
				sendPacket(Msg.NOT_ENOUGH_MP);
				return;
			}
			if(skill.isMagic())
				reduceCurrentMp(calcStat(Stats.MP_MAGIC_SKILL_CONSUME, mpConsume2, null, skill), null);
			else
				reduceCurrentMp(calcStat(Stats.MP_PHYSICAL_SKILL_CONSUME, mpConsume2, null, skill), null);
		}
		callSkill(skill, targets);
	}

	public void altUseSkill(L2Skill skill, L2Character target)
	{
		if(skill == null)
			return;
		int magicId = skill.getId();
		if(isSkillDisabled(magicId))
		{
			if(Config.ALT_SHOW_REUSE_MSG && isPlayer)
				sendPacket(new SystemMessage(SystemMessage.S1_IS_NOT_AVAILABLE_AT_THIS_TIME_BEING_PREPARED_FOR_REUSE).addSkillName(magicId, skill.getDisplayLevel()));
			return;
		}
		if(target == null)
		{
			target = skill.getAimingTarget(this);
			if(target == null)
				return;
		}

		int itemConsume[] = skill.getItemConsume();

		if(itemConsume[0] > 0)
			for(int i = 0; i < itemConsume.length; i++)
				if(!consumeItem(skill.getItemConsumeId()[i], itemConsume[i]))
				{
					sendPacket(Msg.INCORRECT_ITEM_COUNT);
					sendChanges();
					return;
				}

		int level = Math.max(1, getSkillDisplayLevel(magicId));
		_altCastEndTime = 10 + GameTimeController.getGameTicks() + skill.getHitTime() / GameTimeController.MILLIS_IN_TICK;
		_altCastInterruptTime = GameTimeController.getGameTicks() + skill.getSkillInterruptTime() / GameTimeController.MILLIS_IN_TICK;
		Formulas.calcSkillMastery(skill, this);
		long reuseDelay = Formulas.calcSkillReuseDelay(this, skill);
		broadcastPacket(new MagicSkillUse(this, target, skill.getDisplayId(), level, skill.getHitTime(), reuseDelay));
		// Не показывать сообщение для хербов и кубиков
		if(!(skill.getId() >= 4049 && skill.getId() <= 4055 || skill.getId() >= 4164 && skill.getId() <= 4166 || skill.getId() >= 2278 && skill.getId() <= 2285 || skill.getId() == 5115 || skill.getId() == 5116))
			if(!skill.isHandler())
				sendPacket(new SystemMessage(SystemMessage.USE_S1).addSkillName(magicId, (short) level));
			else
				sendPacket(new SystemMessage(SystemMessage.USE_S1).addItemName(skill.getItemConsumeId()[0]));
		// Skill reuse check
		if(reuseDelay > 30000)
			addSkillTimeStamp(skill.getId(), reuseDelay);
		if(reuseDelay > 10)
		{
			disableItem(skill, reuseDelay, reuseDelay);
			disableSkill(skill.getId(), reuseDelay);
		}
		_altSkillTask = ThreadPoolManager.getInstance().scheduleAi(new AltMagicUseTask(target, skill), skill.getHitTime(), isPlayer || isPet || isSummon);
	}

	/**
	 * Break an attack and send Server->Client ActionFailed packet and a System
	 * Message to the L2Character.<BR>
	 * <BR>
	 */
	public void breakAttack()
	{
		if(isAttackingNow())
		{
			abortAttack();

			if(isPlayer)
				sendPacket(Msg.ATTACK_FAILED);
		}
	}

	/**
	 * Break a cast and send Server->Client ActionFailed packet and a System
	 * Message to the L2Character.<BR>
	 * <BR>
	 */
	public void breakCast(boolean force)
	{
		if(isCastingNow() && (force || canAbortCast()))
		{
			abortCast();

			if(isPlayer)
				sendPacket(Msg.CASTING_HAS_BEEN_INTERRUPTED);
		}
	}

	/**
	 * Send a packet to the L2Character AND to all L2Player in the _KnownPlayers
	 * of the L2Character.<BR>
	 * <BR>
	 *
	 * <B><U> Concept</U> :</B><BR>
	 * <BR>
	 * L2Player in the detection area of the L2Character are identified in
	 * <B>_knownPlayers</B>. In order to inform other players of state
	 * modification on the L2Character, server just need to go through
	 * _knownPlayers to send Server->Client Packet<BR>
	 * <BR>
	 *
	 */
	public void broadcastPacket(L2GameServerPacket mov)
	{
		sendPacket(mov);
		broadcastPacketToOthers(mov);
	}

	/**
	 * Send a packet to all L2Player in the _KnownPlayers of the L2Character.<BR>
	 * <BR>
	 *
	 * <B><U> Concept</U> :</B><BR>
	 * <BR>
	 * L2Player in the detection area of the L2Character are identified in
	 * <B>_knownPlayers</B>. In order to inform other players of state
	 * modification on the L2Character, server just need to go through
	 * _knownPlayers to send Server->Client Packet<BR>
	 * <BR>
	 *
	 * <FONT COLOR=#FF0000><B> <U>Caution</U> : This method DOESN'T SEND
	 * Server->Client packet to this L2Character (to do this use function
	 * broadcastPacket)</B></FONT><BR>
	 * <BR>
	 *
	 */
	public final void broadcastPacketToOthers(L2GameServerPacket mov)
	{
		if(!isVisible())
			return;

		FastList<L2Player> others = L2World.getAroundPlayers(this);
		for(FastList.Node<L2Player> n = others.head(), end = others.tail(); (n = n.getNext()) != end;)
		{
			L2Player target = n.getValue();
			if(_objectId != target.getObjectId())
				target.sendPacket(mov);
		}
	}

	/**
	 * Send the Server->Client packet StatusUpdate with current HP and MP to all
	 * other L2Player to inform.<BR>
	 * <BR>
	 *
	 * <B><U> Actions</U> :</B><BR>
	 * <BR>
	 * <li>Create the Server->Client packet StatusUpdate with current HP and MP
	 * </li>
	 * <li>Send the Server->Client packet StatusUpdate with current HP and MP
	 * to all L2Character called _statusListener that must be informed of HP/MP
	 * updates of this L2Character </li>
	 * <BR>
	 * <BR>
	 *
	 * <FONT COLOR=#FF0000><B> <U>Caution</U> : This method DOESN'T SEND CP
	 * information</B></FONT><BR>
	 * <BR>
	 *
	 * <B><U> Overriden in </U> :</B><BR>
	 * <BR>
	 * <li> L2Player : Send current HP,MP and CP to the L2Player and only
	 * current HP, MP and Level to all other L2Player of the Party</li>
	 * <BR>
	 * <BR>
	 *
	 */
	public void broadcastStatusUpdate()
	{
		CopyOnWriteArraySet<L2Character> list = _statusListener;

		if(list == null || list.isEmpty())
			return;

		if(!needStatusUpdate())
		{
			if(Config.DEBUG)
				_log.info("" + getName() + ": saving statusUpdate.");
			return;
		}

		StatusUpdate su = new StatusUpdate(_objectId);
		su.addAttribute(StatusUpdate.CUR_HP, (int) _currentHp);
		su.addAttribute(StatusUpdate.CUR_MP, (int) _currentMp);
		su.addAttribute(StatusUpdate.CUR_CP, (int) _currentCp);

		for(L2Character temp : list)
			if(!Config.FORCE_STATUSUPDATE)
			{
				if(temp.getTarget() == this)
					temp.sendPacket(su);
			}
			else
				temp.sendPacket(su);
	}

	public int calcHeading(int x_dest, int y_dest)
	{
		return (int) (Math.atan2(getY() - y_dest, getX() - x_dest) * HEADINGS_IN_PI) + 32768;
	}

	/**
	 * Calculate the new value of the state with modifiers that will be applied
	 * on the targeted L2Character.<BR>
	 * <BR>
	 *
	 * <B><U> Concept</U> :</B><BR>
	 * <BR>
	 * A L2Character owns a table of Calculators called <B>_calculators</B>.
	 * Each Calculator (a calculator per state) own a table of Func object. A
	 * Func object is a mathematic function that permit to calculate the
	 * modifier of a state (ex : REGENERATE_HP_RATE...) : <BR>
	 * <BR>
	 *
	 * FuncAtkAccuracy -> Math.sqrt(_player.getDEX())*6+_player.getLevel()<BR>
	 * <BR>
	 *
	 * When the calc method of a calculator is launched, each mathematic
	 * function is called according to its priority <B>_order</B>. Indeed, Func
	 * with lowest priority order is executed firsta and Funcs with the same
	 * order are executed in unspecified order. The result of the calculation is
	 * stored in the value property of an Env class instance.<BR>
	 * <BR>
	 *
	 * @param stat
	 *            The stat to calculate the new value with modifiers
	 * @param init
	 *            The initial value of the stat before applying modifiers
	 * @param object
	 *            The L2Charcater whose properties will be used in the
	 *            calculation (ex : CON, INT...)
	 * @param skill
	 *            The L2Skill whose properties will be used in the calculation
	 *            (ex : Level...)
	 *
	 */
	public final double calcStat(Stats stat, double init, L2Character object, L2Skill skill)
	{
		int id = stat.ordinal();
		Calculator c = _calculators[id];
		// If no Func object found, no modifier is applied
		if(c == null || c.size() == 0)
			return init;
		// Create and init an Env object to pass parameters to the Calculator
		Env env = new Env();
		env.character = this;
		env.target = object;
		env.skill = skill;
		env.value = init;
		// Launch the calculation
		c.calc(env);
		if(env.value < 1 && (stat == Stats.STAT_CON || stat == Stats.STAT_DEX || stat == Stats.STAT_INT || stat == Stats.STAT_MEN || stat == Stats.STAT_STR || stat == Stats.STAT_WIT))
			env.value = 1;
		else if(env.value < 0)
			env.value = 0;

		return env.value;
	}

	/**
	 * Return the Attack Speed of the L2Character (delay (in milliseconds)
	 * before next attack).<BR>
	 * <BR>
	 */
	public int calculateAttackSpeed()
	{
		return Formulas.calcPAtkSpd(getPAtkSpd());
	}

	public void callSkill(L2Skill skill, FastList<L2Character> targets)
	{
		try
		{
			boolean needsMirrageCheck = isPlayer && getTarget() != null && skill.isOffensive();
			boolean needsOverhitFlags = skill.isOverhit();

			for(L2Character target : targets)
			{
				if(target.isMonster && needsOverhitFlags)
					target.setOverhitEnabled(true);

				if(needsMirrageCheck && Rnd.chance((int) target.calcStat(Stats.CANCEL_TARGET, 0, null, null)))
				{
					setTarget(null);
					needsMirrageCheck = false;
				}
			}

			if(skill.isOffensive() && _skillsOnMagicAttack != null)
				for(L2Skill sk : _skillsOnMagicAttack.values())
					if(Rnd.chance(sk.getChance()))
						sk.useSkill(this, targets);

			if(skill.getNegateSkill() > 0)
				for(L2Character target : targets)
					for(L2Effect e : target.getAllEffects())
					{
						L2Skill efs = e.getSkill();
						if(efs.getId() == skill.getNegateSkill() && efs.isCancelable() && efs.getPower() <= skill.getNegatePower())
							e.exit();
					}

			if(skill.isOffensive())
				startAttackStanceTask();

			skill.useSkill(this, targets);
		}
		catch(Exception e)
		{
			_log.log(Level.WARNING, "", e);
		}
	}

	/**
	 * Return True if the cast of the L2Character can be aborted.<BR>
	 * <BR>
	 */
	public final boolean canAbortCast()
	{
		return _castInterruptTime > GameTimeController.getGameTicks();
	}

	public boolean checkBlockedStat(Stats stat)
	{
		return _blockedStats != null && _blockedStats.contains(stat);
	}

	public boolean checkReflectSkill(L2Skill skill)
	{
		return Rnd.chance((int) calcStat(skill.isMagic() ? Stats.REFLECT_MAGIC_SKILL : Stats.REFLECT_PHYSIC_SKILL, 0, null, skill));
	}

	public void doCounterAttack(L2Skill skill, L2Character target)
	{
		if(skill.isMagic() || !skill.isOffensive() || skill.getCastRange() > 200)
			return;

		if(Rnd.chance((int) calcStat(Stats.COUNTER_ATTACK, 0, null, skill)))
		{
			double damage = Formulas.calcPhysDam(this, target, null, false, false, false, false);
			if(damage < 1)
				damage = 1;
			sendPacket(new SystemMessage(SystemMessage.YOU_HIT_FOR_S1_DAMAGE).addNumber((int) damage));
			target.reduceCurrentHp(damage, this, true, true, false);
		}
	}

	public void detachAI()
	{
		if(_ai != null)
			_ai.stopAITask();
		_ai = null;
	}

	public final void disableDrop(int time)
	{
		_dropDisabled = System.currentTimeMillis() + time;
	}

	/**
	 * Disable this skill id for the duration of the delay in milliseconds.
	 *
	 * @param skillId
	 * @param delay (seconds * 1000)
	 */
	public void disableSkill(int skillId, long delay)
	{
		if(_disabledSkills == null)
			_disabledSkills = Collections.synchronizedList(new FastList<Integer>());

		_disabledSkills.add(skillId);

		if(delay > 10)
			ThreadPoolManager.getInstance().scheduleAi(new EnableSkill(skillId), delay,  isPlayer || isPet || isSummon);
	}

	public void doAttack(L2Character target)
	{
		if(isAttackingNow())
			return;

		if(target == null)
			return;

		// Get the active weapon item corresponding to the active weapon
		// instance (always equipped in the right hand)
		L2Weapon weaponItem = getActiveWeaponItem();

		if(isAlikeDead() || target.isAlikeDead() || !isInRange(target, 2000))
			return;

		fireMethodInvoked(MethodCollection.onStartAttack, new Object[] { target });

		// Get the Attack Speed of the L2Character (delay (in milliseconds)
		// before next attack)
		// лимит в 0.2 секунды означает скорость атаки 2500
		int sAtk = Math.max(calculateAttackSpeed(), 200) - 50;

		// Get the Attack Reuse Delay of the L2Weapon
		int reuse = (int) (weaponItem != null ? weaponItem.getAttackReuseDelay() * getReuseModifier(target) * 26 / Math.sqrt(getPAtkSpd()) : 0) - 50;
		if(reuse > 0)
			reuse = Math.max(reuse, sAtk);

		_attackEndTime = sAtk + System.currentTimeMillis();
		_attackReuseEndTime = reuse + System.currentTimeMillis();

		// Create a Server->Client packet Attack
		Attack attack = new Attack(this, getChargedSoulShot());

		_isAttackAborted = false;

		setHeading(target, true);

		// Select the type of attack to start
		if(weaponItem == null)
			doAttackHitSimple(attack, target, weaponItem, 1., true, sAtk);
		else
			switch(weaponItem.getItemType())
			{
				case BOW:
					doAttackHitByBow(attack, target, sAtk, reuse, weaponItem);
					break;
				case POLE:
					if(calcStat(Stats.POLE_ATTACK_ANGLE, 0, null, null) == -1)
						doAttackHitSimple(attack, target, weaponItem, 1., true, sAtk);
					else
						doAttackHitByPole(attack, weaponItem, sAtk);
					break;
				case DUAL:
				case DUALFIST:
					doAttackHitByDual(attack, target, weaponItem, sAtk);
					break;
				default:
					doAttackHitSimple(attack, target, weaponItem, 1., true, sAtk);
			}

		// If the Server->Client packet Attack contains at least 1 hit, send the
		// Server->Client packet Attack
		// to the L2Character AND to all L2Player in the _KnownPlayers of the
		// L2Character
		if(attack.hasHits())
			broadcastPacket(attack);

		// Для кайтинга с луком
		if(reuse > 0)
			ThreadPoolManager.getInstance().scheduleAi(new NotifyAITask(CtrlEvent.EVT_READY_TO_ACT), sAtk + 50,  isPlayer || isPet || isSummon);
		ThreadPoolManager.getInstance().scheduleAi(new NotifyAITask(CtrlEvent.EVT_READY_TO_ACT), Math.max(reuse, sAtk) + 50,  isPlayer || isPet || isSummon);
	}

	/**
	 * Launch a Bow attack.<BR>
	 * <BR>
	 *
	 * <B><U> Actions</U> :</B><BR>
	 * <BR>
	 * <li>Calculate if hit is missed or not </li>
	 * <li>Consumme arrows </li>
	 * <li>If hit isn't missed, calculate if shield defense is efficient </li>
	 * <li>If hit isn't missed, calculate if hit is critical </li>
	 * <li>If hit isn't missed, calculate physical damages </li>
	 * <li>If the L2Character is a L2Player, Send a Server->Client packet
	 * SetupGauge </li>
	 * <li>Create a new hit task with Medium priority</li>
	 * <li>Calculate and set the disable delay of the bow in function of the
	 * Attack Speed</li>
	 * <li>Add this hit to the Server-Client packet Attack </li>
	 * <BR>
	 * <BR>
	 *
	 * @param attack
	 *            Server->Client packet Attack in which the hit will be added
	 * @param target
	 *            The L2Character targeted
	 * @param sAtk
	 *            The Attack Speed of the attacker
	 * @param weapon
	 *            L2Weapon item used by this
	 *
	 * @return True if the hit isn't missed
	 *
	 */
	protected boolean doAttackHitByBow(Attack attack, L2Character target, int sAtk, int reuse, L2Weapon weapon)
	{
		int damage1 = 0;
		boolean shld1 = false;
		boolean crit1 = false;

		// Calculate if hit is missed or not
		boolean miss1 = Formulas.calcHitMiss(this, target);

		reduceArrowCount();

		isMoving = false;

		if(!miss1)
		{
			shld1 = Formulas.calcShldUse(this, target);
			crit1 = Formulas.calcCrit(this, target, getCriticalHit(target, null) * 0.01 * target.calcStat(Stats.CRIT_CHANCE_RECEPTIVE, 100, null, null));
			damage1 = (int) Formulas.calcPhysDam(this, target, null, shld1, crit1, false, attack._soulshot);
			if(crit1 && weapon != null)
				weapon.getEffect(true, this, target, true);
		}

		if(isPlayer)
		{
			sendPacket(Msg.GETTING_READY_TO_SHOOT_ARROWS);
			sendPacket(new SetupGauge(SetupGauge.RED, reuse));
		}

		// Create a new hit task with Medium priority
		ThreadPoolManager.getInstance().scheduleAi(new HitTask(target, damage1, crit1, miss1, attack._soulshot, shld1, true), sAtk,  isPlayer || isPet || isSummon);

		attack.addHit(target, damage1, miss1, crit1, shld1);

		return !miss1;
	}

	/**
	 * Launch a Dual attack.<BR>
	 * <BR>
	 *
	 * <B><U> Actions</U> :</B><BR>
	 * <BR>
	 * <li>Calculate if hits are missed or not </li>
	 * <li>If hits aren't missed, calculate if shield defense is efficient
	 * </li>
	 * <li>If hits aren't missed, calculate if hit is critical </li>
	 * <li>If hits aren't missed, calculate physical damages </li>
	 * <li>Create 2 new hit tasks with Medium priority</li>
	 * <li>Add those hits to the Server-Client packet Attack </li>
	 * <BR>
	 * <BR>
	 *
	 * @param attack
	 *            Server->Client packet Attack in which the hit will be added
	 * @param target
	 *            The L2Character targeted
	 * @param weapon
	 *            L2Weapon item used by this
	 *
	 * @return True if hit 1 or hit 2 isn't missed
	 *
	 */
	protected boolean doAttackHitByDual(Attack attack, L2Character target, L2Weapon weapon, int sAtk)
	{
		int damage1 = 0;
		int damage2 = 0;
		boolean shld1 = false;
		boolean shld2 = false;
		boolean crit1 = false;
		boolean crit2 = false;

		boolean miss1 = Formulas.calcHitMiss(this, target);
		boolean miss2 = Formulas.calcHitMiss(this, target);

		if(!miss1)
		{
			shld1 = Formulas.calcShldUse(this, target);
			crit1 = Formulas.calcCrit(this, target, getCriticalHit(target, null) * 0.01 * target.calcStat(Stats.CRIT_CHANCE_RECEPTIVE, 100, null, null));
			damage1 = (int) Formulas.calcPhysDam(this, target, null, shld1, crit1, true, attack._soulshot);
			if(crit1 && weapon != null)
				weapon.getEffect(true, this, target, true);
		}

		if(!miss2)
		{
			shld2 = Formulas.calcShldUse(this, target);
			crit2 = Formulas.calcCrit(this, target, getCriticalHit(target, null) * 0.01 * target.calcStat(Stats.CRIT_CHANCE_RECEPTIVE, 100, null, null));
			damage2 = (int) Formulas.calcPhysDam(this, target, null, shld2, crit2, true, attack._soulshot);
			if(crit1 && weapon != null)
				weapon.getEffect(true, this, target, true);
		}

		if(damage1 + damage2 > 0)
			absorbAndReflect(target, damage1 + damage2);

		// Create a new hit task with Medium priority for hit 1 and for hit 2
		// with a higher delay
		ThreadPoolManager.getInstance().scheduleAi(new HitTask(target, damage1, crit1, miss1, attack._soulshot, shld1, true), sAtk / 2,  isPlayer || isPet || isSummon);
		ThreadPoolManager.getInstance().scheduleAi(new HitTask(target, damage2, crit2, miss2, attack._soulshot, shld2, false), sAtk,  isPlayer || isPet || isSummon);

		attack.addHit(target, damage1, miss1, crit1, shld1);
		attack.addHit(target, damage2, miss2, crit2, shld2);
		return !miss1 || !miss2;
	}

	/**
	 * Launch a Pole attack.<BR>
	 * <BR>
	 *
	 * <B><U> Actions</U> :</B><BR>
	 * <BR>
	 * <li>Get all visible objects in a spheric area near the L2Character to
	 * obtain possible targets </li>
	 * <li>If possible target is the L2Character targeted, launch a simple
	 * attack against it </li>
	 * <li>If possible target isn't the L2Character targeted but is attakable,
	 * launch a simple attack against it </li>
	 * <BR>
	 * <BR>
	 *
	 * @param attack
	 *            Server->Client packet Attack in which the hit will be added
	 * @param weapon
	 *            L2Weapon item used by this
	 *
	 * @return True if one hit isn't missed
	 *
	 */
	protected boolean doAttackHitByPole(Attack attack, L2Weapon weapon, int sAtk)
	{
		boolean hitted = false;
		int angle = (int) calcStat(Stats.POLE_ATTACK_ANGLE, 60, null, null);
		int range = (int) calcStat(Stats.POWER_ATTACK_RANGE, getTemplate().baseAtkRange, null, null);
		double mult = 1.;

		int attackcount = 0;
		// Используем Math.round т.к. обычный кастинг обрезает к меньшему
		// double d = 2.95. int i = (int)d, выйдет что i = 2
		// если 1% угла или 1 дистанции не играет огромной роли, то для
		// количества целей это критично
		int attackcountmax = (int) Math.round(calcStat(Stats.POLE_TARGERT_COUNT, 3, null, null));

		if(isRaid)
			attackcountmax += 12;
		else if(this instanceof L2NpcInstance && getLevel() > 0)
			attackcountmax += getLevel() / 7.5;

		L2Character targ = getAI().getAttackTarget();
		if(targ != null && !targ.isDead() && isInFront(targ, angle))
			if(doAttackHitSimple(attack, targ, weapon, mult, true, sAtk))
			{
				mult *= 0.85;
				attackcount++;
				hitted = true;
			}

		for(L2Character target : getKnownCharacters(range, 400))
			if(attackcount <= attackcountmax)
			{
				if(target != null && !target.isDead() && target.isAutoAttackable(this))
				{
					if(target == getAI().getAttackTarget() || !isInFront(target, angle))
						continue;

					hitted |= doAttackHitSimple(attack, target, weapon, mult, attackcount == 0, sAtk);

					mult *= 0.85;

					attackcount++;
				}
			}
			else
				break;

		// Return true if one hit isn't missed
		return hitted;
	}

	/**
	 * Launch a simple attack.<BR>
	 * <BR>
	 *
	 * <B><U> Actions</U> :</B><BR>
	 * <BR>
	 * <li>Calculate if hit is missed or not </li>
	 * <li>If hit isn't missed, calculate if shield defense is efficient </li>
	 * <li>If hit isn't missed, calculate if hit is critical </li>
	 * <li>If hit isn't missed, calculate physical damages </li>
	 * <li>Create a new hit task with Medium priority</li>
	 * <li>Add this hit to the Server-Client packet Attack </li>
	 * <BR>
	 * <BR>
	 *
	 * @param attack
	 *            Server->Client packet Attack in which the hit will be added
	 * @param target
	 *            The L2Character targeted
	 * @param weapon
	 *            L2Weapon item used by <b>this</b>
	 * @param multiplier
	 *            damage multiplier for polearmers
	 * @return True if the hit isn't missed
	 *
	 */
	protected boolean doAttackHitSimple(Attack attack, L2Character target, L2Weapon weapon, double multiplier, boolean unchargeSS, int sAtk)
	{
		int damage1 = 0;
		boolean shld1 = false;
		boolean crit1 = false;

		// Calculate if hit is missed or not
		boolean miss1 = Formulas.calcHitMiss(this, target);

		// Check if hit isn't missed
		if(!miss1)
		{
			// Calculate if shield defense is efficient
			shld1 = Formulas.calcShldUse(this, target);

			// Calculate if hit is critical
			crit1 = Formulas.calcCrit(this, target, getCriticalHit(target, null) * 0.01 * target.calcStat(Stats.CRIT_CHANCE_RECEPTIVE, 100, null, null));

			// Calculate physical damages
			damage1 = (int) Formulas.calcPhysDam(this, target, null, shld1, crit1, false, attack._soulshot);
			damage1 *= multiplier;

			absorbAndReflect(target, damage1);
			if(crit1 && weapon != null)
				weapon.getEffect(true, this, target, true);
		}
		else if(target.isPlayer)
			if(this instanceof L2NpcInstance)
				target.sendPacket(new SystemMessage(SystemMessage.AVOIDED_S1S_ATTACK).addNpcName(((L2NpcInstance) this).getTemplate().npcId));
			else
				target.sendPacket(new SystemMessage(SystemMessage.AVOIDED_S1S_ATTACK).addString(_name));

		// Create a new hit task with Medium priority
		ThreadPoolManager.getInstance().scheduleAi(new HitTask(target, damage1, crit1, miss1, attack._soulshot, shld1, unchargeSS), sAtk,  isPlayer || isPet || isSummon);

		// Add this hit to the Server-Client packet Attack
		attack.addHit(target, damage1, miss1, crit1, shld1);

		// Return true if hit isn't missed
		return !miss1;
	}

	public long getAnimationEndTime()
	{
		return _animationEndTime;
	}

	public void doCast(L2Skill skill, L2Character target, boolean forceUse)
	{
		// Прерывать дуэли если цель не дуэлянт
		if(getDuel() != null && target.getDuel() != getDuel())
			getDuel().setDuelState(getOwner(), Duel.DuelState.Interrupted);

		if(skill == null)
		{
			sendActionFailed();
			return;
		}

		int itemConsume[] = skill.getItemConsume();

		if(itemConsume[0] > 0)
			for(int i = 0; i < itemConsume.length; i++)
				if(!consumeItem(skill.getItemConsumeId()[i], itemConsume[i]))
				{
					sendPacket(Msg.INCORRECT_ITEM_COUNT);
					sendChanges();
					return;
				}

		int magicId = skill.getId();

		if(magicId == L2Skill.BLINDING_BLOW)
			stopEffect(magicId);

		if(target == null)
		{
			target = skill.getAimingTarget(this);
			if(target == null)
				return;
		}

		fireMethodInvoked(MethodCollection.onStartCast, new Object[] { skill, target, forceUse });

		int level = getSkillDisplayLevel(magicId);
		if(level < 1)
			level = 1;

		int skillTime = skill.isSkillTimePermanent() ? skill.getHitTime() : Formulas.calcMAtkSpd(this, target, skill, skill.getHitTime());
		int skillInterruptTime = skill.isMagic() ? Formulas.calcMAtkSpd(this, target, skill, skill.getSkillInterruptTime()) : 0;

		_animationEndTime = System.currentTimeMillis() + skillTime;

		if(skill.isMagic() && !skill.isSkillTimePermanent() && getChargedSpiritShot() > 0)
		{
			skillTime = (int) (0.70 * skillTime);
			skillInterruptTime = (int) (0.70 * skillInterruptTime);
		}

		//FIXME: для чего тут + 10???
		_castEndTime = 10 + GameTimeController.getGameTicks() + (skillTime + Formulas.calcMAtkSpd(this, target, skill, skill.getCoolTime())) / GameTimeController.MILLIS_IN_TICK;
		_castInterruptTime = GameTimeController.getGameTicks() + skillInterruptTime / GameTimeController.MILLIS_IN_TICK;

		Formulas.calcSkillMastery(skill, this); // Calculate skill mastery for
		// current cast
		long reuseDelay = Formulas.calcSkillReuseDelay(this, skill);

		if(reuseDelay < 500)
			reuseDelay = 500;

		broadcastPacket(new MagicSkillUse(this, target, skill.getDisplayId(), level, skillTime, reuseDelay));

		if(isPlayer)
			if(!skill.isHandler())
				sendPacket(new SystemMessage(SystemMessage.USE_S1).addSkillName(magicId, (short) level));
			else
				sendPacket(new SystemMessage(SystemMessage.USE_S1).addItemName(skill.getItemConsumeId()[0]));

		if(skill.getTargetType() == SkillTargetType.TARGET_HOLY)
			target.getAI().notifyEvent(CtrlEvent.EVT_AGGRESSION, this, 1);

		// Skill reuse check
		if(reuseDelay > 30000)
			addSkillTimeStamp(skill.getId(), reuseDelay);

		if(reuseDelay > 10)
		{
			disableItem(skill, reuseDelay, reuseDelay);
			disableSkill(skill.getId(), reuseDelay);
		}

		double mpConsume1 = skill.getMpConsume1();
		SkillType skillType = skill.getSkillType();

		if(mpConsume1 > 0)
			if(skillType == SkillType.DANCE || skillType == SkillType.SONG)
			{
				int add = -30;
				for(L2Effect e : getAllEffects())
					if((e.getSkill().getSkillType() == SkillType.DANCE || e.getSkill().getSkillType() == SkillType.SONG) && e.getTime() < e.getPeriod() - 30000)
						add += 30;
				mpConsume1 += add;
				mpConsume1 = calcStat(Stats.MP_DANCE_SKILL_CONSUME, mpConsume1, null, skill);
			}
			else if(skill.isMagic())
				reduceCurrentMp(calcStat(Stats.MP_MAGIC_SKILL_CONSUME, mpConsume1, null, skill), null);
			else
				reduceCurrentMp(calcStat(Stats.MP_PHYSICAL_SKILL_CONSUME, mpConsume1, null, skill), null);

		boolean forceBuff = skill.getSkillType() == SkillType.FORCE_BUFF;

		if(forceBuff)
		{
			if(target == this)
			{
				sendPacket(Msg.INCORRECT_TARGET);
				return;
			}
			startForceBuff(target, skill);
		}

		if(skill.getFlyType() != FlyType.NONE)
			ThreadPoolManager.getInstance().scheduleEffect(new FlyToLocationTask(this, target, skill), 50);

		// launch the magic in skillTime milliseconds
		if(skillTime > 50)
		{
			// Send a Server->Client packet SetupGauge with the color of the
			// gauge and the casting time
			if(isPlayer && !forceBuff)
				sendPacket(new SetupGauge(SetupGauge.BLUE, skillTime));

			// Create a task MagicUseTask with Medium priority to launch the
			// MagicSkill at the end of the casting time
			_skillLaunchedTask = ThreadPoolManager.getInstance().scheduleAi(new MagicLaunchedTask(this, forceUse), skillInterruptTime,  isPlayer || isPet || isSummon);
			_skillTask = ThreadPoolManager.getInstance().scheduleAi(new MagicUseTask(target, skill, forceUse), skillTime,  isPlayer || isPet || isSummon);
		}
		else
			onMagicUseTimer(target, skill, forceUse);
	}

	public void startForceBuff(L2Character target, L2Skill skill)
	{
		if(_forceBuff == null)
			_forceBuff = new ForceBuff(this, target, skill);
	}

	public ForceBuff getForceBuff()
	{
		return _forceBuff;
	}

	public void setForceBuff(ForceBuff value)
	{
		_forceBuff = value;
	}

	public void doDie(L2Character killer)
	{
		fireMethodInvoked(MethodCollection.doDie, new Object[] { killer });

		// Set target to null and cancel Attack or Cast
		setTarget(null);

		// Stop movement
		stopMove();

		// Stop HP/MP/CP Regeneration task
		stopHpMpRegeneration();

		_currentHp = 0;

		if(isPlayer && killer instanceof L2Playable)
			_currentCp = 0;

		// Stop all active skills effects in progress on the L2Character
		if(isBlessedByNoblesse() || isSalvation())
		{
			if(isSalvation())
				((L2Player) this).reviveRequest((L2Player) this, 100, false);
			for(L2Effect e : getAllEffects())
				// Noblesse Blessing Buff/debuff effects are retained after
				// death. However, Noblesse Blessing and Lucky Charm are lost as
				// normal.
				if(e.getEffectType() == EffectType.BLESS_NOBLESSE || e.getSkill().getId() == 1325 || e.getSkill().getId() == 2168)
					e.exit();
		}
		else
			stopAllEffects();

		// Send the Server->Client packet StatusUpdate with current HP and MP to
		// all other L2Player to inform
		broadcastStatusUpdate();

		// Notify L2Character AI
		getAI().notifyEvent(CtrlEvent.EVT_DEAD, null, null);

		IScriptHandler handler = ScriptHandler.getInstance().getScriptHandler();
		if(handler != null)
			handler.onDie(this, killer);

		// Notify Quest of character's death
		L2NpcInstance npc = null;
		if(killer instanceof L2NpcInstance)
			npc = (L2NpcInstance) killer;

		if(_NotifyQuestOfDeathList != null)
		{
			for(QuestState qs : getNotifyQuestOfDeath())
				qs.getQuest().notifyDeath(npc, this, qs);
			_NotifyQuestOfDeathList.clear();
		}
	}

	/** Sets HP, MP and CP and revives the L2Character. */
	public void doRevive()
	{
		if(!isTeleporting())
		{
			setIsPendingRevive(false);

			if(isSalvation())
			{
				for(L2Effect e : getAllEffects())
					// Stop Salvation effect
					if(e.getEffectType() == EffectType.SALVATION)
					{
						e.exit();
						break;
					}
				setCurrentCp(getMaxCp());
				setCurrentHp(getMaxHp());
				setCurrentMp(getMaxMp());
			}
			else
			{
				if(isPlayer && Config.RESPAWN_RESTORE_CP >= 0)
					setCurrentCp(getMaxCp() * Config.RESPAWN_RESTORE_CP);

				setCurrentHp(getMaxHp() * Config.RESPAWN_RESTORE_HP);

				if(Config.RESPAWN_RESTORE_MP >= 0)
					setCurrentMp(getMaxMp() * Config.RESPAWN_RESTORE_MP);
			}

			// Start broadcast status
			broadcastPacket(new Revive(this));
		}
		else
			setIsPendingRevive(true);

		if(isPet)
		{
			L2PetInstance pet = (L2PetInstance) this;
			pet.stopDecay();
			pet.setCurrentFed(pet.getMaxFed() / 10);
			pet.startFeed(false);
			setRunning();
		}
	}

	/**
	 * Enable a skill (remove it from _disabledSkills of the L2Character).<BR>
	 * <BR>
	 *
	 * <B><U> Concept</U> :</B><BR>
	 * <BR>
	 * All skills disabled are identified by their skillId in <B>_disabledSkills</B>
	 * of the L2Character <BR>
	 * <BR>
	 *
	 * @param skillId
	 *            The identifier of the L2Skill to enable
	 *
	 */
	public void enableSkill(Integer skillId)
	{
		if(_disabledSkills == null)
			return;

		_disabledSkills.remove(skillId);
		removeSkillTimeStamp(skillId);
	}

	/**
	 * Return a map of 32 bits (0x00000000) containing all abnormal effect in
	 * progress for this L2Character.<BR>
	 * <BR>
	 *
	 * <B><U> Concept</U> :</B><BR>
	 * <BR>
	 * In Server->Client packet, each effect is represented by 1 bit of the map
	 * (ex : BLEEDING = 0x0001 (bit 1), SLEEP = 0x0080 (bit 8)...). The map is
	 * calculated by applying a BINARY OR operation on each effect.<BR>
	 * <BR>
	 *
	 * <B><U> Example of use </U> :</B><BR>
	 * <BR>
	 * <li> Server Packet : CharInfo, NpcInfo, NpcInfoPoly, UserInfo...</li>
	 * <BR>
	 * <BR>
	 */
	public int getAbnormalEffect()
	{
		int ae = _abnormalEffects;
		if(isStunned())
			ae |= ABNORMAL_EFFECT_STUN;
		if(isRooted())
			ae |= ABNORMAL_EFFECT_ROOT;
		if(isSleeping())
			ae |= ABNORMAL_EFFECT_SLEEP;
		if(isConfused())
			ae |= ABNORMAL_EFFECT_CONFUSED;
		if(isMuted() || isPMuted())
			ae |= ABNORMAL_EFFECT_MUTED;
		if(isAfraid())
			ae |= ABNORMAL_EFFECT_AFFRAID;
		if(getEffectByType(EffectType.SILENT_MOVE) != null)
			ae |= ABNORMAL_EFFECT_SILENT_MOVE;
		if(isBigHead())
			ae |= ABNORMAL_EFFECT_BIG_HEAD;
		if(getCustomEffect() > 0)
			ae |= getCustomEffect();
		return ae;
	}

	/**
	 * Return the Accuracy (base+modifier) of the L2Character in function of the
	 * Weapon Expertise Penalty.<BR>
	 * <BR>
	 */
	public int getAccuracy()
	{
		return (int) (calcStat(Stats.ACCURACY_COMBAT, 0, null, null) / getWeaponExpertisePenalty());
	}
	/**
	 * Return the L2CharacterAI of the L2Character and if its null create a new
	 * one.<BR>
	 * <BR>
	 */
	@Override
	public L2CharacterAI getAI()
	{
		if(_ai == null)
			_ai = new L2CharacterAI(this);
		return _ai;
	}

	/**
	 * Возвращает коллекцию скиллов для быстрого перебора
	 */
	public Collection<L2Skill> getAllSkills()
	{
		return _skills.values();
	}

	/**
	 * Возвращает массив скиллов для безопасного перебора
	 */
	public final L2Skill[] getAllSkillsArray()
	{
		if(_skills == null)
			return new L2Skill[0];

		synchronized (_skills)
		{
			return _skills.values().toArray(new L2Skill[_skills.values().size()]);
		}
	}

	/**
	 * Return the Armour Expertise Penalty of the L2Character.<BR>
	 * <BR>
	 */
	public float getArmourExpertisePenalty()
	{
		return 1.f;
	}

	public FastList<L2Player> getAroundPlayers(int radius)
	{
		if(!isVisible())
			return new FastList<L2Player>();

		return L2World.getAroundPlayers(this, radius, Config.PLAYER_VISIBILITY_Z);
	}

	/**
	 * Return the Attack Speed multiplier (base+modifier) of the L2Character to
	 * get proper animations.<BR>
	 * <BR>
	 */
	public final float getAttackSpeedMultiplier()
	{
		return (float) (1.1 * getPAtkSpd() / getTemplate().basePAtkSpd);
	}

	public int getBuffLimit()
	{
		return (int) calcStat(Stats.BUFF_LIMIT, Config.ALT_BUFF_LIMIT, null, null);
	}

	public L2Skill getCastingSkill()
	{
		return _castingSkill;
	}

	/**
	 * If target is L2Character return it, else return null.<BR>
	 * <BR>
	 */
	public final L2Character getCharTarget()
	{
		if(getTarget() == null || !getTarget().isCharacter)
			return null;
		return (L2Character) getTarget();
	}

	/**
	 * Return the COM of the L2Character (base+modifier).<BR>
	 * <BR>
	 */
	public byte getCON()
	{
		return (byte) calcStat(Stats.STAT_CON, _template.baseCON, null, null);
	}

	/**
	 * Return the Critical Hit rate (base+modifier) of the L2Character.<BR>
	 * <BR>
	 */
	public int getCriticalHit(L2Character target, L2Skill skill)
	{
		return (int) calcStat(Stats.CRITICAL_RATE_STATIC, _template.baseCritRate, target, skill);
	}

	/**
	 * Возвращает шанс магического крита в тысячных
	 */
	public double getCriticalMagic(L2Character target, L2Skill skill)
	{
		double val = calcStat(Stats.MCRITICAL_RATE, 8., target, skill);
		if(Config.LIM_MCRIT != 0 && val > Config.LIM_MCRIT)
			val = Config.LIM_MCRIT;
		return val;
	}

	/**
	 * Return the current CP of the L2Character.<BR>
	 * <BR>
	 */
	public final double getCurrentCp()
	{
		return _currentCp;
	}

	/**
	 * Return the current HP of the L2Character.<BR>
	 * <BR>
	 */
	public final double getCurrentHp()
	{
		return _currentHp;
	}

	/**
	 * Return the current MP of the L2Character.<BR>
	 * <BR>
	 */
	public final double getCurrentMp()
	{
		return _currentMp;
	}

	public Location getDestination()
	{
		return new Location(_xTo, _yTo, _zTo);
	}

	public void setDestination(int x, int y, int z)
	{
		_xTo = x;
		_yTo = y;
		_zTo = z;
	}

	/**
	 * Return the DEX of the L2Character (base+modifier).<BR>
	 * <BR>
	 */
	public byte getDEX()
	{
		return (byte) calcStat(Stats.STAT_DEX, _template.baseDEX, null, null);
	}

	/**
	 * Return the Attack Evasion rate (base+modifier) of the L2Character.<BR>
	 * <BR>
	 */
	public int getEvasionRate(L2Character target)
	{
		return (int) (calcStat(Stats.EVASION_RATE, 0, target, null) / getArmourExpertisePenalty());
	}

	/**
	 * Return the orientation of the L2Character.<BR>
	 * <BR>
	 */
	@Override
	public final int getHeading()
	{
		return _heading;
	}

	/**
	 * Return heading to L2Character<BR>
	 * If <b>boolean toChar</b> is true heading calcs this->target, else
	 * target->this.<BR>
	 * <BR>
	 */
	public int getHeadingTo(L2Object target, boolean toChar)
	{
		if(target == null || target == this)
			return -1;

		int dx = target.getX() - getX();
		int dy = target.getY() - getY();
		int heading = (int) (Math.atan2(-dy, -dx) * 32768. / Math.PI + 32768);

		heading = toChar ? target.getHeading() - heading : _heading - heading;

		if(heading < 0)
			heading += 65536;
		return heading;
	}

	/**
	 * Return the INT of the L2Character (base+modifier).<BR>
	 * <BR>
	 */
	public byte getINT()
	{
		return (byte) calcStat(Stats.STAT_INT, _template.baseINT, null, null);
	}

	public FastList<L2Character> getKnownCharacters(int radius)
	{
		if(!isVisible())
			return new FastList<L2Character>();

		return L2World.getAroundCharacters(this, radius, Config.PLAYER_VISIBILITY_Z);
	}

	public FastList<L2Character> getKnownCharacters(int radius, int height)
	{
		if(!isVisible())
			return new FastList<L2Character>();

		return L2World.getAroundCharacters(this, radius, height);
	}

	public FastList<L2NpcInstance> getKnownNpc(int range)
	{
		if(!isVisible())
			return new FastList<L2NpcInstance>();

		return L2World.getAroundNpc(this, range, Config.PLAYER_VISIBILITY_Z);
	}

	/**
	 * Return Skill if the skill is known by the L2Character.<BR>
	 * <BR>
	 *
	 * @param skillId
	 *            The identifier of the L2Skill to check the knowledge
	 *
	 */
	public final L2Skill getKnownSkill(int skillId)
	{
		if(_skills == null)
			return null;

		return _skills.get(skillId);
	}

	/**
	 * Return the Magical Attack range (base+modifier) of the L2Character.<BR>
	 * <BR>
	 */
	public final int getMagicalAttackRange(L2Skill skill)
	{
		if(skill != null)
			return (int) calcStat(Stats.MAGIC_ATTACK_RANGE, skill.getCastRange(), null, skill);
		return getTemplate().baseAtkRange;
	}

	/**
	 * Return the MAtk (base+modifier) of the L2Character for a skill used in
	 * function of abnormal effects in progress.<BR>
	 * <BR>
	 *
	 * <B><U> Example of use </U> :</B><BR>
	 * <BR>
	 * <li> Calculate Magic damage </li>
	 * <BR>
	 * <BR>
	 *
	 * @param target
	 *            The L2Character targeted by the skill
	 * @param skill
	 *            The L2Skill used against the target
	 *
	 */
	public int getMAtk(L2Character target, L2Skill skill)
	{
		if(skill != null && skill.getMatak() > 0)
			return skill.getMatak();
		return (int) calcStat(Stats.MAGIC_ATTACK, _template.baseMAtk, target, skill);
	}

	/**
	 * Return the MAtk Speed (base+modifier) of the L2Character in function of
	 * the Armour Expertise Penalty.<BR>
	 * <BR>
	 */
	public int getMAtkSpd()
	{
		int val = (int) (calcStat(Stats.MAGIC_ATTACK_SPEED, _template.baseMAtkSpd, null, null) / getArmourExpertisePenalty());
		if(Config.LIM_MATK_SPD != 0 && val > Config.LIM_MATK_SPD)
			val = Config.LIM_MATK_SPD;
		return val;
	}

	public double getMAtkSps(L2Character target, L2Skill skill)
	{
		double matk = calcStat(Stats.MAGIC_ATTACK, _template.baseMAtk, target, skill);
		switch(getChargedSpiritShot())
		{
			case 1:
				return matk * 1.41;
			case 2:
				return matk * 2;
			default:
				return matk;
		}
	}

	/**
	 * Return the Max CP (base+modifier) of the L2Character.<BR>
	 * <BR>
	 */
	public final int getMaxCp()
	{
		return (int) calcStat(Stats.MAX_CP, _template.baseCpMax, null, null);
	}

	/**
	 * Return the Max HP (base+modifier) of the L2Character.<BR>
	 * <BR>
	 */
	public int getMaxHp()
	{
		return (int) calcStat(Stats.MAX_HP, _template.baseHpMax, null, null);
	}

	/**
	 * Return the Max MP (base+modifier) of the L2Character.<BR>
	 * <BR>
	 */
	public int getMaxMp()
	{
		return (int) calcStat(Stats.MAX_MP, _template.baseMpMax, null, null);
	}

	/**
	 * Return the MDef (base+modifier) of the L2Character against a skill in
	 * function of abnormal effects in progress.<BR>
	 * <BR>
	 *
	 * <B><U> Example of use </U> :</B><BR>
	 * <BR>
	 * <li> Calculate Magic damage </li>
	 * <BR>
	 * <BR>
	 *
	 * @param target
	 *            The L2Character targeted by the skill
	 * @param skill
	 *            The L2Skill used against the target
	 *
	 */
	public int getMDef(L2Character target, L2Skill skill)
	{
		return Math.max((int) calcStat(Stats.MAGIC_DEFENCE, _template.baseMDef, null, skill), 1);
	}

	/**
	 * Return the MEN of the L2Character (base+modifier).<BR>
	 * <BR>
	 */
	public byte getMEN()
	{
		return (byte) calcStat(Stats.STAT_MEN, _template.baseMEN, null, null);
	}

	public int getMinDistance(L2Object obj)
	{
		int distance = getTemplate().collisionRadius;

		if(obj.isCharacter)
			distance += ((L2Character) obj).getTemplate().collisionRadius;

		return distance;
	}

	public float getMovementSpeedMultiplier()
	{
		return getRunSpeed() * 1f / _template.baseRunSpd;
	}

	/**
	 * Return the RunSpeed (base+modifier) or WalkSpeed (base+modifier) of the
	 * L2Character in function of the movement type.<BR>
	 * <BR>
	 */
	@Override
	public float getMoveSpeed()
	{
		if(isRunning())
			return getRunSpeed();

		return getWalkSpeed();
	}

	/**
	 * Return the MReuse rate (base+modifier) of the L2Character.<BR>
	 * <BR>
	 */
	public final double getMReuseRate(L2Skill skill)
	{
		return calcStat(Stats.MAGIC_REUSE_RATE, _template.baseMReuseRate, null, skill);
	}

	public String getName()
	{
		return _name;
	}

	/**
	 * Return a list of L2Character that attacked.<BR>
	 * <BR>
	 */
	public final ArrayList<QuestState> getNotifyQuestOfDeath()
	{
		if(_NotifyQuestOfDeathList == null)
			_NotifyQuestOfDeathList = new ArrayList<QuestState>();
		return _NotifyQuestOfDeathList;
	}

	/**
	 * Return the PAtk (base+modifier) of the L2Character.<BR>
	 * <BR>
	 */
	public int getPAtk(L2Character target)
	{
		return (int) calcStat(Stats.POWER_ATTACK, _template.basePAtk, target, null);
	}

	/**
	 * Return the PAtk Speed (base+modifier) of the L2Character in function of
	 * the Armour Expertise Penalty.<BR>
	 * <BR>
	 */
	public int getPAtkSpd()
	{
		int val = Math.max((int) (calcStat(Stats.POWER_ATTACK_SPEED, _template.basePAtkSpd, null, null) / getArmourExpertisePenalty()), 1);
		if(Config.LIM_PATK_SPD != 0 && val > Config.LIM_PATK_SPD)
			val = Config.LIM_PATK_SPD;
		return val;
	}

	/**
	 * Return the PDef (base+modifier) of the L2Character.<BR>
	 * <BR>
	 */
	public int getPDef(L2Character target)
	{
		return Math.max((int) calcStat(Stats.POWER_DEFENCE, _template.basePDef, target, null), 1);
	}

	/**
	 * Return the Physical Attack range (base+modifier) of the L2Character.<BR>
	 * <BR>
	 */
	public final int getPhysicalAttackRange()
	{
		return (int) calcStat(Stats.POWER_ATTACK_RANGE, getTemplate().baseAtkRange, null, null);
	}

	/**
	 * Return a Random Damage in function of the weapon.<BR>
	 * <BR>
	 */
	public final int getRandomDamage()
	{
		L2Weapon weaponItem = getActiveWeaponItem();

		if(weaponItem == null)
			return 5 + (int) Math.sqrt(getLevel());

		return weaponItem.getRandomDamage();
	}

	/**
	 * Return the Skill/Spell reuse modifier.<BR>
	 * <BR>
	 */
	public double getReuseModifier(L2Character target)
	{
		return calcStat(Stats.ATK_REUSE, 1, target, null);
	}

	/**
	 * @return the RunSpeed of the L2Character.
	 */
	public int getRunSpeed()
	{
		return getSpeed(_template.baseRunSpd);
	}

	/**
	 * @return the ShieldDef rate (base+modifier) of the L2Character.
	 */
	public final int getShldDef()
	{
		if(isPlayer)
			return (int) calcStat(Stats.SHIELD_DEFENCE, 0, null, null);
		return (int) calcStat(Stats.SHIELD_DEFENCE, _template.baseShldDef, null, null);
	}

	public final short getSkillDisplayLevel(Integer skillId)
	{
		if(_skills == null)
			return -1;

		L2Skill skill = _skills.get(skillId);

		if(skill == null)
			return -1;

		return skill.getDisplayLevel();
	}

	/**
	 * Return the level of a skill owned by the L2Character.<BR>
	 * <BR>
	 *
	 * @param skillId
	 *            The identifier of the L2Skill whose level must be returned
	 *
	 * @return The level of the L2Skill identified by skillId
	 *
	 */
	public final short getSkillLevel(Integer skillId)
	{
		if(_skills == null)
			return -1;

		L2Skill skill = _skills.get(skillId);

		if(skill == null)
			return -1;

		return skill.getLevel();
	}

	public byte getSkillMastery(Integer skillId)
	{
		Byte val = _skillMastery.get(skillId);
		return val == null ? 0 : val;
	}

	public void removeSkillMastery(Integer skillId)
	{
		_skillMastery.remove(skillId);
	}

	/**
	 * Return all skills by type own by the L2Character in a table of L2Skill.<BR>
	 * <BR>
	 *
	 * <B><U> Concept</U> :</B><BR>
	 * <BR>
	 * All skills own by a L2Character are identified in <B>_skills</B> the
	 * L2Character <BR>
	 * <BR>
	 *
	 */
	public final FastTable<L2Skill> getSkillsByType(SkillType type)
	{
		if(_skills == null)
			return new FastTable<L2Skill>();

		FastTable<L2Skill> result = new FastTable<L2Skill>();
		for(L2Skill sk : _skills.values())
			if(sk.getSkillType() == type)
				result.add(sk);

		return result;
	}

	/**
	 * Return the Speed (base+modifier) of the L2Character in function of the
	 * Armour Expertise and Territories Penalty.<BR>
	 * <BR>
	 */
	public final int getSpeed(int baseSpeed)
	{
		if(isFlying())
			return Config.WYVERN_SPEED;

		if(isRiding())
			return Config.STRIDER_SPEED;

		if(isSwimming())
			return getSwimSpeed();

		// err we should be adding TO the persons speed
		// not making it a constant
		double val = calcStat(Stats.RUN_SPEED, baseSpeed, null, null);

		val /= getArmourExpertisePenalty();

		return (int) (val + 0.5);
	}

	/**
	 * Return the list of L2Character that must be informed of HP/MP updates of
	 * this L2Character.<BR>
	 * <BR>
	 *
	 * <B><U> Concept</U> :</B><BR>
	 * <BR>
	 * Each L2Character owns a list called <B>_statusListener</B> that contains
	 * all L2Player to inform of HP/MP updates. Players who must be informed are
	 * players that target this L2Character. When a RegenTask is in progress
	 * sever just need to go through this list to send Server->Client packet
	 * StatusUpdate.<BR>
	 * <BR>
	 *
	 * @return The list of L2Character to inform or null if empty
	 *
	 */
	private CopyOnWriteArraySet<L2Character> getStatusListener()
	{
		return _statusListener;
	}

	/**
	 * Return the STR of the L2Character (base+modifier).<BR>
	 * <BR>
	 */
	public byte getSTR()
	{
		return (byte) calcStat(Stats.STAT_STR, _template.baseSTR, null, null);
	}

	public int getSwimSpeed()
	{
		return (int) calcStat(Stats.RUN_SPEED, Config.SWIMING_SPEED, null, null);
	}

	/**
	 * Return the L2Object targeted or null.<BR>
	 * <BR>
	 */
	public final L2Object getTarget()
	{
		if(target == null)
			return null;

		L2Object t = target.get();
		if(t == null)
			target = null;
		return t;
	}

	/**
	 * Return the identifier of the L2Object targeted or -1.<BR>
	 * <BR>
	 */
	public final int getTargetId()
	{
		if(getTarget() != null)
			return getTarget().getObjectId();
		return -1;
	}

	/**
	 * Return the template of the L2Character.<BR>
	 * <BR>
	 *
	 * <B><U> Concept</U> :</B><BR>
	 * <BR>
	 * Each L2Character owns generic and static properties (ex : all Keltir have
	 * the same number of HP...). All of those properties are stored in a
	 * different template for each type of L2Character. Each template is loaded
	 * once in the server cache memory (reduce memory use). When a new instance
	 * of L2Character is spawned, server just create a link between the instance
	 * and the template This link is stored in <B>_template</B><BR>
	 * <BR>
	 *
	 */
	public L2CharTemplate getTemplate()
	{
		return _template;
	}

	public L2CharTemplate getBaseTemplate()
	{
		return _baseTemplate;
	}

	/**
	 * Return the Title of the L2Character.<BR>
	 * <BR>
	 */
	public String getTitle()
	{
		if(_title != null && _title.length() > 16)
			return _title.substring(0, 16);
		return _title;
	}

	/**
	 * Return the WalkSpeed of the L2Character.<BR>
	 * <BR>
	 */
	public final int getWalkSpeed()
	{
		if(isSwimming())
			return getSwimSpeed();
		return getSpeed(_template.baseWalkSpd);
		// return (getSpeed(_template.baseWalkSpd) * 70) / 100;
	}

	/**
	 * Return the Weapon Expertise Penalty of the L2Character.<BR>
	 * <BR>
	 */
	public float getWeaponExpertisePenalty()
	{
		return 1.f;
	}

	/**
	 * Return the WIT of the L2Character (base+modifier).<BR>
	 * <BR>
	 */
	public byte getWIT()
	{
		return (byte) calcStat(Stats.STAT_WIT, _template.baseWIT, null, null);
	}

	/**
	 * Return True if the L2Character has a L2CharacterAI.<BR>
	 * <BR>
	 */
	@Override
	public boolean hasAI()
	{
		return _ai != null;
	}

	public double headingToRadians(int heading)
	{
		return (heading - 32768) / HEADINGS_IN_PI;
	}

	/**
	 * Return True if the L2Character is dead or use fake death.<BR>
	 * <BR>
	 */
	public final boolean isAlikeDead()
	{
		return _fakeDeath || _currentHp < 1.;
	}

	/**
	 * Return True if the L2Character has aborted its attack.<BR>
	 * <BR>
	 */
	public boolean isAttackAborted()
	{
		return _isAttackAborted;
	}

	/**
	 * Return True if the L2Character is attacking.<BR>
	 * <BR>
	 */
	public final boolean isAttackingNow()
	{
		return _attackEndTime > System.currentTimeMillis();
	}

	/**
	 * Return True if the L2Character is behind the target and can't be seen.<BR>
	 * <BR>
	 */
	public boolean isBehindTarget()
	{
		if(getTarget() != null && getTarget().isCharacter)
		{
			int head = getHeadingTo(getTarget(), true);
			return head != -1 && (head <= 10430 || head >= 55105);
		}

		return false;
	}

	public boolean isToSideOfTarget()
	{
		if(getTarget() != null && getTarget().isCharacter)
		{
			int head = getHeadingTo(getTarget(), true);
			return head != -1 && (head <= 22337 || head >= 43197);
		}

		return false;
	}

	public boolean isToSideOfTarget(L2Object target)
	{
		if(target != null && target.isCharacter)
		{
			int head = getHeadingTo(target, true);
			return head != -1 && (head <= 22337 || head >= 43197);
		}

		return false;
	}

	/**
	 * Return True if the L2Character is behind the target and can't be seen.<BR>
	 * <BR>
	 */
	public boolean isBehindTarget(L2Object target)
	{
		if(target != null && target.isCharacter)
		{
			int head = getHeadingTo(target, true);
			return head != -1 && (head <= 10430 || head >= 55105);
		}
		return false;
	}

	public final boolean isBlessedByNoblesse()
	{
		return _isBlessedByNoblesse > 0;
	}

	public final boolean isSalvation()
	{
		return _isSalvation > 0;
	}

	public final boolean isBuffImmunity()
	{
		return _buffImmunity > 0;
	}

	/**
	 * Return True if the L2Character is dead.<BR>
	 * <BR>
	 */
	public final boolean isDead()
	{
		return _currentHp < 0.5;
	}

	public final boolean isDropDisabled()
	{
		return _dropDisabled > System.currentTimeMillis();
	}

	/**
	 * Return True if the L2Character is flying.<BR>
	 * <BR>
	 */
	public final boolean isFlying()
	{
		return _flying;
	}

	/**
	 * Return True if the L2Character is in combat.<BR>
	 * <BR>
	 */
	public final boolean isInCombat()
	{
		return _stanceTask != null;
	}

	/**
	 * Return True if the target is front L2Character and can be seen.<BR>
	 * <BR>
	 * degrees = 0..180, front->sides->back
	 */
	public boolean isInFront(L2Object target, int degrees)
	{
		int head = getHeadingTo(target, false);
		return head <= 32768 * degrees / 180 || head >= 65536 - 32768 * degrees / 180;
	}

	/**
	 * Return True if the L2Player is invulnerable.<BR>
	 * <BR>
	 */
	public boolean isInvul()
	{
		return _isInvul;
	}

	/**
	 * Check if player is mage
	 */
	public boolean isMageClass()
	{
		return getTemplate().baseMAtk > 3;

	}

	public final boolean isPendingRevive()
	{
		return isDead() && _isPendingRevive;
	}

	/**
	 * Return True if the L2Character is riding.<BR>
	 * <BR>
	 */
	public final boolean isRiding()
	{
		return _riding;
	}

	/**
	 * Return True if the L2Character is running.<BR>
	 * <BR>
	 */
	public final boolean isRunning()
	{
		return _running;
	}

	public boolean isSkillDisabled(Integer skillId)
	{
		return _disabledSkills != null && _disabledSkills.contains(skillId);
	}

	public boolean isSwimming()
	{
		return isPlayer && ((L2Player) this).getWaterTask() != null;
	}

	public final boolean isTeleporting()
	{
		return _isTeleporting;
	}

	public boolean knowsObject(L2Object obj)
	{
		return isInRange(obj, Config.PLAYER_VISIBILITY);
	}

	/**
	 * Возвращает позицию цели, в которой она будет через пол секунды.
	 */
	public Location getIntersectionPoint(L2Character target)
	{
		if(!isInFront(target, 90))
			return new Location(target.getX(), target.getY(), target.getZ());
		double angle = Util.convertHeadingToDegree(target.getHeading()); // угол в градусах
		double radian = Math.toRadians(angle - 90); // угол в радианах
		double range = target.getMoveSpeed() / 2; // расстояние, пройденное за 1 секунду, равно скорости. Берем половину.
		return new Location((int) (target.getX() - range * Math.sin(radian)), (int) (target.getY() + range * Math.cos(radian)), target.getZ());
	}

	public Location applyOffset(Location point, int offset)
	{
		if(_forestalling && isFollow && getFollowTarget() != null && getFollowTarget().isMoving)
			return getIntersectionPoint(getFollowTarget());

		if(offset <= 0)
			return point;

		int dx = point.getX() - getX();
		int dy = point.getY() - getY();
		int dz = point.getZ() - getZ();

		double distance = Math.sqrt(dx * dx + dy * dy + dz * dz);

		if(distance <= offset)
		{
			point.setX(getX());
			point.setY(getY());
			point.setZ(getZ());
			return point;
		}

		if(distance >= 1)
		{
			final double cut = offset / distance;
			int diferentX = point.getX();
			int diferentY = point.getY();
			int diferentZ = point.getZ();
			diferentX -= dx * cut;
			diferentY -= dy * cut;
			diferentZ -= dz * cut;
			point.setX(diferentX);
			point.setY(diferentY);
			point.setZ(diferentZ);

			if(!isFlying() && !isSwimming() && !isInBoat())
				point.setZ(GeoClient.getInstance().getHeight(point.getX(), point.getY(), point.getZ()));
		}

		return point;
	}

	public Location getNextPointToMove()
	{
		synchronized (_targetRecorder)
		{
			if(isSwimming() || isFlying() || isInBoat() || inWater(_xDestination, _yDestination, _zDestination))
				return new Location(_xDestination, _yDestination, _zDestination);
			/*
						if(!isInRangeZ(_xDestination, _yDestination, _zDestination, 2000))
							return isFollow ? null : GeoClient.getInstance().moveCheck(_x, _y, _z, _xDestination, _yDestination, _zDestination);
			*/
			if(_targetRecorder.size() > 0)
				return _targetRecorder.remove(0);

			if(GeoClient.getInstance().canMoveToCoord(getX(), getY(), getZ(), _xDestination, _yDestination, _zDestination))
				return new Location(_xDestination, _yDestination, _zDestination);
			else if(_offset > 200)
				return null;
			else if(_pathfinding)
			{
				_targetRecorder.clear();
				Vector<Location> targets = GeoClient.getInstance().checkMovement(getX(), getY(), getZ(), new Location(_xDestination, _yDestination, _zDestination, 0), this);
				for(Location loc : targets)
					_targetRecorder.add(loc);
				if(_targetRecorder.size() == 0)
					return isFollow ? null : GeoClient.getInstance().moveCheck(getX(), getY(), getZ(), _xDestination, _yDestination, _zDestination);

				if(_targetRecorder.size() > 1)
				{
					if(isPlayer && getAccessLevel() >= 100 && Config.GEODATA_DEBUG)
						spawnWayPoints(_targetRecorder);

					// Удаляем первую точку (начальную позицию чара)
					_targetRecorder.remove(0);

					if(_targetRecorder.isEmpty())
						return null;
					return _targetRecorder.remove(0);
				}

				return GeoClient.getInstance().moveCheck(getX(), getY(), getZ(), _xDestination, _yDestination, _zDestination);
			}
			else
				return isFollow ? null : GeoClient.getInstance().moveCheck(getX(), getY(), getZ(), _xDestination, _yDestination, _zDestination);
		}
	}

	public boolean followToCharacter(L2Character target, int offset, boolean forestalling)
	{
		if(isFollow && target == getFollowTarget() && offset == _offset)
			return true;

		if(!hasAI() || isMovementDisabled() || target == null)
		{
			sendActionFailed();
			return false;
		}

		if(Math.abs(getZ() - target.getZ()) > 1000)
		{
			sendPacket(Msg.CANNOT_SEE_TARGET);
			sendActionFailed();
			return false;
		}

		getAI().clearNextAction();

		isMoving = false;
		isFollow = false;

		if(isFlying() || isInBoat())
		{
			sendActionFailed();
			return false;
		}

		if(offset < 10)
			offset = 10;

		setFollowTarget(target);

		_xDestination = getFollowTarget().getX();
		_yDestination = getFollowTarget().getY();
		_zDestination = getFollowTarget().getZ();

		isFollow = true;

		_pathfinding = true;
		_forestalling = forestalling;
		_targetRecorder.clear();
		Location nextPoint = getNextPointToMove();

		if(nextPoint == null)
		{
			stopMove();
			return false;
		}

		if(_targetRecorder.size() == 0)
			nextPoint = applyOffset(nextPoint, offset);

		_xFrom = getX();
		_yFrom = getY();
		_zFrom = getZ();

		_xTo = nextPoint.getX();
		_yTo = nextPoint.getY();
		_zTo = nextPoint.getZ();

		setHeading(calcHeading(_xTo, _yTo));

		broadcastPacket(new CharMoveToLocation(this));

		_offset = offset;

		_startMoveTick = _followTimestamp = GameTimeController.getGameTicks() - 1;

		isMoving = true;

		_previousSpeed = getMoveSpeed();

		return true;
	}

	public boolean moveToLocation(Location loc, int offset, boolean pathfinding)
	{
		return moveToLocation(loc.getX(), loc.getY(), loc.getZ(), offset, pathfinding);
	}

	public boolean moveToLocation(int x_dest, int y_dest, int z_dest, int offset, boolean pathfinding)
	{
		if(isMoving && _xDestination == x_dest && _yDestination == y_dest && _zDestination == z_dest && _offset == offset && _pathfinding == pathfinding)
			return true;

		if(!hasAI())
			return false;

		getAI().clearNextAction();

		if(isMovementDisabled())
		{
			getAI().setNextAction(nextAction.MOVE, new Location(x_dest, y_dest, z_dest), offset, pathfinding, false);
			sendActionFailed();
			return true;
		}

		if(isPlayer)
			getAI().setIntention(AI_INTENTION_ACTIVE);

		isMoving = false;
		isFollow = false;

		if(isFlying() || isInBoat())
			pathfinding = false;

		if(offset < 0)
			offset = 0;

		_xDestination = x_dest;
		_yDestination = y_dest;
		_zDestination = z_dest;

		_pathfinding = pathfinding;

		_targetRecorder.clear();
		Location nextPoint = getNextPointToMove();

		if(nextPoint == null)
		{
			stopMove();
			return false;
		}

		if(_targetRecorder.size() == 0)
			nextPoint = applyOffset(nextPoint, offset);

		_xFrom = getX();
		_yFrom = getY();
		_zFrom = getZ();

		_xTo = nextPoint.getX();
		_yTo = nextPoint.getY();
		_zTo = nextPoint.getZ();

		setHeading(calcHeading(_xTo, _yTo));

		broadcastPacket(new CharMoveToLocation(this));

		_offset = offset;

		_startMoveTick = _followTimestamp = GameTimeController.getGameTicks() - 1;

		isMoving = true;

		_previousSpeed = getMoveSpeed();

		return true;
	}

	public void updatePosition(int currentGameTick)
	{
		if(isMovementDisabled())
		{
			stopMove();
			return;
		}

		double speed = getMoveSpeed();

		if(_previousSpeed != speed)
		{
			_previousSpeed = speed;

			_startMoveTick = currentGameTick;

			_xFrom = getX();
			_yFrom = getY();
			_zFrom = getZ();

			broadcastPacket(new CharMoveToLocation(this));
		}

		double dx = _xTo - _xFrom;
		double dy = _yTo - _yFrom;
		double dz = _zTo - _zFrom;

		double distance = Math.sqrt(dx * dx + dy * dy + dz * dz);

		double ticksToMove = speed > 0 ? GameTimeController.TICKS_PER_SECOND * distance / speed : 0.0;
		double ticksFromStart = currentGameTick - _startMoveTick;
		double elapsed = ticksToMove > 0 ? ticksFromStart / ticksToMove : 1.0;

		if((1.0 - elapsed) * distance < 3)
		{
			setXYZ(_xTo, _yTo, _zTo);

			if(_targetRecorder.size() == 0)
			{
				// Мы достигли последней точки в списке, а значит и цели
				isMoving = false;

				if(isFollow)
				{
					isFollow = false;
					ThreadPoolManager.getInstance().scheduleGeneral(new NotifyAITask(CtrlEvent.EVT_ARRIVED_TARGET), 1);
				}
				else
					ThreadPoolManager.getInstance().scheduleGeneral(new NotifyAITask(CtrlEvent.EVT_ARRIVED), 1);

				return;
			}

			_startMoveTick = currentGameTick;

			Location nextPoint = getNextPointToMove();

			if(nextPoint == null)
			{
				stopMove();
				return;
			}

			if(_targetRecorder.size() == 0)
				nextPoint = applyOffset(nextPoint, _offset);

			_xFrom = getX();
			_yFrom = getY();
			_zFrom = getZ();

			_xTo = nextPoint.getX();
			_yTo = nextPoint.getY();
			_zTo = nextPoint.getZ();

			setHeading(calcHeading(_xTo, _yTo));

			broadcastPacket(new CharMoveToLocation(this));
			return;
		}

		int new_x = (int) (_xFrom + elapsed * (_xTo - _xFrom) + 0.5);
		int new_y = (int) (_yFrom + elapsed * (_yTo - _yFrom) + 0.5);
		int new_z = (int) (_zFrom + elapsed * (_zTo - _zFrom) + 0.5);

		setXYZ(new_x, new_y, new_z);

		if(isFollow)
		{
			if(getFollowTarget() == null || isFlying() || isInBoat())
			{
				stopMove();
				return;
			}

			if(Math.abs(getZ() - getFollowTarget().getZ()) > 1000)
			{
				sendPacket(Msg.CANNOT_SEE_TARGET);
				stopMove();
				return;
			}

			if(isInRangeZ(getFollowTarget(), _offset))
			{
				if(_forestalling)
				{
					_targetRecorder.clear();
					isMoving = false;
					isFollow = false;
					ThreadPoolManager.getInstance().scheduleGeneral(new NotifyAITask(CtrlEvent.EVT_ARRIVED_TARGET), 1);
				}
				else
				{
					ThreadPoolManager.getInstance().scheduleGeneral(new NotifyAITask(CtrlEvent.EVT_ARRIVED_TARGET), 500);
					_targetRecorder.clear();
					isMoving = false;
					isFollow = false;
				}
				return;
			}

			if((currentGameTick - _followTimestamp) * GameTimeController.MILLIS_IN_TICK > 500 && !getFollowTarget().isInRange(getLoc(), 100))
			{
				_followTimestamp = currentGameTick;
				_startMoveTick = currentGameTick;

				_xDestination = getFollowTarget().getX();
				_yDestination = getFollowTarget().getY();
				_zDestination = getFollowTarget().getZ();

				_targetRecorder.clear();
				Location nextPoint = getNextPointToMove();

				if(nextPoint == null)
				{
					stopMove();
					return;
				}

				if(_targetRecorder.size() == 0)
					nextPoint = applyOffset(nextPoint, _offset);

				_xFrom = getX();
				_yFrom = getY();
				_zFrom = getZ();

				_xTo = nextPoint.getX();
				_yTo = nextPoint.getY();
				_zTo = nextPoint.getZ();

				setHeading(calcHeading(_xTo, _yTo));

				broadcastPacket(new CharMoveToLocation(this));
			}
		}
	}

	private boolean inWater(int x, int y, int z)
	{
		FastList<L2Territory> current_territories = L2World.getTerritory(x, y);
		for(FastList.Node<L2Territory> n = current_territories.head(), end = current_territories.tail(); (n = n.getNext()) != end;)
		{
			L2Territory terr = n.getValue();
			if(terr.getZone() != null && terr.getZone().getType() == ZoneType.water && terr.isInside(x, y, z))
				return true;
		}
		return false;
	}

	public void stopMove()
	{
		if(isMoving)
		{
			isMoving = false;
			setXYZ(getX(), getY(), GeoClient.getInstance().getHeight(getX(), getY(), getZ()));
			broadcastPacket(new StopMove(this));
			broadcastPacket(new ValidateLocation(this));
		}

		if(isFollow)
		{
			isFollow = false;
			if(this instanceof L2Summon)
				setFollowStatus(false);
		}
	}

	public void moveToInABoat(Location destination, Location origin)
	{
		if(isMovementDisabled() || isSitting())
		{
			sendActionFailed();
			return;
		}

		broadcastPacket(new MoveToLocationInVehicle(this, destination, origin));
	}

	/**
	 * Returns true if status update should be done, false if not
	 *
	 * @return boolean
	 */
	protected boolean needStatusUpdate()
	{
		if(Config.FORCE_STATUSUPDATE)
			return true;

		if(!isNpc)
			return true;

		double _intervalHpUpdate = getMaxHp() / 352;

		if(_lastHpUpdate == -99999999)
		{
			_lastHpUpdate = -9999999;
			return true;
		}

		if(getCurrentHp() <= 0 || getMaxHp() < 352)
			return true;

		if(_lastHpUpdate + _intervalHpUpdate < getCurrentHp() && getCurrentHp() > _lastHpUpdate)
		{
			_lastHpUpdate = getCurrentHp();
			return true;
		}

		if(_lastHpUpdate - _intervalHpUpdate > getCurrentHp() && getCurrentHp() < _lastHpUpdate)
		{
			_lastHpUpdate = getCurrentHp();
			return true;
		}
		return false;
	}

	/**
	 * Remove the L2Character from the world when the decay task is launched.<BR>
	 * <BR>
	 */
	public void onDecay()
	{
		decayMe();
		fireMethodInvoked(MethodCollection.onDecay, null);
	}

	/**
	 * Manage Forced attack (ctrl + select target).<BR>
	 * <BR>
	 *
	 * <B><U> Actions</U> :</B><BR>
	 * <BR>
	 * <li>If L2Character or target is in a town area, send a system message
	 * TARGET_IN_PEACEZONE a Server->Client packet ActionFailed </li>
	 * <li>If target is confused, send a Server->Client packet ActionFailed
	 * </li>
	 * <li>If L2Character is a L2ArtefactInstance, send a Server->Client packet
	 * ActionFailed </li>
	 * <li>Send a Server->Client packet MyTargetSelected to start attack and
	 * Notify AI with AI_INTENTION_ATTACK </li>
	 * <BR>
	 * <BR>
	 *
	 * @param player
	 *            The L2Player to attack
	 *
	 */
	@Override
	public void onForcedAttack(L2Player player)
	{
		if(player.isConfused() || player.isBlocked())
		{
			player.sendActionFailed();
			return;
		}

		player.sendPacket(new MyTargetSelected(getObjectId(), player.getLevel() - getLevel()));
		player.getAI().Attack(this, true);
	}

	protected void onHitTimer(L2Character target, int damage, boolean crit, boolean miss, boolean soulshot, boolean shld, boolean unchargeSS)
	{
		if(isAlikeDead())
		{
			sendActionFailed();
			return;
		}

		if(target.isDead() || !isInRange(target, 1500))
		{
			sendActionFailed();
			return;
		}

		// If attack isn't aborted, send a message system (critical hit,
		// missed...) to attacker/target if they are L2Player
		if(!isAttackAborted())
		{
			if(target.isStunned() && Formulas.calcStunBreak(crit))
				target.stopEffects(EffectType.STUN);

			if(isPlayer)
			{
				if(crit)
					sendPacket(Msg.CRITICAL_HIT);

				if(miss)
					sendPacket(Msg.MISSED_TARGET);
				else
					sendPacket(new SystemMessage(SystemMessage.YOU_HIT_FOR_S1_DAMAGE).addNumber(damage));
			}
			else if(isSummon)
				((L2SummonInstance) this).displayHitMessage(damage, crit, miss);

			else if(isPet)
				((L2PetInstance) this).displayHitMessage(damage, crit, miss);

			if(target.isPlayer)
			{
				L2Player enemy = (L2Player) target;

				if(shld && damage > 1)
					enemy.sendPacket(Msg.SHIELD_DEFENSE_HAS_SUCCEEDED);
				else if(shld && damage == 1)
					enemy.sendPacket(Msg.YOUR_EXCELLENT_SHIELD_DEFENSE_WAS_A_SUCCESS);
			}

			// Reduce HP of the target and calculate reflection damage to reduce
			// HP of attacker if necessary
			if(!miss && damage > 0)
			{
				if(target.getForceBuff() != null)
					target.abortCast();

				target.getAI().notifyEvent(CtrlEvent.EVT_ATTACKED, this, 0);

				target.reduceCurrentHp(damage, this, true, true, false);

				// Скиллы, кастуемые при физ атаке
				if(!target.isDead())
				{
					if(_skillsOnAttack != null)
						for(L2Skill skill : _skillsOnAttack.values())
							if((crit || !skill.isOnCrit()) && Rnd.chance(skill.getChance()))
							{
								FastList<L2Character> targets = new FastList<L2Character>();
								targets.add(target);
								broadcastPacket(new MagicSkillLaunched(_objectId, skill.getId(), skill.getLevel(), targets));
								skill.useSkill(this, targets);
							}
					if(target._skillsOnUnderAttack != null)
						for(L2Skill skill : target._skillsOnUnderAttack.values())
							if(Rnd.chance(skill.getChance()))
							{
								FastList<L2Character> targets = new FastList<L2Character>();
								targets.add(this);
								broadcastPacket(new MagicSkillLaunched(target.getObjectId(), skill.getId(), skill.getLevel(), targets));
								skill.useSkill(this, targets);
							}
				}

				// Проверка на мираж
				if(getTarget() != null && isPlayer)
					if(Rnd.chance((int) target.calcStat(Stats.CANCEL_TARGET, 0, null, null)))
						setTarget(null);

				if(soulshot && unchargeSS)
					unChargeShots(false);

				// Manage attack or cast break of the target (calculating rate,
				// sending message...)
				if(Formulas.calcCastBreak(target, crit))
					target.breakCast(false);
			}
		}

		startAttackStanceTask();

		if(checkPvP(target, null))
			startPvPFlag(target);
	}

	public void onMagicUseTimer(L2Character aimingTarget, L2Skill skill, boolean forceUse)
	{
		_castInterruptTime = 0;

		if(_forceBuff != null)
		{
			_forceBuff.delete();
			return;
		}

		if(skill.getCastRange() < Integer.MAX_VALUE && !GeoClient.getInstance().canSeeTarget(this, aimingTarget))
		{
			sendPacket(Msg.CANNOT_SEE_TARGET);
			broadcastPacket(new MagicSkillCanceled(_objectId));
			return;
		}

		if(!skill.checkCondition(this, aimingTarget, forceUse, false, false))
			return;

		int level = getSkillDisplayLevel(skill.getId());
		if(level < 1)
			level = 1;

		FastList<L2Character> targets = skill.getTargets(this, aimingTarget, forceUse);

		double mpConsume2 = skill.getMpConsume2();
		int hpConsume = skill.getHpConsume();

		if(hpConsume > 0)
			synchronized (this)
			{
				reduceCurrentHp(hpConsume >= _currentHp ? _currentHp - 2 : hpConsume, this, true, true, true);
			}

		SkillType skillType = skill.getSkillType();
		if(mpConsume2 > 0)
		{
			if(skillType == SkillType.DANCE || skillType == SkillType.SONG)
			{
				int add = -30;
				for(L2Effect e : getAllEffects())
					if((e.getSkill().getSkillType() == SkillType.DANCE || e.getSkill().getSkillType() == SkillType.SONG) && e.getTime() < e.getPeriod() - 20000)
						add += 30;
				mpConsume2 += add;
				mpConsume2 = calcStat(Stats.MP_DANCE_SKILL_CONSUME, mpConsume2, null, skill);
			}
			else if(skill.isMagic())
				mpConsume2 = calcStat(Stats.MP_MAGIC_SKILL_CONSUME, mpConsume2, null, skill);
			else
				mpConsume2 = calcStat(Stats.MP_PHYSICAL_SKILL_CONSUME, mpConsume2, null, skill);

			if(_currentMp < mpConsume2)
			{
				sendPacket(Msg.NOT_ENOUGH_MP);
				return;
			}
			reduceCurrentMp(mpConsume2, null);
		}

		callSkill(skill, targets);

		int skillCoolTime = Formulas.calcMAtkSpd(this, null, skill, skill.getCoolTime());
		if(skillCoolTime > 0)
			ThreadPoolManager.getInstance().scheduleAi(new Runnable(){
				public void run()
				{
					onCastEndTime();
				}
			}, skillCoolTime,  isPlayer || isPet || isSummon);
		else
			onCastEndTime();
	}

	public void onCastEndTime()
	{
		_castEndTime = 0;
		_castingSkill = null;
		_skillTask = null;
		getAI().notifyEvent(CtrlEvent.EVT_FINISH_CASTING, null, null);
	}

	/**
	 * Reduce the current HP of the L2Character and launch the doDie Task if
	 * necessary.<BR>
	 * <BR>
	 *
	 * <B><U> Overriden in </U> :</B><BR>
	 * <BR>
	 * <li> L2NpcInstance : Update the attacker AggroInfo of the L2NpcInstance
	 * _aggroList</li>
	 * <BR>
	 * <BR>
	 *
	 * @param i
	 *            The HP decrease value
	 * @param attacker
	 *            The L2Character who attacks
	 * @param awake
	 *            The awake state (If True : stop sleeping)
	 * @param standUp
	 *            If True : stand up
	 * @param directHp
	 *            If True : hit on HP instead of CP first
	 *
	 */
	public void reduceCurrentHp(double i, L2Character attacker, boolean awake, boolean standUp, boolean directHp)
	{
		fireMethodInvoked(MethodCollection.ReduceCurrentHp, new Object[] { i, attacker, awake, standUp, directHp, });

		if(isDead() || isInvul())
			return;

		// 5182 = Blessing of protection, работает если разница уровней больше 10 и не в зоне осады
		if(attacker != null && attacker.isPlayer && Math.abs(attacker.getLevel() - getLevel()) > 10)
		{
			// ПК не может нанести урон чару с блессингом
			if(attacker.getKarma() > 0 && getEffectsBySkillId(5182) != null && !isInZone(ZoneType.Siege))
				return;
			// чар с блессингом не может нанести урон ПК
			if(getKarma() > 0 && attacker.getEffectsBySkillId(5182) != null && !attacker.isInZone(ZoneType.Siege))
				return;
		}

		if(awake && isSleeping())
			stopEffects(L2Effect.EffectType.SLEEP);

		if(isMeditated() && attacker != this)
			stopEffects(L2Effect.EffectType.MEDITATION);

		if(standUp && isPlayer)
		{
			standUp();
			if(isFakeDeath() && getEffectByType(L2Effect.EffectType.FAKE_DEATH).getTime() > 2000)
				stopEffects(L2Effect.EffectType.FAKE_DEATH);
		}

		if(attacker != this)
			startAttackStanceTask();

		if(attacker instanceof L2Playable)
		{
			L2Playable pAttacker = (L2Playable) attacker;

			// Flag the attacker if it's a L2Player outside a PvP area
			if(!isDead() && pAttacker.checkPvP(this, null))
				pAttacker.startPvPFlag(this);

			if(isOverhit())
				setOverhitValues(pAttacker, i);
			else
				setOverhitEnabled(false);

			double ii;
			if(!directHp)
			{
				i = _currentCp - i;
				ii = i;

				if(ii < 0)
					ii *= -1;

				if(i < 0)
					i = 0;

				setCurrentCp(i);
			}
			else
				ii = i;

			if(_currentCp == 0 || directHp)
			{
				ii = _currentHp - ii;

				if(ii < 0)
					ii = 0;

				if(isNpc)
					pAttacker.addDamage((L2NpcInstance) this, (int) (_currentHp - ii));

				setCurrentHp(ii);
			}
		}
		else
		{
			i = _currentHp - i;

			if(i < 0)
				i = 0;

			setCurrentHp(i);
		}

		if(isDead())
		{
			// killing is only possible one time
			synchronized (this)
			{
				if(_killedAlready)
					return;

				_killedAlready = true;
			}

			stopMove();
			doDie(attacker);
			_currentHp = 0;
		}
		else
			setOverhitEnabled(false);
	}

	/**
	 * Reduce the current MP of the L2Character.<BR><BR>
	 * @param i The MP decrease value
	 * @param attacker L2Character
	 */
	public void reduceCurrentMp(double i, L2Character attacker)
	{
		if(attacker != null && attacker != this)
		{
			if(isSleeping())
				stopEffects(L2Effect.EffectType.SLEEP);
			if(isMeditated())
				stopEffects(L2Effect.EffectType.MEDITATION);
		}

		// 5182 = Blessing of protection, работает если разница уровней больше 10 и не в зоне осады
		if(attacker != null && attacker.isPlayer && Math.abs(attacker.getLevel() - getLevel()) > 10)
		{
			// ПК не может нанести урон чару с блессингом
			if(attacker.getKarma() > 0 && getEffectsBySkillId(5182) != null && !isInZone(ZoneType.Siege))
				return;
			// чар с блессингом не может нанести урон ПК
			if(getKarma() > 0 && attacker.getEffectsBySkillId(5182) != null && !attacker.isInZone(ZoneType.Siege))
				return;
		}

		i = _currentMp - i;

		if(i < 0)
			i = 0;

		setCurrentMp(i);
	}

	public double relativeSpeed(L2Object target)
	{
		return getMoveSpeed() - target.getMoveSpeed() * Math.cos(headingToRadians(getHeading()) - headingToRadians(target.getHeading()));
	}

	public void removeAllSkills()
	{
		for(L2Skill s : getAllSkillsArray())
			removeSkill(s);
	}

	public void removeBlockStats(FastList<Stats> stats)
	{
		if(_blockedStats != null)
		{
			_blockedStats.removeAll(stats);
			if(_blockedStats.isEmpty())
				_blockedStats = null;
		}
	}

	/**
	 * Возвращает число эффектов соответствующее данному скиллу
	 */
	public int getEffectsCountForSkill(int skill_id)
	{
		if(_effects == null)
			return 0;
		int count = 0;
		for(L2Effect e : _effects)
			if(e.getSkill().getId() == skill_id)
				count++;
		return count;
	}

	public L2Effect getEffectByType(EffectType et)
	{
		if(_effects != null)
			for(L2Effect e : _effects)
				if(e.getEffectType() == et)
					return e;
		return null;
	}

	public FastList<L2Effect> getEffectsBySkill(L2Skill skill)
	{
		if(skill == null)
			return null;
		return getEffectsBySkillId(skill.getId());
	}

	public FastList<L2Effect> getEffectsBySkillId(int skillId)
	{
		if(_effects == null)
			return null;
		FastList<L2Effect> temp = new FastList<L2Effect>();
		for(L2Effect e : _effects)
			if(e.getSkill().getId() == skillId)
				temp.add(e);

		return temp.size() > 0 ? temp : null;
	}

	public L2Effect getEffectByResistType(ResistType rt)
	{
		if(_effects != null)
			for(L2Effect e : _effects)
				if(e.getResistType() == rt)
					return e;
		return null;
	}

	public L2Effect getEffectByIndexAndType(int skill_id, EffectType type)
	{
		if(_effects == null)
			return null;
		for(L2Effect e : _effects)
			if(e.getSkill().getId() == skill_id && e.getEffectType() == type)
				return e;
		return null;
	}

	public ConcurrentLinkedQueue<L2Effect> getAllEffects()
	{
		if(_effects == null)
			return new ConcurrentLinkedQueue<L2Effect>();
		return _effects;
	}

	/**
	 * Возвращает первые эффекты для всех скиллов. Нужно для отображения не
	 * более чем 1 иконки для каждого скилла.
	 */
	public L2Effect[] getAllFirstEffects()
	{
		if(_effects == null)
			return new L2Effect[0];

		FastMap<Integer, L2Effect> temp = new FastMap<Integer, L2Effect>();

		if(_effects != null)
			for(L2Effect ef : _effects)
				if(ef != null)
					temp.put(ef.getSkill().getId(), ef);

		Collection<L2Effect> temp2 = temp.values();

		temp2.toArray(new L2Effect[temp2.size()]);

		return temp2.toArray(new L2Effect[temp2.size()]);
	}

	/**
	 * Ограничение на количество бафов
	 */
	private void checkBuffSlots(L2Effect newEffect)
	{
		if(_effects == null || _effects.size() < 10)
			return;

		if(newEffect.getSkill().isOffensive() || newEffect.getSkill().isToggle() || newEffect.getStackType().equalsIgnoreCase("HpRecoverCast"))
			return;

		int buffs = 0;
		for(L2Effect ef : _effects)
			if(ef != null && ef.isInUse())
				if(!ef.getSkill().isOffensive() && !ef.getSkill().isToggle() && !ef.getStackType().equalsIgnoreCase("HpRecoverCast"))
					buffs++;

		if(buffs < getBuffLimit())
			return;

		int toRemove = 0;

		for(L2Effect ef : _effects)
			if(ef != null && ef.isInUse())
				if(!ef.getSkill().isPvpSkill() && !ef.getSkill().isToggle() && !ef.getStackType().equalsIgnoreCase("HpRecoverCast"))
				{
					toRemove = ef.getSkill().getId();
					break;
				}

		if(toRemove > 0)
			for(L2Effect ef : _effects)
				if(ef != null && ef.isInUse())
					if(ef.getSkill().getId() == toRemove)
						ef.exit();
	}

	public synchronized void addEffect(L2Effect newEffect)
	{
		if(newEffect == null)
			return;

		if(_effects == null)
			_effects = new ConcurrentLinkedQueue<L2Effect>();

		// Хербы при вызванном саммоне делятся с саммоном пополам
		if(Config.HERBS_DIVIDE && (isPet || isSummon || getPet() != null && !getPet().isDead()) && newEffect.getSkill().getId() >= 2278 && newEffect.getSkill().getId() <= 2285)
		{
			newEffect.setPeriod(newEffect.getPeriod() / 2);
			if(!(isPet || isSummon))
				getPet().altUseSkill(newEffect.getSkill(), getPet());
		}

		// затычка на баффы повышающие хп/мп
		double hp = getCurrentHp();
		double mp = getCurrentMp();
		double cp = getCurrentCp();

		// Проверка на имунность к бафам/дебафам
		if(isBuffImmunity() && newEffect.getEffectType() != L2Effect.EffectType.BUFFIMMUNITY)
		{
			SkillType st = newEffect.getSkill().getSkillType();
			if(st.equals(L2Skill.SkillType.BUFF) || st.equals(L2Skill.SkillType.DEBUFF))
				return;
		}

		String stackType = newEffect.getStackType();

		if(stackType.equals("none"))
		{
			// Удаляем такие же эффекты
			for(L2Effect ef : _effects)
				if(ef != null && ef.isInUse() && ef.getStackType().equals("none") && ef.getSkill().getId().equals(newEffect.getSkill().getId()) && ef.getEffectType().equals(newEffect.getEffectType()))
					// Если оставшаяся длительность старого эффекта больше чем длительность нового, то оставляем старый.
					if(newEffect.getTimeLeft() > ef.getTimeLeft())
						ef.exit();
					else
						return;
		}
		else
		{
			// Проверяем, нужно ли накладывать эффект, при совпадении StackType.
			// Новый эффект накладывается только в том случае, если у него больше StackOrder и больше длительность.
			// Если условия подходят - удаляем старый, иначе выходим.
			boolean flag = true;
			int count = 0;
			while(flag)
			{
				flag = false;
				count++;
				if(count > 2)
				{
					System.out.println("------------------------------------------");
					System.out.println("New effect: " + newEffect.getSkill().toString());
					for(L2Effect ef : _effects)
						System.out.println("Effect: " + ef.getSkill().toString());
					System.out.println("------------------------------------------");
					return;
				}
				for(L2Effect ef : _effects)
					if(ef != null && ef.isInUse() && ef.getStackType().equals(stackType))
					{
						if(ef.getSkill().getId() == newEffect.getSkill().getId() && ef.getEffectType() != newEffect.getEffectType())
							break;

						// Эффекты со StackOrder == -1 заменить нельзя (например, Root).
						if(ef.getStackOrder() == -1)
							return;

						// Если новый эффект слабее
						if(newEffect.getStackOrder() < ef.getStackOrder())
						{
							// Если его длительность больше то шедулить
							if(newEffect.getTimeLeft() > ef.getTimeLeft())
							{
								ef.scheduleNext(newEffect);
								newEffect.setInUse(true);
								return;
							}
							// Иначе выход
							return;
						}

						// Если эффекты равны
						else if(newEffect.getStackOrder() == ef.getStackOrder())
						{
							// Если новый эффект короче то выйти
							if(newEffect.getTimeLeft() < ef.getTimeLeft())
								return;

							// Чистим все ненужные зашедуленные
							L2Effect next = ef.getNext();
							while(next != null)
							{
								if(newEffect.getTimeLeft() > next.getTimeLeft())
								{
									next.exit();
									break;
								}
								next = next.getNext();
							}

							// Присоединяем зашедуленные эффекты от старого к новому
							if(ef.getNext() != null && !ef.getNext().isEnded())
								newEffect.scheduleNext(ef.getNext());

							// Отсоединяем зашедуленные от старого
							ef.removeNext();

							// Останавливаем старый
							ef.exit();

							// И пройтись по эффектам еще раз
							flag = true;
							break;
						}

						// Если новый эффект сильнее
						else if(newEffect.getStackOrder() > ef.getStackOrder())
						{
							// Если старый короче то просто остановить его
							if(newEffect.getTimeLeft() > ef.getTimeLeft())
							{
								ef.exit();
								break;
							}
							// Если новый короче то зашедулить старый
							removeStatsOwner(ef);
							_effects.remove(ef);
							newEffect.scheduleNext(ef);
						}
					}
			}
		}

		// Проверяем на лимиты бафов/дебафов
		checkBuffSlots(newEffect);

		// Добавляем новый эффект
		_effects.add(newEffect);

		if(newEffect.getEffector().isPlayer && newEffect.getEffector().getDuel() != null)
			newEffect.getEffector().getDuel().onBuff((L2Player) newEffect.getEffector(), newEffect);

		// Применяем эффект к параметрам персонажа
		addStatFuncs(newEffect.getStatFuncs());

		// Запускаем эффект
		newEffect.setInUse(true);

		// затычка на баффы повышающие хп/мп
		for(Func f : newEffect.getStatFuncs())
			if(f._stat == Stats.MAX_HP)
				setCurrentHp(hp);
			else if(f._stat == Stats.MAX_MP)
				setCurrentMp(mp);
			else if(f._stat == Stats.MAX_CP)
				setCurrentCp(cp);

		// Обновляем иконки
		updateEffectIcons();
	}

	/**
	 * Вызывающий метод синхронизирован, дополнительная синхронизация не нужна.
	 * @see l2r.gameserver.model.L2Effect#stopEffectTask()
	 * @param effect эффект для удаления
	 */
	public void removeEffect(L2Effect effect)
	{
		if(effect == null || _effects == null || !_effects.contains(effect))
			return;

		removeStatsOwner(effect);
		effect.setInUse(false);

		_effects.remove(effect);

		if(effect.getNext() != null)
		{
			L2Effect next = effect.getNext();
			_effects.add(next);
			addStatFuncs(next.getStatFuncs());
			next.updateEffects();
		}

		if(effect.getStackType().equalsIgnoreCase("HpRecoverCast"))
			sendPacket(new ShortBuffStatusUpdate());
		else
			updateEffectIcons();
	}

	public void stopAllEffects()
	{
		if(_effects != null)
			for(L2Effect e : _effects)
				if(e != null)
					e.exit();
	}

	public void stopEffect(Integer skillId)
	{
		if(_effects != null)
			for(L2Effect e : _effects)
				if(e != null && e.getSkill().getId().equals(skillId))
					e.exit();
	}

	public void stopEffects(L2Effect.EffectType type)
	{
		if(_effects != null)
			for(L2Effect e : _effects)
				if(e.getEffectType() == type)
					e.exit();
	}

	/**
	 * Remove a skill from the L2Character and its Func objects from calculator
	 * set of the L2Character.<BR>
	 * <BR>
	 *
	 * <B><U> Concept</U> :</B><BR>
	 * <BR>
	 * All skills own by a L2Character are identified in <B>_skills</B><BR>
	 * <BR>
	 *
	 * <B><U> Actions</U> :</B><BR>
	 * <BR>
	 * <li>Remove the skill from the L2Character _skills </li>
	 * <li>Remove all its Func objects from the L2Character calculator set</li>
	 * <BR>
	 * <BR>
	 *
	 * <B><U> Overriden in </U> :</B><BR>
	 * <BR>
	 * <li> L2Player : Save update in the character_skills table of the database</li>
	 * <BR>
	 * <BR>
	 *
	 * @param skill
	 *            The L2Skill to remove from the L2Character
	 *
	 * @return The L2Skill removed
	 *
	 */
	public L2Skill removeSkill(L2Skill skill)
	{
		if(skill == null)
			return null;

		return removeSkillById(skill.getId());
	}

	public L2Skill removeSkillById(Integer id)
	{
		// Remove the skill from the L2Character _skills
		L2Skill oldSkill = _skills.remove(id);
		if(_skillsOnAttack != null)
			_skillsOnAttack.remove(id);
		if(_skillsOnMagicAttack != null)
			_skillsOnMagicAttack.remove(id);
		if(_skillsOnUnderAttack != null)
			_skillsOnUnderAttack.remove(id);

		// Remove all its Func objects from the L2Character calculator set
		if(oldSkill != null)
			removeStatsOwner(oldSkill);

		return oldSkill;
	}

	/**
	 * Remove a Func from the Calculator set of the L2Character.<BR>
	 * <BR>
	 *
	 * <B><U> Concept</U> :</B><BR>
	 * <BR>
	 * A L2Character owns a table of Calculators called <B>_calculators</B>.
	 * Each Calculator (a calculator per state) own a table of Func object. A
	 * Func object is a mathematic function that permit to calculate the
	 * modifier of a state (ex : REGENERATE_HP_RATE...). To reduce cache memory
	 * use, L2NpcInstances who don't have skills share the same Calculator set
	 * called <B>NPC_STD_CALCULATOR</B>.<BR>
	 * <BR>
	 *
	 * That's why, if a L2NpcInstance is under a skill/spell effect that modify
	 * one of its state, a copy of the NPC_STD_CALCULATOR must be create in its
	 * _calculators before addind new Func object.<BR>
	 * <BR>
	 *
	 * <B><U> Actions</U> :</B><BR>
	 * <BR>
	 * <li>Remove the Func object from _calculators</li>
	 * <BR>
	 * <BR>
	 * <li>If L2Character is a L2NpcInstance and _calculators is equal to
	 * NPC_STD_CALCULATOR, free cache memory and just create a link on
	 * NPC_STD_CALCULATOR in _calculators</li>
	 * <BR>
	 * <BR>
	 *
	 * @param f
	 *            The Func object to remove from the Calculator corresponding to
	 *            the state affected
	 */
	public final synchronized void removeStatFunc(Func f)
	{
		if(f == null)
			return;

		// Select the Calculator of the affected state in the Calculator set
		int stat = f._stat.ordinal();

		if(_calculators[stat] == null)
			return;

		// Remove the Func object from the Calculator
		_calculators[stat].removeFunc(f);

		if(_calculators[stat].size() == 0)
			_calculators[stat] = null;
	}

	/**
	 * Remove a list of Funcs from the Calculator set of the L2Player.<BR>
	 * <BR>
	 *
	 * <B><U> Concept</U> :</B><BR>
	 * <BR>
	 * A L2Character owns a table of Calculators called <B>_calculators</B>.
	 * Each Calculator (a calculator per state) own a table of Func object. A
	 * Func object is a mathematic function that permit to calculate the
	 * modifier of a state (ex : REGENERATE_HP_RATE...). <BR>
	 * <BR>
	 *
	 * <FONT COLOR=#FF0000><B> <U>Caution</U> : This method is ONLY for
	 * L2Player</B></FONT><BR>
	 * <BR>
	 *
	 * <B><U> Example of use </U> :</B><BR>
	 * <BR>
	 * <li> Unequip an item from inventory</li>
	 * <li> Stop an active skill</li>
	 * <BR>
	 * <BR>
	 *
	 * @param funcs
	 *            The list of Func objects to add to the Calculator
	 *            corresponding to the state affected
	 */
	public final synchronized void removeStatFuncs(Func[] funcs)
	{
		for(Func f : funcs)
			removeStatFunc(f);
	}

	/**
	 * Remove all Func objects with the selected owner from the Calculator set
	 * of the L2Character.<BR>
	 * <BR>
	 *
	 * <B><U> Concept</U> :</B><BR>
	 * <BR>
	 * A L2Character owns a table of Calculators called <B>_calculators</B>.
	 * Each Calculator (a calculator per state) own a table of Func object. A
	 * Func object is a mathematic function that permit to calculate the
	 * modifier of a state (ex : REGENERATE_HP_RATE...). To reduce cache memory
	 * use, L2NpcInstances who don't have skills share the same Calculator set
	 * called <B>NPC_STD_CALCULATOR</B>.<BR>
	 * <BR>
	 *
	 * That's why, if a L2NpcInstance is under a skill/spell effect that modify
	 * one of its state, a copy of the NPC_STD_CALCULATOR must be create in its
	 * _calculators before addind new Func object.<BR>
	 * <BR>
	 *
	 * <B><U> Actions</U> :</B><BR>
	 * <BR>
	 * <li>Remove all Func objects of the selected owner from _calculators</li>
	 * <BR>
	 * <BR>
	 * <li>If L2Character is a L2NpcInstance and _calculators is equal to
	 * NPC_STD_CALCULATOR, free cache memory and just create a link on
	 * NPC_STD_CALCULATOR in _calculators</li>
	 * <BR>
	 * <BR>
	 *
	 * <B><U> Example of use </U> :</B><BR>
	 * <BR>
	 * <li> Unequip an item from inventory</li>
	 * <li> Stop an active skill</li>
	 * <BR>
	 * <BR>
	 *
	 * @param owner
	 *            The Object(Skill, Item...) that has created the effect
	 */
	public final void removeStatsOwner(Object owner)
	{
		synchronized (_calculators)
		{
			// Go through the Calculator set
			for(int i = 0; i < _calculators.length; i++)
				if(_calculators[i] != null)
				{
					// Delete all Func objects of the selected owner
					_calculators[i].removeOwner(owner);

					if(_calculators[i].size() == 0)
						_calculators[i] = null;
				}
		}
	}

	/**
	 * Remove the object from the list of L2Character that must be informed of
	 * HP/MP updates of this L2Character.<BR>
	 * <BR>
	 *
	 * <B><U> Concept</U> :</B><BR>
	 * <BR>
	 * Each L2Character owns a list called <B>_statusListener</B> that contains
	 * all L2Player to inform of HP/MP updates. Players who must be informed are
	 * players that target this L2Character. When a RegenTask is in progress
	 * sever just need to go through this list to send Server->Client packet
	 * StatusUpdate.<BR>
	 * <BR>
	 *
	 * <B><U> Example of use </U> :</B><BR>
	 * <BR>
	 * <li> Untarget a PC or NPC</li>
	 * <BR>
	 * <BR>
	 *
	 * @param object
	 *            L2Character to add to the listener
	 *
	 */
	public void removeStatusListener(L2Character object)
	{
		synchronized (this)
		{
			if(getStatusListener() == null)
				return;
			getStatusListener().remove(object);
			if(getStatusListener() != null && getStatusListener().isEmpty())
				setStatusListener(null);
		}
	}

	public void sendActionFailed()
	{
		sendPacket(Msg.ActionFail);
	}

	public L2CharacterAI setAI(L2CharacterAI new_ai)
	{
		if(new_ai == null)
			return _ai = null;
		if(_ai != null)
			_ai.stopAITask();
		_ai = new_ai;
		return _ai;
	}

	/**
	 * Set the current CP of the L2Character.<BR>
	 * <BR>
	 *
	 * <B><U> Actions</U> :</B><BR>
	 * <BR>
	 * <li>Get the Max CP of the L2Character </li>
	 * <li>Set the RegenActive flag to True/False </li>
	 * <li>Launch/Stop the HP/MP/CP Regeneration task with Medium priority</li>
	 * <li>Send the Server->Client packet StatusUpdate with current HP and MP
	 * to all other L2Player to inform</li>
	 * <BR>
	 * <BR>
	 *
	 */
	public final void setCurrentCp(double newCp)
	{
		// Get the Max CP of the L2Character
		int maxCp = getMaxCp();

		if(newCp < 0)
			newCp = 0;

		synchronized (this)
		{
			if(newCp >= maxCp)
			{
				// Set the RegenActive flag to false
				_currentCp = maxCp;
				_flagsRegenActive &= ~CP_REGEN_FLAG;

				// Stop the HP/MP/CP Regeneration task
				if(_flagsRegenActive == 0)
					stopHpMpRegeneration();
			}
			else
			{
				// Set the RegenActive flag to true
				_currentCp = newCp;
				_flagsRegenActive |= CP_REGEN_FLAG;

				// Start the HP/MP/CP Regeneration task with Medium priority
				startHpMpRegeneration();
			}
		}

		// Send the Server->Client packet StatusUpdate with current HP and MP to
		// all other L2Player to inform
		broadcastStatusUpdate();
	}

	/**
	 * Set the current HP of the L2Character.<BR>
	 * <BR>
	 *
	 * <B><U> Actions</U> :</B><BR>
	 * <BR>
	 * <li>Get the Max HP of the L2Character </li>
	 * <li>Set the RegenActive flag to True/False </li>
	 * <li>Launch/Stop the HP/MP/CP Regeneration task with Medium priority</li>
	 * <li>Send the Server->Client packet StatusUpdate with current HP and MP
	 * to all other L2Player to inform</li>
	 * <BR>
	 * <BR>
	 *
	 */
	public final void setCurrentHp(double newHp)
	{
		// Get the Max HP of the L2Character
		int maxHp = getMaxHp();
		double hpStart = _currentHp;

		synchronized (this)
		{
			if(newHp >= maxHp)
			{
				// Set the RegenActive flag to false
				_currentHp = maxHp;
				_flagsRegenActive &= ~HP_REGEN_FLAG;

				// Stop the HP/MP/CP Regeneration task
				if(_flagsRegenActive == 0)
					stopHpMpRegeneration();
			}
			else
			{
				// Set the RegenActive flag to true
				_currentHp = newHp;
				_flagsRegenActive |= HP_REGEN_FLAG;

				// Start the HP/MP/CP Regeneration task with Medium priority
				startHpMpRegeneration();
			}
		}

		checkHpMessages(hpStart, newHp);

		firePropertyChanged(PropertyCollection.HitPoints, hpStart, _currentHp);

		// Send the Server->Client packet StatusUpdate with current HP and MP to
		// all other L2Player to inform
		broadcastStatusUpdate();
	}

	/**
	 * Set the current HP and MP of the L2Character, Launch/Stop a HP/MP/CP
	 * Regeneration Task and send StatusUpdate packet to all other L2Player to
	 * inform (exclusive broadcast).<BR>
	 * <BR>
	 *
	 * <B><U> Actions</U> :</B><BR>
	 * <BR>
	 * <li>Get the Max HP and MP of the L2Character </li>
	 * <li>Set the RegenActive flag to True/False </li>
	 * <li>Launch/Stop the HP/MP/CP Regeneration task with Medium priority</li>
	 * <li>Send the Server->Client packet StatusUpdate with current HP and MP
	 * to all other L2Player to inform</li>
	 * <BR>
	 * <BR>
	 *
	 * <FONT COLOR=#FF0000><B> <U>Caution</U> : This method DOESN'T SEND
	 * Server->Client StatusUpdate packet to the L2Character</B></FONT><BR>
	 * <BR>
	 *
	 * @param newHp
	 *            The new HP value of the L2Player
	 * @param newMp
	 *            The new MP value of the L2Player
	 *
	 */
	public void setCurrentHpMp(double newHp, double newMp)
	{
		// Get the Max HP and MP of the L2Character
		int maxHp = getMaxHp();
		int maxMp = getMaxMp();
		double curHp = _currentHp;

		synchronized (this)
		{
			if(newHp >= maxHp)
			{
				// Set the RegenActive flag to false
				_killedAlready = false;
				_currentHp = maxHp;
				_flagsRegenActive &= ~HP_REGEN_FLAG;
			}
			else
			{
				_currentHp = newHp;

				if(!isDead())
				{
					// Set the RegenActive flag to true
					_killedAlready = false;
					_flagsRegenActive |= HP_REGEN_FLAG;

					// Start the HP/MP/CP Regeneration task with Medium priority
					startHpMpRegeneration();
				}
			}

			if(newMp >= maxMp)
			{
				// Set the RegenActive flag to false
				_currentMp = maxMp;
				_flagsRegenActive &= ~MP_REGEN_FLAG;
			}
			else
			{
				_currentMp = newMp;

				if(!isDead())
				{
					// Set the RegenActive flag to true
					_flagsRegenActive |= MP_REGEN_FLAG;

					// Start the HP/MP/CP Regeneration task with Medium priority
					startHpMpRegeneration();
				}
			}

			// Stop the HP/MP/CP Regeneration task
			if(_flagsRegenActive == 0)
				stopHpMpRegeneration();
		}

		checkHpMessages(curHp, newHp);

		// Send the Server->Client packet StatusUpdate with current HP and MP to
		// all other L2Player to inform
		broadcastStatusUpdate();
	}

	public final void setCurrentMp(double newMp)
	{
		// Get the Max MP of the L2Character
		int maxMp = getMaxMp();

		synchronized (this)
		{
			if(newMp >= maxMp)
			{
				// Set the RegenActive flag to false
				_currentMp = maxMp;
				_flagsRegenActive &= ~MP_REGEN_FLAG;

				// Stop the HP/MP/CP Regeneration task
				if(_flagsRegenActive == 0)
					stopHpMpRegeneration();
			}
			else
			{
				// Set the RegenActive flag to true
				_currentMp = newMp;
				_flagsRegenActive |= MP_REGEN_FLAG;

				// Start the HP/MP/CP Regeneration task with Medium priority
				startHpMpRegeneration();
			}
		}

		// Send the Server->Client packet StatusUpdate with current HP and MP to
		// all other L2Player to inform
		broadcastStatusUpdate();
	}

	/**
	 * Set the current MP of the L2Character.<BR>
	 * <BR>
	 *
	 * <B><U> Actions</U> :</B><BR>
	 * <BR>
	 * <li>Get the Max HP of the L2Character </li>
	 * <li>Set the RegenActive flag to True/False </li>
	 * <li>Launch/Stop the HP/MP/CP Regeneration task with Medium priority</li>
	 * <li>Send the Server->Client packet StatusUpdate with current HP and MP
	 * to all other L2Player to inform</li>
	 * <BR>
	 * <BR>
	 *
	 */
	public final void setCurrentMp(Double newMp)
	{
		setCurrentMp((double) newMp);
	}

	/**
	 * Set the L2Character flying mode to True.<BR>
	 * <BR>
	 */
	public final void setFlying(boolean mode)
	{
		_flying = mode;
	}

	/**
	 * Set the orientation of the L2Character.<BR>
	 * <BR>
	 */
	public final void setHeading(int heading)
	{
		_heading = heading;
	}

	/**
	 * Set the orientation of the L2Character to the target.<BR>
	 * <BR>
	 */
	public final void setHeading(L2Character target, boolean toChar)
	{
		if(target == null || target == this)
			return;

		_heading = (int) (Math.atan2(getY() - target.getY(), getX() - target.getX()) * 32768. / Math.PI) + (toChar ? 32768 : 0);
		if(_heading < 0)
			_heading += 65536;
	}

	public final void setIsBlessedByNoblesse(boolean value)
	{
		if(value)
			_isBlessedByNoblesse++;
		else
			_isBlessedByNoblesse--;
	}

	public final void setIsSalvation(boolean value)
	{
		if(value)
			_isSalvation++;
		else
			_isSalvation--;
	}

	public final void setBuffImmunity(boolean value)
	{
		if(value)
			_buffImmunity++;
		else
			_buffImmunity--;
	}

	/**
	 * Set the invulnerability Flag of the L2Character.<BR>
	 */
	public void setIsInvul(boolean b)
	{
		_isInvul = b;
	}

	public final void setIsPendingRevive(boolean value)
	{
		_isPendingRevive = value;
	}

	public final void setIsTeleporting(boolean value)
	{
		_isTeleporting = value;
	}

	public final void setName(String name)
	{
		_name = name;
	}

	public L2Character getCastingTarget()
	{
		if(castingTarget == null)
			return null;

		L2Character c = castingTarget.get();
		if(c == null)
			castingTarget = null;

		return c;
	}

	public void setCastingTarget(L2Character target)
	{
		castingTarget = target == null ? null : new WeakReference<L2Character>(target);
	}

	/**
	 * Set the L2Character riding mode to True.<BR>
	 * <BR>
	 */
	public final void setRiding(boolean mode)
	{
		_riding = mode;
	}

	/**
	 * Set the L2Character movement type to run and send Server->Client packet
	 * ChangeMoveType to all others L2Player.<BR>
	 * <BR>
	 */
	public final void setRunning()
	{
		if(!_running)
		{
			_running = true;
			broadcastPacket(new ChangeMoveType(this));
		}
	}

	public void setSkillMastery(Integer skill, byte mastery)
	{
		_skillMastery.put(skill, mastery);
	}

	private void setStatusListener(CopyOnWriteArraySet<L2Character> value)
	{
		_statusListener = value;
	}

	public void setTarget(L2Object object)
	{
		if(object != null && !object.isVisible())
			object = null;
		// If object==null, Cancel Attak or Cast
		if(object == null)
		{
			if(getTarget() != null)
				broadcastPacket(new TargetUnselected(this));
			if(isAttackingNow() && getAI().getAttackTarget() == getTarget())
			{
				abortAttack();
				getAI().setIntention(AI_INTENTION_ACTIVE, null, null);
				if(this instanceof L2Player)
				{
					sendMessage(new CustomMessage("l2r.gameserver.model.L2Character.AttackAborted", this));
					sendActionFailed();
				}
			}
			if(isCastingNow() && canAbortCast() && getAI().getAttackTarget() == getTarget())
			{
				abortCast();
				getAI().setIntention(AI_INTENTION_ACTIVE, null, null);
				if(this instanceof L2Player)
				{
					sendMessage(new CustomMessage("l2r.gameserver.model.L2Character.CastingAborted", this));
					sendActionFailed();
				}
			}
		}

		target = object == null ? null : new WeakReference<L2Object>(object);
	}

	/**
	 * Set the template of the L2Character.<BR>
	 * <BR>
	 *
	 * <B><U> Concept</U> :</B><BR>
	 * <BR>
	 * Each L2Character owns generic and static properties (ex : all Keltir have
	 * the same number of HP...). All of those properties are stored in a
	 * different template for each type of L2Character. Each template is loaded
	 * once in the server cache memory (reduce memory use). When a new instance
	 * of L2Character is spawned, server just create a link between the instance
	 * and the template This link is stored in <B>_template</B><BR>
	 * <BR>
	 *
	 * <B><U> Assert </U> :</B><BR>
	 * <BR>
	 * <li> isCharacter</li>
	 * <BR>
	 * <BR
	 *
	 */
	protected void setTemplate(L2CharTemplate template)
	{
		_template = template;
	}

	/**
	 * Set the Title of the L2Character.<BR>
	 * <BR>
	 *
	 * @param title
	 *            The text to set as title
	 *
	 */
	public void setTitle(String title)
	{
		_title = title;
	}

    private L2Character aggressionTarget = null;

	public void setAggressionTarget(L2Character target)
	{
		aggressionTarget = target;
	}

	public L2Character getAggressionTarget()
	{
		return aggressionTarget;
	}

	/**
	 * Set the L2Character movement type to walk and send Server->Client packet
	 * ChangeMoveType to all others L2Player.<BR>
	 * <BR>
	 */
	public void setWalking()
	{
		if(_running)
		{
			_running = false;
			broadcastPacket(new ChangeMoveType(this));
		}
	}

	/**
	 * Active abnormal effects flags in the binary mask and send Server->Client
	 * UserInfo/CharInfo packet.<BR>
	 * <BR>
	 */
	public void startAbnormalEffect(int mask)
	{
		_abnormalEffects |= mask;
		updateAbnormalEffect();
	}

	@Override
	public void startAttackStanceTask()
	{
		if(System.currentTimeMillis() < _stanceInited + 10000)
			return;

		_stanceInited = System.currentTimeMillis();

		// Бесконечной рекурсии не будет, потому что выше проверка на
		// _stanceInited
		if(this instanceof L2Summon && getOwner() != null)
			getOwner().startAttackStanceTask();
		else if(isPlayer && getPet() != null)
			getPet().startAttackStanceTask();

		if(_stanceTask != null)
			_stanceTask.cancel(false);
		else
			broadcastPacket(new AutoAttackStart(getObjectId()));

		_stanceTask = ThreadPoolManager.getInstance().scheduleAi(new CancelAttackStance(), 15000,  isPlayer || isPet || isSummon);
	}

	/**
	 * Start the HP/MP/CP Regeneration task.<BR>
	 * <BR>
	 *
	 * <B><U> Actions</U> :</B><BR>
	 * <BR>
	 * <li>Calculate the regen task period </li>
	 * <li>Launch the HP/MP/CP Regeneration task with Medium priority </li>
	 * <BR>
	 * <BR>
	 *
	 */
	protected synchronized void startHpMpRegeneration()
	{
		if(_regTask == null && !isDead())
		{
			if(Config.DEBUG)
				_log.fine("HP/MP/CP regen started");

			// Get the Regeneration periode
			int period = Formulas.getRegeneratePeriod(this);

			// Create the HP/MP/CP Regeneration task
			_regTask = ThreadPoolManager.getInstance().scheduleEffectAtFixedRate(new RegenTask(), period, period);
		}
	}

	/**
	 * Modify the abnormal effect map according to the mask.<BR>
	 * <BR>
	 */
	public void stopAbnormalEffect(int mask)
	{
		_abnormalEffects &= ~mask;
		updateAbnormalEffect();
	}

	/**
	 * Stop the HP/MP/CP Regeneration task.<BR>
	 * <BR>
	 *
	 * <B><U> Actions</U> :</B><BR>
	 * <BR>
	 * <li>Set the RegenActive flag to False </li>
	 * <li>Stop the HP/MP/CP Regeneration task </li>
	 * <BR>
	 * <BR>
	 *
	 */
	public synchronized void stopHpMpRegeneration()
	{
		if(_regTask != null)
		{
			// Stop the HP/MP/CP Regeneration task
			_regTask.cancel(true);
			_regTask = null;

			// Set the RegenActive flag to false
			_flagsRegenActive = 0;

			if(Config.DEBUG)
				_log.fine("HP/MP/CP regen stop");
		}
	}

	public void block()
	{
		_blocked = true;
	}

	public void unblock()
	{
		_blocked = false;
	}

	public void startConfused()
	{
		if(!_confused)
		{
			_confused = true;
			startAttackStanceTask();
			updateAbnormalEffect();
		}
	}

	public void stopConfused()
	{
		if(_confused)
		{
			_confused = false;
			updateAbnormalEffect();

			breakAttack();
			breakCast(true);
			stopMove();
			getAI().setAttackTarget(null);
		}
	}

	public void startFakeDeath()
	{
		if(!_fakeDeath)
		{
			if(isPlayer)
				((L2Player) this).clearHateList();
			_fakeDeath = true;
			getAI().notifyEvent(CtrlEvent.EVT_FAKE_DEATH, null, null);
			broadcastPacket(new ChangeWaitType(this, ChangeWaitType.WT_START_FAKEDEATH));
			updateAbnormalEffect();
		}
	}

	public void stopFakeDeath()
	{
		if(_fakeDeath)
		{
			_fakeDeath = false;
			broadcastPacket(new ChangeWaitType(this, ChangeWaitType.WT_STOP_FAKEDEATH));
			updateAbnormalEffect();
		}
	}

	public void startFear()
	{
		if(!_afraid)
		{
			_afraid = true;
			breakAttack();
			breakCast(true);
			sendActionFailed();
			stopMove();
			startAttackStanceTask();
			updateAbnormalEffect();
		}
	}

	public void stopFear()
	{
		if(_afraid)
		{
			_afraid = false;
			updateAbnormalEffect();
		}
	}

	public void startMuted()
	{
		if(!_muted)
		{
			_muted = true;
			if(getCastingSkill() != null && getCastingSkill().isMagic())
				breakCast(false);
			startAttackStanceTask();
			updateAbnormalEffect();
		}
	}

	public void stopMuted()
	{
		if(_muted)
		{
			_muted = false;
			updateAbnormalEffect();
		}
	}

	public void startPMuted()
	{
		if(!_pmuted)
		{
			_pmuted = true;
			if(getCastingSkill() != null && !getCastingSkill().isMagic())
				breakCast(false);
			startAttackStanceTask();
			updateAbnormalEffect();
		}
	}

	public void stopPMuted()
	{
		if(_pmuted)
		{
			_pmuted = false;
			updateAbnormalEffect();
		}
	}

	public void startRooted()
	{
		if(!_rooted)
		{
			_rooted = true;
			getAI().clientStopMoving();
			startAttackStanceTask();
			updateAbnormalEffect();
		}
	}

	public void stopRooting()
	{
		if(_rooted)
		{
			_rooted = false;
			updateAbnormalEffect();
		}
	}

	public void startSleeping()
	{
		if(!_sleeping)
		{
			_sleeping = true;
			breakAttack();
			breakCast(true);
			sendActionFailed();
			stopMove();
			startAttackStanceTask();
			updateAbnormalEffect();
		}
	}

	public void stopSleeping()
	{
		if(_sleeping)
		{
			_sleeping = false;
			updateAbnormalEffect();
		}
	}

	public void startStunning()
	{
		if(!_stunned)
		{
			_stunned = true;
			breakAttack();
			breakCast(true);
			sendActionFailed();
			stopMove();
			startAttackStanceTask();
			updateAbnormalEffect();
		}
	}

	public void stopStunning()
	{
		if(_stunned)
		{
			_stunned = false;
			updateAbnormalEffect();
		}
	}

	public void setMeditated(boolean meditated)
	{
		_meditated = meditated;
	}

	public void setParalyzed(boolean paralyzed)
	{
		if(_paralyzed != paralyzed)
		{
			_paralyzed = paralyzed;
			if(paralyzed)
			{
				breakAttack();
				breakCast(true);
				sendActionFailed();
				stopMove();
			}
			updateAbnormalEffect();
		}
	}

	public void setImobilised(boolean imobilised)
	{
		if(_imobilised != imobilised)
		{
			_imobilised = imobilised;
			if(imobilised)
				stopMove();
			updateAbnormalEffect();
		}
	}

	public void setBigHead(boolean bigHead)
	{
		if(_bigHead != bigHead)
		{
			_bigHead = bigHead;
			updateAbnormalEffect();
		}
	}

	/**
	 * Set the overloaded status of the L2Character is overloaded (if True, the
	 * L2Player can't take more item).<BR>
	 * <BR>
	 */
	public void setOverloaded(boolean overloaded)
	{
		_overloaded = overloaded;
	}

	public boolean isConfused()
	{
		return _confused;
	}

	public boolean isFakeDeath()
	{
		return _fakeDeath;
	}

	public boolean isAfraid()
	{
		return _afraid;
	}

	public boolean isBlocked()
	{
		return _blocked;
	}

	public boolean isMuted()
	{
		return _muted;
	}

	public boolean isPMuted()
	{
		return _pmuted;
	}

	public boolean isRooted()
	{
		return _rooted;
	}

	public boolean isSleeping()
	{
		return _sleeping;
	}

	public boolean isStunned()
	{
		return _stunned;
	}

	public boolean isMeditated()
	{
		return _meditated;
	}

	public boolean isParalyzed()
	{
		return _paralyzed;
	}

	public boolean isImobilised()
	{
		return _imobilised;
	}

	public boolean isBigHead()
	{
		return _bigHead;
	}

	/**
	 * Return True if the L2Character is casting.<BR>
	 * <BR>
	 */
	public final boolean isCastingNow()
	{
		return _castEndTime > GameTimeController.getGameTicks();
	}

	/**
	 * Return True if the L2Character can't move (stun, root, sleep, overload,
	 * paralyzed).<BR>
	 * <BR>
	 */
	public boolean isMovementDisabled()
	{
		return isSitting() || isStunned() || isRooted() || isSleeping() || isParalyzed() || isImobilised() || isAlikeDead() || isAttackingNow() || isCastingNow() || _overloaded || _fishing;
	}

	/**
	 * Return True if the L2Character can't use its skills (ex : stun,
	 * sleep...).<BR>
	 * <BR>
	 */
	public boolean isActionsDisabled()
	{
		return isStunned() || isSleeping() || isParalyzed() || isAttackingNow() || isCastingNow() || isInBoat() || isAlikeDead();
	}

	public boolean isPotionsDisabled()
	{
		return isStunned() || isSleeping() || isParalyzed() || isAlikeDead() || isInBoat();
	}

	/**
	 * Return True if the L2Character can't attack (stun, sleep, reuse).<BR>
	 * <BR>
	 */
	public final boolean isAttackingDisabled()
	{
		return _attackReuseEndTime > System.currentTimeMillis();
	}

	/**
	 * Return True if the L2Character can't be controlled by the player
	 * (confused, affraid).<BR>
	 * <BR>
	 */
	public boolean isOutOfControl()
	{
		return isConfused() || isAfraid() || isBlocked();
	}

	public void teleToLocation(Location loc)
	{
		teleToLocation(loc.getX(), loc.getY(), loc.getZ());
	}

	public void teleToLocation(int x, int y, int z)
	{
		z = GeoClient.getInstance().getHeight(x, y, z);

		if(isPlayer && DimensionalRiftManager.getInstance().checkIfInRiftZone(getLoc(), true))
		{
			L2Player player = (L2Player) this;
			if(player.isInParty() && player.getParty().isInDimensionalRift())
			{
				Location newCoords = DimensionalRiftManager.getInstance().getRoom((byte) 0, (byte) 0).getTeleportCoords();
				x = newCoords.getX();
				y = newCoords.getY();
				z = newCoords.getZ();
				player.getParty().getDimensionalRift().usedTeleport(player);
			}
		}

		setTarget(null);
		stopMove();

		if(isPlayer)
		{
			L2Player player = (L2Player) this;

			if(player.isInBoat())
			{
				player.setInBoat(false);
				player.setBoat(null);
				player.setBoatId(0);
			}

			decayMe();

			setXYZInvisible(x, y, z);
			setIsTeleporting(true);

			// Нужно при телепорте с более высокой точки на более низкую, иначе наносится вред от "падения"
			player.setLastClientPosition(null);
			player.setLastServerPosition(null);

			player.sendPacket(new TeleportToLocation(player, x, y, z));
			if(player.getParty() != null)
			{
				Point position = player.getLastPartyPositionSent();
				if(player.getDistance(position.x, position.y) < 500)
				{
					player.getParty().broadcastToPartyMembers(player, new PartyMemberPosition(player));
					position.move(player.getX(), player.getY());
				}
			}
		}
		else
		{
			setXYZ(x, y, z);
			broadcastPacket(new TeleportToLocation(this, x, y, z));
		}
	}

	public void onTeleported()
	{
		if(isPlayer)
		{
			spawnMe();

			L2Player player = (L2Player) this;

			setLastClientPosition(getLoc());
			setLastServerPosition(getLoc());

			setIsTeleporting(false);

			if(_isPendingRevive)
				doRevive();

			if(player.getTrainedBeast() != null)
				player.getTrainedBeast().setXYZ(getX() + Rnd.get(-100, 100), getY() + Rnd.get(-100, 100), getZ());

			sendActionFailed();
		}
	}

	public void sendMessage(CustomMessage message)
	{
		sendMessage(message.toString());
	}

	private long _nonAggroTime;

	public long getNonAggroTime()
	{
		return _nonAggroTime;
	}

	public void setNonAggroTime(long time)
	{
		_nonAggroTime = time;
	}

	public int getCustomEffect()
	{
		return _customEffect;
	}

	public void setCustomEffect(int effect)
	{
		_customEffect = effect;
	}

	@Override
	public String toString()
	{
		return "mob " + getObjectId();
	}

	// ---------------------------- Not Implemented -------------------------------

	@SuppressWarnings("unused")
	public void addExpAndSp(long addToExp, long addToSp)
	{}

	@SuppressWarnings("unused")
	public void addSkillTimeStamp(Integer s, long r)
	{}

	@SuppressWarnings("unused")
	public void broadcastUserInfo()
	{}

	@SuppressWarnings("unused")
	public void checkHpMessages(double currentHp, double newHp)
	{}

	@SuppressWarnings("unused")
	public boolean checkPvP(L2Character target, L2Skill skill)
	{
		return false;
	}

	@SuppressWarnings("unused")
	public boolean consumeItem(int itemConsumeId, int itemCount)
	{
		return true;
	}

	@SuppressWarnings("unused")
	public void doPickupItem(L2Object object)
	{}

	@SuppressWarnings("unused")
	public boolean isFearImmune()
	{
		return false;
	}

	@SuppressWarnings("unused")
	public boolean isLethalImmune()
	{
		return false;
	}

	@SuppressWarnings("unused")
	public boolean getChargedSoulShot()
	{
		return false;
	}

	@SuppressWarnings("unused")
	public int getChargedSpiritShot()
	{
		return 0;
	}

	@SuppressWarnings("unused")
	public Duel getDuel()
	{
		return null;
	}

	@SuppressWarnings("unused")
	public int getIncreasedForce()
	{
		return 0;
	}

	@SuppressWarnings("unused")
	public int getKarma()
	{
		return 0;
	}

	@SuppressWarnings("unused")
	public double getLevelMod()
	{
		return 1;
	}

	@SuppressWarnings("unused")
	public int getNpcId()
	{
		return 0;
	}

	@SuppressWarnings("unused")
	public L2Summon getPet()
	{
		return null;
	}

	@SuppressWarnings("unused")
	public int getPvpFlag()
	{
		return 0;
	}

	@SuppressWarnings("unused")
	public int getTeam()
	{
		return 0;
	}

	@SuppressWarnings("unused")
	public boolean isOverhit()
	{
		return false;
	}

	@SuppressWarnings("unused")
	public boolean isSitting()
	{
		return false;
	}

	@SuppressWarnings("unused")
	public boolean isUndead()
	{
		return false;
	}

	@SuppressWarnings("unused")
	public boolean isUsingDualWeapon()
	{
		return false;
	}

	@SuppressWarnings("unused")
	public void setOverhitEnabled(boolean b)
	{}

	@SuppressWarnings("unused")
	public boolean paralyzeImmune()
	{
		return false;
	}

	@SuppressWarnings("unused")
	public void reduceArrowCount()
	{}

	@SuppressWarnings("unused")
	public void removeSkillTimeStamp(Integer s)
	{}

	@SuppressWarnings("unused")
	public void sendChanges()
	{}

	@SuppressWarnings("unused")
	public void sendMessage(String message)
	{}

	/*@SuppressWarnings("unused")
	@Deprecated
	public void sendMessage(String message, String message_ru)
	{}*/

	public void sendPacket(L2GameServerPacket mov)
	{
	}

	@SuppressWarnings("unused")
	public void setIncreasedForce(int i)
	{}
	@SuppressWarnings("unused")
	public void setOverhitValues(L2Character attacker, double i)
	{}

	public void sitDown()
	{}

	@SuppressWarnings("unused")
	public void standUp()
	{}

	@SuppressWarnings("unused")
	public void startPvPFlag(L2Character target)
	{}

	@SuppressWarnings("unused")
	public boolean unChargeShots(boolean spirit)
	{
		return false;
	}

	@SuppressWarnings("unused")
	public void updateEffectIcons()
	{}

	@SuppressWarnings("unused")
	public void updateStats()
	{}

	@SuppressWarnings("unused")
	public void callMinionsToAssist(L2Character attacker)
	{}

	@SuppressWarnings("unused")
	public boolean hasMinions()
	{
		return false;
	}

	@SuppressWarnings("unused")
	public boolean isCursedWeaponEquipped()
	{
		return false;
	}

	@SuppressWarnings("unused")
	public boolean isHero()
	{
		return false;
	}

	@SuppressWarnings("unused")
	public boolean isInBoat()
	{
		return false;
	}

	@SuppressWarnings("unused")
	public int getAccessLevel()
	{
		return 0;
	}

	@SuppressWarnings("unused")
	public void spawnWayPoints(Vector<Location> recorder)
	{}

	@SuppressWarnings("unused")
	public void setFollowStatus(boolean b)
	{}

	@SuppressWarnings("unused")
	public void setLastClientPosition(Location charPosition)
	{}

	@SuppressWarnings("unused")
	public void setLastServerPosition(Location charPosition)
	{}

	@SuppressWarnings("unused")
	public FastList<L2MinionInstance> getSpawnedMinions()
	{
		return null;
	}

	@SuppressWarnings("unused")
	public boolean hasRandomAnimation()
	{
		return true;
	}

	@SuppressWarnings("unused")
	public boolean hasRandomWalk()
	{
		return true;
	}

	public int getClanCrestId()
	{
		return 0;
	}

	public int getClanCrestLargeId()
	{
		return 0;
	}

	public int getAllyCrestId()
	{
		return 0;
	}

	@SuppressWarnings("unused")
	public void disableItem(L2Skill handler, long timeTotal, long timeLeft)
	{}

	public float getRateAdena()
	{
		return 1.0f;
	}

	public float getRateItems()
	{
		return 1.0f;
	}

	public float getRateExp()
	{
		return 1.0f;
	}

	public float getRateSp()
	{
		return 1.0f;
	}

	public float getRateSpoil()
	{
		return 1.0f;
	}

	@Override
	public float getColRadius()
	{
		return getTemplate().collisionRadius;
	}

	@Override
	public float getColHeight()
	{
		return getTemplate().collisionHeight;
	}

	public boolean canAttackCharacter(L2Character _target)
	{
		return _target.getPlayer() != null;
	}

	public class HateInfo
	{
		public L2NpcInstance npc;
		public int hate;
		public int damage;

		HateInfo(L2NpcInstance attacker)
		{
			npc = attacker;
		}
	}

	protected ConcurrentHashMap<L2NpcInstance, HateInfo> _hateList = null;

	public void addDamage(L2NpcInstance npc, int damage)
	{
		addDamageHate(npc, damage, damage);
	}

	public void addDamageHate(L2NpcInstance npc, int damage, int aggro)
	{
		if(damage <= 0 && aggro <= 0)
			return;

		if(damage > 0 && aggro <= 0)
			aggro = damage;

		if(npc == null)
			return;

		if(_hateList == null)
			_hateList = new ConcurrentHashMap<L2NpcInstance, HateInfo>();

		HateInfo ai = _hateList.get(npc);

		if(ai != null)
		{
			ai.damage += damage;
			ai.hate += aggro;
			if(ai.hate < 0)
				ai.hate = 0;
		}
		else if(aggro > 0)
		{
			ai = new HateInfo(npc);
			ai.damage = damage;
			ai.hate = aggro;
			_hateList.put(npc, ai);
		}
	}

	public ConcurrentHashMap<L2NpcInstance, HateInfo> getHateList()
	{
		if(_hateList == null)
			return new ConcurrentHashMap<L2NpcInstance, HateInfo>();
		return _hateList;
	}

	public void removeFromHatelist(L2NpcInstance npc)
	{
		if(npc != null && _hateList != null)
			_hateList.remove(npc);
	}

	public void clearHateList()
	{
		_hateList = null;
	}

	// --------------------------- End Of Not Implemented ------------------------------

	// --------------------------------- Abstract --------------------------------------

	public abstract byte getLevel();

	public abstract void updateAbnormalEffect();

	public abstract L2ItemInstance getActiveWeaponInstance();

	public abstract L2Weapon getActiveWeaponItem();

	public abstract L2ItemInstance getSecondaryWeaponInstance();

	public abstract L2Weapon getSecondaryWeaponItem();

	// ----------------------------- End Of Abstract -----------------------------------
}
