package l2r.gameserver.model;

import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

import l2r.Config;
import l2r.gameserver.ThreadPoolManager;
import l2r.gameserver.model.L2Skill.SkillType;
import l2r.gameserver.serverpackets.ExOlympiadSpelledInfo;
import l2r.gameserver.serverpackets.MagicEffectIcons;
import l2r.gameserver.serverpackets.PartySpelled;
import l2r.gameserver.serverpackets.SystemMessage;
import l2r.gameserver.skills.Env;
import l2r.gameserver.skills.Stats;
import l2r.gameserver.skills.effects.EffectTemplate;
import l2r.gameserver.skills.effects.ResistType;
import l2r.gameserver.skills.funcs.Func;
import l2r.gameserver.skills.funcs.FuncTemplate;
import l2r.gameserver.skills.funcs.Lambda;

public abstract class L2Effect implements Comparable
{
	protected static final Logger _log = Logger.getLogger(L2Effect.class.getName());

	public static enum EffectState
	{
		CREATED,
		ACTING,
		FINISHING,
		FINISHED
	}

	public static enum EffectType
	{
		BUFF,
		DMG_OVER_TIME,
		DMG_OVER_TIME_LETHAL,
		HEAL_OVER_TIME,
		COMBAT_POINT_HEAL_OVER_TIME,
		MANA_DMG_OVER_TIME,
		MANA_HEAL_OVER_TIME,
		RELAXING,
		STUN,
		ROOT,
		SLEEP,
		HATE,
		FAKE_DEATH,
		UNAGGRO,
		CONFUSE_MOB_ONLY,
		MUTE,
		MUTE_PHYSICAL,
		MUTE_ALL,
		FEAR,
		SILENT_MOVE,
		SEED,
		PARALYZE,
		BLOCK_STAT_FUNCS,
		BLOCK_MAGIC_PHYSIC,
		MANAHEAL,
		HEAL,
		MANAHEAL_PERCENT,
		HEAL_PERCENT,
		LETHALSTRIKE1,
		LETHALSTRIKE2,
		TURNER,
		PETRIFICATION,
		BETRAY,
		WATCHER_GAZE,
		BLESS_NOBLESSE,
		MEDITATION,
		SYMBOL,
		BIG_HEAD,
		GROW,
		INTERRUPT,
		SALVATION,
		BUFFIMMUNITY,
		BATTLE_FORCE,
		SPELL_FORCE,
		TRANSFORMATION,
		DISARM,
		DESTROY_SUMMON
	}

	private static final Func[] _emptyFunctionSet = new Func[0];

	/** Накладывающий эффект */
	protected final L2Character _effector;
	/** Тот, на кого накладывают эффект */
	protected final L2Character _effected;

	protected final L2Skill _skill;

	//or the items that was used.
	//private final L2Item _item;

	// the value of an update
	private final Lambda _lambda;

	// the current state
	protected EffectState _state;

	// period, milliseconds
	private long _period;
	private long _periodStartTime;

	// function templates
	private final FuncTemplate[] _funcTemplates;
	private final Stats[] _stats;

	private final ResistType _resistType;

	// counter
	protected int _count;

	// abnormal effect mask
	private int _abnormalEffect;

	public final class EffectTask implements Runnable
	{
		protected final long _delay;
		protected final long _rate;

		EffectTask(long delay, long rate)
		{
			_delay = delay;
			_rate = rate;
		}

		public long getRate()
		{
			return _rate;
		}

		public void run()
		{
			try
			{
				scheduleEffect();
			}
			catch(Throwable e)
			{
				_log.log(Level.SEVERE, "", e);
			}
		}
	}

	private ScheduledFuture _currentFuture;
	private EffectTask _currentTask;

	/** The Identifier of the stack group */
	private final String _stackType;

	/** The position of the effect in the stack group */
	private final int _stackOrder;

	private boolean _inUse = false;
	private L2Effect _next = null;

	private boolean _skillMastery = false;

	protected L2Effect(Env env, EffectTemplate template)
	{
		_state = EffectState.CREATED;
		_skill = env.skill;
		_effector = env.character;
		_effected = env.target;
		_lambda = template._lambda;
		_funcTemplates = template._funcTemplates;
		_stats = template._stats;
		_count = template._counter;
		_period = template.getPeriod();

		// Check for skill mastery duration time increase
		if(_effector.getSkillMastery(getSkill().getId()) == 2)
		{
			if(_count > 1)
				_count *= 2;
			else
				_period *= 2;
			_skillMastery = true;
			_effector.removeSkillMastery(getSkill().getId());
		}

		if(_skill.getSkillType() == SkillType.BUFF && _period > 30000 && _skill.getId() != 396 && _skill.getId() != 1374)
			_period *= Config.BUFFTIME_MODIFIER;
		if(_skill.getSkillType() == SkillType.SONG || _skill.getSkillType() == SkillType.DANCE)
			_period *= Config.SONGDANCETIME_MODIFIER;

		_abnormalEffect = template._abnormalEffect;
		_stackType = template._stackType;
		_stackOrder = template._stackOrder;
		_resistType = template._resistType;
		_periodStartTime = System.currentTimeMillis();
	}

	public long getPeriod()
	{
		return _period;
	}

	public void setPeriod(long time)
	{
		_period = time;
	}

	public int getCount()
	{
		return _count;
	}

	public void setCount(int newcount)
	{
		_count = newcount;
	}

	public long getTime()
	{
		return System.currentTimeMillis() - _periodStartTime;
	}

	public long getPeriodStartTime()
	{
		return _periodStartTime;
	}

	public long getTimeLeft()
	{
		return getPeriod() * getCount() - getTime();
	}

	public boolean isInUse()
	{
		return _inUse;
	}

	public void setInUse(boolean inUse)
	{
		_inUse = inUse;
		if(_inUse)
			scheduleEffect();
		else if(_state != EffectState.FINISHED)
			_state = EffectState.FINISHING;
	}

	public String getStackType()
	{
		return _stackType;
	}

	public int getStackOrder()
	{
		return _stackOrder;
	}

	public L2Skill getSkill()
	{
		return _skill;
	}

	public L2Character getEffector()
	{
		return _effector;
	}

	public L2Character getEffected()
	{
		return _effected;
	}

	public double calc()
	{
		Env env = new Env();
		env.character = _effector;
		env.target = _effected;
		env.skill = _skill;
		return _lambda.calc(env);
	}

	private synchronized void startEffectTask(long duration)
	{
		if(_currentFuture != null)
		{
			// Cancel the task
			_currentFuture.cancel(false);
			_currentFuture = null;
			_currentTask = null;
		}
		_currentTask = new EffectTask(duration, -1);
		_currentFuture = ThreadPoolManager.getInstance().scheduleEffect(_currentTask, duration);
		updateEffects();
	}

	private synchronized void startEffectTaskAtFixedRate(long delay, long rate)
	{
		if(_currentFuture != null)
		{
			// Cancel the task
			_currentFuture.cancel(false);
			_currentFuture = null;
			_currentTask = null;
		}
		_currentTask = new EffectTask(delay, rate);
		_currentFuture = ThreadPoolManager.getInstance().scheduleEffectAtFixedRate(_currentTask, delay, rate);
		updateEffects();
	}

	/**
	 * Stop the L2Effect task and send Server->Client update packet.<BR><BR>
	 *
	 * <B><U> Actions</U> :</B><BR><BR>
	 * <li>Cancel the effect in the the abnormal effect map of the L2Character </li>
	 * <li>Stop the task of the L2Effect, remove it and update client magic icon </li><BR><BR>
	 *
	 */
	public void exit()
	{
		if(_next != null)
			_next.exit();
		_next = null;

		if(_state == EffectState.FINISHED)
			return;
		if(_state != EffectState.CREATED)
		{
			_state = EffectState.FINISHING;
			scheduleEffect();
		}
		else
			_state = EffectState.FINISHING;
	}

	public boolean isEnded()
	{
		return _state == EffectState.FINISHED || _state == EffectState.FINISHING;
	}

	/**
	 * Stop the task of the L2Effect, remove it and update client magic icon.<BR><BR>
	 *
	 * <B><U> Actions</U> :</B><BR><BR>
	 * <li>Cancel the task </li>
	 * <li>Stop and remove L2Effect from L2Character and update client magic icon </li><BR><BR>
	 *
	 */
	private synchronized void stopEffectTask()
	{
		if(_currentFuture != null)
		{
			// Cancel the task
			_currentFuture.cancel(true);
			_currentFuture = null;
			_currentTask = null;

			_effected.removeEffect(this);
			updateEffects();
		}
	}

	/** returns effect type */
	public abstract EffectType getEffectType();

	/** Notify started */
	public void onStart()
	{
		if(_abnormalEffect != 0)
			getEffected().startAbnormalEffect(_abnormalEffect);
	}

	/**
	 * Cancel the effect in the the abnormal effect map of the effected L2Character.<BR><BR>
	 */
	public void onExit()
	{
		if(_abnormalEffect != 0)
			getEffected().stopAbnormalEffect(_abnormalEffect);
	}

	/** Return true for continuation of this effect */
	public abstract boolean onActionTime();

	public final void rescheduleEffectOnePeriod()
	{
		if(_state != EffectState.ACTING)
			scheduleEffect();
		else
		{
			_currentFuture.cancel(false);
			_currentTask = new EffectTask(getPeriod(), -1);
			_currentFuture = ThreadPoolManager.getInstance().scheduleGeneral(_currentTask, getPeriod());
			_effected.updateEffectIcons();
		}
	}

	public final void scheduleEffect()
	{
		if(_state == EffectState.CREATED)
		{
			_state = EffectState.ACTING;
			onStart();

			// Fake Death и Silent Move не отображаются
			// Отображать сообщение только для первого эффекта скилла
			if(_skill.getId() != 60 && _skill.getId() != 221 && getEffected().getEffectsCountForSkill(getSkill().getId()) == 1)
				getEffected().sendPacket(new SystemMessage(SystemMessage.YOU_CAN_FEEL_S1S_EFFECT).addSkillName(_skill.getId(), _skill.getDisplayLevel()));

			_effected.updateStats(); // Обрабатываем отображение статов

			if(_count > 1)
				startEffectTaskAtFixedRate(_skill.isToggle() ? _period : 5, _period);
			else
				startEffectTask(_period);

			return;
		}

		if(_state == EffectState.ACTING)
		{
			if(_count > 0)
			{
				_count--;
				if(onActionTime())
					return;
			}
			_state = EffectState.FINISHING;
		}

		if(_state == EffectState.FINISHING)
		{
			_state = EffectState.FINISHED;

			// Для ускоренной "остановки" эффекта
			_inUse = false;

			// Cancel the effect in the the abnormal effect map of the L2Character
			onExit();

			// If the time left is equal to zero, send the message
			// Отображать сообщение только для последнего оставшегося эффекта скилла
			if(_count == 0 && getEffected().getEffectsCountForSkill(getSkill().getId()) == 1)
				getEffected().sendPacket(new SystemMessage(SystemMessage.THE_EFFECT_OF_S1_HAS_WORN_OFF).addSkillName(_skill.getId(), _skill.getLevel()));

			// Stop the task of the L2Effect, remove it and update client magic icon
			stopEffectTask();
		}
	}

	public void updateEffects()
	{
		_effected.updateStats();
	}

	public Func[] getStatFuncs()
	{
		if(_funcTemplates == null)
			return _emptyFunctionSet;
		Func[] funcs = new Func[_funcTemplates.length];
		for(int i = 0; i < funcs.length; i++)
		{
			Env env = new Env();
			env.character = _effector;
			env.target = _effected;
			env.skill = _skill;
			Func f = _funcTemplates[i].getFunc(env, this); // effect is owner
			funcs[i] = f;
		}
		return funcs;
	}

	public void addIcon(MagicEffectIcons mi)
	{
		EffectTask task = _currentTask;
		ScheduledFuture future = _currentFuture;
		if(task == null || future == null)
			return;
		if(_state != EffectState.ACTING)
			return;
		int duration = task.getRate() > 0 || _skill.isToggle() ? MagicEffectIcons.INFINITIVE_EFFECT : (int) future.getDelay(TimeUnit.SECONDS);
		mi.addEffect(_skill.getDisplayId(), _skill.getDisplayLevel(), duration);
	}

	public void addPartySpelledIcon(PartySpelled ps)
	{
		EffectTask task = _currentTask;
		ScheduledFuture future = _currentFuture;
		if(task == null || future == null)
			return;
		if(_state != EffectState.ACTING)
			return;
		int duration = task.getRate() > 0 || _skill.isToggle() ? MagicEffectIcons.INFINITIVE_EFFECT : (int) future.getDelay(TimeUnit.SECONDS);
		ps.addPartySpelledEffect(_skill.getDisplayId(), _skill.getDisplayLevel(), duration);
	}

	public void addOlympiadSpelledIcon(L2Player player, ExOlympiadSpelledInfo os)
	{
		EffectTask task = _currentTask;
		ScheduledFuture future = _currentFuture;
		if(task == null || future == null)
			return;
		if(_state != EffectState.ACTING)
			return;
		int duration = task.getRate() > 0 || _skill.isToggle() ? MagicEffectIcons.INFINITIVE_EFFECT : (int) future.getDelay(TimeUnit.SECONDS);
		os.addSpellRecivedPlayer(player);
		os.addEffect(_skill.getDisplayId(), _skill.getDisplayLevel(), duration);
	}

	protected int getLevel()
	{
		return _skill.getLevel();
	}

	public boolean containsStat(Stats stat)
	{
		if(_stats != null)
			for(Stats s : _stats)
				if(s == stat)
					return true;
		return false;
	}

	public final ResistType getResistType()
	{
		return _resistType;
	}

	public boolean isSkillMasteryEffect()
	{
		return _skillMastery;
	}

	public int compareTo(Object obj)
	{
		if(obj.equals(this))
			return 0;
		return 1;
	}

	public void removeNext()
	{
		_next = null;
	}

	public void scheduleNext(L2Effect e)
	{
		if(_next != null)
			_next.exit();
		_next = e;
	}

	public L2Effect getNext()
	{
		return _next;
	}
}