package l2r.gameserver.model;

import java.util.ArrayList;

import l2r.Config;
import l2r.gameserver.cache.Msg;
import l2r.gameserver.model.base.Experience;
import l2r.gameserver.model.entity.DimensionalRift;
import l2r.gameserver.model.entity.SevenSignsFestival.SevenSignsFestival;
import l2r.gameserver.model.instances.L2ItemInstance;
import l2r.gameserver.model.instances.L2NpcInstance;
import l2r.gameserver.model.instances.L2SummonInstance;
import l2r.gameserver.serverpackets.L2GameServerPacket;
import l2r.gameserver.serverpackets.PartySmallWindowAdd;
import l2r.gameserver.serverpackets.PartySmallWindowAll;
import l2r.gameserver.serverpackets.PartySmallWindowDelete;
import l2r.gameserver.serverpackets.PartySmallWindowDeleteAll;
import l2r.gameserver.serverpackets.PartySmallWindowUpdate;
import l2r.gameserver.serverpackets.SystemMessage;
import l2r.gameserver.skills.Stats;
import l2r.gameserver.tables.ItemTable;
import l2r.util.Location;
import l2r.util.Log;
import l2r.util.Rnd;

public class L2Party
{
	private final ArrayList<L2Player> _members = new ArrayList<L2Player>();
	private int _partyLvl = 0;
	private int _partySqLevelSum = 0;
	private int _itemDistribution = 0;
	private int _itemOrder = 0;
	private DimensionalRift _dr;
	private L2CommandChannel _commandChannel;

	public static final int ITEM_LOOTER = 0;
	public static final int ITEM_RANDOM = 1;
	public static final int ITEM_RANDOM_SPOIL = 2;
	public static final int ITEM_ORDER = 3;
	public static final int ITEM_ORDER_SPOIL = 4;

	public float _rateExp;
	public float _rateSp;
	public float _rateDrop;
	public float _rateAdena;
	public float _rateSpoil;

	/**
	 * constructor ensures party has always one member - leader
	 * @param leader создатель парти
	 * @param itemDistribution режим распределения лута
	 */
	public L2Party(L2Player leader, int itemDistribution)
	{
		_itemDistribution = itemDistribution;
		_members.add(leader);
		_partyLvl = leader.getLevel();
		_partySqLevelSum = _partyLvl * _partyLvl;
	}

	/**
	 * @return number of party members
	 */
	public int getMemberCount()
	{
		return _members.size();
	}

	/**
	 * @return all party members
	 */
	public ArrayList<L2Player> getPartyMembers()
	{
		return _members;
	}

	public ArrayList<L2Playable> getPartyMembersWithPets()
	{
		ArrayList<L2Playable> ret = new ArrayList<L2Playable>(_members);
		for(L2Player p : _members)
			if(p.getPet() != null)
				ret.add(p.getPet());
		return ret;
	}

	public L2Player getRandomMember()
	{
		return _members.get(Rnd.get(_members.size()));
	}

	/**
	 * @return random member from party
	 */
	private L2Player getRandomMemberInRange(L2Player player, L2ItemInstance item, int range)
	{
		ArrayList<L2Player> ret = new ArrayList<L2Player>();

		for(L2Player member : _members)
			if(member.isInRange(player, range) && !member.isDead() && member.getInventory().validateCapacity(item) && member.getInventory().validateWeight(item))
				ret.add(member);

		if(ret.size() > 0)
			return ret.get(Rnd.get(ret.size()));

		return null;
	}

	/**
	 * @return next item looter
	 */
	private L2Player getNextLooterInRange(L2Player player, @SuppressWarnings("unused")
	L2ItemInstance item, int range)
	{
		while(true)
		{
			int looter = _itemOrder;
			_itemOrder++;
			if(_itemOrder > _members.size() - 1)
				_itemOrder = 0;

			L2Player ret = looter < _members.size() ? _members.get(looter) : player;

			if(ret.isInRange(player, range) && !ret.isDead())
				return ret;
		}
	}

	/**
	 * true if player is party leader
	 * @param player
	 * @return
	 */
	public boolean isLeader(L2Player player)
	{
		return _members.get(0).equals(player);
	}

	/**
	 * Returns the Object ID for the party leader to be used as a unique identifier of this party
	 * @return int
	 */
	public int getPartyLeaderOID()
	{
		return _members.get(0).getObjectId();
	}

	/**
	 * Возвращает лидера партии
	 * @return L2Player Лидер партии
	 */
	public L2Player getPartyLeader()
	{
		return _members.get(0);
	}

	/**
	 * Broadcasts packet to every party member
	 * @param msg packet to broadcast
	 */
	public void broadcastToPartyMembers(L2GameServerPacket msg)
	{
		synchronized (_members)
		{
			for(L2Player member : _members)
				member.sendPacket(msg);
		}
	}

	/**
	 * Рассылает текстовое сообщение всем членам группы
	 * @param msg сообщение
	 */
	public void broadcastMessageToPartyMembers(String msg)
	{
		broadcastToPartyMembers(SystemMessage.sendString(msg));
	}

	/**
	 * Рассылает пакет всем членам группы исключая указанного персонажа<BR><BR>
	 */
	public void broadcastToPartyMembers(L2Player player, L2GameServerPacket msg)
	{
		synchronized (_members)
		{
			for(L2Player member : _members)
				if(!member.equals(player))
					member.sendPacket(msg);
		}
	}

	public void broadcastToPartyMembersInRange(L2Player player, L2GameServerPacket msg, int range)
	{
		synchronized (_members)
		{
			for(L2Player member : _members)
				if(player.isInRange(member, range))
					member.sendPacket(msg);
		}
	}

	public boolean containsMember(L2Character cha)
	{
		L2Player player = cha.getPlayer();
		if(player == null)
			return false;

		for(L2Player member : _members)
			if(member.getObjectId() == cha.getObjectId())
				return true;

		return false;
	}

	/**
	 * adds new member to party
	 * @param player L2Player to add
	 */
	public void addPartyMember(L2Player player)
	{
		//sends new member party window for all members
		//we do all actions before adding member to a list, this speeds things up a little
		player.sendPacket(new PartySmallWindowAll(_members));

		for(L2Player partyMember : _members)
		player.sendPacket(new SystemMessage(SystemMessage.YOU_HAVE_JOINED_S1S_PARTY).addString(_members.get(0).getName()));
		broadcastToPartyMembers(new SystemMessage(SystemMessage.S1_HAS_JOINED_THE_PARTY).addString(player.getName()));
		broadcastToPartyMembers(new PartySmallWindowAdd(player));

		synchronized (_members)
		{
			_members.add(player);
		}

		recalculatePartyData();

		player.updateEffectIcons();
		if(player.getPet() != null)
			player.getPet().updateEffectIcons();

		// Если партия уже в СС, то вновь прибывшем посылаем пакет открытия окна СС
		if(isInCommandChannel())
			player.sendPacket(Msg.ExMPCCOpen);

		if(isInDimensionalRift())
			_dr.partyMemberInvited();
	}

	/**
	 * removes player from party
	 * @param player L2Player to remove
	 */
	public void removePartyMember(L2Player player)
	{
		if(player == null || !_members.contains(player))
			return;

		synchronized (_members)
		{
			_members.remove(player);
		}

		recalculatePartyData();

		if(player.isFestivalParticipant())
			SevenSignsFestival.getInstance().updateParticipants(player, this);

		// Отсылаемы вышедшему пакет закрытия СС
		if(isInCommandChannel())
			player.sendPacket(Msg.ExMPCCClose);

		player.sendPacket(new SystemMessage(SystemMessage.YOU_HAVE_WITHDRAWN_FROM_THE_PARTY));
		player.sendPacket(new PartySmallWindowDeleteAll());
		player.setParty(null);

		broadcastToPartyMembers(new SystemMessage(SystemMessage.S1_HAS_LEFT_THE_PARTY).addString(player.getName()));
		broadcastToPartyMembers(new PartySmallWindowDelete(player));
		if(isInDimensionalRift())
			_dr.partyMemberExited(player);

		if(player.getDuel() != null)
			player.getDuel().onRemoveFromParty(player);

		if(_members.size() == 1)
		{
			L2Player lastMember = _members.get(0);

			if(lastMember.getDuel() != null)
				lastMember.getDuel().onRemoveFromParty(lastMember);

			// Если в партии остался 1 человек, то удаляем ее из СС
			if(isInCommandChannel())
				_commandChannel.removeParty(this);

			lastMember.setParty(null);
		}
	}

	/**
	 * Change party leader (used for string arguments)
	 * @param name имя нового лидера парти
	 */
	public void changePartyLeader(String name)
	{
		L2Player new_leader = getPlayerByName(name);
		L2Player current_leader = _members.get(0);

		if(new_leader == null || current_leader == null)
			return;

		if(current_leader.equals(new_leader))
		{
			current_leader.sendPacket(new SystemMessage(SystemMessage.YOU_CANNOT_TRANSFER_RIGHTS_TO_YOURSELF));
			return;
		}

		synchronized (_members)
		{
			if(!_members.contains(new_leader))
			{
				current_leader.sendPacket(new SystemMessage(SystemMessage.YOU_CAN_TRANSFER_RIGHTS_ONLY_TO_ANOTHER_PARTY_MEMBER));
				return;
			}

			//меняем местами нового и текущего лидера
			int idx = _members.indexOf(new_leader);
			_members.set(0, new_leader);
			_members.set(idx, current_leader);

			//обновляем инфо о пати у всех пати мемберов и сообщаем о новом лидере
			L2GameServerPacket del_pkt = new PartySmallWindowDeleteAll();
			SystemMessage msg = new SystemMessage(SystemMessage.S1_HAS_BECOME_A_PARTY_LEADER);
			msg.addString(name);
			for(L2Player member : _members)
			{
				member.sendPacket(del_pkt);
				member.sendPacket(new PartySmallWindowAll(_members));
				member.sendPacket(msg);
			}
		}

		if(isInCommandChannel())
			_commandChannel.setChannelLeader(new_leader);
	}

	/**
	 * finds a player in the party by name
	 * @param name имя для поиска
	 * @return найденый L2Player или null если не найдено
	 */
	private L2Player getPlayerByName(String name)
	{
		for(L2Player pl : _members)
			if(pl.getName().equalsIgnoreCase(name))
				return pl;
		return null;
	}

	/**
	 * Oust player from party
	 * @param player L2Player которого выгоняют
	 */
	public void oustPartyMember(L2Player player)
	{
		if(!_members.contains(player))
			return;

		if(isLeader(player))
		{
			removePartyMember(player);
			if(_members.size() > 0)
			{
				SystemMessage msg = new SystemMessage(SystemMessage.S1_HAS_BECOME_A_PARTY_LEADER);
				msg.addString(_members.get(0).getName());
				broadcastToPartyMembers(msg);
				broadcastToPartyMembers(new PartySmallWindowUpdate(_members.get(0)));
			}
		}
		else
			removePartyMember(player);
	}

	/**
	 * Oust player from party Overloaded method that takes player's name as
	 * parameter
	 *
	 * @param name имя игрока для изгнания
	 */
	public void oustPartyMember(String name)
	{
		oustPartyMember(getPlayerByName(name));
	}

	/**
	 * distribute item(s) to party members
	 * @param player
	 * @param item
	 */
	public void distributeItem(L2Player player, L2ItemInstance item)
	{
		distributeItem(player, item, null);
	}

	public void distributeItem(L2Player player, L2ItemInstance item, L2NpcInstance fromNpc)
	{
		L2Player target = player;

		switch(_itemDistribution)
		{
			case ITEM_RANDOM:
			case ITEM_RANDOM_SPOIL:
				target = getRandomMemberInRange(player, item, Config.ALT_PARTY_DISTRIBUTION_RANGE);
				break;
			case ITEM_ORDER:
			case ITEM_ORDER_SPOIL:
				target = getNextLooterInRange(player, item, Config.ALT_PARTY_DISTRIBUTION_RANGE);
				break;
			case ITEM_LOOTER:
			default:
				target = player;
				break;
		}

		if(target == null)
		{
			item.dropToTheGround(player, fromNpc);
			return;
		}

		if(!target.getInventory().validateWeight(item))
		{
			target.sendActionFailed();
			target.sendPacket(Msg.YOU_HAVE_EXCEEDED_THE_WEIGHT_LIMIT);
			item.dropToTheGround(target, fromNpc);
			return;
		}

		if(!target.getInventory().validateCapacity(item))
		{
			target.sendActionFailed();
			target.sendPacket(Msg.YOUR_INVENTORY_IS_FULL);
			item.dropToTheGround(player, fromNpc);
			return;
		}

		if(item.getCount() == 1)
		{
			SystemMessage smsg;
			if(item.getEnchantLevel() > 0)
			{
				smsg = new SystemMessage(SystemMessage.YOU_HAVE_OBTAINED__S1S2);
				smsg.addNumber(item.getEnchantLevel());
				smsg.addItemName(item.getItemId());
			}
			else
			{
				smsg = new SystemMessage(SystemMessage.YOU_HAVE_OBTAINED_S1);
				smsg.addItemName(item.getItemId());
			}

			target.sendPacket(smsg);

			if(item.getEnchantLevel() > 0)
			{
				smsg = new SystemMessage(SystemMessage.S1_HAS_OBTAINED__S2S3);
				smsg.addString(target.getName());
				smsg.addNumber(item.getEnchantLevel());
				smsg.addItemName(item.getItemId());
			}
			else
			{
				smsg = new SystemMessage(SystemMessage.S1_HAS_OBTAINED_S2);
				smsg.addString(target.getName());
				smsg.addItemName(item.getItemId());
			}

			broadcastToPartyMembers(target, smsg);
		}
		else
		{
			SystemMessage smsg = new SystemMessage(SystemMessage.YOU_HAVE_OBTAINED_S2_S1);
			smsg.addItemName(item.getItemId());
			smsg.addNumber(item.getCount());
			target.sendPacket(smsg);

			smsg = new SystemMessage(SystemMessage.S1_HAS_OBTAINED_S3_S2);
			smsg.addString(target.getName());
			smsg.addItemName(item.getItemId());
			smsg.addNumber(item.getCount());
			broadcastToPartyMembers(target, smsg);
		}

		// Remove the L2ItemInstance from the world and send server->client GetItem packets
		item.pickupMe(target);

		L2ItemInstance item2 = target.getInventory().addItem(item);
		Log.LogItem(target, fromNpc, Log.GetItemInPaty, item2);

		target.sendChanges();
	}

	/**
	 * distribute adena to party members
	 * @param adena инстанс адены для распределения
	 */
	public void distributeAdena(L2ItemInstance adena, L2Player player)
	{
		distributeAdena(adena, null, player);
	}

	public void distributeAdena(L2ItemInstance adena, L2NpcInstance fromNpc, L2Player player)
	{
		SystemMessage smsg = new SystemMessage(SystemMessage.YOU_HAVE_OBTAINED_S1_ADENA);

		long totalAdena = adena.getCount();
		int _fromNpc;

		_fromNpc = fromNpc == null ? -1 : fromNpc.getObjectId();

		ArrayList<L2Player> _membersInRange = new ArrayList<L2Player>();

		if(adena.getCount() < _members.size())
			_membersInRange.add(player);
		else
			for(L2Player p : _members)
				if(p.equals(player) || player.isInRange(p, Config.ALT_PARTY_DISTRIBUTION_RANGE) && !p.isDead())
					_membersInRange.add(p);

		long amount = totalAdena / _membersInRange.size();
		long ost = totalAdena % _membersInRange.size();

		for(L2Player member : _membersInRange)
		{
			smsg.addNumber(member.equals(player) ? amount + ost : amount);

			L2ItemInstance newAdena = ItemTable.getInstance().createItem(57, member.getObjectId(), _fromNpc, "L2Party.distributeAdena");
			newAdena.setCount(member.equals(player) ? amount + ost : amount);

			L2ItemInstance item2 = member.getInventory().addItem(newAdena);
			if(fromNpc == null)
				Log.LogItem(member, Log.GetItemInPaty, item2);
			else
				Log.LogItem(member, fromNpc, Log.GetItemInPaty, item2);

			member.sendPacket(smsg);
		}
	}

	/**
	 * Distribute Experience and SP rewards to L2Player Party members in the known area of the last attacker.<BR><BR>
	 *
	 * <B><U> Actions</U> :</B><BR><BR>
	 * <li>Get the L2Player owner of the L2SummonInstance (if necessary) </li>
	 * <li>Calculate the Experience and SP reward distribution rate </li>
	 * <li>Add Experience and SP to the L2Player </li><BR><BR>
	 *
	 * <FONT COLOR=#FF0000><B> <U>Caution</U> : This method DOESN'T GIVE rewards to L2PetInstance</B></FONT><BR><BR>
	 *
	 * @param xpReward The Experience reward to distribute
	 * @param spReward The SP reward to distribute
	 * @param rewardedMembers The list of L2Player to reward and LSummonInstance whose owner must be reward
	 * @param lastAttacker The L2Character that has killed the L2NpcInstance
	 *
	 */
	public void distributeXpAndSp(long xpReward, long spReward, ArrayList<L2Player> rewardedMembers, L2Character lastAttacker, L2NpcInstance target)
	{
		recalculatePartyData();

		ArrayList<L2Player> mtr = new ArrayList<L2Player>();
		int minPartyLevel = -1;
		int maxPartyLevel = -1;
		double partyLvlSum = 0;

		// создаём список тех кто рядом
		for(L2Player member : rewardedMembers)
		{
			if(!lastAttacker.isInRange(member, Config.ALT_PARTY_DISTRIBUTION_RANGE))
				continue;

			if(minPartyLevel == -1 || member.getLevel() < minPartyLevel)
				minPartyLevel = member.getLevel();

			if(maxPartyLevel == -1 || member.getLevel() > maxPartyLevel)
				maxPartyLevel = member.getLevel();

			partyLvlSum += member.getLevel();

			mtr.add(member);
		}

		if(maxPartyLevel - minPartyLevel > 20)
		{
			for(L2Player member : rewardedMembers)
				if(member.getLevel() < maxPartyLevel - 20)
					mtr.remove(member);
			partyLvlSum = 0;
			for(L2Player member : mtr)
				partyLvlSum += member.getLevel();
		}

		// бонус за пати
		double bonus = getExpSpBonus(mtr.size());

		// количество эксп и сп для раздачи на всех
		double XP = xpReward * bonus;
		double SP = spReward * bonus;

		byte targetLevel = target.getLevel();

		for(L2Player member : mtr)
		{
			double lvlPenalty = 1;

			if(member.getLevel() - 9 > targetLevel)
				lvlPenalty = Experience.penaltyModifier(member.getLevel() - targetLevel - 7, 9);

			// отдаем его часть с учетом пенальти
			double memberXp = XP * lvlPenalty * member.getLevel() / partyLvlSum;
			double memberSp = SP * lvlPenalty * member.getLevel() / partyLvlSum;

			// больше чем соло не дадут
			if(memberXp > xpReward)
				memberXp = xpReward;
			if(memberSp > spReward)
				memberSp = spReward;

			// отдаем часть пету или саммону
			if(member.getPet() != null && member.getPet().isSummon)
			{
				float penalty = 1 - ((L2SummonInstance) member.getPet()).getExpPenalty();
				memberXp *= penalty;
				memberSp *= penalty;
			}

			member.addExpAndSp((int) memberXp, (int) memberSp);
		}

		recalculatePartyData();
	}

	public void recalculatePartyData()
	{
		_partySqLevelSum = 0;
		_partyLvl = 0;
		int min = 1000, max = 0;
		float rateExp = 0;
		float rateSp = 0;
		float rateDrop = 0;
		float rateAdena = 0;
		float rateSpoil = 0;
		byte count = 0;
		for(L2Player member : _members)
		{
			int level = member.getLevel();
			_partySqLevelSum += level * level;
			if(level > _partyLvl)
				_partyLvl = level;
			if(level < min)
				min = level;
			if(level > max)
				max = level;
			rateExp += Config.RATE_XP_PARTY;
			rateExp *= member.getBonus().RATE_XP;
			rateSp += Config.RATE_SP_PARTY;
			rateSp *= member.getBonus().RATE_SP;
			rateDrop += Config.RATE_DROP_ITEMS_PARTY;
			rateDrop *= member.getBonus().RATE_DROP_ITEMS;
			rateAdena += Config.RATE_DROP_ADENA_PARTY;
			rateAdena *= member.getBonus().RATE_DROP_ADENA;
			rateSpoil += member.getBonus().RATE_DROP_SPOIL;
			count++;
		}
		_rateExp = rateExp / count;
		_rateSp = rateSp / count;
		_rateDrop = rateDrop / count;
		_rateAdena = rateAdena / count;
		_rateSpoil = rateSpoil / count;
	}

	private static final double[] bonuses = { 1.00, 1.00, 1.30, 1.39, 1.50, 1.54, 1.58, 1.63, 1.67, 1.71 };

	private double getExpSpBonus(int size)
	{
		return bonuses[size];
	}

	public int getLevel()
	{
		return _partyLvl;
	}

	public int getLootDistribution()
	{
		return _itemDistribution;
	}

	public boolean isDistributeSpoilLoot()
	{
		boolean rv = false;

		if(_itemDistribution == ITEM_RANDOM_SPOIL || _itemDistribution == ITEM_ORDER_SPOIL)
			rv = true;

		return rv;
	}

	public boolean isInDimensionalRift()
	{
		return _dr != null;
	}

	public void setDimensionalRift(DimensionalRift dr)
	{
		_dr = dr;
	}

	public DimensionalRift getDimensionalRift()
	{
		return _dr;
	}

	public boolean isInCommandChannel()
	{
		return _commandChannel != null;
	}

	public L2CommandChannel getCommandChannel()
	{
		return _commandChannel;
	}

	public void setCommandChannel(L2CommandChannel channel)
	{
		_commandChannel = channel;
	}

	/**
	 * Телепорт всей пати в одну точку (x,y,z)
	 */
	public void Teleport(int x, int y, int z)
	{
		TeleportParty(getPartyMembers(), new Location(x, y, z));
	}

	/**
	 * Телепорт всей пати в одну точку dest
	 */
	public void Teleport(Location dest)
	{
		TeleportParty(getPartyMembers(), dest);
	}

	/**
	 * Телепорт всей пати на территорию, игроки расставляются рандомно по территории
	 */
	public void Teleport(L2Territory territory)
	{
		RandomTeleportParty(getPartyMembers(), territory);
	}

	/**
	 * Телепорт всей пати на территорию, лидер попадает в точку dest, а все остальные относительно лидера
	 */
	public void Teleport(L2Territory territory, Location dest)
	{
		TeleportParty(getPartyMembers(), territory, dest);
	}

	public static void TeleportParty(ArrayList<L2Player> _members, Location dest)
	{
		for(L2Player _member : _members)
		{
			if(_member == null)
				continue;
			_member.teleToLocation(dest);
		}
	}

	public static void TeleportParty(ArrayList<L2Player> _members, L2Territory territory, Location dest)
	{
		if(!territory.isInside(dest.getX(), dest.getY()))
		{
			Log.add("TeleportParty: dest is out of territory", "errors");
			Thread.dumpStack();
			return;
		}
		int base_x = _members.get(0).getX();
		int base_y = _members.get(0).getY();

		for(L2Player _member : _members)
		{
			if(_member == null)
				continue;
			int diff_x = _member.getX() - base_x;
			int diff_y = _member.getY() - base_y;
			Location loc = new Location(dest.getX() + diff_x, dest.getY() + diff_y, dest.getZ());
			while(!territory.isInside(loc.getX(), loc.getY()))
			{
				diff_x = loc.getX() - dest.getX();
				diff_y = loc.getY() - dest.getY();
				if(diff_x != 0)
					loc.setX(loc.getX() - (diff_x / Math.abs(diff_x)));
				if(diff_y != 0)
					loc.setY(loc.getY() - (diff_y / Math.abs(diff_y)));
			}
			_member.teleToLocation(loc);
		}
	}

	public static void RandomTeleportParty(ArrayList<L2Player> _members, L2Territory territory)
	{
		for(L2Player _member : _members)
		{
			int[] _loc = territory.getRandomPoint();
			if(_member == null || _loc == null)
				continue;
			_member.teleToLocation(_loc[0], _loc[1], _loc[2]);
		}
	}
}