package l2r.gameserver.model;

import java.util.ArrayList;

import l2r.gameserver.ai.DefaultAI;
import l2r.gameserver.model.instances.L2NpcInstance;
import l2r.gameserver.tables.SpawnTable;

/**
 * @author Diamond
 * @Date: 15/5/2007
 * @Time: 10:06:34
 */
public final class L2WorldRegion
{
	L2Object[] _objects = null;

	private int _objectsSize = 0;
	private int _playersSize = 0;

	private int tileX, tileY, tileZ;
	private Boolean _active = false;

	public L2WorldRegion(int pTileX, int pTileY, int pTileZ)
	{
		tileX = pTileX;
		tileY = pTileY;
		tileZ = pTileZ;
	}

	private void switchAI(Boolean isOn)
	{
		if(_objects == null)
		{
			_objectsSize = 0;
			_playersSize = 0;
			return;
		}

		for(L2Character cha : getCharactersList(-1))
			if(isOn)
			{
				if(cha.getAI() instanceof DefaultAI)
					cha.getAI().startAITask();
			}
			else if(cha.hasAI() && !cha.getAI().isGlobalAI() && cha.getAI() instanceof DefaultAI)
			{
				cha.setTarget(null);
				cha.stopMove();
				cha.stopAllEffects();
				cha.getAI().stopAITask();
			}
	}

	private void setActive(boolean value)
	{
		synchronized (_active)
		{
			if(_active == value)
				return;
			_active = value;

			switchAI(value);

			if(value)
				L2World.addActiveRegion(this);
			else
				L2World.removeActiveRegion(this);
		}
	}

	private void changeStatus(boolean status)
	{
		for(L2WorldRegion neighbor : getNeighbors())
			if(neighbor != null && (status || neighbor.areNeighborsEmpty()))
				neighbor.setActive(status);
	}

	public void addToPlayers(L2Object object, L2Character dropper)
	{
		if(_objects == null)
		{
			_objectsSize = 0;
			_playersSize = 0;
			return;
		}

		L2Player player = null;
		if(object.isPlayer)
			player = (L2Player) object;

		// Если object - игрок, показать ему все видимые обьекты в регионе
		if(player != null)
			for(L2Object obj : getObjectsList(object.getReflection()))
			{
				if(obj == null)
					continue;
				// Если это фэйк обсервера - не показывать.
				if(obj.inObserverMode())
				{
					L2WorldRegion[] neighbors = ((L2Player) obj).getObservNeighbors();
					if(neighbors != null)
					{
						boolean flag = false;
						for(L2WorldRegion neighbor : neighbors)
							if(neighbor != null && neighbor.equals(this))
							{
								flag = true;
								break;
							}
						if(flag)
							continue;
					}
				}
				player.addVisibleObject(obj, dropper);
			}

		// Показать обьект всем игрокам в регионе
		for(L2Object obj : getObjectsList(object.getReflection()))
			if(obj != null && obj.isPlayer)
				((L2Player) obj).addVisibleObject(object, dropper);
	}

	public void removeFromPlayers(L2Object object)
	{
		if(_objects == null)
		{
			_objectsSize = 0;
			_playersSize = 0;
			return;
		}

		L2Player player = null;
		if(object.isPlayer)
			player = (L2Player) object;

		// Если object - игрок, убрать у него все видимые обьекты в регионе
		if(player != null)
			for(L2Object obj : getObjectsList(object.getReflection()))
				if(obj != null)
					player.removeVisibleObject(obj);

		// Убрать обьект у всех игроков в регионе
		for(L2Object obj : getObjectsList(object.getReflection()))
			if(obj != null && obj.isPlayer)
				((L2Player) obj).removeVisibleObject(object);
	}

	public void addVisibleObject(L2Object object)
	{
		if(object == null)
			return;
		addObject(object);
		if(object.isPlayer)
		{
			_playersSize++;
			changeStatus(true);
		}
		else if(object.isNpc)
			if(object.getAI() instanceof DefaultAI && !areNeighborsEmpty())
				object.getAI().startAITask();
	}

	public void removeVisibleObject(L2Object object, Boolean move)
	{
		if(object == null)
			return;
		removeObject(object);
		if(object.isPlayer)
		{
			_playersSize--;
			if(_playersSize < 0)
				_playersSize = 0;
			changeStatus(false);
		}
		else if(object.isNpc)
			if(!move && object.getAI() instanceof DefaultAI && !object.getAI().isGlobalAI())
				object.getAI().stopAITask();
	}

	public L2Object[] getObjects()
	{
		if(_objects == null)
		{
			_objects = new L2Object[50];
			_objectsSize = 0;
			_playersSize = 0;
		}
		return _objects;
	}

	private void addObject(L2Object obj)
	{
		if(_objects == null)
		{
			_objects = new L2Object[50];
			_objects[0] = obj;
			_objectsSize = 1;
			return;
		}

		synchronized (_objects)
		{
			if(_objectsSize >= _objects.length)
			{
				L2Object[] temp = new L2Object[_objects.length * 2];
				for(int i = 0; i < _objectsSize; i++)
					temp[i] = _objects[i];
				_objects = temp;
			}

			_objects[_objectsSize] = obj;
			_objectsSize++;
		}
	}

	private void removeObject(L2Object obj)
	{
		if(_objects == null)
		{
			_objectsSize = 0;
			_playersSize = 0;
			return;
		}

		synchronized (_objects)
		{
			if(_objectsSize > 1)
			{
				int k = -1;
				for(int i = 0; i < _objectsSize; i++)
					if(_objects[i] == obj)
						k = i;
				if(k > -1)
				{
					_objects[k] = _objects[_objectsSize - 1];
					_objects[_objectsSize - 1] = null;
					_objectsSize--;
				}
			}
			else if(_objectsSize == 1 && _objects[0] == obj)
			{
				_objects[0] = null;
				_objects = null;
				_objectsSize = 0;
				_playersSize = 0;
			}
		}
	}

	public ArrayList<L2Object> getObjectsList(int reflection)
	{
		ArrayList<L2Object> result = new ArrayList<L2Object>();

		if(_objects == null || _objectsSize == 0)
			return result;

		synchronized (_objects)
		{
			for(int i = 0; i < _objectsSize; i++)
			{
				L2Object obj = _objects[i];
				if(obj != null && (reflection == -1 || obj.getReflection() == reflection))
					result.add(obj);
			}
		}

		return result;
	}

	public ArrayList<L2Character> getCharactersList(int reflection)
	{
		ArrayList<L2Character> result = new ArrayList<L2Character>();

		if(_objects == null || _objectsSize == 0)
			return result;

		synchronized (_objects)
		{
			for(int i = 0; i < _objectsSize; i++)
			{
				L2Object obj = _objects[i];
				if(obj != null && obj.isCharacter && (reflection == -1 || obj.getReflection() == reflection))
					result.add((L2Character) obj);
			}
		}

		return result;
	}

	public ArrayList<L2NpcInstance> getNpcsList(int reflection)
	{
		ArrayList<L2NpcInstance> result = new ArrayList<L2NpcInstance>();

		if(_objects == null || _objectsSize == 0)
			return result;

		synchronized (_objects)
		{
			for(int i = 0; i < _objectsSize; i++)
			{
				L2Object obj = _objects[i];
				if(obj != null && obj.isNpc && (reflection == -1 || obj.getReflection() == reflection))
					result.add((L2NpcInstance) obj);
			}
		}

		return result;
	}

	public ArrayList<L2Player> getPlayersList(int reflection)
	{
		ArrayList<L2Player> result = new ArrayList<L2Player>();

		if(_objects == null || _objectsSize == 0)
			return result;

		synchronized (_objects)
		{
			for(int i = 0; i < _objectsSize; i++)
			{
				L2Object obj = _objects[i];
				if(obj != null && obj.isPlayer && (reflection == -1 || obj.getReflection() == reflection))
					result.add((L2Player) obj);
			}
		}

		return result;
	}

	public void deleteVisibleNpcSpawns()
	{
		if(_objects != null)
		{
			ArrayList<L2NpcInstance> toRemove = new ArrayList<L2NpcInstance>();
			synchronized (_objects)
			{
				for(int i = 0; i < _objectsSize; i++)
				{
					L2Object obj = _objects[i];
					if(obj != null && obj.isNpc)
						toRemove.add((L2NpcInstance) obj);
				}
			}
			for(L2NpcInstance npc : toRemove)
			{
				L2Spawn spawn = npc.getSpawn();
				if(spawn != null)
				{
					npc.deleteMe();
					spawn.stopRespawn();
					SpawnTable.getInstance().deleteSpawn(spawn, false);
				}
			}
		}
	}

	/**
	 * Показывает игроку все видимые обьекты в регионе
	 */
	public void showObjectsToPlayer(L2Player player)
	{
		if(player != null && _objects != null)
			for(L2Object obj : getObjectsList(player.getReflection()))
				if(obj != null)
					player.addVisibleObject(obj, null);
	}

	/**
	 * Убирает у игрока все видимые обьекты в регионе
	 */
	public void removeObjectsFromPlayer(L2Player player)
	{
		if(player != null && _objects != null)
			for(L2Object obj : getObjectsList(player.getReflection()))
				if(obj != null)
					player.removeVisibleObject(obj);
	}

	/**
	 * Убирает обьект у всех игроков в регионе
	 */
	public void removePlayerFromOtherPlayers(L2Object object)
	{
		if(object != null && _objects != null)
			for(L2Object obj : getObjectsList(object.getReflection()))
				if(obj != null && obj.isPlayer)
					((L2Player) obj).removeVisibleObject(object);
	}

	public boolean areNeighborsEmpty()
	{
		if(!isEmpty())
			return false;
		for(L2WorldRegion neighbor : getNeighbors())
			if(neighbor != null && !neighbor.isEmpty())
				return false;
		return true;
	}

	public L2WorldRegion[] getNeighbors()
	{
		return L2World.getNeighbors(tileX, tileY, tileZ);
	}

	public int getObjectsSize()
	{
		return _objectsSize;
	}

	public boolean isEmpty()
	{
		return _playersSize <= 0;
	}

	public boolean isActive()
	{
		return _active;
	}

	public String getName()
	{
		return "(" + tileX + ", " + tileY + ", " + tileZ + ")";
	}
}
