package l2r.gameserver.model;

import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;

import l2r.database.DatabaseUtils;
import l2r.database.FiltredPreparedStatement;
import l2r.database.L2DatabaseFactory;
import l2r.database.ThreadConnection;
import l2r.database.mysql;
import l2r.gameserver.model.instances.L2ItemInstance;
import l2r.gameserver.model.instances.L2ItemInstance.ItemClass;
import l2r.gameserver.model.instances.L2ItemInstance.ItemLocation;
import l2r.gameserver.tables.ItemTable;
import l2r.util.Log;

public abstract class Warehouse
{
	public static enum WarehouseType
	{
		PRIVATE(1),
		CLAN(2),
		CASTLE(3),
		FREIGHT(4);

		private final int _type;

		private WarehouseType(final int type)
		{
			_type = type;
		}

		public int getPacketValue()
		{
			return _type;
		}
	}

	private static final Logger _log = Logger.getLogger(Warehouse.class.getName());

	public abstract int getOwnerId();

	public abstract ItemLocation getLocationType();

	public abstract WarehouseType getWarehouseType();

	public abstract String getLocationId();

	public abstract byte getLocationId(boolean addItem);

	public abstract void setLocationId(L2Player activeChar);

	public L2ItemInstance[] listItems(ItemClass clss)
	{
		final ArrayList<L2ItemInstance> items = new ArrayList<L2ItemInstance>();
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		ResultSet rset = null;

		try
		{
			String _select = clss == ItemClass.ALL ? "" : " AND class = '" + clss.name() + "'";

			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("SELECT object_id FROM items WHERE owner_id=? AND loc=? " + _select + " ORDER BY name ASC LIMIT 200");
			statement.setInt(1, getOwnerId());
			statement.setString(2, getLocationType().name());
			rset = statement.executeQuery();

			L2ItemInstance item;
			while(rset.next())
			{
				final int objectId = rset.getInt(1);
				item = L2ItemInstance.restoreFromDb(objectId);
				if(item == null)
					continue;
				items.add(item);
			}
		}
		catch(final Exception e)
		{
			_log.log(Level.SEVERE, "could not restore warehouse:", e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCSR(con, statement, rset);
		}
		return items.toArray(new L2ItemInstance[items.size()]);
	}

	public int countItems()
	{
		return mysql.simple_get_int("COUNT(object_id)", "items", "owner_id=" + getOwnerId() + " AND loc=" + getLocationType().name());
	}

	public synchronized void addItem(L2ItemInstance newItem)
	{
		if(this instanceof ClanWarehouse)
			Log.add("ClanWarehouse|" + getOwnerId() + "|add|" + newItem.getObjectId() + "|" + newItem.getCount(), "items");
		else if(this instanceof PcWarehouse)
			Log.add("PcWarehouse|" + getOwnerId() + "|add|" + newItem.getObjectId() + "|" + newItem.getCount(), "items");

		L2ItemInstance oldItem = findItemId(newItem.getItemId());

		// non-stackable items are simply added to DB
		if(!newItem.isStackable() || oldItem == null)
		{
			newItem.setOwnerId(getOwnerId());
			newItem.setLocation(getLocationType(), getLocationId(true));
			newItem.updateDatabase();
			return;
		}

		oldItem.setCount(oldItem.getCount() + newItem.getCount());
		oldItem.updateDatabase(true);
		newItem.deleteMe();
	}

	public synchronized L2ItemInstance getItemByObj(int objectId, long count)
	{
		L2ItemInstance item = (L2ItemInstance) L2World.findObject(objectId);
		if(item == null)
			item = L2ItemInstance.restoreFromDb(objectId);

		if(item == null)
		{
			_log.fine("Warehouse.destroyItem: can't destroy objectId: " + objectId + ", count: " + count);
			return null;
		}

		if(item.getLocation() != ItemLocation.CLANWH && item.getLocation() != ItemLocation.WAREHOUSE && item.getLocation() != ItemLocation.FREIGHT)
		{
			_log.warning("WARNING get item not in WAREHOUSE via WAREHOUSE: item objid=" + item.getObjectId() + " ownerid=" + item.getOwnerId());
			return null;
		}

		if(item.getCount() <= count)
		{
			item.setLocation(ItemLocation.VOID, getLocationId(true));
			item.setWhFlag(true);
			item.updateDatabase(true);
			return item;
		}

		item.setCount(item.getCount() - count);
		item.updateDatabase(true);

		L2ItemInstance Newitem = ItemTable.getInstance().createItem(item.getItem().getItemId(), getOwnerId(), objectId, "Warehouse.getItemByObj");
		Newitem.setCount(count);

		if(this instanceof ClanWarehouse)
			Log.add("ClanWarehouse|" + getOwnerId() + "|withdraw|" + item.getItemId() + "|" + item.getObjectId() + "|" + Newitem.getObjectId() + "|" + count, "items");
		else if(this instanceof PcWarehouse)
			Log.add("PcWarehouse|" + getOwnerId() + "|withdraw|" + item.getItemId() + "|" + item.getObjectId() + "|" + Newitem.getObjectId() + "|" + count, "items");

		return Newitem;
	}

	public synchronized void destroyItem(int itemId, long count)
	{
		L2ItemInstance item = findItemId(itemId);

		if(item == null)
		{
			_log.fine("Warehouse.destroyItem: can't destroy itemId: " + itemId + ", count: " + count);
			return;
		}

		if(item.getCount() < count)
			count = item.getCount();

		if(item.getCount() == count)
			item.deleteMe();
		else
		{
			item.setCount(item.getCount() - count);
			item.updateDatabase(true);
		}

		if(this instanceof ClanWarehouse)
			Log.add("ClanWarehouse|" + getOwnerId() + "|destroy item|" + item.getItemId() + "|" + item.getObjectId() + "|" + count, "items");
		else if(this instanceof PcWarehouse)
			Log.add("PcWarehouse|" + getOwnerId() + "|destroy item|" + item.getItemId() + "|" + item.getObjectId() + "|" + count, "items");
	}

	public L2ItemInstance findItemId(final int itemId)
	{
		L2ItemInstance foundItem = null;
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		ResultSet rset = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("SELECT object_id FROM items WHERE owner_id=? AND loc=? AND loc_data IN (" + getLocationId() + ") AND item_id=? ");
			statement.setInt(1, getOwnerId());
			statement.setString(2, getLocationType().name());
			statement.setInt(3, itemId);
			rset = statement.executeQuery();

			if(rset.next())
				foundItem = L2ItemInstance.restoreFromDb(rset.getInt(1));
		}
		catch(Exception e)
		{
			_log.log(Level.WARNING, "could not list warehouse: ", e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCSR(con, statement, rset);
		}
		return foundItem;
	}
}