package l2r.gameserver.model.entity.siege;

import java.sql.ResultSet;
import java.util.List;

import javolution.util.FastList;
import javolution.util.FastMap;
import l2r.database.DatabaseUtils;
import l2r.database.FiltredPreparedStatement;
import l2r.database.L2DatabaseFactory;
import l2r.database.ThreadConnection;
import l2r.gameserver.model.L2Clan;
import l2r.gameserver.model.L2Object;
import l2r.gameserver.model.L2Player;
import l2r.gameserver.model.L2World;
import l2r.gameserver.model.L2Zone;
import l2r.gameserver.model.instances.L2DoorInstance;
import l2r.gameserver.tables.ClanTable;
import l2r.gameserver.tables.DoorTable;
import l2r.gameserver.tables.MapRegionTable;

public abstract class SiegeUnit
{
	protected int _id = 0;

	protected String _name = "";

	/** Clan objectId */
	protected int _ownerId = 0;

	protected L2Zone _zone;

	private List<L2DoorInstance> _doors = new FastList<L2DoorInstance>();

	public SiegeUnit(int siegeUnitId)
	{
		_id = siegeUnitId;

		loadData();
		loadDoor();
	}

	public int getId()
	{
		return _id;
	}

	public String getName()
	{
		return _name;
	}

	public int getOwnerId()
	{
		return _ownerId;
	}

	public L2Zone getZone()
	{
		return _zone;
	}

	public L2Clan getOwner()
	{
		return ClanTable.getInstance().getClan(_ownerId);
	}

	protected abstract void loadData();

	public abstract Siege getSiege();

	public abstract int getSiegeDayOfWeek();

	public abstract int getSiegeHourOfDay();

	// This method sets the siege unit owner; null here means give it back to NPC
	public abstract void changeOwner(L2Clan clan);

	public int getLastSiegeDate()
	{
		return 0;
	}

	public void setLastSiegeDate(@SuppressWarnings("unused")
	int time)
	{}

	public List<L2DoorInstance> getDoors()
	{
		return _doors;
	}

	public L2DoorInstance getDoor(int doorId)
	{
		if(doorId <= 0)
			return null;

		for(int i = 0; i < getDoors().size(); i++)
		{
			L2DoorInstance door = getDoors().get(i);
			if(door.getDoorId() == doorId)
				return door;
		}
		return null;
	}

	/** Return true if object is inside the zone */
	public boolean checkIfInZone(L2Object obj)
	{
		return checkIfInZone(obj.getX(), obj.getY());
	}

	/** Return true if object is inside the zone */
	public boolean checkIfInZone(int x, int y)
	{
		return getZone().checkIfInZone(x, y);
	}

	/** Respawn all doors on siege unit grounds */
	public void spawnDoor(boolean isDoorWeak)
	{
		for(int i = 0; i < getDoors().size(); i++)
		{
			L2DoorInstance door = getDoors().get(i);
			if(door.getCurrentHp() <= 0)
			{
				door.decayMe(); // Kill current if not killed already
				if(isDoorWeak)
					door.setCurrentHp(door.getMaxHp() / 2);
				else
					door.setCurrentHp(door.getMaxHp());
				door.spawnMe(door.getLoc());
				getDoors().set(i, door);
			}
			else if(door.isOpen())
				door.closeMe();
		}
		loadDoorUpgrade(); // Check for any upgrade the doors may have
	}

	public void spawnDoor()
	{
		spawnDoor(false);
	}

	FastMap<Integer, Integer> doorUpgrades = new FastMap<Integer, Integer>();

	// This method upgrade door
	public void upgradeDoor(int doorId, int hp, boolean db)
	{
		L2DoorInstance door = getDoor(doorId);
		if(door == null)
			return;

		if(door.getDoorId() == doorId)
		{
			door.setUpgradeHp(hp);
			door.setCurrentHp(door.getMaxHp());
			door.broadcastStatusUpdate();

			if(db)
				saveDoorUpgrade(doorId, hp);
			doorUpgrades.put(doorId, hp);
		}
	}

	public Integer getDoorUpgrade(int doorId)
	{
		if(doorUpgrades.get(doorId) != null)
			return doorUpgrades.get(doorId);
		return 0;
	}

	// This method is used to begin removing all siege unit upgrades
	public void removeUpgrade()
	{
		removeDoorUpgrade();
	}

	/** Move non clan members off siegeUnit area and to nearest town. */
	public void banishForeigner(L2Player activeChar)
	{
		// Get all players
		for(L2Player player : L2World.getAroundPlayers(activeChar))
		{
			// Skip if player is in clan
			if(player.getClanId() == getOwnerId())
				continue;

			if(checkIfInZone(player))
				player.teleToLocation(MapRegionTable.getInstance().getTeleToLocation(player, MapRegionTable.TeleportWhereType.ClosestTown));
		}
	}

	public void closeDoor(L2Player activeChar, int doorId)
	{
		openCloseDoor(activeChar, doorId, false);
	}

	public void openDoor(L2Player activeChar, int doorId)
	{
		openCloseDoor(activeChar, doorId, true);
	}

	public void openCloseDoor(L2Player activeChar, int doorId, boolean open)
	{
		if(activeChar.getClanId() != getOwnerId())
			return;

		L2DoorInstance door = getDoor(doorId);
		if(door != null)
			if(open)
				door.openMe();
			else
				door.closeMe();
	}

	// This method loads door data from database
	private void loadDoor()
	{
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		ResultSet rset = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("Select id from siege_door where unitId = ?");
			statement.setInt(1, getId());
			rset = statement.executeQuery();

			while(rset.next())
			{
				L2DoorInstance door = DoorTable.getInstance().getDoor(rset.getInt("id"));
				door.spawnMe(door.getLoc());
				_doors.add(door);
			}
		}
		catch(Exception e)
		{
			System.out.println("Exception: loadDoor(): " + e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			DatabaseUtils.closeDatabaseCSR(con, statement, rset);
		}
	}

	// This method loads door upgrade data from database
	private void loadDoorUpgrade()
	{
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		ResultSet rset = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("Select * from siege_doorupgrade where doorId in (Select Id from siege_door where unitId = ?)");
			statement.setInt(1, getId());
			rset = statement.executeQuery();

			while(rset.next())
				upgradeDoor(rset.getInt("doorId"), rset.getInt("hp"), false);
		}
		catch(Exception e)
		{
			System.out.println("Exception: loadDoorUpgrade(): " + e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			DatabaseUtils.closeDatabaseCSR(con, statement, rset);
		}
	}

	private void removeDoorUpgrade()
	{
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("delete from siege_doorupgrade where doorId in (select id from siege_door where unitId=?)");
			statement.setInt(1, getId());
			statement.execute();
		}
		catch(Exception e)
		{
			System.out.println("Exception: removeDoorUpgrade(): " + e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			DatabaseUtils.closeDatabaseCS(con, statement);
		}
	}

	private void saveDoorUpgrade(int doorId, int hp)
	{
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("REPLACE INTO siege_doorupgrade (doorId, hp) values (?,?)");
			statement.setInt(1, doorId);
			statement.setInt(2, hp);
			statement.execute();
		}
		catch(Exception e)
		{
			System.out.println("Exception: saveDoorUpgrade(int doorId, int hp): " + e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			DatabaseUtils.closeDatabaseCS(con, statement);
		}
	}
}