package l2r.gameserver.model.instances;

import java.util.ArrayList;
import java.util.concurrent.Future;
import java.util.logging.Level;
import java.util.logging.Logger;

import l2r.gameserver.ThreadPoolManager;
import l2r.gameserver.model.L2Character;
import l2r.gameserver.model.L2Playable;
import l2r.gameserver.model.L2Player;
import l2r.gameserver.model.L2Skill;
import l2r.gameserver.serverpackets.MagicSkillUse;
import l2r.gameserver.tables.SkillTable;
import l2r.util.Rnd;

public class L2CubicInstance
{
	protected static final Logger _log = Logger.getLogger(L2CubicInstance.class.getName());

	/** Оффсет для корректного сохранения кубиков в базе */
	public static final int CUBIC_STORE_OFFSET = 1000000;

	public static final short STORM_CUBIC = 1;
	public static final short VAMPIRIC_CUBIC = 2;
	public static final short LIFE_CUBIC = 3;
	public static final short VIPER_CUBIC = 4;
	public static final short POLTERGEIST_CUBIC = 5;
	public static final short BINDING_CUBIC = 6;
	public static final short AQUA_CUBIC = 7;
	public static final short SPARK_CUBIC = 8;
	public static final short ATTRACTIVE_CUBIC = 9;

	private L2Player _owner;
	private L2Character _target;

	private int _id;
	private int _level = 1;

	private ArrayList<L2Skill> _skills = new ArrayList<L2Skill>();

	@SuppressWarnings("unchecked")
	private Future _disappearTask;
	@SuppressWarnings("unchecked")
	private Future _actionTask;
	private long _starttime;
	private long _lifetime;

	public L2CubicInstance(final L2Player owner, final int id, final int level, final int lifetime)
	{
		_owner = owner;
		_id = id;
		_level = level;

		switch(_id)
		{
			case STORM_CUBIC:
				_skills.add(SkillTable.getInstance().getInfo(4049, level));
				break;
			case VAMPIRIC_CUBIC:
				_skills.add(SkillTable.getInstance().getInfo(4050, level));
				break;
			case LIFE_CUBIC:
				_skills.add(SkillTable.getInstance().getInfo(4051, level));
				new Heal().run();
				break;
			case VIPER_CUBIC:
				_skills.add(SkillTable.getInstance().getInfo(4052, level));
				break;
			case POLTERGEIST_CUBIC:
				_skills.add(SkillTable.getInstance().getInfo(4053, level));
				_skills.add(SkillTable.getInstance().getInfo(4054, level));
				_skills.add(SkillTable.getInstance().getInfo(4055, level));
				break;
			case BINDING_CUBIC:
				_skills.add(SkillTable.getInstance().getInfo(4164, level));
				break;
			case AQUA_CUBIC:
				_skills.add(SkillTable.getInstance().getInfo(4165, level));
				break;
			case SPARK_CUBIC:
				_skills.add(SkillTable.getInstance().getInfo(4166, level));
				break;
			case ATTRACTIVE_CUBIC:
				_skills.add(SkillTable.getInstance().getInfo(5115, level));
				_skills.add(SkillTable.getInstance().getInfo(5116, level));
				break;
		}
		if(_disappearTask == null)
			_disappearTask = ThreadPoolManager.getInstance().scheduleGeneral(new Disappear(), lifetime); // disappear in 20 mins
		_starttime = System.currentTimeMillis();
		_lifetime = lifetime;
	}

	public void doAction(L2Character target)
	{
		if(_target == target || _owner == target)
			return;
		stopAction();
		_target = target;
		if(_id != ATTRACTIVE_CUBIC && _id != BINDING_CUBIC)
			_actionTask = ThreadPoolManager.getInstance().scheduleGeneral(new Action(), 4000);
		else
			_actionTask = ThreadPoolManager.getInstance().scheduleGeneral(new Action(), 1000);
	}

	public int getId()
	{
		return _id;
	}

	public int getLevel()
	{
		return _level;
	}

	public long lifeLeft()
	{
		return _lifetime - (System.currentTimeMillis() - _starttime);
	}

	public void setLevel(final int level)
	{
		_level = level;
	}

	public void stopAction()
	{
		_target = null;
		if(_actionTask != null)
		{
			_actionTask.cancel(true);
			_actionTask = null;
		}
	}

	public void cancelDisappear()
	{
		if(_disappearTask != null)
		{
			_disappearTask.cancel(true);
			_disappearTask = null;
		}
	}

	public void deleteMe()
	{
		stopAction();
		cancelDisappear();
		_owner = null;
		_skills = null;
	}

	private class Action implements Runnable
	{
		public void run()
		{
			L2Skill skill = _skills.get(Rnd.get(_skills.size()));
			if(_owner.isDead() || _target == null || _target.isDead() || !_owner.isInRangeZ(_target, skill.getCastRange()))
			{
				stopAction();
				if(_owner.isDead())
				{
					_owner.delCubic(_id);
					_owner.broadcastUserInfo();
					cancelDisappear();
				}
				return;
			}
			try
			{
				if(Rnd.chance(50) && skill.checkCondition(_owner, _target, false, false, true))
				{
					_owner.altUseSkill(skill, _target);
					_owner.broadcastPacket(new MagicSkillUse(_owner, _target, skill.getId(), _level, 0, 0));
				}
			}
			catch(final Exception e)
			{
				_log.log(Level.SEVERE, "", e);
			}
			finally
			{
				_actionTask = ThreadPoolManager.getInstance().scheduleGeneral(new Action(), (long) (6. * _owner.calculateAttackSpeed()));
			}
		}
	}

	private class Heal implements Runnable
	{
		public void run()
		{
			if(_owner.isDead())
			{
				stopAction();
				_owner.delCubic(_id);
				_owner.broadcastUserInfo();
				cancelDisappear();
				return;
			}
			try
			{
				L2Skill skill = _skills.get(Rnd.get(_skills.size()));
				L2Character target = _owner;
				if(_owner.getParty() != null)
				{
					for(L2Playable member : _owner.getParty().getPartyMembersWithPets())
						if(member != null && !member.isDead() && member.getCurrentHp() / member.getMaxHp() < target.getCurrentHp() / target.getMaxHp() && _owner.isInRangeZ(member, skill.getCastRange()))
							target = member;
				}
				else if(_owner.getPet() != null && !_owner.getPet().isDead() && _owner.getPet().getCurrentHp() / _owner.getPet().getMaxHp() < _owner.getCurrentHp() / _owner.getMaxHp())
					target = _owner.getPet();
				double hpp = target.getCurrentHp() / target.getMaxHp();
				if(hpp > 0.95)
					return;
				if(Rnd.chance(hpp > 0.6 ? 13 : hpp > 0.3 ? 33 : 53))
				{
					_owner.altUseSkill(skill, target);
					_owner.broadcastPacket(new MagicSkillUse(_owner, target, skill.getId(), 1, 0, 0));
				}
			}
			catch(final Exception e)
			{}
			finally
			{
				_actionTask = ThreadPoolManager.getInstance().scheduleGeneral(new Heal(), (long) (10. * _owner.calculateAttackSpeed()));
			}
		}
	}

	private class Disappear implements Runnable
	{
		public void run()
		{
			stopAction();
			_owner.delCubic(_id);
			_owner.broadcastUserInfo();
		}
	}
}