package l2r.gameserver.model.instances;

import java.util.concurrent.ScheduledFuture;
import java.util.logging.Logger;

import javolution.util.FastList;
import l2r.Config;
import l2r.extensions.scripts.Events;
import l2r.gameserver.ThreadPoolManager;
import l2r.gameserver.ai.CtrlIntention;
import l2r.gameserver.ai.L2CharacterAI;
import l2r.gameserver.ai.L2StaticObjectAI;
import l2r.gameserver.geodata.network.geoclient.GeoClient;
import l2r.gameserver.instancemanager.SiegeManager;
import l2r.gameserver.model.L2Character;
import l2r.gameserver.model.L2Clan;
import l2r.gameserver.model.L2Player;
import l2r.gameserver.model.L2Skill;
import l2r.gameserver.model.L2Territory;
import l2r.gameserver.model.L2World;
import l2r.gameserver.model.entity.ClanHall;
import l2r.gameserver.model.entity.SevenSigns;
import l2r.gameserver.model.entity.siege.Siege;
import l2r.gameserver.model.entity.siege.SiegeClan;
import l2r.gameserver.network.L2GameClient;
import l2r.gameserver.serverpackets.MyTargetSelected;
import l2r.gameserver.serverpackets.StaticObject;
import l2r.gameserver.serverpackets.SystemMessage;
import l2r.gameserver.serverpackets.ValidateLocation;
import l2r.gameserver.templates.L2CharTemplate;
import l2r.gameserver.templates.L2Weapon;

public class L2DoorInstance extends L2Character
{
	protected static Logger _log = Logger.getLogger(L2DoorInstance.class.getName());

	protected final int _doorId;
	protected final String _name;
	private boolean open;
	public boolean geoOpen;
	private boolean _unlockable;
	private boolean _isHPVisible;
	private ClanHall _clanhall;
	public final L2Territory pos;

	protected int _autoActionDelay = -1;
	private ScheduledFuture _autoActionTask;
	private boolean _destroyable;

	@Override
	public L2CharacterAI getAI()
	{
		if(_ai == null)
			_ai = new L2StaticObjectAI(this);
		return _ai;
	}

	public L2DoorInstance(int objectId, L2CharTemplate template, int doorId, String name, boolean unlockable, boolean showHp)
	{
		super(objectId, template);
		_doorId = doorId;
		_name = name;
		_unlockable = unlockable;
		_isHPVisible = showHp;
		geoOpen = true;
		pos = new L2Territory("door_" + doorId);
	}

	public final boolean isUnlockable()
	{
		return _unlockable;
	}

	@Override
	public final byte getLevel()
	{
		return 1;
	}

	/**
	 * @return Returns the doorId.
	 */
	public int getDoorId()
	{
		return _doorId;
	}

	/**
	 * @return Returns the open.
	 */
	public boolean isOpen()
	{
		return open;
	}

	/**
	 * @param open The open to set.
	 */
	public synchronized void setOpen(boolean open)
	{
		this.open = open;
	}

	/**
	 * Sets the delay in milliseconds for automatic opening/closing
	 * of this door instance.
	 * <BR>
	 * <B>Note:</B> A value of -1 cancels the auto open/close task.
	 *
	 * @param actionDelay время задержки между действием
	 */
	public void setAutoActionDelay(int actionDelay)
	{
		if(_autoActionDelay == actionDelay)
			return;

		if(_autoActionTask != null)
		{
			_autoActionTask.cancel(false);
			_autoActionTask = null;
		}

		if(actionDelay > -1)
			_autoActionTask = ThreadPoolManager.getInstance().scheduleGeneralAtFixedRate(new AutoOpenClose(), actionDelay, actionDelay);

		_autoActionDelay = actionDelay;
	}

	public int getDamage()
	{
		int dmg = 6 - (int) Math.ceil(getCurrentHp() / getMaxHp() * 6);
		if(dmg > 6)
			return 6;
		if(dmg < 0)
			return 0;
		return dmg;
	}

	//TODO разобраться
	public boolean isEnemyOf(@SuppressWarnings("unused")
	L2Character cha)
	{
		return true;
	}

	@Override
	public boolean isAutoAttackable(L2Character attacker)
	{
		if(isUnlockable() || isDestroyable())
			return true;
		if(attacker == null)
			return false;
		L2Player player = attacker.getPlayer();
		if(player == null)
			return false;
		L2Clan clan = player.getClan();
		//TODO присвоить осаду обьекту при спавне, чтобы избавиться от перебора
		return !(clan != null && SiegeManager.getSiege(this) == clan.getSiege() && clan.isDefender());
	}

	public boolean isAttackable(L2Character attacker)
	{
		return isAutoAttackable(attacker);
	}

	public boolean isDestroyable()
	{
		return _destroyable;
	}

	public void setDestroyable(boolean val)
	{
		_destroyable = val;
	}

	@Override
	public void updateAbnormalEffect()
	{}

	/**
	 * Return null.<BR><BR>
	 */
	@Override
	public L2ItemInstance getActiveWeaponInstance()
	{
		return null;
	}

	@Override
	public L2Weapon getActiveWeaponItem()
	{
		return null;
	}

	@Override
	public L2ItemInstance getSecondaryWeaponInstance()
	{
		return null;
	}

	@Override
	public L2Weapon getSecondaryWeaponItem()
	{
		return null;
	}

	@Override
	public void onAction(L2Player player)
	{
		if(player == null)
			return;

		if(Events.onAction(player, this))
			return;

		if(this != player.getTarget())
		{
			player.setTarget(this);
			player.sendPacket(new MyTargetSelected(getObjectId(), player.getLevel()));

			if(isAutoAttackable(player))
				//player.sendPacket(new DoorStatusUpdate(this));
				player.sendPacket(new StaticObject(this));

			// correct location
			player.sendPacket(new ValidateLocation(this));
		}
		else
		{
			player.sendPacket(new MyTargetSelected(getObjectId(), player.getLevel()));
			if(isAutoAttackable(player))
			{
				if(Math.abs(player.getZ() - getZ()) < 400) // this max heigth difference might need some tweaking
					player.getAI().Attack(this, false);
				else
					player.sendActionFailed();
			}
			else if(!isInRange(player, INTERACTION_DISTANCE))
				player.getAI().setIntention(CtrlIntention.AI_INTENTION_INTERACT, this);
			else
			{
				if(Config.ALLOW_CH_DOOR_OPEN_ON_CLICK && player.getClan() != null && getClanHall() != null && player.getClanId() == getClanHall().getOwnerId())
					if(!isOpen())
						openMe();
					else
						closeMe();
				player.sendActionFailed();
			}
		}
	}

	public void onActionShift(L2GameClient client)
	{
		L2Player player = client.getActiveChar();

		if(Events.onActionShift(player, this))
			return;

		player.sendActionFailed();
	}

	@Override
	public void broadcastStatusUpdate()
	{
		//DoorStatusUpdate su = new DoorStatusUpdate(this);
		StaticObject su = new StaticObject(this);
		FastList<L2Player> players = L2World.getAroundPlayers(this);
		for(FastList.Node<L2Player> n = players.head(), end = players.tail(); (n = n.getNext()) != end;)
			n.getValue().sendPacket(su);
	}

	public void onOpen()
	{
		scheduleCloseMe(60000);
	}

	public void onClose()
	{
		closeMe();
	}

	/**
	 * Вызывает задание на закрытие двери через заданное время.
	 * @param delay Время в миллисекундах
	 */
	public final void scheduleCloseMe(long delay)
	{
		ThreadPoolManager.getInstance().scheduleGeneral(new CloseTask(), delay);
	}

	public final void closeMe()
	{
		synchronized (this)
		{
			if(!isOpen())
				return;

			open = false;
		}

		setGeoOpen(false);
		broadcastStatusUpdate();
	}

	public final void openMe()
	{
		synchronized (this)
		{
			if(isOpen())
				return;

			open = true;
		}

		setGeoOpen(true);
		broadcastStatusUpdate();
	}

	@Override
	public String toString()
	{
		return "door " + _doorId;
	}

	public String getDoorName()
	{
		return _name;
	}

	public void setClanHall(ClanHall clanhall)
	{
		_clanhall = clanhall;
	}

	public ClanHall getClanHall()
	{
		return _clanhall;
	}

	@Override
	public void doDie(L2Character killer)
	{
		Siege s = SiegeManager.getSiege(this);
		if(s != null)
		{
			for(SiegeClan sc : s.getDefenderClans().values())
			{
				L2Clan clan = sc.getClan();
				if(clan != null)
					for(L2Player player : clan.getOnlineMembers(""))
						if(player != null)
							player.sendPacket(new SystemMessage(SystemMessage.THE_CASTLE_GATE_HAS_BEEN_BROKEN_DOWN));
			}

			for(SiegeClan sc : s.getAttackerClans().values())
			{
				L2Clan clan = sc.getClan();
				if(clan != null)
					for(L2Player player : clan.getOnlineMembers(""))
						if(player != null)
							player.sendPacket(new SystemMessage(SystemMessage.THE_CASTLE_GATE_HAS_BEEN_BROKEN_DOWN));
			}
		}

		setGeoOpen(true);

		super.doDie(killer);
	}

	@Override
	public void spawnMe()
	{
		super.spawnMe();
		closeMe();
		//setGeoOpen(false);
	}

	public boolean isHPVisible()
	{
		return _isHPVisible;
	}

	public void setHPVisible(boolean val)
	{
		_isHPVisible = val;
	}

	private int upgradeHp;

	@Override
	public int getMaxHp()
	{
		return super.getMaxHp() + upgradeHp;
	}

	public void setUpgradeHp(int hp)
	{
		upgradeHp = hp;
	}

	public int getUpgradeHp()
	{
		return upgradeHp;
	}

	@Override
	public int getPDef(L2Character target)
	{
		switch(SevenSigns.getInstance().getSealOwner(SevenSigns.SEAL_STRIFE))
		{
			case SevenSigns.CABAL_DAWN:
				return (int) (super.getPDef(target) * 1.2);
			case SevenSigns.CABAL_DUSK:
				return (int) (super.getPDef(target) * 0.3);
			default:
				return super.getPDef(target);
		}
	}

	@Override
	public int getMDef(L2Character target, L2Skill skill)
	{
		switch(SevenSigns.getInstance().getSealOwner(SevenSigns.SEAL_STRIFE))
		{
			case SevenSigns.CABAL_DAWN:
				return (int) (super.getMDef(target, skill) * 1.2);
			case SevenSigns.CABAL_DUSK:
				return (int) (super.getMDef(target, skill) * 0.3);
			default:
				return super.getMDef(target, skill);
		}
	}

	@Override
	public boolean isLethalImmune()
	{
		return true;
	}

	/**
	 * Двери на осадах уязвимы во время осады.
	 * Остальные двери не уязвимы вообще.
	 * @return инвульная ли дверь.
	 */
	@Override
	public boolean isInvul()
	{
		Siege siege = SiegeManager.getSiege(this);
		return siege == null || !siege.isInProgress();
	}

	@Override
	public boolean isFearImmune()
	{
		return true;
	}

	@Override
	public boolean paralyzeImmune()
	{
		return true;
	}

	private int getDoorHeight()
	{
		return pos.getZmax() - pos.getZmin() & 0xfff0;
	}

	/**
	 * Устанавливает значение закрытости\открытости в геодате<br>
	 * @param val новое значение
	 */
	private void setGeoOpen(boolean val)
	{
		if(geoOpen == val)
			return;

		geoOpen = val;
		if(val)
			GeoClient.getInstance().openDoor(pos);
		else
			GeoClient.getInstance().closeDoor(pos);
	}

	private class CloseTask implements Runnable
	{
		public void run()
		{
			onClose();
		}
	}

	/**
	 * Manages the auto open and closing of a door.
	 */
	private class AutoOpenClose implements Runnable
	{
		public void run()
		{
			String doorAction;
			if(!isOpen())
			{
				doorAction = "opened";
				openMe();
			}
			else
			{
				doorAction = "closed";
				closeMe();
			}

			if(Config.DEBUG)
				_log.info("Auto " + doorAction + " door ID " + _doorId + " (" + _name + ") for " + _autoActionDelay / 1000 / 60 + " minute(s).");
		}
	}
}
