package l2r.gameserver.model.instances;

import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ScheduledFuture;
import java.util.logging.Level;
import java.util.logging.Logger;

import javolution.util.FastList;
import javolution.util.FastMap;
import l2r.Config;
import l2r.extensions.multilang.CustomMessage;
import l2r.gameserver.ThreadPoolManager;
import l2r.gameserver.ai.CtrlEvent;
import l2r.gameserver.ai.CtrlIntention;
import l2r.gameserver.instancemanager.CursedWeaponsManager;
import l2r.gameserver.model.L2Character;
import l2r.gameserver.model.L2Drop;
import l2r.gameserver.model.L2DropData;
import l2r.gameserver.model.L2Manor;
import l2r.gameserver.model.L2Party;
import l2r.gameserver.model.L2Playable;
import l2r.gameserver.model.L2Player;
import l2r.gameserver.model.L2Summon;
import l2r.gameserver.model.base.Experience;
import l2r.gameserver.model.base.ItemToDrop;
import l2r.gameserver.model.instances.L2ItemInstance.ItemLocation;
import l2r.gameserver.model.quest.QuestState;
import l2r.gameserver.serverpackets.SocialAction;
import l2r.gameserver.serverpackets.SystemMessage;
import l2r.gameserver.skills.Stats;
import l2r.gameserver.tables.ItemTable;
import l2r.gameserver.templates.L2Item;
import l2r.gameserver.templates.L2NpcTemplate;
import l2r.util.Location;
import l2r.util.MinionList;
import l2r.util.Rnd;
import l2r.util.Util;

/**
 * This class manages all Monsters.
 *
 * L2MonsterInstance :<BR><BR>
 * <li>L2MinionInstance</li>
 * <li>L2RaidBossInstance </li>
 */
public class L2MonsterInstance extends L2NpcInstance
{
	final static Logger _log = Logger.getLogger(L2MonsterInstance.class.getName());

	public void setOverhitAttacker(L2Character overhitAttacker)
	{
		this.overhitAttacker = new WeakReference<L2Character>(overhitAttacker);
	}

	protected final class RewardInfo
	{
		protected L2Character _attacker;
		protected int _dmg = 0;

		public RewardInfo(final L2Character attacker, final int dmg)
		{
			_attacker = attacker;
			_dmg = dmg;
		}

		public void addDamage(int dmg)
		{
			if(dmg < 0)
				dmg = 0;

			_dmg += dmg;
		}

		@Override
		public int hashCode()
		{
			return _attacker.getObjectId();
		}
	}

	public final class AbsorberInfo
	{
		/** The attacker L2Character concerned by this AbsorberInfo of this L2NpcInstance */
		L2Player _absorber;
		int _crystalId;
		double _absorbedHP;

		AbsorberInfo(final L2Player attacker, final int crystalId, final double absorbedHP)
		{
			_absorber = attacker;
			_crystalId = crystalId;
			_absorbedHP = absorbedHP;
		}

		/**
		 * Return the Identifier of the absorber L2Character.<BR><BR>
		 */
		@Override
		public int hashCode()
		{
			return _absorber.getObjectId();
		}
	}

	private Boolean _dead = false;
	private Object _dieLock = new Object();

	/** True if an over-hit enabled skill has successfully landed on the L2NpcInstance */
	private boolean _overhit;

	/** Stores the extra (over-hit) damage done to the L2NpcInstance when the attacker uses an over-hit enabled skill */
	private double _overhitDamage;

	/** Stores the attacker who used the over-hit enabled skill on the L2NpcInstance */
	private WeakReference<L2Character> overhitAttacker;

	protected final MinionList minionList;

	@SuppressWarnings("unchecked")
	private ScheduledFuture minionMaintainTask;

	private static final int MONSTER_MAINTENANCE_INTERVAL = 1000;

	/** Максимальный уровень мобов */
	private static final int MONSTER_MAX_LEVEL = 100;

	private HashSet<L2ItemInstance> _inventory;

	/** crops */
	private L2ItemInstance[] _harvestItems;
	private Object _harvestLock = new Object();
	private L2Item _seeded;
	private L2Player _seeder;

	/** True if a Soul Crystal was successfully used on the L2NpcInstance */
	private boolean _absorbed;

	/** The table containing all L2Player that successfully absorbed the soul of this L2NpcInstance */
	private ConcurrentHashMap<L2Player, AbsorberInfo> _absorbersList;

	/** Table containing all Items that a Dwarf can Sweep on this L2NpcInstance */
	private L2ItemInstance[] _sweepItems;
	private Object _sweepLock = new Object();

	/** Static tables containing all mobIDs that can have their souls absorbed, divided by max soul crystal level */
	private static final int[] _absorbingMOBS_level4 = {
			20583,
			20584,
			20585,
			20586,
			20587,
			20588,
			20636,
			20637,
			20638,
			20639,
			20640,
			20641,
			20642,
			20643,
			20644,
			20645 };
	private static final int[] _absorbingMOBS_level8 = { 20646, 20647, 20648, 20649, 20650, 21006, 21007, 21008 };
	private static final int[] _absorbingMOBS_level10 = {
			20627,
			20628,
			20629,
			20674,
			20761,
			20762,
			20821,
			20823,
			20826,
			20827,
			20828,
			20829,
			20830,
			20831,
			20858,
			20859,
			20860,
			21009,
			21010,
			21062,
			21063,
			21068,
			21070 };
	private static final int[] _absorbingMOBS_level12 = { 29022, 25163, 25269, 25453, 25328, 25109, 29020, 25283, 25286, 22215, 22216, 22217, 29065 };
	private static final int[] _absorbingMOBS_level13 = { 25338, 29019, 25319, 29028, 29046, 29047 };

	/** Soul Crystal Basic Informations */
	// First ID of each soul crystal
	private static final short _REDCRYSTAL_OFFSET = 4629;
	private static final short _REDCRYSTAL_LVL11 = 5577;
	private static final short _REDCRYSTAL_LVL12 = 5580;
	private static final short _REDCRYSTAL_LVL13 = 5908;
	private static final short _REDCRYSTAL_LVL14 = 9570;
	private static final short _GREENCRYSTAL_OFFSET = 4640;
	private static final short _GREENCRYSTAL_LVL11 = 5578;
	private static final short _GREENCRYSTAL_LVL12 = 5581;
	private static final short _GREENCRYSTAL_LVL13 = 5911;
	private static final short _GREENCRYSTAL_LVL14 = 9572;
	private static final short _BLUECRYSTAL_OFFSET = 4651;
	private static final short _BLUECRYSTAL_LVL11 = 5579;
	private static final short _BLUECRYSTAL_LVL12 = 5582;
	private static final short _BLUECRYSTAL_LVL13 = 5914;
	private static final short _BLUECRYSTAL_LVL14 = 9571;

	// Max number of levels a soul crystal may reach
	private static final short _MAX_CRYSTALS_LEVEL = 14;
	// Chances for each event of a soul crystal
	private static final float _LEVEL_CHANCE = 0.325f;
	/** End of Soul Crystal Basic Informations */

	// For ALT_GAME_MATHERIALSDROP
	protected static final L2DropData[] _matdrop = new L2DropData[] {
	//                                           Item              Price Chance
			new L2DropData(1864, 1, 1, 50000, 1), // Stem              100   5%
			new L2DropData(1865, 1, 1, 25000, 1), // Varnish           200   2.5%
			new L2DropData(1866, 1, 1, 16666, 1), // Suede             300   1.6666%
			new L2DropData(1867, 1, 1, 33333, 1), // Animal Skin       150   3.3333%
			new L2DropData(1868, 1, 1, 50000, 1), // Thread            100   5%
			new L2DropData(1869, 1, 1, 25000, 1), // Iron Ore          200   2.5%
			new L2DropData(1870, 1, 1, 25000, 1), // Coal              200   2.5%
			new L2DropData(1871, 1, 1, 25000, 1), // Charcoal          200   2.5%
			new L2DropData(1872, 1, 1, 50000, 1), // Animal Bone       150   5%
			new L2DropData(1873, 1, 1, 10000, 1), // Silver Nugget     500   1%
			new L2DropData(1874, 1, 1, 1666, 20), // Oriharukon Ore    3000  0.1666%
			new L2DropData(1875, 1, 1, 1666, 20), // Stone of Purity   3000  0.1666%
			new L2DropData(1876, 1, 1, 5000, 20), // Mithril Ore       1000  0.5%
			new L2DropData(1877, 1, 1, 1000, 20), // Adamantite Nugget 5000  0.1%
			new L2DropData(4039, 1, 1, 833, 40), //  Mold Glue         6000  0.0833%
			new L2DropData(4040, 1, 1, 500, 40), //  Mold Lubricant    10000 0.05%
			new L2DropData(4041, 1, 1, 217, 40), //  Mold Hardener     23000 0.0217%
			new L2DropData(4042, 1, 1, 417, 40), //  Enria             12000 0.0417%
			new L2DropData(4043, 1, 1, 833, 40), //  Asofe             6000  0.0833%
			new L2DropData(4044, 1, 1, 833, 40) //   Thons             6000  0.0833%
	};

	protected static final L2DropData[] _herbs = new L2DropData[] {
	//                                            Name      Chance
			// HP/MP herbs
			new L2DropData(8600, 1, 1, 150000, 1), // of Life            10%
			new L2DropData(8601, 1, 1, 60000, 1), //  Greater of Life    4%
			new L2DropData(8602, 1, 1, 16000, 1), //  Superior of Life   0.8%
			new L2DropData(8603, 1, 1, 150000, 1), // of Mana            10%
			new L2DropData(8604, 1, 1, 60000, 1), //  Greater of Mana    4%
			new L2DropData(8605, 1, 1, 16000, 1), //  Superior of Mana   0.8%
			//new L2DropData(8614, 1, 1, 2000, 1), // f Recovery         0.2%

			// Speed herb
			new L2DropData(8611, 1, 1, 150000, 1), // of Speed           15%

			// Warrior herbs
			new L2DropData(8606, 1, 1, 150000, 1), // of Power           15%
			new L2DropData(8608, 1, 1, 150000, 1), // of Atk. Spd.       15%
			new L2DropData(8610, 1, 1, 150000, 1), // of Critical Attack 15%
			//new L2DropData(8612, 1, 1, 2000, 1), //   of Warrior       0.2%

			// Mystic herbs
			new L2DropData(8607, 1, 1, 150000, 1), // of Magic           15%
			new L2DropData(8609, 1, 1, 150000, 1), // of Casting Speed   15%
	//new L2DropData(8613, 1, 1, 2000, 1), //     of Mystic          0.2%
	};

	protected static final L2DropData[] _lifestones = new L2DropData[] {
	//
			new L2DropData(8723, 1, 1, 100, 44, 46), // Life Stone: level 46
			new L2DropData(8724, 1, 1, 100, 47, 49), // Life Stone: level 49
			new L2DropData(8725, 1, 1, 100, 50, 52), // Life Stone: level 52
			new L2DropData(8726, 1, 1, 100, 53, 55), // Life Stone: level 55
			new L2DropData(8727, 1, 1, 100, 56, 58), // Life Stone: level 58
			new L2DropData(8728, 1, 1, 100, 59, 61), // Life Stone: level 61
			new L2DropData(8729, 1, 1, 100, 62, 66), // Life Stone: level 64
			new L2DropData(8730, 1, 1, 100, 67, 72), // Life Stone: level 67
			new L2DropData(8731, 1, 1, 100, 73, 75), // Life Stone: level 70
			new L2DropData(8732, 1, 1, 100, 76, 79), // Life Stone: level 76
			new L2DropData(9573, 1, 1, 100, 76, MONSTER_MAX_LEVEL), // Life Stone: level 80
			new L2DropData(8733, 1, 1, 50, 44, 46), // Mid-Grade Life Stone: level 46
			new L2DropData(8734, 1, 1, 50, 47, 49), // Mid-Grade Life Stone: level 49
			new L2DropData(8735, 1, 1, 50, 50, 52), // Mid-Grade Life Stone: level 52
			new L2DropData(8736, 1, 1, 50, 53, 55), // Mid-Grade Life Stone: level 55
			new L2DropData(8737, 1, 1, 50, 56, 58), // Mid-Grade Life Stone: level 58
			new L2DropData(8738, 1, 1, 50, 59, 61), // Mid-Grade Life Stone: level 61
			new L2DropData(8739, 1, 1, 50, 62, 66), // Mid-Grade Life Stone: level 64
			new L2DropData(8740, 1, 1, 50, 67, 72), // Mid-Grade Life Stone: level 67
			new L2DropData(8741, 1, 1, 50, 73, 75), // Mid-Grade Life Stone: level 70
			new L2DropData(8742, 1, 1, 50, 76, 79), // Mid-Grade Life Stone: level 76
			new L2DropData(9574, 1, 1, 50, 80, MONSTER_MAX_LEVEL), // Mid-Grade Life Stone: level 80
			new L2DropData(8743, 1, 1, 10, 44, 46), // High-Grade Life Stone: level 46
			new L2DropData(8744, 1, 1, 10, 47, 49), // High-Grade Life Stone: level 49
			new L2DropData(8745, 1, 1, 10, 50, 52), // High-Grade Life Stone: level 52
			new L2DropData(8746, 1, 1, 10, 53, 55), // High-Grade Life Stone: level 55
			new L2DropData(8747, 1, 1, 10, 56, 58), // High-Grade Life Stone: level 58
			new L2DropData(8748, 1, 1, 10, 59, 61), // High-Grade Life Stone: level 61
			new L2DropData(8749, 1, 1, 10, 62, 66), // High-Grade Life Stone: level 64
			new L2DropData(8750, 1, 1, 10, 67, 72), // High-Grade Life Stone: level 67
			new L2DropData(8751, 1, 1, 10, 73, 75), // High-Grade Life Stone: level 70
			new L2DropData(8752, 1, 1, 10, 76, 79), // High-Grade Life Stone: level 76
			new L2DropData(9575, 1, 1, 10, 80, MONSTER_MAX_LEVEL) // High-Grade Life Stone: level 80
	};

	protected static final L2DropData[] _toplifestones = new L2DropData[] {
	//
			new L2DropData(8753, 1, 1, 1000000, 44, 46), // Top-Grade Life Stone: level 46
			new L2DropData(8754, 1, 1, 1000000, 47, 49), // Top-Grade Life Stone: level 49
			new L2DropData(8755, 1, 1, 1000000, 50, 52), // Top-Grade Life Stone: level 52
			new L2DropData(8756, 1, 1, 1000000, 53, 55), // Top-Grade Life Stone: level 55
			new L2DropData(8757, 1, 1, 1000000, 56, 58), // Top-Grade Life Stone: level 58
			new L2DropData(8758, 1, 1, 1000000, 59, 61), // Top-Grade Life Stone: level 61
			new L2DropData(8759, 1, 1, 1000000, 62, 66), // Top-Grade Life Stone: level 64
			new L2DropData(8760, 1, 1, 1000000, 67, 72), // Top-Grade Life Stone: level 67
			new L2DropData(8761, 1, 1, 1000000, 73, 75), // Top-Grade Life Stone: level 70
			new L2DropData(8762, 1, 1, 1000000, 76, 79), // Top-Grade Life Stone: level 76
			new L2DropData(9576, 1, 1, 1000000, 80, MONSTER_MAX_LEVEL) // Top-Grade Life Stone: level 80
	};

	protected static final L2DropData[] _raiditems = new L2DropData[] {
	//
			new L2DropData(9814, 1, 2, 300000, 40, 80), // Memento Mori
			//new L2DropData(8158, 1, 3, 300000, 40, 49), // Memento Mori: C-Grade
			//new L2DropData(8159, 1, 3, 300000, 50, 60), // Memento Mori: B-Grade
			//new L2DropData(8160, 1, 3, 300000, 61, 76), // Memento Mori: A-Grade
			new L2DropData(9815, 1, 2, 300000, 40, 80), // Dragon Heart
			//new L2DropData(8161, 1, 3, 300000, 40, 49), // Dragon Heart: C-Grade
			//new L2DropData(8162, 1, 3, 300000, 50, 60), // Dragon Heart: B-Grade
			//new L2DropData(8163, 1, 3, 300000, 61, 76), // Dragon Heart: A-Grade
			new L2DropData(9816, 1, 2, 300000, 40, 80), // Earth Egg
			//new L2DropData(8164, 1, 3, 300000, 40, 49), // Earth Egg: C-Grade
			//new L2DropData(8165, 1, 3, 300000, 50, 60), // Earth Egg: B-Grade
			//new L2DropData(8166, 1, 3, 300000, 61, 76), // Earth Egg: A-Grade
			new L2DropData(9817, 1, 2, 300000, 40, 80), // Nonliving Nucleus
			//new L2DropData(8167, 1, 3, 300000, 40, 49), // Nonliving Nucleus: C-Grade
			//new L2DropData(8168, 1, 3, 300000, 50, 60), // Nonliving Nucleus: B-Grade
			//new L2DropData(8169, 1, 3, 300000, 61, 76), // Nonliving Nucleus: A-Grade
			new L2DropData(9818, 1, 2, 300000, 40, 80), // Angelic Essence
			//new L2DropData(8170, 1, 3, 300000, 40, 49), // Angelic Essence: C-Grade
			//new L2DropData(8171, 1, 3, 300000, 50, 60), // Angelic Essence: B-Grade
			//new L2DropData(8172, 1, 3, 300000, 61, 76), // Angelic Essence: A-Grade
			new L2DropData(8176, 1, 2, 300000, 40, 80) // Destruction Tombstone
	};

	/**
	 * Constructor of L2MonsterInstance (use L2Character and L2NpcInstance constructor).<BR><BR>
	 *
	 * <B><U> Actions</U> :</B><BR><BR>
	 * <li>Call the L2Character constructor to set the _template of the L2MonsterInstance (copy skills from template to object and link _calculators to NPC_STD_CALCULATOR) </li>
	 * <li>Set the name of the L2MonsterInstance</li>
	 * <li>Create a RandomAnimation Task that will be launched after the calculated delay if the server allow it </li><BR><BR>
	 *
	 * @param objectId Identifier of the object to initialized
	 * @param template to apply to the NPC
	 */
	public L2MonsterInstance(int objectId, L2NpcTemplate template)
	{
		super(objectId, template);
		minionList = new MinionList(this);
	}

	@Override
	public boolean isMovementDisabled()
	{
		// Невозможность ходить для этих мобов
		return getNpcId() == 18344 || getNpcId() == 18345 || super.isMovementDisabled();
	}

	@Override
	public boolean isLethalImmune()
	{
		return getNpcId() == 22215 || getNpcId() == 22216 || getNpcId() == 22217 || super.isLethalImmune();
	}

	/**
	 * Return True if the attacker is not another L2MonsterInstance.<BR><BR>
	 */
	@Override
	public boolean isAutoAttackable(L2Character attacker)
	{
		return !attacker.isMonster;
	}

	@Override
	public void onSpawn()
	{
		_dead = false;
		super.onSpawn();
		spawnMinions();

		// Clear mob spoil, absorbs, seed
		setSpoiled(false);
		_sweepItems = null;
		_absorbed = false;
		_absorbersList = null;
		_seeded = null;
		_seeder = null;
	}

	protected int getMaintenanceInterval()
	{
		return MONSTER_MAINTENANCE_INTERVAL;
	}

	public void spawnMinions()
	{
		if(getTemplate().getMinionData() != null)
			try
			{
				minionMaintainTask = ThreadPoolManager.getInstance().scheduleGeneral(new Runnable(){
					public void run()
					{
						try
						{
							minionList.maintainMinions();
						}
						catch(Throwable e)
						{
							_log.log(Level.SEVERE, "", e);
						}
					}
				}, getMaintenanceInterval());
			}
			catch(NullPointerException e)
			{}
	}

	public Location getMinionPosition()
	{
		return minionList.getMinionPosition();
	}

	@Override
	public void callMinionsToAssist(L2Character attacker)
	{
		if(minionList.hasMinions())
			for(L2MinionInstance minion : minionList.getSpawnedMinions())
				if(minion != null && minion.getAI().getIntention() != CtrlIntention.AI_INTENTION_ATTACK && !minion.isDead())
					minion.getAI().notifyEvent(CtrlEvent.EVT_AGGRESSION, attacker, Rnd.get(1, 100));
	}

	public void setDead(boolean dead)
	{
		_dead = dead;
	}

	public void removeMinions()
	{
		if(minionMaintainTask != null)
			minionMaintainTask.cancel(true);

		minionList.maintainLonelyMinions();
	}

	@Override
	public FastList<L2MinionInstance> getSpawnedMinions()
	{
		return minionList.getSpawnedMinions();
	}

	public int getTotalSpawnedMinionsInstances()
	{
		return minionList.countSpawnedMinions();
	}

	public int getTotalSpawnedMinionsGroups()
	{
		return minionList.lazyCountSpawnedMinionsGroups();
	}

	public void notifyMinionDied(L2MinionInstance minion)
	{
		minionList.removeSpawnedMinion(minion);
	}

	public void notifyMinionSpawned(L2MinionInstance minion)
	{
		minionList.addSpawnedMinion(minion);
	}

	@Override
	public boolean hasMinions()
	{
		return minionList.hasMinions();
	}

	@Override
	public void deleteMe()
	{
		if(hasMinions())
			removeMinions();
		super.deleteMe();
	}

	@Override
	public void doDie(L2Character killer)
	{
		if(minionMaintainTask != null)
			minionMaintainTask.cancel(true);

		synchronized (_dieLock)
		{
			if(_dead)
				return;
			_dead = true;
		}

		if(this instanceof L2ChestInstance && !((L2ChestInstance) this).isFake())
		{
			super.doDie(killer);
			return;
		}

		// Notify the Quest Engine of the L2NpcInstance death if necessary
		try
		{
			if(killer instanceof L2Playable)
			{
				// Get the L2Player that killed the L2NpcInstance
				final L2Player player = killer.getPlayer();
				if(player == null) // маловероятно но бывает
					return;

				if(Config.KILL_COUNTER)
					player.incrementKillsCounter(getNpcId());
				getTemplate().killscount++;

				if(player.getParty() != null)
					for(L2Player pl : player.getParty().getPartyMembers())
					{
						ArrayList<QuestState> ql = pl.getQuestsForKills(this);
						if(ql != null && !pl.isDead() && player.isInRange(pl, Config.ALT_PARTY_DISTRIBUTION_RANGE))
							for(QuestState qs : ql)
								if(qs.getQuest().isParty() || pl.equals(player))
									qs.getQuest().notifyKill(this, qs);
					}
				else
				{
					ArrayList<QuestState> ql = player.getQuestsForKills(this);
					if(ql != null)
						for(QuestState qs : ql)
							qs.getQuest().notifyKill(this, qs);
				}
			}
		}
		catch(final Exception e)
		{
			e.printStackTrace();
		}

		// Distribute Exp and SP rewards to L2Player (including Summon owner) that hit the L2NpcInstance and to their Party members
		try
		{
			calculateRewards(killer);
		}
		catch(final Exception e)
		{
			_log.log(Level.SEVERE, "", e);
			e.printStackTrace();
		}

		// Enhance soul crystals of the attacker if this L2NpcInstance had its soul absorbed
		try
		{
			if(killer instanceof L2Playable)
				levelSoulCrystals(killer.getPlayer());
		}
		catch(final Exception e)
		{
			_log.log(Level.SEVERE, "", e);
		}

		super.doDie(killer);
	}

	private void calculateRewards(L2Character lastAttacker)
	{
		ConcurrentHashMap<L2Playable, AggroInfo> aggroList = getAggroList();

		L2Player player = lastAttacker.getPlayer();
		if(player == null || aggroList.size() == 0)
			return;

		int npcID = getTemplate().npcId;

		FastMap<L2Character, RewardInfo> rewards = new FastMap<L2Character, RewardInfo>().setShared(true);

		for(AggroInfo info : aggroList.values())
		{
			L2Character attacker = info.attacker;
			L2Character owner = null;

			if(attacker.isSummon)
				owner = attacker.getOwner();

			int damage = info.damage;

			if(npcID <= 0 || damage <= 1)
				continue;

			synchronized (this)
			{
				RewardInfo reward = rewards.get(attacker);
				if(attacker.isPlayer || attacker.isPet)
				{
					if(reward == null)
						rewards.put(attacker, new RewardInfo(attacker, damage));
					else
						reward.addDamage(damage);
				}
				else if(owner != null)
				{
					//TODO Проверить нужно ли нам учитывать разницу в уровне суммона и хозяина.
					int levelsDiff = owner.getLevel() - attacker.getLevel();
					if(levelsDiff < 0)
						levelsDiff = 0;
					damage = damage - levelsDiff * 50;
					if(damage < 0)
						damage = 0;
					reward = rewards.get(owner);
					if(reward == null)
						rewards.put(owner, new RewardInfo(owner, damage));
					else
						reward.addDamage(damage);
				}
			}
		}

		for(FastMap.Entry<L2Character, RewardInfo> e = rewards.head(), end = rewards.tail(); e != null && (e = e.getNext()) != end && e != null;)
		{
			RewardInfo reward = e.getValue();
			if(reward == null)
				continue;

			L2Character attacker = reward._attacker;

			if(attacker == null)
				continue;

			if(attacker.isDead())
				continue;

			L2Party attackerParty = null;
			if(attacker.isPlayer)
				attackerParty = ((L2Player) attacker).getParty();

			int maxHp = getMaxHp();

			if(attackerParty == null)
			{
				int damage = reward._dmg;
				float penalty = 0;
				if(attacker.isPlayer)
				{
					L2Summon summon = attacker.getPet();
					if(summon != null && summon.isSummon)
						penalty = ((L2SummonInstance) summon).getExpPenalty();
				}

				if(damage > maxHp)
					damage = maxHp;

				if(damage > 0)
				{
					int diff = attacker.getLevel() - getLevel();

					//kamael exp penalty
					if(attacker.getLevel() > 77 && diff > 3 && diff <= 5)
						diff += 3;

					long xp = 0;
					long sp = 0;

					if(attacker.knowsObject(this))
					{
						long[] tmp = calculateExpAndSp(diff, damage);
						xp = tmp[0];
						xp *= 1 - penalty;
						sp = tmp[1];
					}

					if(isOverhit() && player == getOverhitAttacker())
					{
						int overHitExp = calculateOverhitExp(xp);
						player.sendPacket(new SystemMessage(SystemMessage.OVER_HIT));
						player.sendPacket(new SystemMessage(SystemMessage.ACQUIRED_S1_BONUS_EXPERIENCE_THROUGH_OVER_HIT).addNumber(overHitExp));
						xp += overHitExp;
					}

					attacker.addExpAndSp(xp, sp);
				}
				rewards.remove(attacker);
			}
			else
			{
				int partyDmg = 0;
				float partyMul = 1.f;
				int partylevel = 1;

				ArrayList<L2Player> rewardedMembers = new ArrayList<L2Player>();

				for(L2Player partyMember : attackerParty.getPartyMembers())
				{
					if(partyMember == null || partyMember.isDead())
						continue;

					RewardInfo ai = rewards.get(partyMember);
					if(ai != null)
						partyDmg += ai._dmg;

					L2Summon pet = partyMember.getPet();
					if(pet != null)
						//final RewardInfo aiPet = rewards.get(pet);
						//if(aiPet != null)
						//	partyDmg += aiPet.dmg;
						rewards.remove(pet);

					if(!partyMember.isDead())
						rewardedMembers.add(partyMember);

					rewards.remove(partyMember);
					if(partyMember.isInRange(lastAttacker, 2700) && partyMember.getLevel() > partylevel)
						partylevel = partyMember.getLevel();
				}

				if(partyDmg < maxHp)
					partyMul = (float) partyDmg / maxHp;
				else
					partyDmg = maxHp;

				if(partyDmg > 0)
				{
					long xp = 0;
					long sp = 0;

					if(attacker.knowsObject(this))
					{
						int diff = partylevel - getLevel();

						//kamael exp penalty
						if(partylevel > 77 && diff > 3 && diff <= 5)
							diff += 3;

						long[] tmp = calculateExpAndSp(diff, partyDmg);
						xp = tmp[0];
						sp = tmp[1];
					}

					xp *= partyMul;
					sp *= partyMul;

					// Check for an over-hit enabled strike
					// (When in party, the over-hit exp bonus is given to the whole party and splitted proportionally through the party members)
					if(isOverhit() && player == getOverhitAttacker())
					{
						int overHitExp = calculateOverhitExp(xp);
						attackerParty.broadcastToPartyMembersInRange(player, new SystemMessage(SystemMessage.ACQUIRED_S1_BONUS_EXPERIENCE_THROUGH_OVER_HIT).addNumber(overHitExp), 2700);
						xp += overHitExp;
					}

					attackerParty.distributeXpAndSp(xp, sp, rewardedMembers, lastAttacker, this);
				}
			}
		}

		// Check the drop of a cursed weapon
		CursedWeaponsManager.getInstance().dropAttackable(this, player);

		// Manage Base, Quests and Special Events drops of the L2NpcInstance
		doItemDrop(lastAttacker);

		// Manage Sweep drops of the L2NpcInstance
		if(isSpoiled())
			doSweepDrop(lastAttacker);

		if(!isRaid && !(this instanceof L2MinionInstance)) // С рейдов падают только топовые лайфстоны, с миньонов вообще ничего дополнительно не падает
		{
			int leveldiff = calculateLevelDiffForDrop();
			double chancemod = leveldiff > 1 ? Experience.penaltyModifier(leveldiff - 1, 9) : 1.;
			double RATE_HP = ((L2NpcTemplate) _template).rateHp;

			// Дополнительный дроп материалов
			if(Config.ALT_GAME_MATHERIALSDROP && chancemod > 0 && (!isSeeded() || _seeded.isAltSeed()))
				for(L2DropData d : _matdrop)
					if(getLevel() >= d.getMinLevel())
					{
						int count = Util.rollDrop(d.getMinDrop(), d.getMaxDrop(), d.getChance() * chancemod * RATE_HP);
						if(count > 0)
							dropItem(lastAttacker.getPlayer(), d.getItemId(), count);
					}

			// Хербы
			if(((L2NpcTemplate) _template).isDropHerbs)
				for(L2DropData h : _herbs)
					if(Rnd.get(1, L2Drop.MAX_CHANCE) <= h.getChance() * chancemod)
						dropItem(lastAttacker.getPlayer(), h.getItemId(), 1);

			// Лайфстоуны
			if(chancemod > 0)
				for(L2DropData l : _lifestones)
					if(getLevel() >= l.getMinLevel() && getLevel() <= l.getMaxLevel() && Rnd.get(1, L2Drop.MAX_CHANCE) <= l.getChance() * Config.RATE_DROP_ITEMS * chancemod * RATE_HP)
						dropItem(lastAttacker.getPlayer(), l.getItemId(), 1);
		}
		else if(isRaid)
		{
			// Лайфстоуны с рейдов
			for(L2DropData l : _toplifestones)
				if(getLevel() >= l.getMinLevel() && getLevel() <= l.getMaxLevel() && Rnd.get(1, L2Drop.MAX_CHANCE) <= l.getChance() * Config.RATE_DROP_ITEMS)
					dropItem(lastAttacker.getPlayer(), l.getItemId(), 1);
			// предметы для изучения клановых скилов (не падают с эпиков)
			if(!(this instanceof L2BossInstance))
				for(L2DropData l : _raiditems)
					if(getLevel() >= l.getMinLevel() && getLevel() <= l.getMaxLevel() && Rnd.get(1, L2Drop.MAX_CHANCE) <= l.getChance())
					{
						int mod = (int) (getLevel() / 20 * Config.RATE_DROP_ITEMS);
						dropItem(lastAttacker.getPlayer(), l.getItemId(), Rnd.get(l.getMinDrop() * mod, l.getMaxDrop() * mod));
					}
		}

		clearAggroList();
	}

	private int calculateLevelDiffForDrop()
	{
		if(Config.DEEPBLUE_DROP_RULES)
		{
			int mobLevel = getLevel();
			int level = mobLevel, maxDamage = 0;

			for(AggroInfo aggro : getAggroList().values())
				if(aggro.attacker != null)
					if(aggro.damage > maxDamage)
					{
						maxDamage = aggro.damage;
						level = aggro.attacker.getLevel();
					}

			// According to official data (Prima), deep blue mobs are 9 or more levels below players
			if(level - mobLevel > (isRaid ? Config.RAID_MAX_LEVEL_DIFF : 8))
				return level - mobLevel - 8;
		}

		return 0;
	}

	public void giveItem(L2ItemInstance item, boolean store)
	{
		if(_inventory == null)
			_inventory = new HashSet<L2ItemInstance>();

		_inventory.add(item);

		if(store)
		{
			item.setOwnerId(getNpcId());
			item.setLocation(ItemLocation.MONSTER);
			item.updateDatabase();
		}
	}

	/**
	 * Может вернуть null если инвентарь пуст
	 */
	public HashSet<L2ItemInstance> getInventory()
	{
		return _inventory;
	}

	public void truncateInventory()
	{
		_inventory.clear();
		_inventory = null;
	}

	@Override
	public void onRandomAnimation()
	{
		// Action id для живности 1-3
		broadcastPacket(new SocialAction(getObjectId(), Rnd.get(1, 3)));
	}

	@Override
	public int getKarma()
	{
		return 0;
	}

	/**
	 * Activate the absorbed soul condition on the L2NpcInstance.<BR><BR>
	 */
	public void absorbSoul()
	{
		_absorbed = true;
	}

	/**
	 * Return True if the L2NpcInstance had his soul absorbed.<BR><BR>
	 */
	public boolean isAbsorbed()
	{
		return _absorbed;
	}

	/**
	 * Adds an attacker that successfully absorbed the soul of this L2NpcInstance into the _absorbersList.<BR><BR>
	 *
	 * params:  attacker  - a valid L2Player
	 *      condition - an integer indicating the event when mob dies. This should be:
	 *              = 0   - "the crystal scatters";
	 *              = 1   - "the crystal failed to absorb. nothing happens";
	 *              = 2   - "the crystal resonates because you got more than 1 crystal on you";
	 *              = 3   - "the crystal cannot absorb the soul because the mob level is too low";
	 *              = 4   - "the crystal successfuly absorbed the soul";
	 */
	public void addAbsorber(final L2Player attacker, final int crystalId)
	{
		// The attacker must not be null
		if(attacker == null)
			return;

		// This L2NpcInstance must be of one type in the _absorbingMOBS_levelXX tables.
		// OBS: This is done so to avoid triggering the absorbed conditions for mobs that can't be absorbed.
		boolean validTarget = false;
		for(final int mobId : _absorbingMOBS_level4)
			if(getNpcId() == mobId)
			{
				validTarget = true;
				break;
			}
		if(!validTarget)
			for(final int mobId : _absorbingMOBS_level8)
				if(getNpcId() == mobId)
				{
					validTarget = true;
					break;
				}
		if(!validTarget)
			for(final int mobId : _absorbingMOBS_level10)
				if(getNpcId() == mobId)
				{
					validTarget = true;
					break;
				}
		if(!validTarget)
			for(final int mobId : _absorbingMOBS_level12)
				if(getNpcId() == mobId)
				{
					validTarget = true;
					break;
				}
		if(!validTarget)
			for(final int mobId : _absorbingMOBS_level13)
				if(getNpcId() == mobId)
				{
					validTarget = true;
					break;
				}
		if(!validTarget)
			return;

		// If we have no _absorbersList initiated, do it
		AbsorberInfo ai = null;
		if(_absorbersList == null)
			_absorbersList = new ConcurrentHashMap<L2Player, AbsorberInfo>();
		else
			ai = _absorbersList.get(attacker);

		// If the L2Character attacker isn't already in the _absorbersList of this L2NpcInstance, add it
		if(ai == null)
		{
			ai = new AbsorberInfo(attacker, crystalId, getCurrentHp());
			_absorbersList.put(attacker, ai);
		}
		else
		{
			ai._absorber = attacker;
			ai._crystalId = crystalId;
			ai._absorbedHP = getCurrentHp();
		}

		absorbSoul();
	}

	/**
	 * Calculate the leveling chance of Soul Crystals based on the attacker that killed this L2NpcInstance
	 *
	 * @param attacker The player that last hitted (killed) this L2NpcInstance
	 *
	 */
	private void levelSoulCrystals(final L2Character attacker)
	{
		// Only L2Player can absorb a soul
		if(attacker == null || !attacker.isPlayer)
		{
			_absorbed = false;
			_absorbersList = null;
			return;
		}

		// Init some useful vars
		boolean success = false;
		boolean levelPartyCrystals = false;
		final int thisMobId = getNpcId();
		final L2Player killer = (L2Player) attacker;

		// Check if this L2NpcInstance isn't within any of the groups of mobs that can be absorbed
		int minCrystalLevel = 0;
		int maxCrystalLevel = 0;
		for(final int mobId : _absorbingMOBS_level4)
			if(thisMobId == mobId)
			{
				maxCrystalLevel = 4;
				success = true;
				break;
			}
		if(!success)
			for(final int mobId : _absorbingMOBS_level8)
				if(thisMobId == mobId)
				{
					maxCrystalLevel = 8;
					success = true;
					break;
				}
		if(!success)
			for(final int mobId : _absorbingMOBS_level10)
				if(thisMobId == mobId)
				{
					maxCrystalLevel = 10;
					success = true;
					break;
				}
		if(!success)
			for(final int mobId : _absorbingMOBS_level12)
				if(thisMobId == mobId)
				{
					minCrystalLevel = 10;
					maxCrystalLevel = 12;
					success = true;
					levelPartyCrystals = true;
					break;
				}
		if(!success)
			for(final int mobId : _absorbingMOBS_level13)
				if(thisMobId == mobId)
				{
					minCrystalLevel = 12;
					maxCrystalLevel = 13;
					success = true;
					levelPartyCrystals = true;
					break;
				}

		// If this is not a valid L2NpcInstance, clears the _absorbersList and just return
		if(!success)
		{
			_absorbed = false;
			_absorbersList = null;
			return;
		}

		// If this mob is a boss, then skip some checkings
		success = true;
		if(!levelPartyCrystals)
		{
			// Fail if this L2NpcInstance isn't absorbed or there's no one in its _absorbersList
			if(!_absorbed || _absorbersList == null)
			{
				_absorbed = false;
				_absorbersList = null;
				return;
			}

			// Fail if the killer isn't in the _absorbersList of this L2NpcInstance and mob is not boss
			final AbsorberInfo ai = _absorbersList.get(killer);
			if(ai == null || ai._absorber.getObjectId() != killer.getObjectId())
				success = false;

			// Check if the soul crystal was used when HP of this L2NpcInstance wasn't higher than half of it
			if(ai != null && ai._absorbedHP > getMaxHp() / 2)
				success = false;

			if(!success)
			{
				_absorbed = false;
				_absorbersList = null;
				return;
			}
		}

		_absorbed = false;
		_absorbersList = null;

		// Init some more info now
		final float dice = Rnd.get();

		// Now we got four choices:
		// 1- The Monster level is too low for the crystal. Nothing happens.
		// 2- Everything is correct, but it failed. Nothing happens. (57.5%)
		// 3- Everything is correct, but it failed. The crystal scatters. A sound event is played. (10%)
		// 4- Everything is correct, the crystal level up. A sound event is played. (32.5%)

		ArrayList<L2Player> players;
		if(levelPartyCrystals && killer.isInParty())
			players = killer.getParty().getPartyMembers();
		else
			(players = new ArrayList<L2Player>()).add(killer);

		for(final L2Player player : players)
		{
			int oldCrystalId = 0;
			int newCrystalId = 0;
			short crystalsCount = 0;
			byte crystalLevel = 0;
			boolean canIncreaseCrystal = false;
			boolean resonated = false;

			// Check how many soul crystals the player has in his inventory and which soul crystal he has
			for(final L2ItemInstance item : player.getInventory().getItems())
			{
				final int itemId = item.getItemId();
				if(!(itemId == _REDCRYSTAL_LVL11 || itemId == _BLUECRYSTAL_LVL11 || itemId == _GREENCRYSTAL_LVL11 || itemId == _REDCRYSTAL_LVL12 || itemId == _BLUECRYSTAL_LVL12 || itemId == _GREENCRYSTAL_LVL12 || itemId == _REDCRYSTAL_LVL13 || itemId == _BLUECRYSTAL_LVL13 || itemId == _GREENCRYSTAL_LVL13))
					if(!(itemId >= _REDCRYSTAL_OFFSET || itemId >= _GREENCRYSTAL_OFFSET || itemId >= _BLUECRYSTAL_OFFSET) || !(itemId <= _REDCRYSTAL_OFFSET + 10 || itemId <= _GREENCRYSTAL_OFFSET + 10 || itemId <= _BLUECRYSTAL_OFFSET + 10))
						continue;

				if(++crystalsCount > 1)
				{
					resonated = true;
					break;
				}

				int crystalOffset = 0;
				for(crystalLevel = 0; crystalLevel <= _MAX_CRYSTALS_LEVEL; crystalLevel++)
				{
					// Get a crystal offset to calculate the new crystal ID
					// If the crystal level is way too high for this mob, say that we can't increase it
					// Define crystals' IDs based on the crystalOffset found

					crystalOffset = getNextLevelCrystalId(itemId, crystalLevel);
					if(crystalOffset == 0)
						continue;

					canIncreaseCrystal = crystalLevel >= minCrystalLevel && crystalLevel < maxCrystalLevel;
					oldCrystalId = itemId;
					newCrystalId = crystalOffset;
					break;
				}
			}

			if(crystalsCount < 1)
				continue;

			// The player has more than one soul crystal with him, the crystal resonates and we skip this player
			if(resonated)
			{
				// player.sendPacket(new PlaySound("ItemSound2.broken_key"));
				player.sendPacket(new SystemMessage(SystemMessage.THE_SOUL_CRYSTALS_CAUSED_RESONATION_AND_FAILED_AT_ABSORBING_A_SOUL));
				continue;
			}

			// The soul crystal stage of the player is way too high, refuse to increase it
			if(!canIncreaseCrystal)
			{
				player.sendPacket(new SystemMessage(SystemMessage.THE_SOUL_CRYSTAL_IS_REFUSING_TO_ABSORB_A_SOUL));
				continue;
			}

			// If the killer succeeds or it is a boss mob, level up the crystal
			if(levelPartyCrystals || dice < _LEVEL_CHANCE)
			{
				L2ItemInstance oldCrystal = player.getInventory().getItemByItemId(oldCrystalId);
				if(oldCrystal == null)
					continue;

				L2ItemInstance newCrystal = ItemTable.getInstance().createItem(newCrystalId, player.getObjectId(), getNpcId(), "L2NpcInstance.levelSoulCrystals");
				player.getInventory().addItem(newCrystal);
				player.getInventory().destroyItem(oldCrystal, 1, true);
				// player.sendPacket(new PlaySound("ItemSound3.sys_enchant_success"));
				player.sendPacket(new SystemMessage(SystemMessage.EARNED_S1).addItemName(newCrystalId));
				player.sendPacket(new SystemMessage(SystemMessage.THE_SOUL_CRYSTAL_SUCCEEDED_IN_ABSORBING_A_SOUL));
				continue;
			}

			player.sendPacket(new SystemMessage(SystemMessage.THE_SOUL_CRYSTAL_IS_REFUSING_TO_ABSORB_A_SOUL));
		}
	}

	private int getNextLevelCrystalId(int crystalId, final int crystalLevel)
	{
		switch(crystalLevel)
		{
			case 10:
				if(crystalId == _REDCRYSTAL_OFFSET + 10)
					return _REDCRYSTAL_LVL11;
				if(crystalId == _GREENCRYSTAL_OFFSET + 10)
					return _GREENCRYSTAL_LVL11;
				if(crystalId == _BLUECRYSTAL_OFFSET + 10)
					return _BLUECRYSTAL_LVL11;
				break;
			case 11:
				if(crystalId == _REDCRYSTAL_LVL11)
					return _REDCRYSTAL_LVL12;
				if(crystalId == _GREENCRYSTAL_LVL11)
					return _GREENCRYSTAL_LVL12;
				if(crystalId == _BLUECRYSTAL_LVL11)
					return _BLUECRYSTAL_LVL12;
				break;
			case 12:
				if(crystalId == _REDCRYSTAL_LVL12)
					return _REDCRYSTAL_LVL13;
				if(crystalId == _GREENCRYSTAL_LVL12)
					return _GREENCRYSTAL_LVL13;
				if(crystalId == _BLUECRYSTAL_LVL12)
					return _BLUECRYSTAL_LVL13;
				break;
			case 13:
				if(crystalId == _REDCRYSTAL_LVL13)
					return _REDCRYSTAL_LVL14;
				if(crystalId == _GREENCRYSTAL_LVL13)
					return _GREENCRYSTAL_LVL14;
				if(crystalId == _BLUECRYSTAL_LVL13)
					return _BLUECRYSTAL_LVL14;
				break;
		}
		if(crystalId == _REDCRYSTAL_OFFSET + crystalLevel || crystalId == _GREENCRYSTAL_OFFSET + crystalLevel || crystalId == _BLUECRYSTAL_OFFSET + crystalLevel)
			return ++crystalId;

		return 0;
	}

	public L2ItemInstance[] takeHarvest()
	{
		synchronized (_harvestLock)
		{
			final L2ItemInstance[] harvest = _harvestItems;
			_harvestItems = null;
			_seeded = null;
			return harvest;
		}
	}

	public void setSeeded(final short id, L2Player player)
	{
		if(player == null)
			return;

		_seeded = ItemTable.getInstance().getTemplate(id);
		_seeder = player;
		// Количество всходов зависит от множителя HP конкретного моба
		int count = (int) getTemplate().rateHp;
		// Для мобов с множителем HP < 1 число всходов = 1
		if(count == 0)
			count = 1;
		final List<L2ItemInstance> harvested = new FastList<L2ItemInstance>();
		final L2ItemInstance crops = ItemTable.getInstance().createItem(L2Manor.getInstance().getCropType(id), 0, getNpcId(), "L2NpcInstance.setSeeded");
		// Количество всходов от xHP до (xHP + xHP/2)
		if(count > 1)
			crops.setCount(Rnd.get(count * Config.RATE_MANOR, (int) Math.round(1.5 * count * Config.RATE_MANOR)));
		else
			crops.setCount(Config.RATE_MANOR);
		harvested.add(crops);

		if(harvested.size() > 0)
			_harvestItems = harvested.toArray(new L2ItemInstance[harvested.size()]);
	}

	public boolean isSeeded(L2Player seeder)
	{
		return _seeder != null && seeder == _seeder;
	}

	public boolean isSeeded()
	{
		return _seeded != null;
	}

	/** True if a Dwarf has used Spoil on this L2NpcInstance */
	private boolean _isSpoiled;

	/**
	 * Return True if this L2NpcInstance has drops that can be sweeped.<BR><BR>
	 */
	public boolean isSpoiled()
	{
		return _isSpoiled;
	}

	/**
	 * Set the spoil state of this L2NpcInstance.<BR><BR>
	 */
	public void setSpoiled(boolean isSpoiled)
	{
		_isSpoiled = isSpoiled;
	}

	public void doItemDrop(L2Character lastAttacker)
	{
		L2Character mostHated = getMostHated();
		if(mostHated != null)
			lastAttacker = mostHated;

		L2Player player = lastAttacker.getPlayer();

		if(player == null)
			return;

		int levelDiff = calculateLevelDiffForDrop();

		if(getTemplate().getDropData() != null)
		{
			FastList<ItemToDrop> drops = getTemplate().getDropData().rollDrop(levelDiff, isRaid, player);
			for(ItemToDrop drop : drops)
			{
				// Если в моба посеяно семя, причем не альтернативное - не давать никакого дропа, кроме адены.
				if(_seeded != null && !_seeded.isAltSeed() && drop.itemId != 57 && drop.itemId != 6360 && drop.itemId != 6361 && drop.itemId != 6362)
					continue;
				dropItem(player, drop.itemId, drop.count);
			}
		}

		if(_inventory != null)
		{
			for(L2ItemInstance drop : _inventory)
				if(drop != null)
				{
					player.sendMessage(new CustomMessage("l2r.gameserver.model.instances.L2MonsterInstance.ItemBelongedToOther", player).addString(drop.getItem().getName()));
					dropItem(player, drop);
				}
			truncateInventory();
		}
	}

	private void doSweepDrop(final L2Character lastAttacker)
	{
		final L2Player player = lastAttacker.getPlayer();

		if(player == null)
			return;

		final int levelDiff = calculateLevelDiffForDrop();

		final ArrayList<L2ItemInstance> spoiled = new ArrayList<L2ItemInstance>();

		if(getTemplate().getDropData() != null)
		{
			final FastList<ItemToDrop> spoils = getTemplate().getDropData().rollSpoil(levelDiff, isRaid, player);
			for(final ItemToDrop spoil : spoils)
			{
				final L2ItemInstance dropit = ItemTable.getInstance().createItem(spoil.itemId, lastAttacker.getObjectId(), getNpcId(), "L2NpcInstance.doSweepDrop");
				dropit.setCount(spoil.count);
				spoiled.add(dropit);
			}
		}

		if(spoiled.size() > 0)
			_sweepItems = spoiled.toArray(new L2ItemInstance[spoiled.size()]);
	}

	private long[] calculateExpAndSp(long diff, long damage)
	{
		long xp = getExpReward() * damage / getMaxHp();
		long sp = getSpReward() * damage / getMaxHp();

		if(diff > 5)
		{
			double mod = Math.pow(.83, diff - 5);
			xp *= mod;
			sp *= mod;
		}

		if(xp < 0)
			xp = 0;
		if(sp < 0)
			sp = 0;

		return new long[] { xp, sp };
	}

	@Override
	public void setOverhitEnabled(final boolean status)
	{
		_overhit = status;
	}

	@Override
	public void setOverhitValues(L2Character attacker, double damage)
	{
		// Calculate the over-hit damage
		// Ex: mob had 10 HP left, over-hit skill did 50 damage total, over-hit damage is 40
		double overhitDmg = (getCurrentHp() - damage) * -1;
		if(overhitDmg < 0)
		{
			// we didn't killed the mob with the over-hit strike. (it wasn't really an over-hit strike)
			// let's just clear all the over-hit related values
			setOverhitEnabled(false);
			_overhitDamage = 0;
			setOverhitAttacker(null);
			return;
		}
		setOverhitEnabled(true);
		_overhitDamage = overhitDmg;
		setOverhitAttacker(attacker);
	}

	public L2Character getOverhitAttacker()
	{
		if(overhitAttacker == null)
			return null;

		L2Character c = overhitAttacker.get();
		if(c == null)
			overhitAttacker = null;

		return c;
	}

	public double getOverhitDamage()
	{
		return _overhitDamage;
	}

	@Override
	public boolean isOverhit()
	{
		return _overhit;
	}

	public int calculateOverhitExp(final long normalExp)
	{
		double overhitPercentage = getOverhitDamage() * 100 / getMaxHp();
		if(overhitPercentage > 25)
			overhitPercentage = 25;
		double overhitExp = overhitPercentage / 100 * normalExp;
		return (int) Math.round(overhitExp);
	}

	/**
	 * Return True if a Dwarf use Sweep on the L2NpcInstance and if item can be spoiled.<BR><BR>
	 */
	public boolean isSweepActive()
	{
		return _sweepItems != null && _sweepItems.length > 0;
	}

	/**
	 * Return table containing all L2ItemInstance that can be spoiled.<BR><BR>
	 */
	public L2ItemInstance[] takeSweep()
	{
		synchronized (_sweepLock)
		{
			final L2ItemInstance[] sweep = _sweepItems.clone();
			_sweepItems = null;
			return sweep;
		}
	}

	@Override
	public boolean isInvul()
	{
		return false;
	}
}