package l2r.gameserver.model.instances;

import java.util.concurrent.ScheduledFuture;
import java.util.logging.Level;
import java.util.logging.Logger;

import l2r.gameserver.ThreadPoolManager;
import l2r.gameserver.idfactory.IdFactory;
import l2r.gameserver.instancemanager.QuestManager;
import l2r.gameserver.instancemanager.RaidBossSpawnManager;
import l2r.gameserver.model.L2Character;
import l2r.gameserver.model.L2Playable;
import l2r.gameserver.model.L2Player;
import l2r.gameserver.model.quest.Quest;
import l2r.gameserver.model.quest.QuestState;
import l2r.gameserver.serverpackets.SystemMessage;
import l2r.gameserver.tables.NpcTable;
import l2r.gameserver.templates.L2NpcTemplate;
import l2r.util.MinionList;
import l2r.util.Rnd;

/**
 * This class manages all RaidBoss.
 * In a group mob, there are one master called RaidBoss and several slaves called Minions.
 */
public class L2RaidBossInstance extends L2MonsterInstance
{
	protected static Logger _log = Logger.getLogger(L2RaidBossInstance.class.getName());

	@SuppressWarnings( { "unused", "unchecked" })
	private ScheduledFuture minionUnspawnTask;

	private static final int RAIDBOSS_MAINTENANCE_INTERVAL = 60000;
	private static final int MINION_UNSPAWN_INTERVAL = 5000; //time to unspawn minions when boss is dead, msec

	private RaidBossSpawnManager.StatusEnum _raidStatus;

	/**
	 * Constructor of L2RaidBossInstance (use L2Character and L2NpcInstance constructor).<BR><BR>
	 *
	 * <B><U> Actions</U> :</B><BR><BR>
	 * <li>Call the L2Character constructor to set the _template of the L2RaidBossInstance (copy skills from template to object and link _calculators to NPC_STD_CALCULATOR) </li>
	 * <li>Set the name of the L2RaidBossInstance</li>
	 * <li>Create a RandomAnimation Task that will be launched after the calculated delay if the server allow it </li><BR><BR>
	 *
	 * @param objectId Identifier of the object to initialized
	 * @param L2NpcTemplate Template to apply to the NPC
	 */
	public L2RaidBossInstance(int objectId, L2NpcTemplate template)
	{
		super(objectId, template);
	}

	@Override
	protected int getMaintenanceInterval()
	{
		return RAIDBOSS_MAINTENANCE_INTERVAL;
	}

	protected int getMinionUnspawnInterval()
	{
		return MINION_UNSPAWN_INTERVAL;
	}

	protected int getKilledInterval()
	{
		return Rnd.get(240000, 360000); //4-6 minutes to respawn
	}

	@Override
	public void notifyMinionDied(L2MinionInstance minion)
	{
		ThreadPoolManager.getInstance().scheduleGeneral(new maintainKilledMinion(minionList, minion), getKilledInterval());
		super.notifyMinionDied(minion);
	}

	static class maintainKilledMinion implements Runnable
	{
		private MinionList _minionList;
		private L2MinionInstance _minion;

		maintainKilledMinion(MinionList minionList, L2MinionInstance minion)
		{
			_minionList = minionList;
			_minion = minion;
		}

		public void run()
		{
			try
			{
				//check leader
				if(_minion.getLeader() != null && !_minion.getLeader().isDead())
					_minionList.spawnSingleMinionSync(_minion.getNpcId());
			}
			catch(Throwable e)
			{
				_log.log(Level.SEVERE, "", e);
				e.printStackTrace();
			}
		}
	}

	@Override
	public void reduceCurrentHp(double i, L2Character attacker, boolean awake, boolean standUp, boolean directHp)
	{
		if(attacker == null || attacker.getPlayer() == null || (attacker == this || i > getMaxHp() / 10) && !attacker.getPlayer().isGM())
			return;
		super.reduceCurrentHp(i, attacker, awake, standUp, directHp);
	}

	@Override
	public void doDie(L2Character killer)
	{
		L2Player player;

		if(killer instanceof L2Playable)
		{
			player = killer.getPlayer();
			if(player.getParty() != null)
				player.getParty().broadcastToPartyMembers(new SystemMessage(SystemMessage.CONGRATULATIONS_YOUR_RAID_WAS_SUCCESSFUL));
			else
				player.sendPacket(new SystemMessage(SystemMessage.CONGRATULATIONS_YOUR_RAID_WAS_SUCCESSFUL));

			Quest q = QuestManager.getQuest(508);
			if(q != null)
			{
				String qn = q.getName();
				if(player.getClan() != null && player.getClan().getLeader().isOnline() && player.getClan().getLeader().getPlayer().getQuestState(qn) != null)
				{
					QuestState st = player.getClan().getLeader().getPlayer().getQuestState(qn);
					st.getQuest().onKill(this, st);
				}
			}
		}

		RaidBossSpawnManager.getInstance().updateStatus(this, true);

		if(hasMinions())
			minionUnspawnTask = ThreadPoolManager.getInstance().scheduleGeneral(new Runnable(){
				public void run()
				{
					try
					{
						removeMinions();
					}
					catch(Throwable e)
					{
						_log.log(Level.SEVERE, "", e);
						e.printStackTrace();
					}
				}
			}, getMinionUnspawnInterval());

		int boxId = 0;
		switch(getNpcId())
		{
			case 25035: // Shilens Messenger Cabrio
				boxId = 31027;
				break;
			case 25054: // Demon Kernon
				boxId = 31028;
				break;
			case 25126: // Golkonda, the Longhorn General
				boxId = 31029;
				break;
			case 25220: // Death Lord Hallate
				boxId = 31030;
				break;
		}

		if(boxId != 0)
		{
			L2NpcTemplate boxTemplate = NpcTable.getTemplate(boxId);
			if(boxTemplate != null)
			{
				final L2NpcInstance box = new L2NpcInstance(IdFactory.getInstance().getNextId(), boxTemplate);
				box.spawnMe(getLoc());

				ThreadPoolManager.getInstance().scheduleGeneral(new Runnable(){
					public void run()
					{
						box.deleteMe();
					}
				}, 60000);
			}
		}

		super.doDie(killer);
	}

	@Override
	public void onSpawn()
	{
		RaidBossSpawnManager.getInstance().updateStatus(this, false);
		super.onSpawn();

		if(getNpcId() == 25286 || getNpcId() == 25283)
			return;
		getSpawn().stopRespawn();
	}

	public void setRaidStatus(RaidBossSpawnManager.StatusEnum status)
	{
		_raidStatus = status;
	}

	public RaidBossSpawnManager.StatusEnum getRaidStatus()
	{
		return _raidStatus;
	}

	@Override
	public boolean isFearImmune()
	{
		return true;
	}

	@Override
	public boolean paralyzeImmune()
	{
		return true;
	}

	@Override
	public boolean isLethalImmune()
	{
		return true;
	}

	@Override
	public boolean hasRandomWalk()
	{
		return false;
	}
}