package l2r.gameserver.model.quest;

import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import javolution.util.FastList;
import javolution.util.FastMap;
import l2r.Config;
import l2r.gameserver.model.L2Character;
import l2r.gameserver.model.L2Object;
import l2r.gameserver.model.L2Party;
import l2r.gameserver.model.L2Player;
import l2r.gameserver.model.L2World;
import l2r.gameserver.model.base.Experience;
import l2r.gameserver.model.instances.L2ItemInstance;
import l2r.gameserver.model.instances.L2NpcInstance;
import l2r.gameserver.serverpackets.ExShowQuestMark;
import l2r.gameserver.serverpackets.NpcSay;
import l2r.gameserver.serverpackets.PlaySound;
import l2r.gameserver.serverpackets.QuestList;
import l2r.gameserver.serverpackets.StatusUpdate;
import l2r.gameserver.serverpackets.SystemMessage;
import l2r.gameserver.serverpackets.TutorialEnableClientEvent;
import l2r.gameserver.serverpackets.TutorialShowHtml;
import l2r.gameserver.serverpackets.TutorialShowQuestionMark;
import l2r.gameserver.tables.ItemTable;
import l2r.gameserver.templates.L2Item;
import l2r.util.Files;
import l2r.util.Log;
import l2r.util.Rnd;

public final class QuestState
{
	protected static Logger _log = Logger.getLogger(Quest.class.getName());

	/** Player who engaged the quest */
	private L2Player _player;

	/** Quest associated to the QuestState */
	private Quest _quest;

	/** State of the quest */
	private State _state;

	/** List of couples (variable for quest,value of the variable for quest) */
	private Map<String, String> _vars;

	/**
	 * Constructor of the QuestState : save the quest in the list of quests of the player.<BR/><BR/>
	 *
	 * <U><I>Actions :</U></I><BR/>
	 * <LI>Save informations in the object QuestState created (Quest, Player, Completion, State)</LI>
	 * <LI>Add the QuestState in the player's list of quests by using setQuestState()</LI>
	 * <LI>Add drops gotten by the quest</LI>
	 * <BR/>
	 * @param quest : quest associated with the QuestState
	 * @param player : L2Player pointing out the player
	 * @param state : state of the quest
	 */
	QuestState(Quest quest, L2Player player, State state)
	{
		_quest = quest;
		_player = player;

		// Save the state of the quest for the player in the player's list of quest onwed
		_player.setQuestState(this);

		// set the state of the quest
		_state = state;
	}

	/**
	 * Add XP and SP as quest reward
	 * <br><br>
	 * Метод учитывает рейты!
	 */
	public void addExpAndSp(long exp, long sp)
	{
		addExpAndSp(exp, sp, false);
	}

	/**
	 * Add XP and SP as quest reward
	 * <br><br>
	 * Метод учитывает рейты!
	 * 3-ий параметр true/false показывает является ли квест на профессию 
	 * и рейты учитываются в завимисомти от параметра RateQuestsRewardOccupationChange
	 */
	public void addExpAndSp(long exp, long sp, boolean rate)
	{
		if(!rate || rate && Config.RATE_QUESTS_OCCUPATION_CHANGE)
			_player.addExpAndSp((long) (exp * getRateQuestsReward()), (long) (sp * getRateQuestsReward()), true, false);
		else
			_player.addExpAndSp(exp, sp, true, false);
	}

	/**
	 * Add player to get notification of characters death
	 * @param character : L2Character of the character to get notification of death
	 */
	public void addNotifyOfDeath(L2Character character)
	{
		if(character == null)
			return;
		character.addNotifyQuestOfDeath(this);
	}

	public void addRadar(int x, int y, int z)
	{
		_player.radar.addMarker(x, y, z);
	}

	public void clearRadar()
	{
		_player.radar.removeAllMarkers();
	}

	/**
	 * Destroy element used by quest when quest is exited
	 * @param repeatable
	 * @return QuestState
	 */
	public QuestState exitCurrentQuest(boolean repeatable)
	{
		// Clean drops
		if(_quest.getItems() != null)
			// Go through values of class variable "drops" pointing out mobs that drop for quest
			for(Integer itemId : _quest.getItems())
			{
				// Get [item from] / [presence of the item in] the inventory of the player
				L2ItemInstance item = _player.getInventory().getItemByItemId(itemId);
				if(item == null || itemId == 57)
					continue;
				long count = item.getCount();
				// If player has the item in inventory, destroy it (if not gold)
				_player.getInventory().destroyItemByItemId(itemId, count, true);
				_player.getWarehouse().destroyItem(itemId, count);
			}

		// If quest is repeatable, delete quest from list of quest of the player and from database (quest CAN be created again => repeatable)
		if(repeatable)
		{
			_player.delQuestState(_quest.getName());
			Quest.deleteQuestInDb(this);
			_vars = null;
		}
		else
		{ // Otherwise, delete variables for quest and update database (quest CANNOT be created again => not repeatable)
			if(_vars != null)
				for(String var : _vars.keySet())
					unset(var);
			setState(_quest.COMPLETED);
			Quest.updateQuestInDb(this);
		}
		_player.sendPacket(new QuestList(_player));
		return this;
	}

	/**
	 * Return the value of the variable of quest represented by "var"
	 * @param var : name of the variable of quest
	 * @return Object
	 */
	public Object get(String var)
	{
		if(_vars == null)
			return null;
		return _vars.get(var);
	}

	/**
	 * Return the value of the variable of quest represented by "var"
	 * @param var : String designating the variable for the quest
	 * @return int
	 */
	public int getInt(String var)
	{
		int varint = 0;
		try
		{
			varint = Integer.parseInt(_vars.get(var));
		}
		catch(Exception e)
		{
			_log.finer(getPlayer().getName() + ": variable " + var + " isn't an integer: " + varint + e);
		}
		return varint;
	}

	/**
	 * Return item number which is equipped in selected slot
	 * @return int
	 */
	public int getItemEquipped(int loc)
	{
		return getPlayer().getInventory().getPaperdollItemId(loc);
	}

	/**
	 * NOTE: This is to be deprecated; replaced by
	 * Quest.getPcSpawn(L2Player) For now, I shall leave it as is Return a
	 * QuestPcSpawn for current player instance
	 */
	//TODO снести
	public final QuestPcSpawn getPcSpawn()
	{
		return QuestPcSpawnManager.getInstance().getPcSpawn(getPlayer());
	}

	/**
	 * Return the L2Player
	 * @return L2Player
	 */
	public L2Player getPlayer()
	{
		return _player;
	}

	/**
	 * Return the quest
	 * @return Quest
	 */
	public Quest getQuest()
	{
		return _quest;
	}

	/**
	 * Return the quantity of one sort of item hold by the player
	 * @param itemId : ID of the item wanted to be count
	 * @return int
	 */
	public long getQuestItemsCount(int itemId)
	{
		return _player.getInventory().getCountOf(itemId);
	}

	/**
	 * Return the QuestTimer object with the specified name
	 * @return QuestTimer<BR> Return null if name does not exist
	 */
	public final QuestTimer getQuestTimer(String name)
	{
		return getQuest().getQuestTimer(name, null, getPlayer());

	}

	/**
	 * Возвращает случайное значение<BR>
	 * Будьте внимательны, этот метод возвращает число от 0 до max.<br>
	 * Где 0 включительно, а max <b>исключительно</b>.<br>
	 * Этот метод устаревший, используйте Rnd.chance() или Rnd.get()
	 * @param max : max value for randomisation
	 * @return int
	 */
	@Deprecated
	public int getRandom(int max)
	{
		return Rnd.get(max);
	}

	/**
	 * Return the state of the quest
	 * @return State
	 */
	public State getState()
	{
		return _state;
	}

	/**
	 * Return ID of the state of the quest
	 * @return String
	 */
	public String getStateId()
	{
		return _state.getName();
	}

	/**
	 * Добавить предмет игроку
	 * By default if item is adena rates 'll be applyed, else no
	 * @param itemId
	 * @param count
	 */
	public void giveItems(int itemId, long count)
	{
		if(itemId == 57)
			giveItems(itemId, count, 0, true);
		else
			giveItems(itemId, count, 0, false);
	}

	/**
	 * Добавить предмет игроку
	 * @param itemId
	 * @param count
	 * @param rate - учет квестовых рейтов
	 */
	public void giveItems(int itemId, long count, boolean rate)
	{
		giveItems(itemId, count, 0, rate);
	}

	/**
	 * Добавить предмет игроку
	 * @param itemId
	 * @param count
	 * @param enchantlevel
	 * @param rate - учет квестовых рейтов
	 */
	public void giveItems(int itemId, long count, int enchantlevel, boolean rate)
	{
		if(count <= 0)
			count = 1;

		if(rate)
			count = count * (long) getRateQuestsReward();

		// Get template of item
		L2Item template = ItemTable.getInstance().getTemplate(itemId);
		if(template == null)
			return;

		if(template.isStackable())
		{
			L2ItemInstance item = ItemTable.getInstance().createItem(itemId, _player.getObjectId(), 0, "QuestState.giveItems");

			// Set quantity of item
			item.setCount(count);

			// Add items to player's inventory
			_player.getInventory().addItem(item);

			if(enchantlevel > 0)
				item.setEnchantLevel(enchantlevel);

			Log.LogItem(_player, Log.GetQuestItem, item);
		}
		else
			for(int i = 0; i < count; i++)
			{
				L2ItemInstance item = ItemTable.getInstance().createItem(itemId, _player.getObjectId(), 0, "QuestState.giveItems");

				// Set quantity of item
				item.setCount(1);

				// Add items to player's inventory
				_player.getInventory().addItem(item);

				if(enchantlevel > 0)
					item.setEnchantLevel(enchantlevel);

				Log.LogItem(_player, Log.GetQuestItem, item);
			}

		// If item for reward is gold, send message of gold reward to client
		if(template.getItemId() == 57)
			_player.sendPacket(new SystemMessage(SystemMessage.EARNED_S1_ADENA).addNumber((int) count));
		else if(count > 1)
		{
			SystemMessage smsg = new SystemMessage(SystemMessage.EARNED_S2_S1_s);
			smsg.addItemName(template.getItemId());
			smsg.addNumber((int) count);
			_player.sendPacket(smsg);
		}
		else
		{
			SystemMessage smsg = new SystemMessage(SystemMessage.EARNED_S1);
			smsg.addItemName(template.getItemId());
			_player.sendPacket(smsg);
		}

		StatusUpdate su = new StatusUpdate(_player.getObjectId());
		su.addAttribute(StatusUpdate.CUR_LOAD, _player.getCurrentLoad());
		_player.sendPacket(su);
	}

	/**
	 * Этот метод рассчитывает количество дропнутых вещей в зависимости от рейтов.
	 * <br><br>
	 * Следует учесть, что контроль за верхним пределом вещей в квестах, в которых
	 * нужно набить определенное количество предметов не осуществляется.
	 * <br><br>
	 * Ни один из передаваемых параметров не должен быть равен 0
	 *
	 * @param count количество при рейтах 1х
	 * @param calcChance шанс при рейтах 1х, в процентах
	 * @return количество вещей для дропа, может быть 0
	 */
	public int rollDrop(int count, double calcChance, boolean prof)
	{
		if(calcChance <= 0 || count <= 0)
			return 0;
		return rollDrop(count, count, calcChance, prof);
	}

	/**
	 * Этот метод рассчитывает количество дропнутых вещей в зависимости от рейтов.
	 * <br><br>
	 * Следует учесть, что контроль за верхним пределом вещей в квестах, в которых
	 * нужно набить определенное количество предметов не осуществляется.
	 * <br><br>
	 * Ни один из передаваемых параметров не должен быть равен 0
	 *
	 * @param min минимальное количество при рейтах 1х
	 * @param max максимальное количество при рейтах 1х
	 * @param calcChance шанс при рейтах 1х, в процентах
	 * @param prof - учитывать дроп по параметру "рейт дропа для квестов на профу"
	 * @return количество вещей для дропа, может быть 0
	 */
	public int rollDrop(int min, int max, double calcChance, boolean prof)
	{
		if(calcChance <= 0 || min <= 0 || max <= 0)
			return 0;
		int dropmult = 1;
		calcChance *= getRateQuestsDrop(prof);
		if(calcChance > 100)
		{
			if((int) Math.ceil(calcChance / 100) <= calcChance / 100)
				calcChance = Math.nextUp(calcChance);
			dropmult = (int) Math.ceil(calcChance / 100);
			calcChance = calcChance / dropmult;
		}
		return Rnd.chance(calcChance) ? Rnd.get(min * dropmult, max * dropmult) : 0;
	}

	public float getRateQuestsDrop(boolean prof)
	{
		if(prof)
			return Config.RATE_QUESTS_DROP_PROF * _player.getBonus().RATE_QUESTS_DROP;
		return Config.RATE_QUESTS_DROP * _player.getBonus().RATE_QUESTS_DROP;
	}

	public float getRateQuestsReward()
	{
		return Config.RATE_QUESTS_REWARD * _player.getBonus().RATE_QUESTS_REWARD;
	}

	/**
	 * Этот метод рассчитывает количество дропнутых вещей в зависимости от рейтов и дает их,
	 * проверяет максимум, а так же проигрывает звук получения вещи.
	 * <br><br>
	 * Ни один из передаваемых параметров не должен быть равен 0
	 *
	 * @param itemId id вещи
	 * @param min минимальное количество при рейтах 1х
	 * @param max максимальное количество при рейтах 1х
	 * @param limit максимум таких вещей
	 * @param calcChance
	 * @return true если после выполнения количество достигло лимита
	 */
	public boolean rollAndGive(int itemId, int min, int max, int limit, double calcChance)
	{
		if(calcChance <= 0 || min <= 0 || max <= 0 || limit <= 0 || itemId <= 0)
			return false;
		return rollAndGive(itemId, min, max, limit, calcChance, false);
	}

	/**
	 * Этот метод рассчитывает количество дропнутых вещей в зависимости от рейтов и дает их,
	 * проверяет максимум, а так же проигрывает звук получения вещи.
	 * <br><br>
	 * Ни один из передаваемых параметров не должен быть равен 0
	 *
	 * @param itemId id вещи
	 * @param min минимальное количество при рейтах 1х
	 * @param max максимальное количество при рейтах 1х
	 * @param limit максимум таких вещей
	 * @param calcChance
	 * @param QuestProf
	 * @return true если после выполнения количество достигло лимита
	 */
	public boolean rollAndGive(int itemId, int min, int max, int limit, double calcChance, boolean prof)
	{
		if(calcChance <= 0 || min <= 0 || max <= 0 || limit <= 0 || itemId <= 0)
			return false;
		long count = rollDrop(min, max, calcChance, prof);
		if(count > 0)
		{
			L2ItemInstance already = _player.getInventory().getItemByItemId(itemId);
			long alreadyCount = 0;
			if(already != null)
				alreadyCount = already.getCount();
			if(alreadyCount + count > limit)
				count = limit - alreadyCount;
			if(count > 0)
			{
				giveItems(itemId, count, false);
				if(count + alreadyCount < limit)
					playSound("ItemSound.quest_itemget");
				else
				{
					playSound("ItemSound.quest_middle");
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * Этот метод рассчитывает количество дропнутых вещей в зависимости от рейтов и дает их,
	 * а так же проигрывает звук получения вещи.
	 * <br><br>
	 * Следует учесть, что контроль за верхним пределом вещей в квестах, в которых
	 * нужно набить определенное количество предметов не осуществляется.
	 * <br><br>
	 * Ни один из передаваемых параметров не должен быть равен 0
	 *
	 * @param itemId id вещи
	 * @param min минимальное количество при рейтах 1х
	 * @param max максимальное количество при рейтах 1х
	 * @param calcChance
	 */
	public void rollAndGive(int itemId, int min, int max, double calcChance)
	{
		if(calcChance <= 0 || min <= 0 || max <= 0 || itemId <= 0)
			return;
		rollAndGive(itemId, min, max, calcChance, false);
	}

	/**
	 * Этот метод рассчитывает количество дропнутых вещей в зависимости от рейтов и дает их,
	 * а так же проигрывает звук получения вещи.
	 * <br><br>
	 * Следует учесть, что контроль за верхним пределом вещей в квестах, в которых
	 * нужно набить определенное количество предметов не осуществляется.
	 * <br><br>
	 * Ни один из передаваемых параметров не должен быть равен 0
	 *
	 * @param itemId id вещи
	 * @param min минимальное количество при рейтах 1х
	 * @param max максимальное количество при рейтах 1х
	 * @param prof - учитывать дроп по параметру "рейт дропа для квестов на профу"
	 * @param calcChance
	 */
	public void rollAndGive(int itemId, int min, int max, double calcChance, boolean prof)
	{
		if(calcChance <= 0 || min <= 0 || max <= 0 || itemId <= 0)
			return;
		int count = rollDrop(min, max, calcChance, prof);
		if(count > 0)
		{
			giveItems(itemId, count, false);
			playSound("ItemSound.quest_itemget");
		}
	}

	/**
	 * Этот метод рассчитывает количество дропнутых вещей в зависимости от рейтов и дает их,
	 * а так же проигрывает звук получения вещи.
	 * <br><br>
	 * Следует учесть, что контроль за верхним пределом вещей в квестах, в которых
	 * нужно набить определенное количество предметов не осуществляется.
	 * <br><br>
	 * Ни один из передаваемых параметров не должен быть равен 0
	 *
	 * @param itemId id вещи
	 * @param count количество при рейтах 1х
	 * @param calcChance
	 */
	public boolean rollAndGive(int itemId, int count, double calcChance)
	{
		if(calcChance <= 0 || count <= 0 || itemId <= 0)
			return false;
		return rollAndGive(itemId, count, calcChance, false);
	}

	/**
	 * Этот метод рассчитывает количество дропнутых вещей в зависимости от рейтов и дает их,
	 * а так же проигрывает звук получения вещи.
	 * <br><br>
	 * Следует учесть, что контроль за верхним пределом вещей в квестах, в которых
	 * нужно набить определенное количество предметов не осуществляется.
	 * <br><br>
	 * Ни один из передаваемых параметров не должен быть равен 0
	 *
	 * @param itemId id вещи
	 * @param count количество при рейтах 1х
	 * @param calcChance
	 * @param prof - учитывать дроп по параметру "рейт дропа для квестов на профу"
	 */
	public boolean rollAndGive(int itemId, int count, double calcChance, boolean prof)
	{
		if(calcChance <= 0 || count <= 0 || itemId <= 0)
			return false;
		int countToDrop = rollDrop(count, calcChance, prof);
		if(countToDrop > 0)
		{
			giveItems(itemId, countToDrop, false);
			playSound("ItemSound.quest_itemget");
			return true;
		}
		return false;
	}

	/**
	 * Return true if quest completed, false otherwise
	 * @return boolean
	 */
	public boolean isCompleted()
	{
		return getStateId().equalsIgnoreCase("Completed");
	}

	/**
	 * Return true if quest started, false otherwise
	 * @return boolean
	 */
	public boolean isStarted()
	{
		return !(getStateId().equals("Start") || getStateId().equals("Completed"));
	}

	public void killNpcByObjectId(int _objId)
	{
		L2Object obj = L2World.findObject(_objId);
		if(obj instanceof L2NpcInstance)
			((L2NpcInstance) obj).doDie(null);
		else
			_log.warning("Attemp to kill object that is not npc in quest " + getQuest().getQuestIntId());
	}

	public void monsterSay(L2NpcInstance npc, int chatType, String text)
	{
		NpcSay cs = null;
		if(npc == null)
			return;
		cs = new NpcSay(npc, chatType, text);
		List<L2Player> nearbyPlayers = new FastList<L2Player>();
		switch(chatType)
		{
			case 0:
				for(L2Player player : npc.getAroundPlayers(1500))
					nearbyPlayers.add(player);
				for(L2Player nearbyPlayer : nearbyPlayers)
				{
					if(nearbyPlayer == null)
						continue;
					nearbyPlayer.sendPacket(cs);
				}
				break;
			case 1:
				if(Config.GLOBAL_CHAT <= Experience.MAX_LEVEL)
					for(L2Player player : L2World.getAllPlayers())
						nearbyPlayers.add(player);
				else
					for(L2Player player : L2World.getAroundPlayers(npc))
						nearbyPlayers.add(player);
				for(L2Player nearbyPlayer : nearbyPlayers)
				{
					if(nearbyPlayer == null)
						continue;
					nearbyPlayer.sendPacket(cs);
				}
				break;
			case 2:
				_player.sendPacket(cs);
				break;
		}
	}

	/**
	 * Return value of parameter "val" after adding the couple (var,val) in class variable "vars".<BR><BR>
	 * <U><I>Actions :</I></U><BR>
	 * <LI>Initialize class variable "vars" if is null</LI>
	 * <LI>Initialize parameter "val" if is null</LI>
	 * <LI>Add/Update couple (var,val) in class variable FastMap "vars"</LI>
	 * <LI>If the key represented by "var" exists in FastMap "vars", the couple (var,val) is updated in the database. The key is known as
	 * existing if the preceding value of the key (given as result of function put()) is not null.<BR>
	 * If the key doesn't exist, the couple is added/created in the database</LI>
	 * @param var : String indicating the name of the variable for quest
	 * @param val : String indicating the value of the variable for quest
	 * @return String (equal to parameter "val")
	 */
	public String set(String var, String val)
	{
		if(_vars == null)
			_vars = new FastMap<String, String>();
		if(val == null)
			val = "";
		_vars.put(var, val);
		Quest.updateQuestVarInDb(this, var, val);
		if(var.equals("cond"))
			_player.sendPacket(new QuestList(_player));
		return val;
	}

	/**
	 * Add parameter used in quests.
	 * @param var : String pointing out the name of the variable for quest
	 * @param val : String pointing out the value of the variable for quest
	 * @return String (equal to parameter "val")
	 */
	String setInternal(String var, String val)
	{
		if(_vars == null)
			_vars = new FastMap<String, String>();
		if(val == null)
			val = "";
		_vars.put(var, val);
		return val;
	}

	/**
	 * Return state of the quest after its initialization.<BR><BR>
	 * <U><I>Actions :</I></U>
	 * <LI>Remove drops from previous state</LI>
	 * <LI>Set new state of the quest</LI>
	 * <LI>Add drop for new state</LI>
	 * <LI>Update information in database</LI>
	 * <LI>Send packet QuestList to client</LI>
	 * @param state
	 * @return object
	 */
	public Object setState(State state)
	{
		if(state == null)
			return null;

		// set new state
		_state = state;

		if(getQuest().getQuestIntId() < 999 && isStarted())
			_player.sendPacket(new ExShowQuestMark(getQuest().getQuestIntId()));

		Quest.updateQuestInDb(this);
		_player.sendPacket(new QuestList(_player));
		return state;
	}

	public void removeRadar(int x, int y, int z)
	{
		_player.radar.removeMarker(x, y, z);
	}

	/**
	 * Send a packet in order to play sound at client terminal
	 * @param sound
	 */
	public void playSound(String sound)
	{
		_player.sendPacket(new PlaySound(sound));
	}

	public void playTutorialVoice(String voice)
	{
		_player.sendPacket(new PlaySound(2, voice, 0, 0, _player.getLoc()));
	}

	public void onTutorialClientEvent(int number)
	{
		_player.sendPacket(new TutorialEnableClientEvent(number));
	}

	public void showQuestionMark(int number)
	{
		_player.sendPacket(new TutorialShowQuestionMark(number));
	}

	public void showTutorialHTML(String html)
	{
		String text = Files.read("data/scripts/quests/_255_Tutorial/" + html, _player);
		if(text == null || text.equalsIgnoreCase(""))
			text = "<html><body>File data/scripts/quests/_255_Tutorial/" + html + " not found or file is empty.</body></html>";
		_player.sendPacket(new TutorialShowHtml(text));
	}

	/**
	 * Start a timer for quest.<BR><BR>
	 * @param name<BR> The name of the timer. Will also be the value for event of onEvent
	 * @param time<BR> The milisecond value the timer will elapse
	 */
	public void startQuestTimer(String name, long time)
	{
		getQuest().startQuestTimer(name, time, null, getPlayer());
	}

	/**
	 * Удаляет указанные предметы из инвентаря игрока, и обновляет инвентарь
	 * @param itemId : id удаляемого предмета
	 * @param count : число удаляемых предметов<br>
	 * Если count передать -1, то будут удалены все указанные предметы.
	 * @return Количество удаленных предметов
	 */
	public long takeItems(int itemId, long count)
	{
		// Get object item from player's inventory list
		L2ItemInstance item = _player.getInventory().getItemByItemId(itemId);
		if(item == null)
			return 0;
		// Tests on count value in order not to have negative value
		if(count < 0 || count > item.getCount())
			count = item.getCount();

		// Destroy the quantity of items wanted
		_player.getInventory().destroyItemByItemId(itemId, count, true);
		// Send message of destruction to client
		if(itemId == 57)
			_player.sendPacket(new SystemMessage(SystemMessage.S1_ADENA_DISAPPEARED).addNumber(count));
		else if(count == 1)
			_player.sendPacket(new SystemMessage(SystemMessage.S1_HAS_DISAPPEARED).addItemName(itemId));
		else
			_player.sendPacket(new SystemMessage(SystemMessage.S2_S1_HAS_DISAPPEARED).addItemName(itemId).addNumber(count));

		return count;
	}

	/**
	 * Remove the variable of quest from the list of variables for the quest.<BR><BR>
	 * <U><I>Concept : </I></U>
	 * Remove the variable of quest represented by "var" from the class variable FastMap "vars" and from the database.
	 * @param var : String designating the variable for the quest to be deleted
	 * @return String pointing out the previous value associated with the variable "var"
	 */
	public String unset(String var)
	{
		if(_vars == null)
			return null;
		String old = _vars.remove(var);
		if(old != null)
			Quest.deleteQuestVarInDb(this, var);
		return old;
	}

	private boolean checkPartyMember(L2Player member, State st, int maxrange, L2Object rangefrom)
	{
		if(member == null)
			return false;
		if(rangefrom != null && maxrange > 0 && !member.isInRange(rangefrom, maxrange))
			return false;
		if(st != null)
		{
			QuestState qs = member.getQuestState(getQuest().getName());
			if(qs == null || qs.getState() != st)
				return false;
		}
		return true;
	}

	public FastList<L2Player> getPartyMembers(State st, int maxrange, L2Object rangefrom)
	{
		FastList<L2Player> result = new FastList<L2Player>();
		L2Party party = getPlayer().getParty();
		if(party == null)
		{
			if(checkPartyMember(getPlayer(), st, maxrange, rangefrom))
				result.add(getPlayer());
			return result;
		}

		for(L2Player _member : party.getPartyMembers())
			if(checkPartyMember(_member, st, maxrange, rangefrom))
				result.add(getPlayer());

		return result;
	}

	public L2Player getRandomPartyMember()
	{
		return getRandomPartyMember(null);
	}

	public L2Player getRandomPartyMember(State st)
	{
		return getRandomPartyMember(st, 0, null);
	}

	public L2Player getRandomPartyMember(State st, int maxrangefromplayer)
	{
		return getRandomPartyMember(st, maxrangefromplayer, getPlayer());
	}

	public L2Player getRandomPartyMember(State st, int maxrange, L2Object rangefrom)
	{
		FastList<L2Player> list = getPartyMembers(st, maxrange, rangefrom);
		if(list.size() == 0)
			return null;
		return list.get(Rnd.get(list.size()));
	}
}