package l2r.gameserver.tables;

import java.sql.ResultSet;
import java.util.logging.Logger;

import l2r.database.DatabaseUtils;
import l2r.database.FiltredPreparedStatement;
import l2r.database.L2DatabaseFactory;
import l2r.database.ThreadConnection;
import l2r.gameserver.handler.IScriptHandler;
import l2r.gameserver.handler.ScriptHandler;
import l2r.gameserver.instancemanager.CastleManager;
import l2r.gameserver.instancemanager.ClanHallManager;
import l2r.gameserver.instancemanager.FortressManager;
import l2r.gameserver.instancemanager.SiegeManager;
import l2r.gameserver.instancemanager.TownManager;
import l2r.gameserver.model.L2Character;
import l2r.gameserver.model.L2Clan;
import l2r.gameserver.model.L2Player;
import l2r.gameserver.model.L2World;
import l2r.gameserver.model.L2Zone;
import l2r.gameserver.model.Reflection;
import l2r.gameserver.model.L2Zone.ZoneType;
import l2r.gameserver.model.base.Race;
import l2r.gameserver.model.entity.siege.Siege;
import l2r.gameserver.model.instances.L2NpcInstance;
import l2r.util.Location;
import l2r.util.Rnd;

@SuppressWarnings( { "nls", "unqualified-field-access", "boxing" })
public class MapRegionTable
{
	private final static Logger _log = Logger.getLogger(MapRegionTable.class.getName());

	private static MapRegionTable _instance;

	private final int[][] _regions = new int[L2World.WORLD_SIZE_X][L2World.WORLD_SIZE_Y];

	public static enum TeleportWhereType
	{
		Castle,
		ClanHall,
		ClosestTown,
		SecondClosestTown,
		Headquarter,
		Fortress
	}

	public static MapRegionTable getInstance()
	{
		if(_instance == null)
			_instance = new MapRegionTable();
		return _instance;
	}

	private MapRegionTable()
	{
		int count2 = 0;

		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		ResultSet rset = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("SELECT * FROM mapregion");
			rset = statement.executeQuery();
			while(rset.next())
			{
				int y = rset.getInt("y10_plus");

				for(int i = 0; i < L2World.WORLD_SIZE_X; i++)
				{
					int region = rset.getInt("x" + (Math.abs(L2World.MAP_MIN_X >> 15) + 10 + i));
					_regions[i][y] = region;
					count2++;
				}
			}

			_log.fine("Loaded " + count2 + " mapregions.");
		}
		catch(Exception e)
		{
			_log.warning("error while creating map region data: " + e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCSR(con, statement, rset);
		}
	}

	public final int getMapRegion(int posX, int posY)
	{
		int tileX = posX - L2World.MAP_MIN_X >> 15;
		int tileY = posY - L2World.MAP_MIN_Y >> 15;
		return _regions[tileX][tileY];
	}

	public Location getTeleToLocation(L2Character activeChar, TeleportWhereType teleportWhere)
	{
		L2Player player = activeChar.getPlayer();

		if(player != null)
		{
			IScriptHandler handler = ScriptHandler.getInstance().getScriptHandler();
			if(handler != null)
			{
				Location loc = handler.onEscape(player);
				if(loc != null)
					return loc;
			}

			if(player.getReflection() != 0)
			{
				Reflection r = ReflectionTable.getInstance().getById(player.getReflection());
				if(r != null && r.getCore() != null)
					return r.getCore();
				player.setReflection(0);
			}

			L2Clan clan = player.getClan();

			if(clan != null)
			{
				// If teleport to clan hall
				if(teleportWhere == TeleportWhereType.ClanHall && clan.getHasHideout() != 0)
					return ClanHallManager.getInstance().getClanHall(clan.getHasHideout()).getZone().getSpawn();

				// If teleport to castle
				if(teleportWhere == TeleportWhereType.Castle && clan.getHasCastle() != 0)
					return CastleManager.getInstance().getCastleByIndex(clan.getHasCastle()).getZone().getSpawn();

				// If teleport to fortress
				if(teleportWhere == TeleportWhereType.Fortress && clan.getHasFortress() != 0)
					return FortressManager.getInstance().getFortressByIndex(clan.getHasFortress()).getZone().getSpawn();

				// Checking if in siege
				Siege siege = SiegeManager.getSiege(activeChar);
				if(siege != null && siege.isInProgress())
				{
					if(teleportWhere == TeleportWhereType.Headquarter)
					{
						// Check if player's clan is attacker
						L2NpcInstance flag = siege.getHeadquarter(player.getClan());
						if(flag != null)
							// Спаун рядом с флагом
							return Rnd.coordsRandomize(flag.getLoc(), 49, 51);
						return TownManager.getInstance().getClosestTown(activeChar).getSpawn();
					}

					// Check if player's clan is defender
					if(siege.getDefenderClan(player.getClan()) != null && siege.getSiegeUnit() != null && siege.getSiegeUnit().getZone() != null)
						return player.getKarma() > 1 ? siege.getSiegeUnit().getZone().getPKSpawn() : siege.getSiegeUnit().getZone().getSpawn();
					return player.getKarma() > 1 ? TownManager.getInstance().getClosestTown(activeChar).getPKSpawn() : TownManager.getInstance().getClosestTown(activeChar).getSpawn();
				}
			}

			// Светлые эльфы не могут воскрешаться в городе темных
			if(player.getRace() == Race.elf && TownManager.getInstance().getClosestTown(activeChar).getTownId() == 3)
				return player.getKarma() > 1 ? TownManager.getInstance().getTown(2).getPKSpawn() : TownManager.getInstance().getTown(2).getSpawn();

			// Темные эльфы не могут воскрешаться в городе светлых
			if(player.getRace() == Race.darkelf && TownManager.getInstance().getClosestTown(activeChar).getTownId() == 2)
				return player.getKarma() > 1 ? TownManager.getInstance().getTown(3).getPKSpawn() : TownManager.getInstance().getTown(3).getSpawn();

			L2Zone battle = activeChar.getZone(ZoneType.battle_zone);

			// If in battle zone
			if(battle != null && battle.getRestartPoints() != null)
				return player.getKarma() > 1 ? battle.getPKSpawn() : battle.getSpawn();

			// If in pease zone
			if(activeChar.isInZone(ZoneType.peace_zone) && activeChar.getZone(ZoneType.peace_zone).getRestartPoints() != null)
				return player.getKarma() > 1 ? activeChar.getZone(ZoneType.peace_zone).getPKSpawn() : activeChar.getZone(ZoneType.peace_zone).getSpawn();

			// If in offshore zone == pease zone options.
			if(activeChar.isInZone(ZoneType.offshore) && activeChar.getZone(ZoneType.offshore).getRestartPoints() != null)
				return player.getKarma() > 1 ? activeChar.getZone(ZoneType.offshore).getPKSpawn() : activeChar.getZone(ZoneType.offshore).getSpawn();

			return player.getKarma() > 1 ? TownManager.getInstance().getClosestTown(activeChar).getPKSpawn() : TownManager.getInstance().getClosestTown(activeChar).getSpawn();
		}

		return TownManager.getInstance().getClosestTown(activeChar).getSpawn();
	}
}