package l2r.gameserver.tables;

import java.sql.ResultSet;
import java.util.HashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import l2r.database.DatabaseUtils;
import l2r.database.FiltredPreparedStatement;
import l2r.database.L2DatabaseFactory;
import l2r.database.ThreadConnection;
import l2r.gameserver.model.L2Character;
import l2r.gameserver.model.L2PetData;
import l2r.gameserver.model.L2Player;
import l2r.gameserver.model.L2Summon;
import l2r.gameserver.model.instances.L2ItemInstance;

public class PetDataTable
{
	private static final Logger _log = Logger.getLogger(PetDataTable.class.getName());

	private static PetDataTable _instance = new PetDataTable();
	private HashMap<Integer, L2PetData> _pets;

	public static final short[] _itemControlIds = { 2375, 3500, 3501, 3502, 4422, 4423, 4424, 4425, 6648, 6649, 6650, 9882 };

	public static PetDataTable getInstance()
	{
		return _instance;
	}

	public void reload()
	{
		_instance = new PetDataTable();
	}

	private PetDataTable()
	{
		_pets = new HashMap<Integer, L2PetData>(1200, 0.95f);
		FillPetDataTable();
	}

	public L2PetData getInfo(int petNpcId, Byte level)
	{
		return _pets.get(petNpcId * 100 + level);
	}

	private void FillPetDataTable()
	{
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		ResultSet rset = null;
		L2PetData petData;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("SELECT id, level, exp, hp, mp, patk, pdef, matk, mdef, acc, evasion, crit, speed, atk_speed, cast_speed, max_meal, battle_meal, normal_meal, loadMax, hpregen, mpregen FROM pet_data");
			rset = statement.executeQuery();
			while(rset.next())
			{
				petData = new L2PetData();
				petData.setID(rset.getInt("id"));
				petData.setLevel(rset.getInt("level"));
				petData.setExp(rset.getInt("exp"));
				petData.setHP(rset.getInt("hp"));
				petData.setMP(rset.getInt("mp"));
				petData.setPAtk(rset.getInt("patk"));
				petData.setPDef(rset.getInt("pdef"));
				petData.setMAtk(rset.getInt("matk"));
				petData.setMDef(rset.getInt("mdef"));
				petData.setAccuracy(rset.getInt("acc"));
				petData.setEvasion(rset.getInt("evasion"));
				petData.setCritical(rset.getInt("crit"));
				petData.setSpeed(rset.getInt("speed"));
				petData.setAtkSpeed(rset.getInt("atk_speed"));
				petData.setCastSpeed(rset.getInt("cast_speed"));
				petData.setFeedMax(rset.getInt("max_meal"));
				petData.setFeedBattle(rset.getInt("battle_meal"));
				petData.setFeedNormal(rset.getInt("normal_meal"));
				petData.setMaxLoad(rset.getInt("loadMax"));
				petData.setHpRegen(rset.getInt("hpregen"));
				petData.setMpRegen(rset.getInt("mpregen"));

				_pets.put(petData.getID() * 100 + petData.getLevel(), petData);
			}
		}
		catch(Exception e)
		{
			_log.warning("Cannot fill up PetDataTable: " + e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCSR(con, statement, rset);
		}
		_log.config("PetDataTable: Loaded " + _pets.size() + " pets.");
	}

	public static boolean isPetControlItem(L2ItemInstance item)
	{
		for(short id : _itemControlIds)
			if(item.getItemId() == id)
				return true;
		return false;
	}

	public static void deletePet(L2ItemInstance item, L2Character owner)
	{
		{
			int petObjectId = 0;
			ThreadConnection con = null;
			FiltredPreparedStatement statement = null;
			ResultSet rset = null;
			try
			{
				con = L2DatabaseFactory.getInstance().getConnection();
				statement = con.prepareStatement("SELECT objId FROM pets WHERE item_obj_id=?");
				statement.setInt(1, item.getObjectId());
				rset = statement.executeQuery();
				while(rset.next())
					petObjectId = rset.getInt("objId");
				DatabaseUtils.closeDatabaseSR(statement, rset);

				L2Summon summon = owner.getPet();
				if(summon != null && summon.getObjectId() == petObjectId)
					summon.unSummon();

				L2Player player = owner.getPlayer();
				if(player != null && player.isMounted() && player.getMountObjId() == petObjectId)
					player.setMount(0, 0);

				//	 if it's a pet control item, delete the pet
				statement = con.prepareStatement("DELETE FROM pets WHERE item_obj_id=?");
				statement.setInt(1, item.getObjectId());
				statement.execute();
			}
			catch(Exception e)
			{
				_log.log(Level.WARNING, "could not restore pet objectid:", e);
			}
			finally
			{
				DatabaseUtils.closeDatabaseCSR(con, statement, rset);
			}
		}
	}

	public static void unSummonPet(L2ItemInstance oldItem, L2Character owner)
	{
		int petObjectId = 0;
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		ResultSet rset = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("SELECT objId FROM pets WHERE item_obj_id=?");
			statement.setInt(1, oldItem.getObjectId());
			rset = statement.executeQuery();

			while(rset.next())
				petObjectId = rset.getInt("objId");

			if(owner == null)
				return;

			L2Summon summon = owner.getPet();
			if(summon != null && summon.getObjectId() == petObjectId)
				summon.unSummon();

			L2Player player = owner.getPlayer();
			if(player != null && player.isMounted() && player.getMountObjId() == petObjectId)
				player.setMount(0, 0);
		}
		catch(Exception e)
		{
			_log.log(Level.WARNING, "could not restore pet objectid:", e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCSR(con, statement, rset);
		}
	}

	public static int getSummonId(L2ItemInstance item)
	{
		int npcId = 0;

		switch(item.getItemId())
		{
			// wolf pet a
			case 2375:
				npcId = 12077;
				break;

			// Sin Eater
			case 4425:
				npcId = 12564;
				break;

			// hatchling of wind
			case 3500:
				npcId = 12311;
				break;

			// hatchling of star
			case 3501:
				npcId = 12312;
				break;

			// hatchling of twilight
			case 3502:
				npcId = 12313;
				break;
			//	wind strider
			case 4422:
				npcId = 12526;
				break;
			//	Star strider
			case 4423:
				npcId = 12527;
				break;
			// Twilight strider
			case 4424:
				npcId = 12528;
				break;
			// Baby Buffalo
			case 6648:
				npcId = 12780;
				break;
			// Baby Kookaburra
			case 6650:
				npcId = 12781;
				break;
			// Baby Cougar
			case 6649:
				npcId = 12782;
				break;
			// Great Wolf
			case 9882:
				npcId = 16025;
				break;
			// unknown item id.. should never happen
			default:
		}
		return npcId;
	}

}
