package l2r.gameserver.tables;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.logging.Level;
import java.util.logging.Logger;

import javolution.util.FastMap;
import l2r.database.DatabaseUtils;
import l2r.database.FiltredPreparedStatement;
import l2r.database.FiltredStatement;
import l2r.database.L2DatabaseFactory;
import l2r.database.ThreadConnection;
import l2r.gameserver.model.L2Clan;
import l2r.gameserver.model.L2Player;
import l2r.gameserver.model.L2Skill;
import l2r.gameserver.model.L2SkillLearn;
import l2r.gameserver.model.L2Skill.SkillType;
import l2r.gameserver.model.base.ClassId;
import l2r.gameserver.model.base.L2EnchantSkillLearn;
import l2r.util.Log;

@SuppressWarnings( { "nls", "unqualified-field-access", "boxing" })
public class SkillTreeTable
{
	public static final int NORMAL_ENCHANT_COST_MULTIPLIER = 1;
	public static final int SAFE_ENCHANT_COST_MULTIPLIER = 3;

	public static final int NORMAL_ENCHANT_BOOK = 6622;
	public static final int SAFE_ENCHANT_BOOK = 9627;
	public static final int CHANGE_ENCHANT_BOOK = 9626;
	public static final int UNTRAIN_ENCHANT_BOOK = 9625;

	private static final Logger _log = Logger.getLogger(SkillTreeTable.class.getName());

	private static final SkillTreeTable _instance = new SkillTreeTable();

	private static FastMap<ClassId, ArrayList<L2SkillLearn>> _skillTrees;
	private static ArrayList<FastMap<Integer, FastMap<Integer, L2SkillLearn>>> _skillCostTable;
	private static FastMap<Integer, ArrayList<L2EnchantSkillLearn>> _enchant;
	private static ArrayList<L2SkillLearn> _fishingSkills;
	private static ArrayList<L2SkillLearn> _clanSkills;

	public static SkillTreeTable getInstance()
	{
		return _instance;
	}

	/**
	 * Return the minimum level needed to have this Expertise.<BR><BR>
	 *
	 * @param grade The grade level searched
	 *
	 */
	public static short getExpertiseLevel(int grade)
	{
		if(grade <= 0)
			return 0;

		for(L2SkillLearn sl : SkillTreeTable._skillTrees.get(ClassId.fighter))
			// TODO: переписать нафиг
			if(sl.id == 239 && sl.skillLevel == grade)
				return sl.minLevel;

		throw new Error("Expertise not found for grade " + grade);
	}

	public static int getMinSkillLevel(int skillID, ClassId classID, int skillLVL)
	{
		if(skillID > 0 && skillLVL > 0)
			for(L2SkillLearn sl : SkillTreeTable._skillTrees.get(classID))
				if(sl.skillLevel == skillLVL && sl.id == skillID)
					return sl.minLevel;

		return 0;
	}

	private SkillTreeTable()
	{
		_skillTrees = new FastMap<ClassId, ArrayList<L2SkillLearn>>();
		_fishingSkills = new ArrayList<L2SkillLearn>();
		_clanSkills = new ArrayList<L2SkillLearn>();
		int classintid = 0;
		int count = 0;

		ThreadConnection con = null;
		FiltredPreparedStatement classliststatement = null;
		FiltredPreparedStatement skilltreestatement = null;
		ResultSet classlist = null, skilltree = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			classliststatement = con.prepareStatement("SELECT * FROM class_list ORDER BY id");
			skilltreestatement = con.prepareStatement("SELECT class_id, skill_id, level, name, sp, min_level, rep FROM skill_trees where class_id=? AND class_id >= 0 ORDER BY skill_id, level");
			classlist = classliststatement.executeQuery();
			while(classlist.next())
			{
				classintid = classlist.getInt("id");
				ClassId classId = ClassId.values()[classintid];
				ArrayList<L2SkillLearn> list = new ArrayList<L2SkillLearn>();

				skilltreestatement.setInt(1, classintid);
				skilltree = skilltreestatement.executeQuery();
				addSkills(con, skilltree, list);

				_skillTrees.put(ClassId.values()[classintid], list);
				count += list.size();

				ClassId secondparent = classId.getParent((byte) 1);
				if(secondparent == classId.getParent((byte) 0))
					secondparent = null;

				classId = classId.getParent((byte) 0);
				while(classId != null)
				{
					ArrayList<L2SkillLearn> parentList = _skillTrees.get(classId);
					list.addAll(parentList);
					classId = classId.getParent((byte) 0);
					if(classId == null && secondparent != null)
					{
						classId = secondparent;
						secondparent = secondparent.getParent((byte) 1);
					}
				}

				//_log.config("SkillTreeTable: skill tree for class " + classintid + " has " + list.size() + " skills");
			}
			DatabaseUtils.closeDatabaseSR(classliststatement, classlist);
			classliststatement = null;
			classlist = null;
			DatabaseUtils.closeDatabaseSR(skilltreestatement, skilltree);
			loadFishingSkills(con);
			loadClanSkills(con);
			loadEnchants(con);
		}
		catch(Exception e)
		{
			_log.log(Level.SEVERE, "error while creating skill tree for classId " + classintid, e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseSR(classliststatement, classlist);
			DatabaseUtils.closeDatabaseSR(skilltreestatement, skilltree);
			DatabaseUtils.closeConnection(con);
		}

		loadSkillCostTable();

		_log.config("SkillTreeTable: Loaded " + count + " skills.");
		_log.config("SkillTreeTable: Loaded " + _fishingSkills.size() + " fishing skills.");
		_log.config("SkillTreeTable: Loaded " + _clanSkills.size() + " clan skills.");
		_log.config("SkillTreeTable: Loaded " + _enchant.size() + " enchanted skills.");
	}

	private void loadFishingSkills(ThreadConnection con) throws SQLException
	{
		FiltredPreparedStatement statement = null;
		ResultSet skilltree = null;
		try
		{
			statement = con.prepareStatement("SELECT class_id, skill_id, level, name, sp, min_level, rep FROM skill_trees WHERE class_id=-1 ORDER BY skill_id, level");
			skilltree = statement.executeQuery();
			addSkills(con, skilltree, _fishingSkills);
		}
		finally
		{
			DatabaseUtils.closeDatabaseSR(statement, skilltree);
		}
	}

	private void loadClanSkills(ThreadConnection con) throws SQLException
	{
		FiltredPreparedStatement statement = null;
		ResultSet skilltree = null;
		try
		{
			statement = con.prepareStatement("SELECT class_id, skill_id, level, name, sp, min_level, rep FROM skill_trees WHERE class_id=-2 ORDER BY skill_id, level");
			skilltree = statement.executeQuery();
			addSkills(con, skilltree, _clanSkills);
		}
		finally
		{
			DatabaseUtils.closeDatabaseSR(statement, skilltree);
		}
	}

	private void addSkills(ThreadConnection con, ResultSet skilltree, ArrayList<L2SkillLearn> dest) throws SQLException
	{
		while(skilltree.next())
		{
			short id = skilltree.getShort("skill_id");
			byte lvl = skilltree.getByte("level");
			String name = skilltree.getString("name");
			if(lvl == 1)
			{
				L2Skill s = SkillTable.getInstance().getInfo(id, 1);
				if(s == null || s.getSkillType() == SkillType.NOTDONE)
					Log.add(name + " - " + id, "unimplemented_skills", "");
			}
			byte minLvl = skilltree.getByte("min_level");
			int cost;
			short itemId = 0;
			int itemCount = 0;
			if(skilltree.getInt("class_id") == -2)
				cost = skilltree.getInt("rep");
			else
				cost = skilltree.getInt("sp");
			FiltredPreparedStatement statement2 = con.prepareStatement("SELECT item_id, item_count FROM skill_spellbooks WHERE skill_id=? AND level=?");
			statement2.setInt(1, id);
			statement2.setInt(2, lvl);
			ResultSet itemIdCount = statement2.executeQuery();
			if(itemIdCount.next())
			{
				itemId = itemIdCount.getShort("item_id");
				itemCount = itemIdCount.getInt("item_count");
			}
			statement2.close();
			L2SkillLearn skl = new L2SkillLearn(id, lvl, minLvl, name, cost, itemId, itemCount, skilltree.getInt("class_id") == -1, skilltree.getInt("class_id") == -2, skilltree.getInt("class_id") == -4);
			dest.add(skl);
		}
	}

	private void loadSkillCostTable()
	{
		_skillCostTable = new ArrayList<FastMap<Integer, FastMap<Integer, L2SkillLearn>>>(ClassId.values().length + 1);
		for(ClassId cid : ClassId.values())
			_skillCostTable.add(cid.getId(), new FastMap<Integer, FastMap<Integer, L2SkillLearn>>());

		for(ClassId classId : _skillTrees.keySet())
		{
			FastMap<Integer, FastMap<Integer, L2SkillLearn>> skt = _skillCostTable.get(classId.getId());

			ArrayList<L2SkillLearn> lst = _skillTrees.get(classId);
			for(L2SkillLearn skl : lst)
			{
				FastMap<Integer, L2SkillLearn> skillmap = skt.get((int) skl.getId());
				if(skillmap == null)
				{
					skillmap = new FastMap<Integer, L2SkillLearn>();
					skt.put((int) skl.getId(), skillmap);
				}
				skillmap.put((int) skl.getLevel(), skl);
			}
		}
	}

	private void loadEnchants(ThreadConnection con) throws SQLException
	{
		_enchant = new FastMap<Integer, ArrayList<L2EnchantSkillLearn>>();
		FiltredStatement statement = null;
		ResultSet skilltree = null;
		try
		{
			statement = con.createStatement();
			skilltree = statement.executeQuery("SELECT skill_id, level, name, base_lvl, enchant_type, sp, min_skill_lvl FROM enchant_skill_trees ORDER BY skill_id, level");
			int cnt = 0;
			while(skilltree.next())
			{
				int id = skilltree.getInt("skill_id");
				int level = skilltree.getInt("level");
				String name = skilltree.getString("name");
				int base = skilltree.getInt("base_lvl");
				String type = skilltree.getString("enchant_type");
				int sp = skilltree.getInt("sp");
				int min = skilltree.getInt("min_skill_lvl");
				L2EnchantSkillLearn e = new L2EnchantSkillLearn(id, level, name, type, min, base, sp);

				ArrayList<L2EnchantSkillLearn> t = _enchant.get(id);
				if(t == null)
					t = new ArrayList<L2EnchantSkillLearn>();
				t.add(e);
				_enchant.put(id, t);

				try
				{
					SkillTable.getInstance().getInfo(id, convertEnchantLevel(base, level)).setDisplayLevel((short) level);
					// 101-130 = base+1-base+30
					// 201-230 = base+31-base+60
					// 301-330 = base+61-base+90
					// 401-430 = base+91-base+120
					// 501-530 = base+121-base+150
					// 601-630 = base+151-base+180
				}
				catch(NullPointerException x)
				{}

				cnt++;
			}
			_log.config("SkillTreeTable: Loaded " + cnt + " skills enchants.");
		}
		finally
		{
			DatabaseUtils.closeDatabaseSR(statement, skilltree);
		}
	}

	public void reloadEnchant()
	{
		for(ArrayList<L2EnchantSkillLearn> t : _enchant.values())
			if(t != null)
				for(L2EnchantSkillLearn e : t)
					try
					{
						SkillTable.getInstance().getInfo(e.getId(), e.getBaseLevel() + e.getLevel() - 100).setDisplayLevel((short) e.getLevel());
					}
					catch(NullPointerException x)
					{}
	}

	public ArrayList<L2SkillLearn> getAvailableSkills(L2Player cha, ClassId classId)
	{
		ArrayList<L2SkillLearn> result = new ArrayList<L2SkillLearn>();
		ArrayList<L2SkillLearn> skills = _skillTrees.get(classId);
		if(skills == null)
		{
			// the skilltree for this class is undefined, so we give an empty list
			_log.warning("Skilltree for class " + classId + " is not defined !");
			return new ArrayList<L2SkillLearn>(0);
		}

		Collection<L2Skill> oldSkills = cha.getAllSkills();
		for(L2SkillLearn temp : skills)
			if(temp.minLevel <= cha.getLevel())
			{
				boolean knownSkill = false;
				for(L2Skill s : oldSkills)
				{
					if(knownSkill)
						break;
					if(s.getId() == temp.id)
					{
						knownSkill = true;
						if(s.getLevel() == temp.skillLevel - 1)
							// this is the next level of a skill that we know
							result.add(temp);
					}
				}
				if(!knownSkill && temp.skillLevel == 1)
					// this is a new skill
					result.add(temp);
			}
		return result;
	}

	public L2SkillLearn[] getAvailableClanSkills(L2Clan clan)
	{
		ArrayList<L2SkillLearn> result = new ArrayList<L2SkillLearn>();
		ArrayList<L2SkillLearn> skills = _clanSkills;

		if(skills == null)
			return new L2SkillLearn[0];

		L2Skill[] oldSkills = clan.getAllSkills();

		for(L2SkillLearn temp : skills)
			if(temp.minLevel <= clan.getLevel())
			{
				boolean knownSkill = false;

				for(int j = 0; j < oldSkills.length && !knownSkill; j++)
					if(oldSkills[j].getId() == temp.id)
					{
						knownSkill = true;

						if(oldSkills[j].getLevel() == temp.skillLevel - 1)
							// this is the next level of a skill that we know
							result.add(temp);
					}

				if(!knownSkill && temp.skillLevel == 1)
					// this is a new skill
					result.add(temp);
			}

		return result.toArray(new L2SkillLearn[result.size()]);
	}

	public ArrayList<L2Skill> getSkillsToEnchant(L2Player cha)
	{
		ArrayList<L2Skill> result = new ArrayList<L2Skill>();

		Collection<L2Skill> skills = cha.getAllSkills();
		if(skills.size() == 0)
			return result;

		for(L2Skill s : skills)
			if(_enchant.get(s.getId()) != null && s.getLevel() < SkillTable.getInstance().getMaxLevel(s.getId(), s.getLevel()))
				result.add(s);

		return result;
	}

	public static ArrayList<L2EnchantSkillLearn> getFirstEnchantsForSkill(int skillid)
	{
		ArrayList<L2EnchantSkillLearn> result = new ArrayList<L2EnchantSkillLearn>();

		ArrayList<L2EnchantSkillLearn> enchants = _enchant.get(skillid);
		if(enchants == null)
			return result;

		for(L2EnchantSkillLearn e : enchants)
			if(e.getLevel() % 100 == 1)
				result.add(e);

		return result;
	}

	public static ArrayList<L2EnchantSkillLearn> getEnchantsForChange(int skillid, int level)
	{
		ArrayList<L2EnchantSkillLearn> result = new ArrayList<L2EnchantSkillLearn>();

		ArrayList<L2EnchantSkillLearn> enchants = _enchant.get(skillid);
		if(enchants == null)
			return result;

		for(L2EnchantSkillLearn e : enchants)
			if(e.getLevel() % 100 == level % 100)
				result.add(e);

		return result;
	}

	public static L2EnchantSkillLearn getSkillEnchant(int skillid, int level)
	{
		ArrayList<L2EnchantSkillLearn> enchants = _enchant.get(skillid);
		if(enchants == null)
			return null;

		for(L2EnchantSkillLearn e : enchants)
			if(e.getLevel() == level)
				return e;
		return null;
	}

	/**
	 * Преобразует уровень скила из клиентского представления в серверное
	 * @param baseLevel базовый уровень скила - максимально возможный без заточки
	 * @param level - текущий уровень скила
	 * @return уровень скила
	 */
	public static int convertEnchantLevel(int baseLevel, int level)
	{
		if(level < 100)
			return level;
		return baseLevel + ((level - level % 100) / 100 - 1) * 30 + level % 100;
	}

	public static L2SkillLearn getSkillLearn(int skillid, short level, ClassId classid, L2Clan clan)
	{
		if(clan != null)
		{
			L2SkillLearn[] clskills = getInstance().getAvailableClanSkills(clan);
			for(L2SkillLearn tmp : clskills)
				if(tmp.id == skillid && tmp.skillLevel == level)
					return tmp;
		}
		else
		{
			if(_fishingSkills != null)
				for(L2SkillLearn tmp : _fishingSkills)
					if(tmp.id == skillid && tmp.skillLevel == level)
						return tmp;

			//		for(ArrayList<L2SkillLearn> tmp1 : _skillTrees.get(classid))
			for(L2SkillLearn tmp : _skillTrees.get(classid))
				if(tmp.id == skillid && tmp.skillLevel == level)
					return tmp;
		}
		return null;
	}

	public L2SkillLearn[] getAvailableFishingSkills(L2Player cha)
	{
		ArrayList<L2SkillLearn> result = new ArrayList<L2SkillLearn>();
		if(_fishingSkills == null)
		{
			_log.warning("Fishing skills not defined!");
			return new L2SkillLearn[0];
		}

		Collection<L2Skill> oldSkills = cha.getAllSkills();

		for(L2SkillLearn temp : _fishingSkills)
			if(temp.minLevel <= cha.getLevel())
			{
				boolean knownSkill = false;
				for(L2Skill s : oldSkills)
				{
					if(knownSkill)
						break;
					if(s.getId() == temp.id)
					{
						knownSkill = true;
						if(s.getLevel() == temp.skillLevel - 1)
							result.add(temp);
					}
				}

				if(!knownSkill && temp.skillLevel == 1)
					result.add(temp);
			}
		return result.toArray(new L2SkillLearn[result.size()]);
	}

	public byte getMinLevelForNewSkill(L2Player cha, ClassId classId)
	{
		byte minlevel = 0;
		//ArrayList<L2SkillLearn> result = new ArrayList<L2SkillLearn>();
		ArrayList<L2SkillLearn> skills = _skillTrees.get(classId);
		if(skills == null)
		{
			// the skilltree for this class is undefined, so we give an empty list
			_log.warning("Skilltree for class " + classId + " is not defined !");
			return minlevel;
		}

		//L2Skill[] oldSkills = cha.getAllSkills();

		for(L2SkillLearn temp : skills)
			if(temp.minLevel > cha.getLevel())
				if(minlevel == 0 || temp.minLevel < minlevel)
					minlevel = temp.minLevel;
		return minlevel;
	}

	public int getSkillCost(L2Player player, L2Skill skill)
	{
		// TODO снести этот костыль
		switch(skill.getId())
		{
			// Рыбацкие скилы
			case 1312:
			case 1313:
			case 1314:
			case 1315:
			case 1368:
			case 1369:
			case 1370:
			case 1371:
			case 1372:
				// Скилы трансформации
			case 617:
			case 618:
			case 541:
			case 544:
			case 547:
			case 550:
			case 553:
			case 556:
				return 0;
		}

		FastMap<Integer, FastMap<Integer, L2SkillLearn>> skt = _skillCostTable.get(player.getActiveClass());
		if(skt == null)
			return Integer.MAX_VALUE;
		FastMap<Integer, L2SkillLearn> skillmap = skt.get(skill.getId());
		if(skillmap == null)
			return Integer.MAX_VALUE;
		L2SkillLearn skl = skillmap.get(1 + Math.max(player.getSkillLevel(skill.getId()), 0));
		if(skl == null)
			return Integer.MAX_VALUE;
		return skl.getSpCost();
	}

	public int getSkillRepCost(L2Clan clan, L2Skill skill)
	{
		int min = 100000000;
		int lvl = clan.getLeader().getPlayer().getSkillLevel(skill.getId());

		if(lvl > 0)
			lvl += 1;
		else
			lvl = 1;
		if(_clanSkills != null)
			for(L2SkillLearn tmp : _clanSkills)
			{
				if(tmp.id != skill.getId())
					continue;
				if(tmp.skillLevel != lvl)
					continue;
				if(tmp.minLevel > clan.getLevel())
					continue;
				min = Math.min(min, Math.round(tmp._repCost));
			}
		return min;
	}

	/**
	 * Возвращает true если скилл может быть изучен данным классом
	 * @param player
	 * @param skill_id
	 * @param skill_level
	 * @return true/false
	 */
	public boolean isSkillPossible(L2Player player, int skillid, short level)
	{
		for(L2SkillLearn tmp : _clanSkills)
			if(tmp.id == skillid && tmp.skillLevel <= level)
				return true;

		ArrayList<L2SkillLearn> skills = _skillTrees.get(ClassId.values()[player.getActiveClass()]);
		for(L2SkillLearn skilllearn : skills)
			if(skilllearn.id == skillid && skilllearn.skillLevel <= level)
				return true;

		return false;
	}
}