package l2r.gameserver.templates;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import javolution.util.FastMap;
import l2r.extensions.scripts.Script;
import l2r.extensions.scripts.Scripts;
import l2r.gameserver.model.L2Drop;
import l2r.gameserver.model.L2DropData;
import l2r.gameserver.model.L2MinionData;
import l2r.gameserver.model.L2Skill;
import l2r.gameserver.model.L2Skill.SkillType;
import l2r.gameserver.model.base.ClassId;
import l2r.gameserver.model.quest.Quest;

/**
 * This cl contains all generic data of a L2Spawn object.<BR><BR>
 *
 * <B><U> Data</U> :</B><BR><BR>
 * <li>npcId, type, name, sex</li>
 * <li>revardExp, revardSp</li>
 * <li>aggroRange, factionId, factionRange</li>
 * <li>rhand, lhand, armor</li>
 * <li>_drops</li>
 * <li>_minions</li>
 * <li>_teachInfo</li>
 * <li>_skills</li>
 * <li>_questsStart</li><BR><BR>
 */
public final class L2NpcTemplate extends L2CharTemplate
{
	private static final Logger _log = Logger.getLogger(L2NpcTemplate.class.getName());

	public static enum ShotsType
	{
		NONE,
		SOUL,
		SPIRIT,
		BSPIRIT,
		SOUL_SPIRIT,
		SOUL_BSPIRIT
	}

	public final int npcId;
	public String type;
	public String ai_type;
	public final String name;
	public String title;
	public final String sex;
	public final byte level;
	public final int revardExp;
	public final int revardSp;
	public final short aggroRange;
	public final int rhand;
	public final int lhand;
	public final int armor;
	public final String factionId;
	public final short factionRange;
	public final String jClass;
	public int displayId = 0;
	public boolean isDropHerbs = false;
	public final ShotsType shots;
	public final boolean isRaid;
	private StatsSet _AIParams = null;

	/** fixed skills*/
	private short race = 0;
	public double rateHp = 1;

	/** The object containing all Item that can be dropped by L2NpcInstance using this L2NpcTemplate*/
	private L2Drop _drop = null;
	public int killscount = 0;

	/** The table containing all Minions that must be spawn with the L2NpcInstance using this L2NpcTemplate*/
	private final List<L2MinionData> _minions = new ArrayList<L2MinionData>(0);

	private List<ClassId> _teachInfo;
	private HashMap<Integer, L2Skill> _skills;
	private HashMap<SkillType, L2Skill[]> _skillsByType;
	private final static HashMap<Integer, L2Skill> _emptySkills = new HashMap<Integer, L2Skill>(0);
	private Map<Quest.QuestEventType, Quest[]> _questEvents;

	/**
	 * Constructor of L2Character.<BR><BR>
	 * @param set The StatsSet object to transfer data to the method
	 */
	@SuppressWarnings("unchecked")
	public L2NpcTemplate(StatsSet set, StatsSet AIParams)
	{
		super(set);
		npcId = set.getInteger("npcId");
		displayId = set.getInteger("displayId");
		type = set.getString("type");
		ai_type = set.getString("ai_type");
		name = set.getString("name");
		title = set.getString("title");
		sex = set.getString("sex");
		level = set.getByte("level");
		revardExp = set.getInteger("revardExp");
		revardSp = set.getInteger("revardSp");
		aggroRange = set.getShort("aggroRange");
		rhand = set.getInteger("rhand");
		lhand = set.getInteger("lhand");
		armor = set.getInteger("armor");
		jClass = set.getString("jClass");
		String f = set.getString("factionId", null);
		if(f == null)
			factionId = "";
		else
			factionId = f.intern();
		factionRange = set.getShort("factionRange");
		_teachInfo = null;
		isDropHerbs = set.getBool("isDropHerbs");
		shots = set.getEnum("shots", ShotsType.class, ShotsType.NONE);
		_AIParams = AIParams;

		Class _this = null;
		try
		{
			_this = Class.forName("l2r.gameserver.model.instances." + type + "Instance");
		}
		catch(ClassNotFoundException e)
		{
			Script sc = Scripts.getInstance().getClasses().get("npc.model.instances." + type + "Instance");
			if(sc != null)
				_this = sc.getRawClass();
		}
		isRaid = _this != null && raidboss.isAssignableFrom(_this);
	}

	private static Class raidboss = null;
	static
	{
		try
		{
			raidboss = Class.forName("l2r.gameserver.model.instances.L2RaidBossInstance");
		}
		catch(ClassNotFoundException e)
		{}
	}

	public L2NpcTemplate(StatsSet set)
	{
		this(set, null);
	}

	public void addTeachInfo(ClassId classId)
	{
		if(_teachInfo == null)
			_teachInfo = new ArrayList<ClassId>();
		_teachInfo.add(classId);
	}

	public ClassId[] getTeachInfo()
	{
		if(_teachInfo == null)
			return null;
		return _teachInfo.toArray(new ClassId[_teachInfo.size()]);
	}

	public boolean canTeach(ClassId classId)
	{
		if(_teachInfo == null)
			return false;
		return _teachInfo.contains(classId);
	}

	public void addDropData(L2DropData drop)
	{
		if(_drop == null)
			_drop = new L2Drop();
		_drop.addData(drop);
	}

	public void addRaidData(L2MinionData minion)
	{
		_minions.add(minion);
	}

	public void addSkill(L2Skill skill)
	{
		if(_skills == null)
			_skills = new HashMap<Integer, L2Skill>();
		if(_skillsByType == null)
			_skillsByType = new HashMap<SkillType, L2Skill[]>();

		_skills.put(skill.getId(), skill);

		L2Skill[] skilllist;
		if(_skillsByType.get(skill.getSkillType()) != null)
		{
			skilllist = new L2Skill[_skillsByType.get(skill.getSkillType()).length + 1];
			System.arraycopy(_skillsByType.get(skill.getSkillType()), 0, skilllist, 0, _skillsByType.get(skill.getSkillType()).length);
		}
		else
			skilllist = new L2Skill[1];

		skilllist[skilllist.length - 1] = skill;

		if(skill.getTargetType() != L2Skill.SkillTargetType.TARGET_NONE && skill.getSkillType() != L2Skill.SkillType.NOTDONE && skill.isActive())
			_skillsByType.put(skill.getSkillType(), skilllist);
	}

	public L2Skill[] getSkillsByType(SkillType type)
	{
		if(_skillsByType == null)
			return new L2Skill[0];

		L2Skill[] ret = _skillsByType.get(type);

		return ret != null ? ret : new L2Skill[0];
	}

	/**
	 * Return the list of all possible drops of this L2NpcTemplate.<BR><BR>
	 */
	public L2Drop getDropData()
	{
		return _drop;
	}

	/**
	 * Обнуляет дроплист моба
	 */
	public void clearDropData()
	{
		_drop = null;
	}

	/**
	 * Return the list of all Minions that must be spawn with the L2NpcInstance using this L2NpcTemplate.<BR><BR>
	 */
	public List<L2MinionData> getMinionData()
	{
		return _minions;
	}

	public HashMap<Integer, L2Skill> getSkills()
	{
		return _skills == null ? _emptySkills : _skills;
	}

	public void addQuestEvent(Quest.QuestEventType EventType, Quest q)
	{
		if(_questEvents == null)
			_questEvents = new FastMap<Quest.QuestEventType, Quest[]>();

		if(_questEvents.get(EventType) == null)
			_questEvents.put(EventType, new Quest[] { q });
		else
		{
			Quest[] _quests = _questEvents.get(EventType);
			int len = _quests.length;

			// if only one registration per npc is allowed for this event type
			// then only register this NPC if not already registered for the specified event.
			// if a quest allows multiple registrations, then register regardless of count
			if(EventType.isMultipleRegistrationAllowed() || len < 1)
			{
				Quest[] tmp = new Quest[len + 1];
				for(int i = 0; i < len; i++)
				{
					if(_quests[i].getName().equals(q.getName()))
					{
						_quests[i] = q;
						return;
					}
					tmp[i] = _quests[i];
				}
				tmp[len] = q;
				_questEvents.put(EventType, tmp);
			}
			else
				_log.warning("Quest event not allowed in multiple quests. Skipped addition of Event Type \"" + EventType + "\" for NPC \"" + name + "\" and quest \"" + q.getName() + "\".");
		}
	}

	public Quest[] getEventQuests(Quest.QuestEventType EventType)
	{
		if(_questEvents == null)
			return null;
		return _questEvents.get(EventType);
	}

	public short getRace()
	{
		return race;
	}

	public void setRace(short newrace)
	{
		race = newrace;
	}

	public boolean isUndead()
	{
		return race == 1;
	}

	public void setRateHp(double newrate)
	{
		rateHp = newrate;
	}

	@Override
	public String toString()
	{
		return "Npc template " + name + "[" + npcId + "]";
	}

	@Override
	public int getNpcId()
	{
		return npcId;
	}

	public final String getJClass()
	{
		return jClass;
	}

	public final StatsSet getAIParams()
	{
		return _AIParams;
	}
}
