package l2r.gameserver.templates;

import l2r.gameserver.model.L2Character;
import l2r.gameserver.model.L2Skill;
import l2r.gameserver.model.L2Skill.SkillType;
import l2r.gameserver.serverpackets.SystemMessage;
import l2r.gameserver.tables.SkillTable;
import l2r.util.Rnd;

public final class L2Weapon extends L2Item
{
	private final int _soulShotCount;
	private final int _spiritShotCount;
	private final int _pDam;
	private final int _rndDam;
	private final int _atkReuse;
	private final int _mpConsume;
	private final int _mDam;

	private L2Skill _enchant4Skill = null; // skill that activates when item is enchanted +4 (for duals)
	private final boolean _sa;

	public enum WeaponType
	{
		NONE(1, "Shield"),
		SWORD(2, "Sword"),
		BLUNT(3, "Blunt"),
		DAGGER(4, "Dagger"),
		BOW(5, "Bow"),
		POLE(6, "Pole"),
		ETC(7, "Etc"),
		FIST(8, "Fist"),
		DUAL(9, "Dual Sword"),
		DUALFIST(10, "Dual Fist"),
		BIGSWORD(11, "Big Sword"), // Two Handed Swords
		PET(12, "Pet"),
		ROD(13, "Rod"),
		BIGBLUNT(14, "Big Blunt"),
		CROSSBOW(15, "Crossbow"),
		RAPIER(16, "Rapier"),
		ANCIENTSWORD(17, "Ancient Sword"); // Kamael 2h sword

		private final int _id;
		private final String _name;

		private WeaponType(int id, String name)
		{
			_id = id;
			_name = name;
		}

		public int mask()
		{
			return 1 << _id;
		}

		@Override
		public String toString()
		{
			return _name;
		}
	}

	/**
	 * Constructor for Weapon.<BR><BR>
	 * <U><I>Variables filled :</I></U><BR>
	 * <LI>_soulShotCount & _spiritShotCount</LI>
	 * <LI>_pDam & _mDam & _rndDam</LI>
	 * <LI>_critical</LI>
	 * <LI>_hitModifier</LI>
	 * <LI>_avoidModifier</LI>
	 * <LI>_shieldDes & _shieldDefRate</LI>
	 * <LI>_atkSpeed & _AtkReuse</LI>
	 * <LI>_mpConsume</LI>
	 * @param type : L2ArmorType designating the type of armor
	 * @param set : StatsSet designating the set of couples (key,value) caracterizing the armor
	 * @see L2Item constructor
	 */
	public L2Weapon(WeaponType type, StatsSet set)
	{
		super(type, set);
		_soulShotCount = set.getInteger("soulshots");
		_spiritShotCount = set.getInteger("spiritshots");
		_pDam = set.getInteger("p_dam");
		_rndDam = set.getInteger("rnd_dam");
		_atkReuse = set.getInteger("atk_reuse", type == WeaponType.BOW ? 1500 : type == WeaponType.CROSSBOW ? 1200 : 0);
		_mpConsume = set.getInteger("mp_consume");
		_mDam = set.getInteger("m_dam");
		_sa = !_icon.endsWith("i00");

		int sId = set.getInteger("enchant4_skill_id");
		int sLv = set.getInteger("enchant4_skill_lvl");
		if(sId > 0 && sLv > 0)
			_enchant4Skill = SkillTable.getInstance().getInfo(sId, sLv);
	}

	/**
	 * Returns the type of Weapon
	 * @return L2WeaponType
	 */
	@Override
	public WeaponType getItemType()
	{
		return (WeaponType) super.type;
	}

	/**
	 * Returns the ID of the Etc item after applying the mask.
	 * @return int : ID of the Weapon
	 */
	@Override
	public int getItemMask()
	{
		return getItemType().mask();
	}

	/**
	 * Returns the quantity of SoulShot used.
	 * @return int
	 */
	public int getSoulShotCount()
	{
		return _soulShotCount;
	}

	/**
	 * Returns the quatity of SpiritShot used.
	 * @return int
	 */
	public int getSpiritShotCount()
	{
		return _spiritShotCount;
	}

	/**
	 * Returns the physical damage.
	 * @return int
	 */
	public int getPDamage()
	{
		return _pDam;
	}

	/**
	 * Returns the random damage inflicted by the weapon
	 * @return int
	 */
	public int getRandomDamage()
	{
		return _rndDam;
	}

	/**
	 * Return the Attack Reuse Delay of the L2Weapon.<BR><BR>
	 * @return int
	 */
	public int getAttackReuseDelay()
	{
		return _atkReuse;
	}

	/**
	 * Returns the magical damage inflicted by the weapon
	 * @return int
	 */
	public int getMDamage()
	{
		return _mDam;
	}

	/**
	 * Returns the MP consumption with the weapon
	 * @return int
	 */
	public int getMpConsume()
	{
		return _mpConsume;
	}

	public void getEffect(boolean critNotCast, L2Character effector, L2Character effected, boolean offensive)
	{
		if(effector == null || effected == null)
			return;

		if(_skillOnAction == null)
			return;

		if(_skillOnCritNotCast != critNotCast || _offensive != offensive)
			return;

		if(!Rnd.chance(_chance))
			return;

		if((effected.isRaid || effected.isPlayer) && (_skillOnAction.getSkillType() == SkillType.PARALYZE || _skillOnAction.getSkillType() == SkillType.ROOT || _skillOnAction.getSkillType() == SkillType.STUN || _skillOnAction.getSkillType() == SkillType.MUTE))
			_skillOnAction.getEffects(effector, effected, true);
		else
			_skillOnAction.getEffects(effector, effected, false);

		effector.sendPacket(new SystemMessage(SystemMessage.S1_HAS_SUCCEEDED).addString("Your weapon's special ability"));
	}

	/**
	 * @return skill that player get when has equipped weapon +4  or more  (for duals SA)
	 */
	public L2Skill getEnchant4Skill()
	{
		return _enchant4Skill;
	}

	public boolean isSa()
	{
		return _sa;
	}
}