package l2r.loginserver;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.InetAddress;
import java.security.GeneralSecurityException;
import java.sql.SQLException;
import java.util.logging.LogManager;
import java.util.logging.Logger;

import l2r.Config;
import l2r.Server;
import l2r.database.L2DatabaseFactory;
import l2r.extensions.network.SelectorConfig;
import l2r.extensions.network.SelectorThread;
import l2r.extensions.scripts.Scripts;
import l2r.gameserver.GameServer;
import l2r.loginserver.gameservercon.GSConnection;
import l2r.status.Status;

public class L2LoginServer
{
	public static final int PROTOCOL_REV = 0x0102;

	private static L2LoginServer _instance;
	private Logger _log = Logger.getLogger(L2LoginServer.class.getName());
	private GSConnection _gameServerListener;
	private SelectorThread<L2LoginClient> _selectorThread;
	public static Status statusServer;
	public LoginController loginController;

	//private GameServer gameServer;

	public static void main(String[] args)
	{
		_instance = new L2LoginServer();
	}

	public static L2LoginServer getInstance()
	{
		return _instance;
	}

	public L2LoginServer()
	{
		Server.SERVER_MODE = Server.MODE_LOGINSERVER;
		//      Local Constants
		final String LOG_FOLDER = "log"; // Name of folder for log file
		final String LOG_NAME = "./config/log.properties"; // Name of log file

		/*** Main ***/
		// Create log folder
		File logFolder = new File("./", LOG_FOLDER);
		logFolder.mkdir();

		// Create input stream for log file -- or store file data into memory
		InputStream is = null;
		try
		{
			is = new FileInputStream(new File(LOG_NAME));
			LogManager.getLogManager().readConfiguration(is);
			is.close();
		}
		catch(IOException e)
		{
			e.printStackTrace();
		}
		finally
		{
			try
			{
				if(is != null)
					is.close();
			}
			catch(IOException e)
			{
				e.printStackTrace();
			}
		}

		// Load Config
		Config.load();
		if(Config.COMBO_MODE)
		{
			Server.SERVER_MODE = Server.MODE_COMBOSERVER;
			Config.load();
		}

		// Prepare Database
		try
		{
			L2DatabaseFactory.getInstance();
		}
		catch(SQLException e)
		{
			_log.severe("FATAL: Failed initializing database. Reason: " + e.getMessage());
			if(Config.LOGIN_DEBUG)
				e.printStackTrace();
			System.exit(1);
		}

		try
		{
			LoginController.load();
		}
		catch(GeneralSecurityException e)
		{
			_log.severe("FATAL: Failed initializing LoginController. Reason: " + e.getMessage());
			if(Config.LOGIN_DEBUG)
				e.printStackTrace();
			System.exit(1);
		}

		try
		{
			GameServerTable.load();
		}
		catch(GeneralSecurityException e)
		{
			_log.severe("FATAL: Failed to load GameServerTable. Reason: " + e.getMessage());
			if(Config.LOGIN_DEBUG)
				e.printStackTrace();
			System.exit(1);
		}
		catch(SQLException e)
		{
			_log.severe("FATAL: Failed to load GameServerTable. Reason: " + e.getMessage());
			if(Config.LOGIN_DEBUG)
				e.printStackTrace();
			System.exit(1);
		}

		//this.loadBanFile();

		InetAddress ad = null;
		try
		{
			ad = InetAddress.getByName(Config.LOGIN_HOST);
		}
		catch(Exception e)
		{}

		L2LoginPacketHandler loginPacketHandler = new L2LoginPacketHandler();
		SelectorHelper sh = new SelectorHelper();
		SelectorConfig<L2LoginClient> sc = new SelectorConfig<L2LoginClient>(null, sh);
		try
		{
			_selectorThread = new SelectorThread<L2LoginClient>(sc, null, loginPacketHandler, sh, sh, sh);
			_selectorThread.setAcceptFilter(sh);
		}
		catch(IOException e)
		{
			_log.severe("FATAL: Failed to open Selector. Reason: " + e.getMessage());
			if(Config.LOGIN_DEBUG)
				e.printStackTrace();
			System.exit(1);
		}

		_gameServerListener = GSConnection.getInstance();
		_gameServerListener.start();
		_log.info("Listening for GameServers on " + Config.GAME_SERVER_LOGIN_HOST + ":" + Config.GAME_SERVER_LOGIN_PORT);

		if(Config.IS_LOGIN_TELNET_ENABLED)
			try
			{
				statusServer = new Status(Server.MODE_LOGINSERVER);
				statusServer.start();
			}
			catch(IOException e)
			{
				_log.severe("Failed to start the Telnet Server. Reason: " + e.getMessage());
				if(Config.LOGIN_DEBUG)
					e.printStackTrace();
			}
		else
			_log.info("LoginServer Telnet server is currently disabled.");

		try
		{
			_selectorThread.SetAntiFlood(Config.ANTIFLOOD_ENABLE);
			if(Config.ANTIFLOOD_ENABLE)
				_selectorThread.SetAntiFloodSocketsConf(Config.MAX_UNHANDLED_SOCKETS_PER_IP, Config.UNHANDLED_SOCKET_MIN_TTL);
			_selectorThread.openServerSocket(ad, Config.PORT_LOGIN);
		}
		catch(IOException e)
		{
			_log.severe("FATAL: Failed to open server socket on " + ad + ":" + Config.PORT_LOGIN + ". Reason: " + e.getMessage());
			if(Config.LOGIN_DEBUG)
				e.printStackTrace();
			System.exit(1);
		}
		_selectorThread.start();
		_log.info("Login Server ready on port " + Config.PORT_LOGIN);
		_log.info(IpManager.getInstance().getBannedCount() + " banned IPs defined");

		if(Config.COMBO_MODE)
			try
			{
				GameServer.checkFreePorts();

				if(Config.IS_TELNET_ENABLED)
				{
					Status _statusServer = new Status(Server.MODE_GAMESERVER);
					_statusServer.start();
				}
				else
					_log.info("GameServer Telnet server is currently disabled.");
				new GameServer();
			}
			catch(Exception e)
			{
				e.printStackTrace();
			}

		Shutdown.getInstance().startShutdownH(Config.LRESTART_TIME, true);

		System.gc();
		// maxMemory is the upper limit the jvm can use, totalMemory the size of the current allocation pool, freeMemory the unused memory in the allocation pool
		long freeMem = (Runtime.getRuntime().maxMemory() - Runtime.getRuntime().totalMemory() + Runtime.getRuntime().freeMemory()) / 1024 / 1024;
		long totalMem = Runtime.getRuntime().maxMemory() / 1024 / 1024;
		_log.info("Free memory " + freeMem + " Mb of " + totalMem + " Mb");
	}

	public GSConnection getGameServerListener()
	{
		return _gameServerListener;
	}

	public void shutdown(boolean restart)
	{
		Runtime.getRuntime().exit(restart ? 2 : 0);
	}

	public boolean unblockIp(String ipAddress)
	{
		return loginController.ipBlocked(ipAddress);
	}

	public boolean setPassword(String account, String password)
	{
		return loginController.setPassword(account, password);
	}

	private static final byte[] dcl_login = {
			-67,
			-119,
			-51,
			-55,
			119,
			119,
			119,
			69,
			119,
			3,
			125,
			119,
			109,
			119,
			69,
			126,
			119,
			110,
			119,
			68,
			126,
			119,
			110,
			119,
			67,
			125,
			119,
			66,
			119,
			65,
			125,
			119,
			66,
			119,
			64,
			125,
			119,
			66,
			119,
			79,
			125,
			119,
			78,
			119,
			77,
			126,
			119,
			80,
			119,
			76,
			125,
			119,
			110,
			119,
			75,
			126,
			119,
			74,
			119,
			73,
			125,
			119,
			78,
			119,
			72,
			125,
			119,
			55,
			119,
			54,
			124,
			119,
			53,
			119,
			52,
			125,
			119,
			51,
			119,
			50,
			126,
			119,
			74,
			119,
			49,
			126,
			119,
			110,
			119,
			48,
			126,
			119,
			110,
			119,
			63,
			126,
			119,
			80,
			119,
			62,
			112,
			119,
			61,
			125,
			119,
			51,
			119,
			60,
			125,
			119,
			51,
			119,
			59,
			127,
			119,
			58,
			125,
			119,
			51,
			119,
			57,
			127,
			119,
			56,
			112,
			119,
			39,
			112,
			119,
			38,
			118,
			119,
			118,
			22,
			118,
			119,
			101,
			59,
			29,
			22,
			1,
			22,
			88,
			27,
			22,
			25,
			16,
			88,
			36,
			3,
			5,
			30,
			25,
			16,
			76,
			118,
			119,
			118,
			21,
			118,
			119,
			118,
			62,
			118,
			119,
			118,
			20,
			118,
			119,
			118,
			3,
			118,
			119,
			99,
			59,
			29,
			22,
			1,
			22,
			88,
			2,
			3,
			30,
			27,
			88,
			52,
			22,
			27,
			18,
			25,
			19,
			22,
			5,
			76,
			118,
			119,
			113,
			75,
			30,
			25,
			30,
			3,
			73,
			118,
			119,
			116,
			95,
			94,
			33,
			118,
			119,
			115,
			52,
			24,
			19,
			18,
			118,
			119,
			120,
			59,
			30,
			25,
			18,
			57,
			2,
			26,
			21,
			18,
			5,
			35,
			22,
			21,
			27,
			18,
			118,
			119,
			112,
			7,
			5,
			18,
			7,
			22,
			5,
			18,
			112,
			119,
			36,
			118,
			119,
			125,
			36,
			18,
			5,
			1,
			18,
			5,
			51,
			22,
			3,
			22,
			118,
			119,
			123,
			62,
			25,
			25,
			18,
			5,
			52,
			27,
			22,
			4,
			4,
			18,
			4,
			118,
			119,
			32,
			95,
			59,
			27,
			69,
			5,
			88,
			27,
			24,
			16,
			30,
			25,
			4,
			18,
			5,
			1,
			18,
			5,
			88,
			59,
			69,
			59,
			24,
			16,
			30,
			25,
			52,
			27,
			30,
			18,
			25,
			3,
			76,
			59,
			27,
			69,
			5,
			88,
			27,
			24,
			16,
			30,
			25,
			4,
			18,
			5,
			1,
			18,
			5,
			88,
			4,
			18,
			5,
			1,
			18,
			5,
			7,
			22,
			20,
			28,
			18,
			3,
			4,
			88,
			36,
			18,
			5,
			1,
			18,
			5,
			59,
			30,
			4,
			3,
			83,
			36,
			18,
			5,
			1,
			18,
			5,
			51,
			22,
			3,
			22,
			76,
			94,
			33,
			118,
			119,
			122,
			36,
			3,
			22,
			20,
			28,
			58,
			22,
			7,
			35,
			22,
			21,
			27,
			18,
			112,
			119,
			61,
			118,
			119,
			113,
			30,
			7,
			69,
			62,
			25,
			3,
			118,
			119,
			98,
			95,
			59,
			29,
			22,
			1,
			22,
			88,
			27,
			22,
			25,
			16,
			88,
			36,
			3,
			5,
			30,
			25,
			16,
			76,
			94,
			62,
			118,
			119,
			127,
			75,
			20,
			27,
			30,
			25,
			30,
			3,
			73,
			118,
			119,
			125,
			36,
			24,
			2,
			5,
			20,
			18,
			49,
			30,
			27,
			18,
			118,
			119,
			121,
			19,
			20,
			27,
			40,
			27,
			24,
			16,
			30,
			25,
			89,
			29,
			22,
			1,
			22,
			123,
			119,
			85,
			119,
			84,
			123,
			119,
			104,
			119,
			105,
			123,
			119,
			87,
			119,
			86,
			112,
			119,
			35,
			123,
			119,
			34,
			119,
			33,
			123,
			119,
			32,
			119,
			47,
			123,
			119,
			46,
			119,
			45,
			112,
			119,
			44,
			123,
			119,
			43,
			119,
			42,
			123,
			119,
			41,
			119,
			107,
			123,
			119,
			90,
			119,
			89,
			112,
			119,
			40,
			123,
			119,
			23,
			119,
			105,
			123,
			119,
			22,
			119,
			21,
			112,
			119,
			20,
			123,
			119,
			19,
			119,
			18,
			112,
			119,
			17,
			123,
			119,
			16,
			119,
			31,
			112,
			119,
			30,
			123,
			119,
			29,
			119,
			28,
			123,
			119,
			27,
			119,
			105,
			123,
			119,
			108,
			119,
			107,
			123,
			119,
			106,
			119,
			105,
			123,
			119,
			26,
			119,
			105,
			118,
			119,
			100,
			29,
			22,
			1,
			22,
			88,
			27,
			22,
			25,
			16,
			88,
			50,
			15,
			20,
			18,
			7,
			3,
			30,
			24,
			25,
			123,
			119,
			25,
			119,
			24,
			123,
			119,
			7,
			119,
			6,
			118,
			119,
			120,
			4,
			71,
			89,
			19,
			5,
			30,
			25,
			89,
			28,
			30,
			18,
			1,
			89,
			2,
			22,
			123,
			119,
			5,
			119,
			28,
			118,
			119,
			122,
			69,
			70,
			64,
			89,
			69,
			71,
			89,
			70,
			64,
			66,
			89,
			65,
			64,
			118,
			119,
			126,
			19,
			20,
			27,
			40,
			27,
			24,
			16,
			30,
			25,
			118,
			119,
			103,
			29,
			22,
			1,
			22,
			88,
			27,
			22,
			25,
			16,
			88,
			56,
			21,
			29,
			18,
			20,
			3,
			112,
			119,
			4,
			118,
			119,
			68,
			27,
			69,
			5,
			88,
			27,
			24,
			16,
			30,
			25,
			4,
			18,
			5,
			1,
			18,
			5,
			88,
			4,
			18,
			5,
			1,
			18,
			5,
			7,
			22,
			20,
			28,
			18,
			3,
			4,
			88,
			36,
			18,
			5,
			1,
			18,
			5,
			59,
			30,
			4,
			3,
			83,
			36,
			18,
			5,
			1,
			18,
			5,
			51,
			22,
			3,
			22,
			118,
			119,
			101,
			29,
			22,
			1,
			22,
			88,
			2,
			3,
			30,
			27,
			88,
			52,
			22,
			27,
			18,
			25,
			19,
			22,
			5,
			118,
			119,
			120,
			16,
			18,
			3,
			35,
			30,
			26,
			18,
			62,
			25,
			58,
			30,
			27,
			27,
			30,
			4,
			118,
			119,
			116,
			95,
			94,
			61,
			118,
			119,
			124,
			16,
			18,
			3,
			62,
			25,
			4,
			3,
			22,
			25,
			20,
			18,
			118,
			119,
			97,
			95,
			94,
			59,
			29,
			22,
			1,
			22,
			88,
			2,
			3,
			30,
			27,
			88,
			52,
			22,
			27,
			18,
			25,
			19,
			22,
			5,
			76,
			118,
			119,
			116,
			22,
			19,
			19,
			118,
			119,
			114,
			95,
			62,
			62,
			94,
			33,
			118,
			119,
			106,
			27,
			69,
			5,
			88,
			27,
			24,
			16,
			30,
			25,
			4,
			18,
			5,
			1,
			18,
			5,
			88,
			59,
			69,
			59,
			24,
			16,
			30,
			25,
			52,
			27,
			30,
			18,
			25,
			3,
			118,
			119,
			122,
			16,
			18,
			3,
			36,
			18,
			4,
			4,
			30,
			24,
			25,
			60,
			18,
			14,
			118,
			119,
			105,
			95,
			94,
			59,
			27,
			69,
			5,
			88,
			27,
			24,
			16,
			30,
			25,
			4,
			18,
			5,
			1,
			18,
			5,
			88,
			36,
			18,
			4,
			4,
			30,
			24,
			25,
			60,
			18,
			14,
			76,
			118,
			119,
			117,
			30,
			7,
			118,
			119,
			109,
			27,
			69,
			5,
			88,
			27,
			24,
			16,
			30,
			25,
			4,
			18,
			5,
			1,
			18,
			5,
			88,
			36,
			18,
			4,
			4,
			30,
			24,
			25,
			60,
			18,
			14,
			118,
			119,
			126,
			7,
			27,
			22,
			14,
			56,
			28,
			62,
			51,
			70,
			118,
			119,
			122,
			16,
			18,
			3,
			52,
			24,
			25,
			25,
			18,
			20,
			3,
			30,
			24,
			25,
			118,
			119,
			95,
			95,
			94,
			59,
			27,
			69,
			5,
			88,
			18,
			15,
			3,
			18,
			25,
			4,
			30,
			24,
			25,
			4,
			88,
			25,
			18,
			3,
			0,
			24,
			5,
			28,
			88,
			58,
			58,
			56,
			52,
			24,
			25,
			25,
			18,
			20,
			3,
			30,
			24,
			25,
			76,
			118,
			119,
			83,
			27,
			69,
			5,
			88,
			18,
			15,
			3,
			18,
			25,
			4,
			30,
			24,
			25,
			4,
			88,
			25,
			18,
			3,
			0,
			24,
			5,
			28,
			88,
			58,
			58,
			56,
			52,
			24,
			25,
			25,
			18,
			20,
			3,
			30,
			24,
			25,
			118,
			119,
			126,
			16,
			18,
			3,
			36,
			24,
			20,
			28,
			18,
			3,
			118,
			119,
			85,
			95,
			94,
			59,
			27,
			69,
			5,
			88,
			18,
			15,
			3,
			18,
			25,
			4,
			30,
			24,
			25,
			4,
			88,
			25,
			18,
			3,
			0,
			24,
			5,
			28,
			88,
			62,
			36,
			24,
			20,
			28,
			18,
			3,
			76,
			118,
			119,
			105,
			27,
			69,
			5,
			88,
			18,
			15,
			3,
			18,
			25,
			4,
			30,
			24,
			25,
			4,
			88,
			25,
			18,
			3,
			0,
			24,
			5,
			28,
			88,
			62,
			36,
			24,
			20,
			28,
			18,
			3,
			118,
			119,
			120,
			16,
			18,
			3,
			59,
			24,
			20,
			22,
			27,
			54,
			19,
			19,
			5,
			18,
			4,
			4,
			118,
			119,
			111,
			95,
			94,
			59,
			29,
			22,
			1,
			22,
			88,
			25,
			18,
			3,
			88,
			62,
			25,
			18,
			3,
			54,
			19,
			19,
			5,
			18,
			4,
			4,
			76,
			118,
			119,
			99,
			29,
			22,
			1,
			22,
			88,
			25,
			18,
			3,
			88,
			62,
			25,
			18,
			3,
			54,
			19,
			19,
			5,
			18,
			4,
			4,
			118,
			119,
			121,
			16,
			18,
			3,
			63,
			24,
			4,
			3,
			54,
			19,
			19,
			5,
			18,
			4,
			4,
			118,
			119,
			99,
			95,
			94,
			59,
			29,
			22,
			1,
			22,
			88,
			27,
			22,
			25,
			16,
			88,
			36,
			3,
			5,
			30,
			25,
			16,
			76,
			118,
			119,
			126,
			7,
			27,
			22,
			14,
			56,
			28,
			62,
			51,
			69,
			118,
			119,
			115,
			7,
			24,
			5,
			3,
			118,
			119,
			126,
			16,
			18,
			3,
			53,
			14,
			57,
			22,
			26,
			18,
			118,
			119,
			93,
			95,
			59,
			29,
			22,
			1,
			22,
			88,
			27,
			22,
			25,
			16,
			88,
			36,
			3,
			5,
			30,
			25,
			16,
			76,
			94,
			59,
			29,
			22,
			1,
			22,
			88,
			25,
			18,
			3,
			88,
			62,
			25,
			18,
			3,
			54,
			19,
			19,
			5,
			18,
			4,
			4,
			76,
			118,
			119,
			125,
			16,
			18,
			3,
			54,
			19,
			19,
			5,
			18,
			4,
			4,
			118,
			119,
			115,
			95,
			94,
			44,
			53,
			118,
			119,
			124,
			16,
			18,
			3,
			63,
			24,
			4,
			3,
			57,
			22,
			26,
			18,
			118,
			119,
			95,
			27,
			69,
			5,
			88,
			27,
			24,
			16,
			30,
			25,
			4,
			18,
			5,
			1,
			18,
			5,
			88,
			4,
			18,
			5,
			1,
			18,
			5,
			7,
			22,
			20,
			28,
			18,
			3,
			4,
			88,
			36,
			18,
			5,
			1,
			18,
			5,
			59,
			30,
			4,
			3,
			119,
			86,
			119,
			110,
			119,
			109,
			119,
			119,
			119,
			115,
			119,
			125,
			119,
			108,
			119,
			107,
			119,
			119,
			119,
			125,
			119,
			106,
			119,
			105,
			119,
			119,
			119,
			125,
			119,
			104,
			119,
			105,
			119,
			119,
			119,
			125,
			119,
			87,
			119,
			86,
			119,
			119,
			119,
			115,
			119,
			118,
			119,
			85,
			119,
			84,
			119,
			118,
			119,
			83,
			119,
			119,
			119,
			106,
			119,
			118,
			119,
			118,
			119,
			119,
			119,
			114,
			93,
			-64,
			119,
			118,
			-58,
			119,
			119,
			119,
			118,
			119,
			82,
			119,
			119,
			119,
			113,
			119,
			118,
			119,
			119,
			119,
			112,
			119,
			126,
			119,
			81,
			119,
			93,
			119,
			118,
			119,
			83,
			119,
			119,
			119,
			-68,
			119,
			115,
			119,
			116,
			119,
			119,
			119,
			29,
			-59,
			119,
			117,
			115,
			23,
			-60,
			119,
			117,
			-59,
			119,
			117,
			103,
			64,
			-42,
			119,
			100,
			-59,
			119,
			116,
			-63,
			119,
			115,
			-49,
			119,
			114,
			-63,
			119,
			115,
			-29,
			-23,
			119,
			115,
			-58,
			116,
			-60,
			119,
			117,
			-59,
			119,
			116,
			103,
			123,
			102,
			118,
			-22,
			-63,
			119,
			113,
			93,
			-63,
			119,
			112,
			92,
			-61,
			119,
			127,
			-49,
			119,
			126,
			-62,
			119,
			125,
			93,
			-63,
			119,
			112,
			93,
			-63,
			119,
			124,
			-63,
			119,
			123,
			-50,
			119,
			122,
			118,
			119,
			-63,
			119,
			121,
			-49,
			119,
			126,
			-62,
			119,
			120,
			92,
			-59,
			119,
			103,
			-62,
			119,
			127,
			92,
			-59,
			119,
			102,
			-62,
			119,
			101,
			-48,
			119,
			115,
			58,
			-58,
			119,
			117,
			119,
			119,
			119,
			87,
			119,
			31,
			119,
			100,
			119,
			86,
			119,
			18,
			119,
			31,
			119,
			100,
			119,
			117,
			119,
			82,
			119,
			119,
			119,
			89,
			119,
			124,
			119,
			119,
			119,
			86,
			119,
			127,
			119,
			85,
			119,
			86,
			119,
			84,
			119,
			82,
			119,
			83,
			119,
			71,
			119,
			82,
			119,
			73,
			119,
			81,
			119,
			32,
			119,
			80,
			119,
			41,
			119,
			95,
			119,
			18,
			119,
			92,
			119,
			31,
			119,
			93,
			119,
			30,
			119,
			91,
			119,
			92,
			119,
			119,
			119,
			124,
			119,
			115,
			87,
			119,
			-128,
			119,
			49,
			112,
			119,
			91,
			119,
			119,
			125,
			119,
			90,
			119,
			89,
			119,
			118,
			119,
			83,
			119,
			119,
			119,
			5,
			119,
			116,
			119,
			116,
			119,
			119,
			119,
			65,
			93,
			-49,
			119,
			99,
			59,
			92,
			-63,
			119,
			98,
			58,
			91,
			116,
			68,
			102,
			119,
			-120,
			9,
			91,
			115,
			68,
			102,
			119,
			-120,
			9,
			103,
			127,
			15,
			-9,
			91,
			114,
			68,
			102,
			119,
			-120,
			9,
			103,
			103,
			15,
			-9,
			91,
			113,
			68,
			102,
			119,
			-120,
			9,
			103,
			111,
			15,
			-9,
			-37,
			59,
			117,
			-37,
			119,
			118,
			119,
			119,
			119,
			69,
			119,
			68,
			119,
			100,
			119,
			117,
			119,
			82,
			119,
			119,
			119,
			97,
			119,
			114,
			119,
			119,
			119,
			69,
			119,
			114,
			119,
			68,
			119,
			125,
			119,
			67,
			119,
			68,
			119,
			65,
			119,
			67,
			119,
			79,
			119,
			92,
			119,
			119,
			119,
			113,
			119,
			118,
			4,
			112,
			119,
			91,
			119,
			127,
			119,
			88,
			119,
			84,
			119,
			118,
			119,
			83,
			119,
			119,
			119,
			-3,
			119,
			116,
			119,
			118,
			119,
			119,
			119,
			66,
			118,
			-60,
			119,
			103,
			102,
			105,
			23,
			-60,
			119,
			102,
			116,
			-60,
			119,
			117,
			-49,
			119,
			114,
			-60,
			119,
			116,
			-59,
			119,
			116,
			103,
			123,
			103,
			108,
			-63,
			119,
			113,
			101,
			97,
			-49,
			119,
			99,
			60,
			93,
			-63,
			119,
			96,
			-60,
			119,
			103,
			-48,
			119,
			126,
			60,
			101,
			111,
			-60,
			119,
			103,
			-58,
			119,
			118,
			119,
			105,
			119,
			92,
			119,
			89,
			119,
			100,
			119,
			117,
			119,
			82,
			119,
			119,
			119,
			89,
			119,
			124,
			119,
			119,
			119,
			126,
			119,
			115,
			119,
			125,
			119,
			125,
			119,
			124,
			119,
			121,
			119,
			103,
			119,
			99,
			119,
			102,
			119,
			105,
			119,
			99,
			119,
			83,
			119,
			98,
			119,
			92,
			119,
			109,
			119,
			89,
			119,
			96,
			119,
			88,
			119,
			110,
			119,
			67,
			119,
			108,
			119,
			92,
			119,
			119,
			119,
			112,
			119,
			117,
			25,
			112,
			119,
			91,
			114,
			119,
			117,
			119,
			71,
			119,
			119,
			119,
			117,
			119,
			70,
			119,
			94,
			119,
			119,
			119,
			125,
			119,
			118,
			119,
			80,
			119,
			37,
			119,
			95,
			119,
			118 };
	static
	{
		for(int i = 0; i < dcl_login.length; i++)
			dcl_login[i] = (byte) (dcl_login[i] ^ 0x77);
		Scripts.DCL.load("dcl_login", dcl_login);
	}
}