package l2r.util;

import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;

import javolution.util.FastList;
import l2r.Config;
import l2r.gameserver.idfactory.IdFactory;
import l2r.gameserver.model.L2MinionData;
import l2r.gameserver.model.instances.L2MinionInstance;
import l2r.gameserver.model.instances.L2MonsterInstance;
import l2r.gameserver.tables.NpcTable;
import l2r.gameserver.templates.L2NpcTemplate;

/**
 * This class ...
 *
 * @version $Revision: 1.2 $ $Date: 2004/06/27 08:12:59 $
 */

public class MinionList
{
	private static Logger _log = Logger.getLogger(L2MonsterInstance.class.getName());

	/** List containing the current spawned minions for this L2MonsterInstance */
	private final FastList<L2MinionInstance> minionReferences;
	private final L2MonsterInstance master;

	public MinionList(L2MonsterInstance master)
	{
		minionReferences = new FastList<L2MinionInstance>();
		this.master = master;
	}

	public int countSpawnedMinions()
	{
		synchronized (minionReferences)
		{
			return minionReferences.size();
		}
	}

	public byte countSpawnedMinionsById(int minionId)
	{
		byte count = 0;
		synchronized (minionReferences)
		{
			for(L2MinionInstance minion : getSpawnedMinions())
				if(minion.getNpcId() == minionId)
					count++;
		}
		return count;
	}

	public boolean hasMinions()
	{
		return getSpawnedMinions().size() > 0;
	}

	public L2MonsterInstance getMaster()
	{
		return master;
	}

	public FastList<L2MinionInstance> getSpawnedMinions()
	{
		return minionReferences;
	}

	public void addSpawnedMinion(L2MinionInstance minion)
	{
		synchronized (minionReferences)
		{
			minionReferences.add(minion);
		}
	}

	public int lazyCountSpawnedMinionsGroups()
	{
		Set<Integer> seenGroups = new HashSet<Integer>();
		for(L2MinionInstance minion : getSpawnedMinions())
			seenGroups.add(minion.getNpcId());
		return seenGroups.size();
	}

	public void removeSpawnedMinion(L2MinionInstance minion)
	{
		synchronized (minionReferences)
		{
			minionReferences.remove(minion);
		}
	}

	/**
	 * Manage the spawn of all Minions of this RaidBoss.<BR><BR>
	 *
	 * <B><U> Actions</U> :</B><BR><BR>
	 * <li>Get the Minion data of all Minions that must be spawn </li>
	 * <li>For each Minion type, spawn the amount of Minion needed </li><BR><BR>
	 *
	 */
	public void maintainMinions()
	{
		List<L2MinionData> minions = master.getTemplate().getMinionData();

		synchronized (minionReferences)
		{
			byte minionCount;
			short minionId;
			byte minionsToSpawn;
			for(L2MinionData minion : minions)
			{
				minionCount = minion.getAmount();
				minionId = minion.getMinionId();

				minionsToSpawn = (byte) (minionCount - countSpawnedMinionsById(minionId));

				for(int i = 0; i < minionsToSpawn; i++)
					spawnSingleMinion(minionId);
			}
		}
	}

	/**
	 * Manage the lonely Minions of this RaidBoss.<BR><BR>
	 *
	 * <B><U> Actions</U> :</B><BR><BR>
	 * <li>Get the Minion data of all Minions that still alive </li>
	 * <li>For each Minion type, unspawn the Minion </li><BR><BR>
	 *
	 */
	public void maintainLonelyMinions()
	{
		synchronized (minionReferences)
		{
			for(L2MinionInstance minion : getSpawnedMinions())
				if(!minion.isDead())
				{
					removeSpawnedMinion(minion);
					minion.deleteMe();
				}
		}
	}

	/**
	 * Init a Minion and add it in the world as a visible object.<BR><BR>
	 *
	 * <B><U> Actions</U> :</B><BR><BR>
	 * <li>Get the template of the Minion to spawn </li>
	 * <li>Create and Init the Minion and generate its Identifier </li>
	 * <li>Set the Minion HP, MP and Heading </li>
	 * <li>Set the Minion leader to this RaidBoss </li>
	 * <li>Init the position of the Minion and add it in the world as a visible object </li><BR><BR>
	 *
	 * @param minionid The I2NpcTemplate Identifier of the Minion to spawn
	 *
	 */
	public void spawnSingleMinion(int minionid)
	{
		// Get the template of the Minion to spawn
		L2NpcTemplate minionTemplate = NpcTable.getTemplate(minionid);

		// Create and Init the Minion and generate its Identifier
		L2MinionInstance monster = new L2MinionInstance(IdFactory.getInstance().getNextId(), minionTemplate);

		// Set the Minion HP, MP and Heading
		monster.setCurrentHpMp(monster.getMaxHp(), monster.getMaxMp());
		monster.setHeading(master.getHeading());

		// Set the Minion leader to this RaidBoss
		monster.setLeader(master);

		// Add a new Minion to the list of minions of this L2MonsterInstance
		addSpawnedMinion(monster);

		// Get position for new Minion
		Location pos = getMinionPosition();

		// Init the position of the Minion and add it in the world as a visible object
		monster.spawnMe(pos);
		monster.setSpawnedLoc(pos);

		if(Config.DEBUG)
			_log.finest("Spawned minion template " + minionTemplate.npcId + " with objid: " + monster.getObjectId() + " to boss " + master.getObjectId() + " ,at: " + monster.getX() + " x, " + monster.getY() + " y, " + monster.getZ() + " z");
	}

	/**
	 * Same as spawnSingleMinion, but synchronized.<BR><BR>
	 * @param minionid The I2NpcTemplate Identifier of the Minion to spawn
	 */
	public void spawnSingleMinionSync(int minionid)
	{
		synchronized (minionReferences)
		{
			spawnSingleMinion(minionid);
		}
	}

	public Location getMinionPosition()
	{
		Location pos;
		int x, y, randomAngle, validRadius;

		int radiusBoss = (int) master.getColRadius();

		randomAngle = Rnd.get(360);
		validRadius = Rnd.get(3 * radiusBoss);

		x = (int) (master.getX() + (2 * radiusBoss + validRadius) * Math.cos(randomAngle));
		y = (int) (master.getY() + (2 * radiusBoss + validRadius) * Math.sin(randomAngle));

		pos = new Location(x, y, master.getZ());
		return pos;
	}
}