package l2r.gameserver.network;

import static l2r.gameserver.network.L2GameClient.GameClientState.IN_GAME;

import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.util.concurrent.RejectedExecutionException;
import java.util.logging.Logger;

import l2r.Config;
import l2r.extensions.network.HeaderInfo;
import l2r.extensions.network.IClientFactory;
import l2r.extensions.network.IMMOExecutor;
import l2r.extensions.network.IPacketHandler;
import l2r.extensions.network.MMOConnection;
import l2r.extensions.network.ReceivablePacket;
import l2r.extensions.network.TCPHeaderHandler;
import l2r.gameserver.ThreadPoolManager;
import l2r.gameserver.clientpackets.*;
import l2r.gameserver.model.L2Player;
import l2r.util.Log;

/**
 * Stateful Packet Handler<BR>
 * The Stateful approach prevents the server from handling inconsistent packets, examples:<BR>
 * <li>Clients sends a MoveToLocation packet without having a character attached. (Potential errors handling the packet).</li>
 * <li>Clients sends a RequestAuthLogin being already authed. (Potential exploit).</li>
 * <BR><BR>
 * Note: If for a given exception a packet needs to be handled on more then one state, then it should be added to all these states.
 */
public final class L2GamePacketHandler extends TCPHeaderHandler<L2GameClient> implements IPacketHandler<L2GameClient>, IClientFactory<L2GameClient>, IMMOExecutor<L2GameClient>
{
	private static final Logger _log = Logger.getLogger(L2GamePacketHandler.class.getName());

	/**
     */
	public L2GamePacketHandler()
	{
		super(null);
	}

	// implementation
	public ReceivablePacket<L2GameClient> handlePacket(ByteBuffer data, L2GameClient client)
	{
		//ByteBuffer temp = data.duplicate();
		//int tid = temp.get() & 0xff;
		//int tsize = temp.remaining();
		//_log.warning("Packet: " + Integer.toHexString(tid));
		//byte[] tarray = new byte[tsize];
		//temp.get(tarray);
		//_log.warning(printData(tarray, tsize));
		L2Player activeChar;
		int id = data.get() & 0xFF;
		if(client == null)
			return null;
		if(client.getUPTryes() > 4)
			return null;

		ReceivablePacket<L2GameClient> msg = null;
		int id2;
		int id3;

		switch(client.getState())
		{
			case CONNECTED:
				switch(id)
				{
					case 0x00:
						msg = new ProtocolVersion();
						break;
					case 0x08:
						msg = new AuthLogin();
						break;
                    case 0x09:
						msg = new Logout();
						break;
					case 0xca:
						msg = new GameGuardReply();
						break;
					default:
						_log.severe("Unknown packet on state: CONNECTED, id: " + Integer.toHexString(id) + " from " + client.getConnection().getSocket().getInetAddress().getHostAddress());
						if(Config.PROTECT_ENABLE && !client.HWID.isEmpty())
							_log.severe("Client HWID: " + client.HWID);
						break;
				}
				break;
			case AUTHED:
				switch(id)
				{
					case 0x09:
						msg = new Logout();
						break;
                    case 0x0b:
						msg = new CharacterCreate(); //RequestCharacterCreate();
						break;
                    case 0x0c:
						msg = new CharacterDelete(); //RequestCharacterDelete();
						break;					
					case 0x0d:
						msg = new CharacterSelected(); //CharacterSelect();
						break;
					case 0x0e:
						msg = new NewCharacter(); //RequestNewCharacter();
						break;
					case 0x62:
						msg = new CharacterRestore(); //RequestCharacterRestore();
						break;
					case 0xca:
						msg = new GameGuardReply();
						break;
					default:
						_log.severe("Unknown packet on state: AUTHED, id: " + Integer.toHexString(id));
						break;
				}
				break;
			case IN_GAME:
				switch(id)
				{
					case 0x01:
						msg = new MoveBackwardToLocation();
						break;
                    case 0x02:
						//	msg = new ?();
						break;
                    case 0x03:
						activeChar = client.getActiveChar();
						if(activeChar != null && activeChar.isBotLevel() > 0)
							activeChar.setIsBot(1);
						msg = new EnterWorld();
						break;
					case 0x04:
						msg = new Action();
						break;
					case 0x05:
						msg = new RequestStopPledgeWar();
						break;
					case 0x06:
						//	msg = RequestSCCheck(); // ? Format: cdx
						break;
					case 0x07:
						msg = new GameGuardReply();
						// сдесь совсем другой пакет ResponseAuthGameGuard[cddddd] (c) Drin
						break;
					case 0x08:
						msg = new AuthLogin();
						break;
                    case 0x09:
						msg = new Logout();
						break;
                    case 0x0a:
						msg = new AttackRequest();
						break;
                    case 0x0f:
						msg = new RequestItemList();
						break;
                    case 0x10:
                        //msg = new RequestEquipItem();
                        break;
                    case 0x11:
						msg = new RequestUnEquipItem();
						break;
					case 0x12:
						msg = new RequestDropItem();
						break;
					case 0x14:
						msg = new UseItem(); //RequestUseItem();
						break;
                    case 0x15:
						msg = new TradeRequest();
						break;
					case 0x16:
						msg = new AddTradeItem();
						break;
					case 0x17:
						msg = new TradeDone(); //RequestTradeDone();
						break;
                    case 0x18:
                        //  msg = new?();
                        break;
                    case 0x19:
                        //  msg = new ?();
                        break;
                    case 0x1a:
				        msg = new DummyPacket();
				        break;
			        case 0x1b:
				        msg = new RequestSocialAction();
				        break;
                    case 0x1c:
				        msg = new ChangeMoveType2();
				        break;
					case 0x1d:
						msg = new ChangeWaitType2();
						break;
					case 0x1e:
						msg = new RequestSellItem();
						break;
					case 0x1f:
						msg = new RequestBuyItem();
						break;
					case 0x20:
						msg = new RequestLinkHtml();
						break;
					case 0x21:
						msg = new RequestBypassToServer();
						break;
					case 0x22:
						msg = new RequestBBSwrite();
						break;
					case 0x23:
						msg = new RequestCreatePledge();
						break;
					case 0x24:
						msg = new RequestJoinPledge(); //RequestBBSWrite();
						break;
					case 0x25:
						msg = new RequestAnswerJoinPledge();
						break;
					case 0x26:
						msg = new RequestWithdrawalPledge();
						break;
					case 0x27:
						msg = new RequestOustPledgeMember();
						break;
                    case 0x28:
                         //msg = new RequestDismissPledge();
                        break;
					case 0x29:
						msg = new RequestJoinParty();
						break;
					case 0x2a:
						msg = new RequestAnswerJoinParty();
						break;
                    case 0x2b:
						msg = new RequestWithDrawalParty();
						break;
					case 0x2c:
						msg = new RequestOustPartyMember();
						break;
                    case 0x2d:
						//	msg = new RequestDismissParty(); // Format: c ?
						break;
                    case 0x2e:
						msg = new RequestMagicSkillList();
						break;
                    case 0x2f:
						msg = new RequestMagicSkillUse();
						break;
					case 0x30:
						 msg = new Appearing();
						break;
					case 0x31:
                        if(Config.ALLOW_WAREHOUSE)
						msg = new SendWareHouseDepositList();
						break;
					case 0x32:
						msg = new SendWareHouseWithDrawList();
						break;
					case 0x33:
						msg = new RequestShortCutReg();
						break;
					case 0x35:
						msg = new RequestShortCutDel();
						break;
					case 0x36:
						msg = new CannotMoveAnymore();
						break;
					case 0x37:
						msg = new RequestTargetCanceld();
						break;
                    case 0x38:
                        msg = new Say2C();
                        break;                
					case 0x3c:
						msg = new RequestPledgeMemberList();
						break;
					case 0x3e:
						msg = new RequestMagicSkillList();
						break;
					case 0x3f:
						msg = new RequestSkillList();
						break;
					case 0x40:
						msg = new AnswerTradeRequest();
						break;
                    case 0x41:
						//msg = new MoveWithDelta();    ... unused ?? or only on ship ??
						break;
					case 0x42:
						msg = new RequestGetOnVehicle();
						break;
					case 0x43:
						msg = new RequestGetOffVehicle();
						break;
					case 0x44:
						msg = new AnswerTradeRequest();
						break;
					case 0x45:
						msg = new RequestActionUse();
						break;
                    case 0x46:
                        msg = new RequestRestart();
                        break;
					case 0x47:
						// msg = new RequestSiegeInfo();
						break;
					case 0x48:
						msg = new ValidatePosition(); //RequestTargetCancel();
						break;
                    case 0x4a:
                        msg = new StartRotatingC();
                        break;
					case 0x4b:
						msg = new FinishRotating();
						break;
					case 0x4c:
						//	msg = new ?();
						break;
                    case 0x4d:
						msg = new RequestStartPledgeWar();
						break;
					case 0x4e:
						// msg = new ?();
						break;
					case 0x4f:
						msg = new RequestStopPledgeWar(); // Format: c ?
						break;
					// -- maybe GM packet's
					case 0x39:
						if(!data.hasRemaining())
						{
							handleIncompletePacket(client);
							break;
						}
						id2 = data.get() & 0xff;
						switch(id2)
						{
							case 0x00:
								//	msg = new SendCharacterInfo(); // Format: ccS ?
								break;
							case 0x01:
								//	msg = new SendSummonCmd(); // Format: ccS ?
								break;
							case 0x02:
								//	msg = new SendServerStatus(); // Format: cc ?
								break;
							case 0x03:
								//	msg = new SendL2ParamSetting(); // Format: ccdd ?
								break;
							default:
								activeChar = client.getActiveChar();
								int size = data.remaining();
								_log.warning("Unknown Packet: 0x39:" + Integer.toHexString(id2) + ", from ip: " + client.getIpAddr() + ", Char: " + activeChar != null ? activeChar.toString() : "null" + ", Login: " + client.getLoginName());
								Log.add("Unknown Packet: 0x39:" + Integer.toHexString(id2) + ", from ip: " + client.getIpAddr() + ", Char: " + activeChar != null ? activeChar.toString() : "null" + ", Login: " + client.getLoginName(), "unknown_packets");
								byte[] array = new byte[size];
								data.get(array);
								_log.warning(printData(array, size));
								client.addUPTryes();
								client.closeNow(false);
								break;
						}
						break;
					case 0x50:
						//	msg = new RequestReplyStopPledgeWar(); // Format: c ?
						break;
					case 0x51:
						//	msg = new RequestSurrendPledgeWar();
						break;
					case 0x52:
						//	msg = new RequestReplySurrendPledgeWar(); // Format: cddd ?
						break;
                    case 0x53:
						msg = new RequestSetPledgeCrest();
						break;
					case 0x54:
						//msg = new ?();
						break;
                    case 0x55:
						msg = new RequestGiveNickName();
						break;
					case 0x56:
						//msg = new RequestActionUse();
						break;
					case 0x57:
						msg = new RequestShowBoard();
						break;
					case 0x58:
						msg = new RequestEnchantItem(); // Format: c ?
						break;
					case 0x59:
						msg = new RequestDestroyItem();
						break;
					case 0x5a:
						//	msg = new RequestSEKCustom(); // ?
						// Format: cdd, SlotNum : %d Direction : %d
						break;
					case 0x5b:
						msg = new SendBypassBuildCmd();
						break;
					case 0x5c:
						msg = new RequestGetOnVehicle();
						break;
					case 0x5d:
						msg = new RequestGetOffVehicle();
						break;
					case 0x5e:
						msg = new RequestFriendInvite();
						break;
					case 0x5f:
						msg = new RequestAnswerFriendInvite();
						break;
					case 0x60:
						msg = new RequestFriendList();
						break;
					case 0x61:
						msg = new RequestFriendDel();
						break;
					case 0x62:
						msg = new CharacterRestore();
						break;
					case 0x63:
						msg = new RequestQuestList(); //RequestDestroyQuest();
						break;
					case 0x64:
						msg = new RequestQuestAbort();
						break;
					case 0x65:
						// msg = new RequestPledgeInfo();
						break;
					case 0x66:
						msg = new RequestPledgeInfo();
						break;
					case 0x67:
						//msg = new RequestPledgeExtendedInfo();
						break;
					case 0x68:
						msg = new RequestPledgeCrest();
						break;
					case 0x69:
						//msg = new RequestSurrenderPersonally();
						break;
					case 0x6a:
						//msg = new RequestRide();
						break;
					case 0x6b:
						msg = new RequestAquireSkillInfo();
						break;
					case 0x6c:
						msg = new RequestAquireSkill();
						break;
					case 0x6d:
						msg = new RequestRestartPoint();
						break;
					case 0x6e:
						msg = new RequestGMCommand();
						break;
					case 0x6f:
						msg = new RequestPartyMatchConfig();
						break;
					case 0x70:
						msg = new RequestPartyMatchList();
						break;
					case 0x71:
						msg = new RequestPartyMatchDetail();
						break;
					case 0x72:
						msg = new RequestCrystallizeItem(); // ?
						break;
					case 0x73:
						msg = new RequestPrivateStoreManage();
						break;
					case 0x74:
						msg = new SetPrivateStoreList();
						break;
					case 0x75:
						// msg = new RequestPrivateStoreManageCancel();
						break;
					case 0x76:
						msg = new RequestPrivateStoreQuitSell();
						break;
					case 0x77:
						msg = new SetPrivateStoreMsgSell();
						break;
					case 0x78:
						//msg = new RequestPrivateStoreList();
						break;
					case 0x79:
						msg = new SendPrivateStoreBuyBuyList();
						break;
					case 0x7a:
						//msg = new ReviveReplay();
						break;
                    case 0x7b:
                        msg = new RequestTutorialLinkHtml();
                        break;
					case 0x7c:
						msg = new RequestTutorialPassCmdToServer();
						break;
					case 0x7d:
						msg = new RequestTutorialQuestionMark();
						break;
					case 0x7e:
						msg = new RequestTutorialClientEvent();
						break;
					case 0x7f:
						msg = new RequestPetition();
						break;
					case 0x80:
						msg = new RequestPetitionCancel();
						break;
					case 0x81:
						msg = new RequestGmList();
						break;
					case 0x82:
						msg = new RequestJoinAlly();
						break;
					case 0x83:
						msg = new RequestAnswerJoinAlly();
						break;
					case 0x84:
						msg = new AllyLeave();
						break;
					case 0x85:
						msg = new AllyDismiss();
						break;
					case 0x86:
						msg = new RequestDismissAlly();
						break;
					case 0x87:
						msg = new RequestSetAllyCrest();
						break;
					case 0x88:
						msg = new RequestAllyCrest();
						break;
					case 0x89:
						msg = new RequestChangePetName();
						break;
					case 0x8a:
						msg = new RequestPetUseItem();
						break;
					case 0x8b:
						msg = new RequestGiveItemToPet();
						break;
					case 0x8c:
						msg = new RequestGetItemFromPet();
						break;
                    case 0x8e:
						msg = new RequestAllyInfo();
						break;
					case 0x8f:
						msg = new RequestPetGetItem();
						break;
					case 0x90:
						msg = new RequestPrivateStoreBuyManage();
						break;
					case 0x91:
						msg = new SetPrivateStoreListBuy();
						break;
					case 0x92:
						//msg = new RequestPrivateStoreBuyManageCancel();
						break;
					case 0x93:
						msg = new RequestPrivateStoreQuitBuy();
						break;
					case 0x94:
						msg = new SetPrivateStoreMsgBuy();
						break;
					case 0x95:
						//msg = new RequestPrivateStoreBuyList();
						break;
					case 0x96:
						msg = new RequestSellToPrivateStoreBuy(); // ?
						break;
					case 0x97:
						//msg = new SendTimeCheckPacket();
						break;
					case 0x98:
				    //msg = new RequestStartAllianceWar();
				        if(client.getActiveChar() != null)
					        _log.warning("Client " + client.getLoginName() + "/" + client.getActiveChar().getName() + " sends obsolete RequestStartAllianceWar packet");
				        break;
			        case 0x99:
				        //msg = new RequestReplyStartAllianceWar();
				        if(client.getActiveChar() != null)
					        _log.warning("Client " + client.getLoginName() + "/" + client.getActiveChar().getName() + " sends obsolete RequestReplyStartAllianceWar packet");
				        break;
			        case 0x9a:
				        //msg = new RequestStopAllianceWar();
				        if(client.getActiveChar() != null)
				        	_log.warning("Client " + client.getLoginName() + "/" + client.getActiveChar().getName() + " sends obsolete RequestStopAllianceWar packet");
			        	break;
			        case 0x9b:
	        			//msg = new RequestReplyStopAllianceWar();
   	        			if(client.getActiveChar() != null)
           					_log.warning("Client " + client.getLoginName() + "/" + client.getActiveChar().getName() + " sends obsolete RequestReplyStopAllianceWar packet");
        				break;
		        	case 0x9c:
				        //msg = new RequestSurrenderAllianceWar();
				        if(client.getActiveChar() != null)
				        	_log.warning("Client " + client.getLoginName() + "/" + client.getActiveChar().getName() + " sends obsolete RequestSurrenderAllianceWar packet");
				        break;
			        case 0x9d:
				        // RequestSkillCoolTime
				        activeChar = client.getActiveChar();
						if(activeChar != null)
						activeChar.startBotCheck();
						msg = !Config.KILL_BOTS_WITH_SKILLCOOLTIME && activeChar != null && activeChar.isBot() ? null : new RequestSkillCoolTime();
                        break;
					case 0x9e:
						msg = new RequestPackageSendableItemList(); // ?
						break;
					case 0x9f:
						msg = new RequestPackageSend(); // ?
						break;
					case 0xa0:
						msg = new RequestBlock();
						break;
					case 0xa1:
						//msg = new RequestCastleSiegeInfo();
						break;
					case 0xa2:
						msg = new RequestSiegeAttackerList();
						break;
					case 0xa3:
						msg = new RequestSiegeDefenderList();
						break;
					case 0xa4:
						msg = new RequestJoinSiege();
						break;
					case 0xa5:
						msg = new RequestConfirmSiegeWaitingList();
						break;
					case 0xa6:
						//msg = RequestSetCastleSiegeTime();
						break;
					case 0xa7:
						msg = new RequestMultiSellChoose();
						break;
					case 0xa8:
						//msg = new NetPing();
						break;
					case 0xa9:
						//msg = new RequestRemainTime();
						break;
					case 0xaa:
						msg = new BypassUserCmd(); // format: cd ?
						break;
					case 0xab:
						msg = new SnoopQuit();
						break;
					case 0xac:
						msg = new RequestRecipeBookOpen();
						break;
					case 0xad:
						msg = new RequestRecipeBookDestroy();
						break;
					case 0xae:
						msg = new RequestRecipeItemMakeInfo();
						break;
					case 0xaf:
						msg = new RequestRecipeItemMakeSelf();
						break;
					case 0xb0:
						msg = new RequestRecipeShopManageList();
						break;
					case 0xb1:
						msg = new RequestRecipeShopMessageSet(); // format: cddd ?
						break;
					case 0xb2:
						msg = new RequestRecipeShopListSet(); //Format: c
						break;
					case 0xb3:
						msg = new RequestRecipeShopManageQuit();
						break;
					case 0xb4:
						msg = new RequestRecipeShopManageCancel();
						break;
					case 0xb5:
						msg = new RequestRecipeShopMakeInfo();
						break;
					case 0xb6:
						msg = new RequestRecipeShopMakeItem();
						break;
					case 0xb7:
						msg = new RequestRecipeShopManagePrev();
						break;
					case 0xb8:
						msg = new ObserverReturn();
						break;
					case 0xb9:
						msg = new RequestEvaluate();
						break;
					case 0xba:
						msg = new RequestHennaList();
						break;
					case 0xbb:
						msg = new RequestHennaItemInfo();
						break;
					case 0xbc:
						msg = new RequestHennaEquip();
						break;
					case 0xbd:
						//msg = new RequestHennaUnequipList();
						break;
					case 0xbe:
						//msg = new RequestHennaUnequipInfo();
						break;
					case 0xbf:
						//msg = new RequestHennaUnequip();
						break;
					case 0xc0:
						msg = new RequestPledgePower(); // format: cd ?
						break;
					case 0xc1:
						msg = new RequestMakeMacro(); //RequestObserverEndPacket();
						break;
					case 0xc2:
						msg = new RequestDeleteMacro(); // Recommend
						break;
					case 0xc3:
						msg = new RequestBuyProcure(); //RequestHennaItemList();
						break;
					case 0xc4:
						msg = new RequestBuySeed();
						break;
					case 0xc5:
						msg = new DlgAnswer();
						break;
					case 0xc6:
						msg = new RequestPreviewItem();
						break;
					case 0xc7:
						msg = new RequestSSQStatus();
						break;
					case 0xc8:
						//msg = new PetitionVote();
						break;
					case 0xca:
						msg = new GameGuardReply();
						break;
					case 0xcc:
						msg = new RequestSendFriendMsg();
						break;
					case 0xcd:
						msg = new RequestShowMiniMap();
						break;
					case 0xce:
						//msg = new RequestSendMsnChatLog();
						break;
					case 0xcf:
						msg = new RequestRecordInfo(); // ?
						break;
			case 0xd0: // 656 protocol
				id2 = data.getShort() & 0xffff;
				switch(id2)
				{
					case 0x01:
						msg = new RequestOustFromPartyRoom();
						break;
					case 0x02:
						msg = new RequestDismissPartyRoom();
						break;
					case 0x03:
						msg = new RequestWithdrawPartyRoom();
						break;
					case 0x04:
						msg = new RequestChangePartyLeader();
						break;
					case 0x05:
						msg = new RequestAutoSoulShot();
						break;
					case 0x06:
						msg = new RequestExEnchantSkillInfo();
						break;
					case 0x07:
						msg = new RequestExEnchantSkill();
						break;
					case 0x08:
						if(client.getActiveChar() != null)
							if(client.getActiveChar().isBotLevel() == 2)
								client.getActiveChar().setIsBot(0);
							else if(client.getActiveChar().isBotLevel() == 1)
								client.getActiveChar().setIsBot(2);
						msg = new RequestManorList();
						break;
					case 0x09:
						msg = new RequestProcureCropList();
						break;
					case 0x0a:
						msg = new RequestSetSeed();
						break;
					case 0x0b:
						msg = new RequestSetCrop();
						break;
					case 0x0c:
						msg = new RequestWriteHeroWords();
						break;
					case 0x0d:
						msg = new RequestExMPCCAskJoin();
						break;
					case 0x0e:
						msg = new RequestExMPCCAcceptJoin();
						break;
					case 0x0f:
						msg = new RequestExMPCCExit();
						break;
					case 0x10:
						msg = new RequestPledgeCrestLarge();
						break;
					case 0x11:
						msg = new RequestSetPledgeCrestLarge();
						break;
					case 0x12:
						msg = new RequestOlympiadObserverEnd();
						break;
					case 0x13:
						msg = new RequestOlympiadMatchList();
						break;
					case 0x16:
						//Interlude РћС‚РєСЂС‹С‚РёРµ РѕРєРЅР° Waiting List РІ Party
						int sz = data.remaining();
						_log.warning("Party Waiting List:" + Integer.toHexString(id));
						byte[] arr = new byte[sz];
						data.get(arr);
						_log.warning(printData(arr, sz));
						break;
					case 0x17:
						//Interlude Р—Р°РєСЂС‹С‚РёРµ РѕРєРЅР° Party
						break;
					case 0x18:
						msg = new RequestGetBossRecord();
						break;
					case 0x19:
						msg = new RequestPledgeSetAcademyMaster();
						break;
					case 0x1a:
						msg = new RequestPledgePowerGradeList();
						break;
					case 0x1b:
						msg = new RequestPledgeMemberPowerInfo();
						break;
					case 0x1c:
						msg = new RequestPledgeSetMemberPowerGrade();
						break;
					case 0x1d: // target clan member
						msg = new RequestPledgeMemberInfo();
						break;
					case 0x1e:
						msg = new RequestPledgeWarList();
						break;
					case 0x22:
						msg = new RequestCursedWeaponList();
						break;
					case 0x23:
						msg = new RequestCursedWeaponLocation();
						break;
					case 0x24:
						msg = new RequestPledgeReorganizeMember();
						break;
					// couldnt find it 0x25 :S
					case 0x26:
						msg = new RequestExMPCCShowPartyMembersInfo();
						break;
					case 0x27:
						msg = new RequestDuelStart(data, client);
						break;
					case 0x28:
						msg = new RequestDuelAnswerStart(data, client);
						break;
					case 0x29:
						msg = new RequestConfirmTargetItem();
						break;
					case 0x2a:
						msg = new RequestConfirmRefinerItem();
						break;
					case 0x2b:
						msg = new RequestConfirmGemStone();
						break;
					case 0x2c:
						msg = new RequestRefine();
						break;
					case 0x2d:
						msg = new RequestConfirmCancelItem();
						break;
					case 0x2e:
						msg = new RequestRefineCancel();
						break;
					//case 0x2f:
						//msg = new RequestExMagicSkillUseGround();
						//break;
					case 0x30:
						msg = new RequestDuelSurrender(data, client);
						break;
                    default:
							int size = data.remaining();
							activeChar = client.getActiveChar();
							_log.warning("Unknown Packet: 0xd0:" + Integer.toHexString(id2) + ", from ip: " + client.getIpAddr() + ", Char: " + activeChar != null ? activeChar.toString() : "null" + ", Login: " + client.getLoginName());
							Log.add("Unknown Packet: 0xd0:" + Integer.toHexString(id2) + ", from ip: " + client.getIpAddr() + ", Char: " + activeChar != null ? activeChar.toString() : "null" + ", Login: " + client.getLoginName(), "unknown_packets");
							if(Config.PROTECT_ENABLE && !client.HWID.isEmpty())
								_log.severe("Client HWID: " + client.HWID);
							byte[] array = new byte[size];
							data.get(array);
							_log.warning(printData(array, size));
							client.addUPTryes();
							client.closeNow(false);
							break;
						}
						break;

					default:
					{
						try
						{
							int sz = data.remaining();
							activeChar = client.getActiveChar();
							_log.warning("Unknown Packet:" + Integer.toHexString(id) + ", from ip: " + client.getIpAddr() + ", Char: " + activeChar != null ? activeChar.toString() : "null" + ", Login: " + client.getLoginName());
							byte[] arr = new byte[sz];
							data.get(arr);
							System.out.println("packet " + Integer.toHexString(id) + " on " + client.getState().toString());
							if(Config.PROTECT_ENABLE && !client.HWID.isEmpty())
								_log.severe("Client HWID: " + client.HWID);
							_log.warning(printData(arr, sz));
							client.addUPTryes();
							client.closeNow(false);
						}
						catch(Exception e)
						{
							e.printStackTrace();
						}
						break;
					}
				}
				break;
		}
		return msg;
	}

	// impl
	public L2GameClient create(MMOConnection<L2GameClient> con)
	{
		return new L2GameClient(con);
	}

	public void execute(ReceivablePacket<L2GameClient> rp)
	{
		try
		{
			if(rp.getClient().getState() == IN_GAME)
				ThreadPoolManager.getInstance().executePacket(rp);
			else
				ThreadPoolManager.getInstance().executeIOPacket(rp);
		}
		catch(RejectedExecutionException e)
		{
			// if the server is shutdown we ignore
			if(!ThreadPoolManager.getInstance().isShutdown())
				_log.severe("Failed executing: " + rp.getClass().getSimpleName() + " for Client: " + rp.getClient().toString());
		}
	}

	public static String printData(byte[] data, int len)
	{
		StringBuffer result = new StringBuffer();

		int counter = 0;

		for(int i = 0; i < len; i++)
		{
			if(counter % 16 == 0)
				result.append(fillHex(i, 4) + ": ");

			result.append(fillHex(data[i] & 0xff, 2) + " ");
			counter++;
			if(counter == 16)
			{
				result.append("   ");
				int charpoint = i - 15;
				for(int a = 0; a < 16; a++)
				{
					int t1 = data[charpoint++];
					if(t1 > 0x1f && t1 < 0x80)
						result.append((char) t1);
					else
						result.append('.');
				}
				result.append("\n");
				counter = 0;
			}
		}

		int rest = data.length % 16;
		if(rest > 0)
		{
			for(int i = 0; i < 17 - rest; i++)
				result.append("   ");

			int charpoint = data.length - rest;
			for(int a = 0; a < rest; a++)
			{
				int t1 = data[charpoint++];
				if(t1 > 0x1f && t1 < 0x80)
					result.append((char) t1);
				else
					result.append('.');
			}
			result.append("\n");
		}

		return result.toString();
	}

	private static String fillHex(int data, int digits)
	{
		String number = Integer.toHexString(data);

		for(int i = number.length(); i < digits; i++)
			number = "0" + number;

		return number;
	}

	public void handleIncompletePacket(L2GameClient client)
	{
		_log.warning("Packet not completed. Maybe cheater. IP:" + client.getIpAddr());
		if(Config.PROTECT_ENABLE && !client.HWID.isEmpty())
			_log.severe("Client HWID: " + client.HWID);
		if(client.getUPTryes() > 4)
		{
			L2Player activeChar = client.getActiveChar();
			if(activeChar == null)
			{
				_log.warning("Too many incomplete packets, connection closed. IP: " + client.getIpAddr() + ", account:" + client.getLoginName());
				client.closeNow(true);
				return;
			}
			_log.warning("Too many incomplete packets, connection closed. IP: " + client.getIpAddr() + ", account:" + client.getLoginName() + ", character:" + activeChar.getName());
			activeChar.logout(false, false, true);
		}
		else
			client.addUPTryes();
	}

	@SuppressWarnings("unchecked")
	@Override
	public HeaderInfo handleHeader(SelectionKey key, ByteBuffer buf)
	{
		if(buf.remaining() >= 2)
		{
			int dataPending = (buf.getShort() & 0xffff) - 2;
			L2GameClient client = ((MMOConnection<L2GameClient>) key.attachment()).getClient();
			return getHeaderInfoReturn().set(0, dataPending, false, client);
		}
		L2GameClient client = ((MMOConnection<L2GameClient>) key.attachment()).getClient();
		return getHeaderInfoReturn().set(2 - buf.remaining(), 0, false, client);
	}
}
